#ifndef  _SB_DMBUFFER_H
#define  _SB_DMBUFFER_H

// Inventor
#include <Inventor/SbLinear.h>

// Digital Media
#include <dmedia/dmedia.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include <vl/vl.h>

// GL
#include <GL/glu.h>
#include <GL/glx.h>

// System V
#include <ulocks.h>

// DMU
#include "libdmu/dmuGLTex2d.h"
#include "libdmu/dmuGLCtx.h"

/**
 * Wrapper class for DMBuffers:
 * 
 * Maintains two active buffers: <i>working</i> and <i>current</i>.
 * <p>
 * The <i>current</i> buffer is always available for access and contains
 * the last <u>complete</u> image.
 * The <i>working</i> buffer is allocatable for user/engine modification.
 * Once released, it is made the <i>current</i> buffer
 * <p>SbDMBuffer's are created automatically by all existing DMBufferEngines.
 * <p><i>Note: These access buffers should be thread-safe</i>
 *
 * @man dmbuffer
 * @man vlDMPoolGetParams
 * @man dmBufferCreatePool
 * @man dmBufferAllocate
 * @man dmICSend
 * @man dmICReceive
 * @man vlEventRecv
 * @man vlEventToDMBuffer
 * @man vlDMBufferSend
 * @man dmBufferMapData
 * @see SbDMBufferHelper.html
 */
class SbDMBuffer {

  public:

    // We Rely on you...
    /**
     * Constructor
     *
     * This constructor should be called if you have already created
     * an instance of both a dmuGLContext and dmuTexture.
     *
     * @param display	The X display
     * @param size	The size of each pool entry
     */
    SbDMBuffer( dmuGLCtx *dmuGLContext, dmuGLTex2d *dmuObject );

    // We Alloc
    /**
     * Constructor
     *
     * This constructor should be called if you want the SbDMBuffer
     * to create and manage an offscreen pbuffer.
     *
     * <b>Currently unimplemented</b>
     *
     * @param display	The X display
     * @param size	The size of each pool entry
     */
    SbDMBuffer( Display *display, const SbVec2s &size );

    /**
     * Destructor
     */
    ~SbDMBuffer( void );

  // Map Existing Data
    /**
     * Maps and locks the current buffer and returns a pointer to
     * its data. (Data is modifiable)
     * @param size	(Returns) Image dimensions
     * @param pixels	(Returns) Pointer to pixel data
     * @return		Pointer to pixel data
     * @man dmBufferMapData
     */
    unsigned char *	 getValue(SbVec2s &size, unsigned char **pixels);

    /**
     * Maps, locks, and <b>makes current</b> for rendering
     * the current buffer. Provides direct access to
     * the allocated dmuGLCtx object. (Data is modifiable)
     *
     * @param size	(Returns) Image dimensions
     * @return		The allocated dmuGLCtx
     * @man DMbuffer
     */
    dmuGLCtx *		 getValue(SbVec2s &s);

    /**
     * Releases the lock from a previous call to getValue
     * @see		getValue
     */
    void		 releaseValue( void );

    /**
     * Called by SoDMBufferVideoTexture to update the contents
     * of the current dmuGLTexture object.
     *
     * @see		SoDMBufferVideoTexture.html
     */
    // DMU stuff..
    void 		 defineTex( void );

    // Allocate & Modify Content
    // Map pixels for glDrawPixels, etc.
    /**
     * Allocates, Maps, locks, and <b>makes current</b> for rendering
     * the working buffer. Providing direct access to
     * a newly allocated DMbuffer from the buffer pool.
     *
     * <p>
     * <b>Until allocatable SbDMBuffer's are implemented, this is
     * essentially the same as getValue(SbVec2s &,unsigned char **)</b></p>
     *
     * @param size	(Returns) Image dimensions
     * @param pixels	(Returns) Pointer to pixel data
     * @return		Pointer to pixel data
     *
     * @man dmBufferAllocate
     * @man dmBufferMapData
     * @man glXMakeCurrent
     */
    unsigned char *	 startEditing(SbVec2s &s, unsigned char **pixels);

    // Make current render context
    /**
     * Allocates, Maps, locks, and <b>makes current</b> for rendering
     * the working buffer. Providing direct access to
     * a newly allocated DMbuffer from the buffer pool.
     *
     * <p>
     * <b>Until allocatable SbDMBuffer's are implemented, this is
     * essentially the same as getValue(SbVec2s &)</b></p>
     *
     * @param size	(Returns) Image dimensions
     * @return		The allocated dmuGLCtx
     *
     * @man dmBufferAllocate
     * @man dmBufferMapData
     * @man glXMakeCurrent
     */
    dmuGLCtx *		 startEditing(SbVec2s &s);

  
    // Unmap / return context
    /**
     * Releases the lock from a prvious call to startEditing 
     * making the working buffer the current buffer.
     * @see		startEditing
     */
    void       		 finishEditing( void );

  private:
    Bool		GRABBED;
    dmuGLTex2d		*dmuTexture;
    dmuGLCtx		*dmuGLContext;
};
#endif

