#ifndef _SO_DM_BUFFER_TEXTURE2_H
#define _SO_DM_BUFFER_TEXTURE2_H

#include <Inventor/nodes/SoTexture2.h>
#include <Inventor/fields/SoSFMatrix.h>
#include "SoSFDMBufferImage.h"

//////////////////////////////////////////////////////////////////////////////
//
//  Class: SoDMBufferTexture2
//
//  Texture node.
//
//////////////////////////////////////////////////////////////////////////////

/**
 * SoDMBufferTexture2 is a subclass of the SoTexture2 node which
 * provides texture mapping capabilities in an Inventor scene graph.
 * <p>
 * The <b>image</b> field is an SoSFDMBufferImage which (once assigned
 * to an SbDMBuffer) provides dynamic (DMbuffer) texture mapping in
 * Inventor on O2 systems.
 * <p>
 * The <b>transform</b> field is used to automatically adjust texture
 * coordinates for DMbuffer texture pixel data which does not have
 * dimensions which are a power of two.
 * <p>
 * For example: If <b>image</b> is connected to the <b>textureBits</b>
 * field of an SoDMBufferMovieEngine having a movie with dimensions,
 * 196x115, the SoDMBufferMovieEngine will allocate an SbDMBuffer with
 * dimensions 256x256 and render the movie in the lower right-hand corner.
 * For this reason, the SoDMBufferMovieEngine provides a <b>scale</b>
 * field which can be connected to the <b>transform</b> field. To
 * compensate for the ill-fitting render, the SoDMBufferMovieEngine
 * class would set its <b>scale</b> field as follows:
 * <p>
 * <pre>
 *   float scale[4][4] = { rx,  0,  0,  0,
 *                          0, ry,  0,  0,
 *                          0,  0,  1,  0,
 *                          0,  0,  0,  1 };
 * </pre>
 * <p>
 * Where rx = is the ration of the movie's width to the allocated 
 * SbDMBuffer width (196/256 = 0.76562); and ry = is the ratio of
 * the movie's height to the SbDMBuffer height (115/256 = 0.44921).
 * <p>
 * This allows the SoDMBufferTexture2 node to adjust the GL texture
 * matrix stack in such a way that no adjustments should need to be
 * made to down-stream objects -- in short, it does what you would expect.
 * <h2>Example Uses:</h2>
 * <p>
 * <b>Video Texture</b>
 * <pre>
 *    SoDMBufferTexture2 *texture = new SoDMBufferTexture2();
 *    sep->addChild(texture);
 *
 *    SoDMBufferVideoEngine *videoIn = new SoDMBufferVideoEngine();
 *    texture->image.connectFrom( &videoIn->textureBits );
 *    texture->transform.connectFrom( &videoIn->scale );
 * </pre>
 * <p>
 * <b>Movie Texture</b>
 * <pre>
 *    SoDMBufferTexture2 *texture = new SoDMBufferTexture2();
 *    sep->addChild(texture);
 *
 *    SoDMBufferMovieEngine *movieIn = new SoDMBufferMovieEngine();
 *    texture->image.connectFrom( &movieIn->textureBits );
 *    texture->transform.connectFrom( &movieIn->scale );
 * </pre>
 * <b>Renderable Texture</b>
 * <pre>
 *    float scale[4][4] = { rx,  0,  0,  0,
 *                          0, ry,  0,  0,
 *                          0,  0,  1,  0,
 *                          0,  0,  0,  1 };
 *
 *    SbMatrix renderMatrix;
 *    renderMatrix.setValue( scale );
 *
 *    SbDMBuffer *renderDMBuffer;
 *    renderDMBuffer = new SbDMBuffer( SbVec2s( WIDTH, HEIGHT) );
 *
 *    Display *display = NULL;
 *    openStandardPBuffer( display, renderDMBuffer );
 *    renderDMBuffer->setBufferCount( BUFFERCOUNT );
 *    renderDMBuffer->createPool();
 *
 *    SoDMBufferTexture2 *texture = new SoDMBufferTexture2();
 *    sep->addChild(texture);
 *
 *    SoSFDMBufferImage *image = new SoSFDMBufferImage();
 *    texture->image.connectFrom( &renderDMBuffer );
 *    texture->transform.connectFrom( &renderMatrix );
 * </pre>
 *
 * <p>
 * <b>Fields from class SoTexture2:</b>
 * <pre>
 *    SoSFEnum            wrapS
 *    SoSFEnum            wrapT
 *    SoSFEnum            model
 *    SoSFColor           blendColor
 * </pre>
 *
 * @man SoTexture2
 * @man SoSFImage
 * @man dmbuffer
 * @see SbDMBuffer.html
 * @see SoDMBufferVideoEngine.html
 * @see SoDMBufferMovieEngine.html
 * @see SoDMBufferDMICMovieEngine.html
 */
class SoDMBufferTexture2 : public SoTexture2 {

    SO_NODE_HEADER(SoDMBufferTexture2);

  public:

    // Fields.
    /**
     * Must be connected to an SoSFDMBufferImage (such as the output
     * of an SoDMBufferVideoEngine)
     *
     * @see		SoSFDMBufferImage.html
     */
    SoSFDMBufferImage	dmBufferImage;

    /**
     * May be connected to an SoSFMatrix (such as the scale output
     * of an SoDMBufferVideoEngine)
     *
     * @see		SoSFDMBufferImage.html
     */
    SoSFMatrix		transform;

    // Constructor
    SoDMBufferTexture2();
    
  SoEXTENDER public:
    virtual void	doAction(SoAction *action);
    virtual void	GLRender(SoGLRenderAction *action);

  SoINTERNAL public:
    static void		initClass();

  protected:
    virtual ~SoDMBufferTexture2();
};

#endif 
