#ifndef  _SO_SF_DMBUFFER_VIDEO_ENGINE_H
#define  _SO_SF_DMBUFFER_VIDEO_ENGINE_H

#include <Inventor/fields/SoSFString.h>

// DMUIV
#include "SoDMBufferEngine.h"

// DMU
#include "libdmu/dmuVideoIn.h"
#include "libdmu/dmuGLTex2d.h"
#include "libdmu/dmuGLCtx.h"

/**
 * SoDMBufferVideoEngine feeds a constant stream of video
 * from the currently selected video device (unless otherwise specified)
 * to it's outputs: <b>textureBits</b> and <b>linearBits</b>. These
 * outputs are typically tied to an SoSFDMBufferImage (such as the
 * image field of an SoDMBufferTexture2 node.
 * <p>
 * The scale varable is adjusted, based on the size of the incident
 * video, to properly scale downstream textures when connected to an
 * SoDMBufferTexture2 object.
 * <p>
 * <b>Fields from class SoDMBufferEngine:</b>
 * <pre>
 *    SoEngineOutput textureBits;
 *    SoEngineOutput linearBits;
 *    SoEngineOutput scale;
 * </pre>
 * <p>
 * <b>Technical Notes:</b>
 * <p>
 * Video processing is done in a separate thread.
 * <p>
 * Default Video Parameters:
 * <pre>
 *    VIDEO_WIDTH  = 512;
 *    VIDEO_HEIGHT = 512;
 *    VIDEO_BUFFER_COUNT =   5;
 * </pre>
 *
 * Video Layout Parameters:
 * <pre>
 *    LAYOUT_GRAPHICS,
 *    LAYOUT_LINEAR,
 *    LAYOUT_BOTH
 * </pre>
 *
 */
class SoDMBufferVideoEngine : public SoDMBufferEngine {

    SO_ENGINE_HEADER( SoDMBufferVideoEngine );

  public:

    SoSFString	videoSource;

    const short VIDEO_WIDTH  = 512;
    const short VIDEO_HEIGHT = 512;
    const short VIDEO_BUFFER_COUNT =   5;

    enum { LAYOUT_GRAPHICS,
	   LAYOUT_LINEAR };

    static void 	initClass();

    /**
     * Constructor
     * <p>
     * The default values (assumed by this constructor) are appropriate for most applications.
     * <p>
     * (see the Insight Digital Media Programmer's Guide for information concerning video parameters)
     *
     * @param layout		Dictates active engine outputs.
     * @param videoSource	Source for video. NULL= selectable via vcp(1); specify others by name: "Composite".
     * @param originX		Video horizontal origin
     * @param originY		Video vertical origin
     * @param zoomNumerator	Video zoom (fractional) numerator
     * @param zoomDenominator	Video zoom (fractional) denominator
     * @param aspectNumerator	Video aspect (fractional) numerator
     * @param aspectDenominator	Video aspect (fractional) denominator
     * @param width		Video and SbDMBuffer size/width (must be power of 2)
     * @param height		Video and SbDMBuffer size/height (must be power of 2)
     * @param bufferCount	DMbufferpool size
     * @man vl
     * @man vlSetControl
     */
    SoDMBufferVideoEngine( int layout = LAYOUT_GRAPHICS,
			   const char* videoSource = NULL, 
			   int originX = 0,
			   int originY = 32,
			   int zoomNumerator = -1,
			   int zoomDenominator = -1,
			   int aspectNumerator = -1,
			   int aspectDenominator = -1,
			   int width = VIDEO_WIDTH, 
			   int height = VIDEO_HEIGHT,
			   int bufferCount = VIDEO_BUFFER_COUNT );

    /**
     * Stop dynamically updating the current video image.
     */
    virtual void start( void );

    /**
     * Resume dynamically updating the current video image.
     */
    virtual void stop( void );

    void inputChanged(SoField *whichField);


  protected:

    dmuVideoIn *videoIn;
    dmuGLCtx   *videoContext;
    dmuGLTex2d *texDrain;


  int layout;
  int originX;
  int originY;
  int zoomNumerator;
  int zoomDenominator;
  int aspectNumerator;
  int aspectDenominator;
  int width;
  int height;
  int bufferCount;


  private:

    ~SoDMBufferVideoEngine();
    int VIDEO_LAYOUT;

};

#endif
