#ifndef  _SO_SF_DMBUFFER_IMAGE_H
#define  _SO_SF_DMBUFFER_IMAGE_H

// We can't really inherit from SoSFImage since it didn't
// bother to declare most of its functions virtual. May
// as well create a new class and avoid any mis-understanding.

// Inventor
#include <Inventor/fields/SoSubField.h>

// DMUIV
#include "SbDMBuffer.h"

/**
 * A field containing an SbDMBuffer. The SoSFDMBufferImage does <i>not</i> create
 * or manage an SbDMBuffer; it simply provides access to an assigned SbDMBuffer.
 * Classes which wish to have an SoSFDMBufferImage field, must
 * initialize its contents with a call to setValue().
 * <p>
 * Once initialized, the contents of this field may be fetched by
 * using one of three access methods: getImage (overloaded twice)
 * or getBuffer. Once the contents of the field have been used (e.g.
 * a call has been made to glXAssociateDMPbufferSGIX or glDrawPixels)
 * the image should be released -- DMbuffers typically have highly
 * dynamic contents, releasing an SoSFDMBufferImage makes it's
 * associated DMBufferpool entry available for use again.
 * <p>
 * This field can also be used to drive DMbufferpools by allocating
 * a buffer from the pool for rendering. This is a three step process:
 * a call is made to startEditing reserving an SbDMBuffer working
 * the image is rendered to the returned pixel data pointer, and a call
 * is made to finishEditing which frees any previous data and makes
 * avialable the newly rendered content.
 *
 * @man SoTexture2
 * @man SoSFImage
 * @man dmbuffer
 * @see SbDMBuffer.html
 */
class SoSFDMBufferImage: public SoSField  {

    SO_SFIELD_HEADER(SoSFDMBufferImage, SbDMBuffer *, SbDMBuffer *);

  public:

    // getValue returns the size, number of components and a 
    // pointer to a mapped dmbuffer image.

    /**
     * Maps and locks the current buffer and returns a pointer to
     * its data. (Data is modifiable)
     * @return		Pointer to pixel data
     *
     * @see SbDMBuffer.html#getValue
     */
    const unsigned char *	getImage( void );

    /**
     * Maps and locks the current buffer and returns a pointer to
     * its data. (Data is modifiable)
     *
     * @param size	(Returns) Image dimensions
     * @return		Pointer to pixel data
     *
     * @see SbDMBuffer.html#getValue
     */
    const unsigned char *	getImage(SbVec2s &size);

    /**
     * Maps and locks the current buffer and returns a pointer to
     * the SbDMBuffer. (Data is modifiable)
     *
     * @param size	(Returns) Image dimensions
     * @return		Pointer to the SbDMBuffer
     *
     * @see SbDMBuffer.html#getValue
     */
    SbDMBuffer *		getBuffer( SbVec2s &size );

  
    /**
     * Releases the lock from a previous call to getValue
     * @see #getValue
     * @see SbDMBuffer.html#releaseValue
     */
    void			release( void );

    /**
     * Allocates, Maps, locks, and <b>makes current</b> for rendering
     * the working buffer. Providing direct access to
     * a newly allocated DMbuffer from the buffer pool.
     *
     * <p>
     * <b>Until allocatable SbDMBuffer's are implemented, this is
     * essentially the same as getValue(SbVec2s &)</b></p>
     *
     * @param size	(Returns) Image dimensions
     * @return		The allocated dmuGLCtx
     *
     * @man dmBufferAllocate
     * @man dmBufferMapData
     * @man glXMakeCurrent
     *
     * @see SbDMBuffer#startEditing
     */
    SbDMBuffer *		startEditing(SbVec2s &s);

    /**
     * Releases the lock from a prvious call to startEditing 
     * making the working buffer the current buffer.
     * @see startEditing
     * @see SbDMBuffer#finishEditing
     */
    void			finishEditing( void );

  SoINTERNAL public:
    static void			initClass();
};

#endif







