#ifndef PARAMETRIC_CURVE_H
#define PARAMETRIC_CURVE_H

// Parametric Curves
#include "Primitive.h"

// STL
#include "list.h"

// X11
#include <X11/Xlib.h>

class ParametricCurve {
  public:

    typedef struct {
	Primitive::Point point;
	list<Primitive::Point> weight;
    } knotInfo;

    typedef list<knotInfo>::iterator KNOT;
    typedef list<Primitive::Point>::iterator WEIGHT;

    ParametricCurve( void );
    ~ParametricCurve( void );

    // Definition
    virtual void addKnot( KNOT *whichKnot, int x, int y);
    virtual void insertKnotBefore(KNOT whichKnot, int x, int y );
    virtual void removeKnot( KNOT whichKnot );
    virtual void finish( void );

    // Query Curve
    virtual void getFirstKnot( KNOT *whichKnot, int *x, int *y );
    virtual void getFirstWeight( KNOT whichKnot, 
				 WEIGHT *whichWeight,
				 int *x, int *y );
    virtual Bool getKnotAtPoint( KNOT *whichKnot,
				 int x, int y, int radius = 5 );
    virtual Bool getWeightAtPoint( KNOT whichKnot, 
				   WEIGHT *whichWeight, 
				   int x, int y, int radius = 5 );
    virtual Bool getNextKnot( KNOT *whichKnot, int *x, int *y);
    virtual Bool getNextWeight( KNOT whichKnot,
				WEIGHT *whichWeight, int *x, int *y);
    virtual Bool getPoints(int (**point)[2], int *num_points );

    // Modify Curve
    virtual void moveKnot(KNOT whichKnot, int x, int y);
    virtual void moveWeight(KNOT whichKnot, WEIGHT whichWeight, int x, int y);
    virtual void moveCurve(int startx, int starty,
			   int currentx, int currenty);

    // Direct Drawing Support
    typedef void (*DirectRenderCallback)( int x, int y, Bool firstPoint,
					  void *callData );

    virtual void registerDirectRenderCallback(DirectRenderCallback callback,
					      void *userCallData);
    virtual void render( void ) = 0;
    virtual Bool isDirectDraw( void );
    virtual void enableInterp(Bool onOff) { interpolate = onOff;}
    virtual void setInterpolationStepSize(int size) { interpolationStepSize = size;}

    // Query State
    virtual Bool pointIsOnCurve( int x, int y, int radius = 5 );
    virtual Bool isDefined( void );
    
  protected:
    // Can't use Inventor's SbMatrix class since
    // it only allows SbVec3f multiplication (need
    // 4 parameter vector matrix support)
    class Matrix {
      public:
	Matrix(float a11, float a12, float a13, float a14,
	       float a21, float a22, float a23, float a24,
	       float a31, float a32, float a33, float a34,
	       float a41, float a42, float a43, float a44);
	void multMatrixVec(float s[4], float d[4]);
	void multVecMatrix(float s[4], float d[4]);
	const float * getValue( void );

      private:
	float values[4][4];
    };

    Matrix *CurveMatrix;

    Bool dirty;
    Bool defined;
    list<knotInfo> knot;

    int (*point)[2];

    int num_points, point_array_length;
    int last_knot;
    int last_weight;

    virtual void addWeight( KNOT whichKnot,
			    WEIGHT *whichWeight, int x, int y);
    virtual void insertWeightBefore( KNOT whichKnot,
				     WEIGHT whichWeight,int x, int y );
    virtual void removeWeight( KNOT whichKnot,
			       WEIGHT whichWeight );

    virtual Bool buildGeometryMatrix( KNOT whichKnot,
				      float (*geometryVectorX)[4][4]) = 0;

    virtual void calculateSpline( KNOT startKnot );

    // Direct Render Support
    DirectRenderCallback directRender;
    Bool                interpolate;
    int                 interpolationStepSize;

    virtual void renderDirect( KNOT startKnot ) = 0;

  void *userCallData;
  
};

#endif
