#ifndef  _SO_SF_DMBUFFER_ENGINE_H
#define  _SO_SF_DMBUFFER_ENGINE_H

// Inventor
#include <Inventor/engines/SoSubEngine.h>
#include <Inventor/fields/SoSFTrigger.h>

// DMUIV
#include "SoSFDMBufferImage.h"

// STL
#include <list.h>

// DMU
#include "libdmu/dmuObj.h"

/**
 * SoDMBufferEngine is the base class for each of the
 * DMBuffer engines.
 * <p>
 * Subclasses should support both DM_IMAGE_LAYOUT_GRAPHICS 
 * and DM_IMAGE_LAYOUT_LINEAR formats. The <b>textureBits</b>
 * and <b>linearBits</b> engine outputs should provide access to
 * the indicated layout (i.e. GRAPHICS - <b>textureBits</b>, LINEAR -
 * <b>linearBits</b>.
 * <p>
 * Additionally, where appropriate, the engine should provide a
 * texture scaling matrix. (see discussion in SbDMBufferTexture2).
 * 
 * @see SoDMBufferTexture2.html
 * @see SoDMBufferVideoEngine.html
 * @see SoDMBufferMovieEngine.html
 * @see SoDMBufferDMICMovieEngine.html
 */
class SoDMBufferEngine : public SoEngine {

    SO_ENGINE_ABSTRACT_HEADER( SoDMBufferEngine );

  public:

    typedef struct {
	dmuObj *dmuobj;
	Bool started;
    } BurnItem;

    /**
     * (SoSFDMBufferImage) DM_IMAGE_LAYOUT_GRAPHICS access to engine's
     * image.
     * @see SoSFDMBufferImage.html
     */
    SoEngineOutput textureBits;

    /**
     * (SoSFDMBufferImage) DM_IMAGE_LAYOUT_LINEAR access to engine's image. 
     * @see SoSFDMBufferImage.html
     */
    SoEngineOutput linearBits;

    /**
     * (SbMatrix) access to engine's SoSFDMBufferImage
     * <p> Discussion of proper use can be found in SoDMBufferTexture2
     * documentation.
     *
     * @see SoSFDMBufferTexture2.html
     * @see SoSFDMBufferImage.html
     */
    SoEngineOutput scale;

    /**
     * Static initialization routine. (Used by some engines to create
     * separate data acquisition threads -- SoDMBufferVideoEngine and
     * SoDMBufferDMICMovieEngine)
     *
     * @see SoDMBufferVideoEngine.html
     * @see SoDMBufferMovieEngine.html
     * @see SoDMBufferDMICMovieEngine.html
     */
    static void initClass();
    SoDMBufferEngine( void );

    /**
     * Stop dynamically updating the current image.
     *
     * @see SoSFDMBufferImage.html
     * @see SoSFDMBufferTexture2.html
     * @see SoDMBufferVideoEngine.html
     * @see SoDMBufferMovieEngine.html
     * @see SoDMBufferDMICMovieEngine.html
     */
    virtual void stop( void ) = 0;

    /**
     * Resume dynamically updating the current image.
     *
     * @see SoSFDMBufferImage.html
     * @see SoSFDMBufferTexture2.html
     * @see SoDMBufferVideoEngine.html
     * @see SoDMBufferMovieEngine.html
     * @see SoDMBufferDMICMovieEngine.html
     */
    virtual void start( void ) = 0;

  protected:
    SoSFTrigger	 frameTrigger;
    SbMatrix	 scaleMatrix;
    SbDMBuffer   *engineDMBuffer;

    virtual void evaluate();

    static void registerBurnItem( dmuObj *dmuobj );

  private:
    static pid_t child;

    static void burn( void *);
    static unsigned int burnCount;
    static usptr_t *arena;
    static usema_t *burnSema;
    static list<BurnItem *> burnStack;

    static SoSFTrigger *globalFrameTrigger;
};

#endif
