/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* gvplay_j = Plays compressed video fields from disk to
 *            video-out and to the gfx monitor.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-gfx_frd or	- Display a full frame by doubling scan lines. 
 *      -gfx_fri or	- Display a full frame by interlacing fields.
 *      -gfx_half>  	- Display half height field. 
 *     <-single>  	- Force single buffering (default is double buffered)
 *     fileName 	- file created with vgrab_j.
 *
 *
 * This program plays fields of compresed video from disk to video-out and gfx
 * in real time.  The video is stored as compressed jpeg on disk and 
 * decompressedinto YUV 4:2:2 before being drawn to gfx and sent to video-out. 
 *
 * By default, the gfx window is the size of a field of video (i.e. 720x243 for
 * CCIR 525).  -gfx_half forces the gfx window to be half sized (360x243 for
 * CCIR 525).  -gfx_frd and -gfx_fri cause the window to display an entire frame
 * (720x486 for CCIR 525).  For -gfx_fri, single buffering is always forced
 * on and the SGI interlace extension is used.
 *
 * DMbufferpools provide the application with special purpose memory that
 * can be shared between the gfx, video, and compression devices.  A
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of four DMbufferpools:
 *      _inpool - This pool contains "INBUFS" DMbuffers, used as input for the
 *                decompressor. Compressed fields are read into this pool.
 *     _outpool - This pool contains "OUTBUFS" DMbuffers, used to hold
 *                output from the decompressor (YUV 422 video fields).
 *		  This pool is also attached to the VL_MEM node that drives
 *		  video-output.  This allows these dmbuffers to be sent
 *		  directly to video out with no additional copying.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 * 	- After starting the program, a blank window appears.  
 *	- Click with the left mouse button to start play.  Click again to stop,
 *	  then again to start again.  
 *	- Each button click causes the main thread to enter the routine doit().
 *      - Each time play is started,  a read thread (see ReadIt) and display
 * 	  thread (see PutField) are sproc'd from inside the routine doit().
 *	- While things are playing, the main thread is in  XtAppMainLoop(), 
 *	  waiting for another button click.
 *
 *    Read Thread:
 * 	- The read thread (see ReadIt) reads compressed fields of video 
 *	  off of disk into dmbuffers allocated from _inpool.
 *	- The compressed fields are then sent to the decompressor.
 *	- When play stops, the read thread exits first.
 *
 *    Display Thread:
 *	- The display thread (see PutField) continually checks for valid output
 *	  from the decompressor.  
 *	- The uncompressed output (residing in a dmbuffer) is then sent to 
 *	  video-out.  
 *	- When a dmbuffer is sent to video-out (using vlDMBufferSend) its video 
 *        display time is calcualted using vlGetFrontierMSC() and saved away. 
 *	- The display thread then draws the field to the gfx monitor (using 
 *	  glDrawPixels) when the corresponding video-out display time has been 
 *	  reached (by sampling dmGetUST).  
 *      - This keeps video play  and play to the gfx monitor in sync.  
 *      - **** WHEN RUNNING IN DOUBLE BUFFERED MODE, this seems to work best 
 *	  when the gfx timing is set to 75 Hz (see xscreen).
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/jinfo.h"
#include "common/fileutils.h"

#define INBUFS 12 
#define OUTBUFS 12
#define GFXBUFS 12
#define FILL_CHECK 60
#define VL_COMPRESSION_JPEG 2

/* All global variables in this file are preceded by an "_".  */

/* This structure contains dmbuffer/usr pairs used to sync. play between
 * the gfx monitor and video out.
 */
typedef struct {
    DMbuffer buf;
    stamp_t ust;
} GFXBuf; 

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

/* File  globals	*/
jFieldInfo_t *_indexbuffer; 	/* Array of compressed field sizes */
fileHeader_t  _f;

/* dmIC related defines */
DMimageconverter _vice;         /* Handle to decompressor               */
DMbufferpool     _inpool;       /* dmbuffer pool used as decompressor input */
DMbufferpool     _outpool;      /* dmbuffer pool used as decompressor output. */
                                /* AND is also attached to a VL_MEM node*/
int              _inpoolfd;     /* fd used to select on _inpool         */
int              _outpoolfd;    /* fd used to select on _outpool        */

/* gfx declarations.	*/
Display   *_dpy;
Widget     _glw;		/* GL drawing widget.		*/
GLXContext _context;		/* GL context.			*/
int        _dbuffer;  		/* Double buffering flag.	*/
int        _inter;	 	/* Interlacing flag.		*/ 
int        _winx;		/* Length of gfx window.	*/
int        _winy;		/* Width of gfx window.		*/

volatile int _IamPlaying = 0;
int _CurField = 0;		/* Current field being read.    */
int _transferSize; 		/* Bytes in uncompressed image 	*/

/* Routine declarations	*/
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void initVL();
void initGL();
void initCompression();
void PutField();
void ReadIt();
int sginap (long );

main(int argc, char *argv[])
{
    int i,n;
    char *Fname;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"doit", doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n";

    _inter = 0; _dbuffer = 1; _winx =1; _winy = 1;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-single") == 0) {
            _dbuffer = 0;
        }
        else if (strcmp (argv[i], "-gfx_fri") == 0) {
            _inter = 1; _dbuffer = 0; _winy = 2; _winx = 1;
	}
        else if (strcmp (argv[i], "-gfx_frd") == 0) {
            _inter = 0;             ; _winy = 2; _winx = 1;
        }
        else if (strcmp (argv[i], "-gfx_half") == 0) {
            _inter = 0;             ; _winy = 1; _winx = 2;
        }
        else {
            Fname = argv[i];
        }
    }

    /* Get file information.	*/
    fileInit(Fname, &_f, FALSE);
    _indexbuffer = createIndexBuffer(_f.Nfields);
    readIndexBuffer(&_f, _indexbuffer);

    ERROR ((_f.packing != VL_COMPRESSION_JPEG), "File not a jpeg file");

    _winx = _f.uSize/_winx;
    _winy = _f.vSize*_winy;

    /* Set up video hardware.	*/
    initVL();

    /* Set up compression hardware.	*/
    initCompression();

    fprintf(stderr, "Playing  %d fields of %dx%d \n", _f.Nfields, _f.uSize, _f.vSize);
    fprintf(stderr, "Header size = %d, quality = %d\n", _f.headerSize, _f.qf);
    fprintf(stderr, "Adjusted transfer size = %d\n", _transferSize);
    fprintf(stderr, "Gfx window size = %dx%d\n", _winx, _winy);

    /* Do that nasty Motif stuff then head off to the Xt main loop.     */

    toplevel = XtAppInitialize(&app_context, "4DgiftsGlx",
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv,
			       NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight, _winy);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);
    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);
}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the initialize call back routine for our widget.   This
     * is called automatically from the Xt loop.
     */
    float xdecim, ydecim;
    Arg args[18];
    XVisualInfo *vi;

    _dpy = XtDisplay(_glw);

    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(w, args, 1);

    /* create a visual context */
    _context = glXCreateContext(XtDisplay(_glw), vi, 0, GL_TRUE);
    GLwDrawingAreaMakeCurrent (_glw, _context);

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glColor4f(0.0, 0.0, 0.0, 1.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_winx, 0.0, (double)_winy, -1.0, 1.0);

    if (_inter) {
        glEnable(GL_INTERLACE_SGIX);
        glPixelZoom (1.0, -1.0);
    }
    else {
	xdecim = ((float)(_winx))/_f.uSize;
	ydecim = ((float)(_winy))/_f.vSize;
	glPixelZoom (xdecim, -ydecim);
    }

    glRasterPos2i (0, _winy);
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the expose callback routine, called whenever an expose happens.  
     * Note, if real time play is happening then this routine cannot try to 
     * grab the gfx context.
     */
    static int firstExpose = 1;

    if (firstExpose) {
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);

        if (_dbuffer) {
	    glXSwapBuffers(_dpy,XtWindow(_glw));
            glClear(GL_COLOR_BUFFER_BIT);
	}
 	firstExpose = 0;
	fprintf(stderr, "*****> Click in window to start play, click again to stop, etc.\n");
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routine is called each time the left mouse button is clicked.
     * If things are stopped then the read and display threads are sproc'd.
     * If things are playing then the read and display thread are stopped
     * by setting _IamPlaying to 0.
     */
    int readPID, putPID;

    if (!_IamPlaying) {
        /* Things are stopped, start play	*/
	_IamPlaying = 1;
        GC (glXMakeCurrent(_dpy, None, NULL));
        readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
        putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);
    }
    else {
        /* Setting _IamPlaying to 0 will stop the ReadIt().  Upon exiting,
         * ReadIt() sets _IamPlaying to -1, this will stop PutField().  Upon
         * exiting, PutField sets _IamPlaying to -2.
         */
        _IamPlaying = 0;
        while (_IamPlaying != -2) sginap(1);  /* Wait for both threads.	*/
        _IamPlaying = 0;
	VC  (vlEndTransfer(_svr, _path));
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));

        /* Make sure _CurField is pointing at an even field     */
        if ((_CurField % 2) != 0) {
            _CurField++;
            if (_CurField >= _f.Nfields) _CurField = 0;
        }
    }
}

void PutField()
{
    /* This routine has two main tasks.  It checks for decompressor output
     * by doing select on "icfd", the file descriptor for the decompressor's 
     * output * queue.  The dmbuffer containing the uncompressed output is 
     * then sent to video. After sending the dmbuffer to video-out, it saves 
     * it in go[tail].buf along with the ust for when this field will appear 
     * at video out.  So the "go" queue contains dmbuffers which have been 
     * sent to video out and are scheduled to go to gfx.  Each time through 
     * the loop, go[head].ust is tested against the current system UST to see 
     * if it is time to draw the next field to gfx.  This keeps video and gfx 
     * play in sync.
     *
     * The ReadIt() thread is constantly reading fields off of disk and
     * placing them in the input pool of the decompressor. 
     */

    GFXBuf go[GFXBUFS];	    /* (dmbuf,ust) pairs that need to go to gfx */
    int head = 0;           /* The head of the go queue - the next field for gfx output */
    int tail = 0;           /* The tail of the go queue - the field just sent to video  */
    int count = 0;
    int i, scan;
    stamp_t cur_ust;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    USTMSCpair pair;
    double ust_per_msc;
    void *dataPtr;
    fd_set fds;
    int icfd;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();

    GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
    ust_per_msc = vlGetUSTPerMSC(_svr, _path, _src);
    icfd = dmICGetDstQueueFD(_vice);  /* to select on available uncompressed output. */

    /* If interlacing, the first field starts on a different line
     * depending if 525 or 625 timing.
     */
    scan = (_f.vSize == 288) ? 0 : 1;

    while (_IamPlaying != -1) {

        dmGetUST((unsigned long long *)&cur_ust);

        /* If the current ust has passed the video-out ust for the head
         * of the go list then draw the head of the go list to gfx.
         */
        if ((head != tail) && (cur_ust > go[head].ust)) {
	    dataPtr = dmBufferMapData(go[head].buf);

	    if (_inter) {
	        glRasterPos2i (0, _winy - scan);
	        scan = (scan == 1) ? 0 : 1; 
	    }

	    glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

            if (_dbuffer) {
                 if (!_inter)  glRecti(0, _winy, _winx, _winy-1);
                 glXSwapBuffers(_dpy,XtWindow(_glw));
            }

	    dmBufferFree(go[head].buf);  /* This dmbuffer can now be freed.	*/
            head++;
            if (head == GFXBUFS) head = 0; 
	}

        /* The rest of the loop deals with getting decompressor output and sending
         * it to video-out.
         */
        /* Test for a full queue (gfx output buffers).  If there
         * is one then go ahead and wait for the next compressed
         * field via select.
         */
        i = tail + 1;
        if (i == GFXBUFS) i = 0;

        if (i == head) {        /* No room in go to queue up next entry */
            continue;
            /* fprintf(stderr, "FULL QUEUE\n"); */
        }

        /* Wait for valid output from decompressor.	*/
	FD_ZERO(&fds);
	FD_SET(icfd, &fds);
	OC (select(icfd+1, &fds, NULL, NULL, NULL));

	/* Get ptr to output data  */
	DC( dmICReceive(_vice, &(go[tail].buf)) );

        if (count == 0) {
	    VC (vlBeginTransfer(_svr, _path, 0, NULL));
	    vlGetUSTMSCPair(_svr, _path, _drn, VL_ANY, _src, &pair);
        }

        /* Get MSC for the sample and check for drops   */
        newmsc = vlGetFrontierMSC(_svr, _path, _src);
        if ((oldmsc != -1) && (newmsc != oldmsc + 1) && (count > FILL_CHECK)) {
	    fprintf(stderr, "MSC drop old = %lld, new = %lld\n", oldmsc, newmsc);
        }
        oldmsc = newmsc;

        /* Send the dmbufer to video out and calculate the video-out ust for
         * the field.  This is used later to draw the field to gfx.
         */
	VC ( vlDMBufferSend(_svr, _path, go[tail].buf) );
        go[tail].ust = pair.ust + (newmsc - pair.msc)*ust_per_msc;

        /* Point the tail at the next free location in go queue */
        tail++;
        if (tail == GFXBUFS) tail = 0;
        count++;
    } 
    
    /* Helps with field swap problem	*/
    if ((count%2) != 0) {
        VC ( vlDMBufferSend(_svr, _path, go[head].buf) );
    }

    /* We have to free any dmbuffer in use.  The go queue
     * contains entries that have been sent to video but not yet
     * drawn to gfx.  
     */

    while (head != tail) {
	dmBufferFree(go[head].buf);
	head++;
	if (head == GFXBUFS) head = 0;
    }

    GC (glXMakeCurrent(_dpy, None, NULL));  /* Surrender gl context */
    _IamPlaying--;                          /* Tell doit() all is done */
    _exit(0);
}

void ReadIt()
{
    /* This thread simply reads compressed fields from disk and
     * sends them to the decompressor.
     */
    int nextRead;
    int dataSize;
    fd_set fds;
    unsigned char *dataPtr;
    DMbuffer buf;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();

    OC (lseek64 (_f.fd, _f.headerSize + _indexbuffer[_CurField].offset, SEEK_SET));

    while (_IamPlaying) {
        /* Wait for an input buffer to be available */ 
        FD_ZERO(&fds);
        FD_SET(_inpoolfd, &fds);
        OC (select(_inpoolfd+1, NULL, &fds, NULL, NULL));

        DC(dmBufferAllocate(_inpool, &buf));
        dataPtr = dmBufferMapData(buf);
    
        /* Read the compressed field */
        nextRead = _indexbuffer[_CurField+1].offset - _indexbuffer[_CurField].offset; 
        OC (read(_f.fd, dataPtr, nextRead));
    
        /* Set the buffer size and send it to decompressor  */
        nextRead -= _indexbuffer[_CurField].pad;
        dmBufferSetSize(buf, nextRead);

        /* Wait till there's room in _outpool before sending*/
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

	/* Send the field to be decompressed.	*/
        DC (dmICSend(_vice, buf, 0,(DMbuffer *)NULL) );
        dmBufferFree(buf);

        _CurField += 1;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    OC (lseek64 (_f.fd, _f.headerSize + _indexbuffer[_CurField].offset, SEEK_SET));
	    fprintf(stderr, "Done pass through file\n");
        }
   }
   _IamPlaying--;	/* Let PutField() know its time to exit	*/
   _exit(0); 
}


void initVL()
{
    /* Initialize the video stuff.  In particular, set up
     * a VL_MEM to VL_VIDEO path.  The VL_MEM node will be
     * attached to _outpool (see initCompression).  This allows
     * dmbuffers containing the uncompressed fields to be sent directly
     * to video out.
     */
    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }

    ERROR((devPtr == NULL), "No mvp device");

    /* Create mem to vid path.			*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set three data attributes on the memory node.    */
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);

    _transferSize = _f.uSize*_f.vSize*2;
}


void initCompression (void)
{
    /* This routine intializes the real time decompressor.  This includes
     * setting up _inpool and _outpool.  These are the two DMbufferpools used
     * to hold input/output from the decompressor.  _outpool is also attached
     * to the VL_MEM node _src.  This allows decompressed results to go directly
     * to video out.
     */
    DMparams* p;
    int numIC, i;

    /* Find real-time jpeg compressor and try to open it */
    numIC = dmICGetNum();
    dmParamsCreate(&p);

    for (i = 0; i < numIC; i++) {

        /* Get description of next converter */
        if (dmICGetDescription(i, p) != DM_SUCCESS) {
            continue;
        }

        /* Look for the right combination */
        if ((dmParamsGetInt(p, DM_IC_ID) == 'jpeg') &&
            (dmParamsGetEnum(p, DM_IC_CODE_DIRECTION) == DM_IC_CODE_DIRECTION_DECODE) &&
            (dmParamsGetEnum(p, DM_IC_SPEED) == DM_IC_SPEED_REALTIME)) {
	    break;
        }
    }
    dmParamsDestroy(p);

    if (dmICCreate(i, &_vice) != DM_SUCCESS) {
	fprintf(stderr,"Cannot open an appropriate real-time compressor\n");
	exit(1);
    }

    /* set the params for dmIC source */
    dmParamsCreate(&p);
    DC(dmSetImageDefaults(p,_f.uSize,_f.vSize,DM_IMAGE_PACKING_CbYCrY));
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetSrcParams(_vice, p) );

    /* set the params for dmIC destionation.  These are the same as for 
     * dmIC source except for DM_IMAGE_COMPRESSION 
     */
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_UNCOMPRESSED);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetDstParams(_vice, p) );
    dmParamsDestroy(p);

    /* Create buffer pool for compressor input */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,INBUFS,_transferSize,DM_FALSE,DM_TRUE) );
    DC( dmICGetSrcPoolParams(_vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_inpool) );
    dmParamsDestroy(p);

    _inpoolfd = dmBufferGetPoolFD(_inpool);
    DC (dmBufferSetPoolSelectSize(_inpool, _transferSize) );

    /* Create buffer pool for dmIC compressor output 
     * This pool will also be used to drive the VL_MEM node, allowing
     * video fields to go from the decompressor to video without any
     * additional copying.
     */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,OUTBUFS,_transferSize,DM_FALSE,DM_TRUE));
    DC( dmICGetDstPoolParams(_vice, p) );
    VC( vlDMPoolGetParams(_svr, _path, _src, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_outpool) );

    /* tell dmIC to use this for its output */
    DC( dmICSetDstPool(_vice, _outpool) );
    dmParamsDestroy(p);

    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _transferSize) );
} 
