/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vgrab_j = Grabs live video in and writes to disk as compressed jpeg files. 
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-n fieldsToCapture>     - Number of fields to capture, default = 200.
 *                                This arg should be an even number.
 *     <-qf qualityFactor>      - Number between 1-100 that governs the quality 
 *				  of the compression.  A higher number implies
 *				  higher quality.  Default is 75.
 *     fileName                 - File for results.
 *
 * This is a real simple program to capture fields of compressed video 
 * to disk in real time.  
 *
 * DMbufferpools provide the application with special purpose memory that
 * can be shared between the gfx, video, and compression devices.  A
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of two DMbufferpools:
 *      _inpool - This pool contains "BUFS" DMbuffers, used as input for the
 *                compressor.  Fields from video input are fed into this pool. 
 *     _outpool - This pool contains "BUFS" DMbuffers, used to hold
 *                output from the compressor.  Fields from this DMbufferpool are
 *		  written to disk.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() is called to
 *	  perform the real work. 
 *	- doit() sits in a tight loop checking constantly checking for
 *	  two conditions:
 *		1. Have any new fields arrived from my input?  If yes,
 *		   then send the field to the compressor.
 *		2. Has the compressor finished compressing a field sent 
 *		   to it in step 1?  If yes, then write the compressed
 *		   result to a file.
 *	- Once the desired number of fields have been compressed and
 *	  written to disk the program exits.
 *	- Each compressed field is padded out to "COMPRESSED_PAD" when 
 *	  written to disk - this facilitates direct I/O in other apps.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>

#include <vl/vl.h>
#include <vl/dev_mvp.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/jinfo.h"
#include "common/fileutils.h"

#define BUFS 12
#define MAX_FIELDS 50000
#define VL_COMPRESSION_JPEG 2


/* All global variables are preceded by an "_"	*/

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

int _transferSize;	        /* Video field size in bytes.	   */
DMimageconverter _vice;		/* Handle to compressor.	   */
DMbufferpool _inpool;		/* Input dmbuffer pool for compressor.    */
DMbufferpool _outpool;	        /* Output dmbuffer pool from compressor.  */
char *_padBuffer;
jFieldInfo_t *_indexBuffer;
fileHeader_t *_f;


/* Routine declarations */
void doit();
void initVL();
void initCompression();
int sginap (long );

main(int argc, char *argv[])
{
    int i;
    int n = 200;
    int qf = 75;
    char *Fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            n = atoi(argv[i]);
            if (n > MAX_FIELDS) n = MAX_FIELDS; 
        }
        else if (strcmp (argv[i], "-qf") == 0) {
           ++i;
           qf = atoi(argv[i]);
        }
	else {
	    Fname = argv[i];
        }
    }

    _f = fileCreate(Fname, FALSE);
    _f->packing = VL_COMPRESSION_JPEG;
    _f->compressed = TRUE;
    _f->pad = COMPRESSED_PAD;
    _f->Nfields = n;
    _f->qf = qf;
    _indexBuffer = createIndexBuffer(_f->Nfields);

    initVL();		/* Initialize video input */
    initCompression();	/* Initialize compressor and set up its DMbufferpools */
    _padBuffer   = (char * )memalign(_f->headerSize, COMPRESSED_PAD);
    memset(_padBuffer, 0xff, COMPRESSED_PAD);

    fprintf(stderr, "Capturing %d fields at %dx%d with quality = %d\n", _f->Nfields, _f->uSize, _f->vSize, _f->qf);
    doit();
}

void doit()
{
    /* The work routine.  It simply loops checking for the arrival of new video
     * fields (which are then sent to the decompressor) AND also checks for 
     * any compressed results coming back from the compressor (which are then 
     * written to disk).
     */

    VLEvent event;
    DMbuffer buf;
    USTMSCpair oldpair, newpair;
    int i, compressedDataSize;
    int fieldsCaptured = 0;
    int fieldsCompressed = 0;
    int queued = 0;
    int pad = 0;
    int tmp;
    off64_t  bytesCaptured = 0;
    char *dataPtr;

    setscheduling();  /* Raise priority if root or setuid */

    lseek64 (_f->fd, (off64_t)_f->headerSize, SEEK_SET);

    VC (vlSelectEvents(_svr, _path, VLTransferCompleteMask));

    VC (vlBeginTransfer(_svr, _path, 0, NULL)); 

    /* Simply loop, grabbing video fields into dmbuffers, sending 
     * the fields to the compressor,  and then writing the compressed
     * result to disk.
     */
    while (fieldsCompressed != _f->Nfields) { 

	/* Any video fields to queue up?  Make sure there is
         * a place to put them in the input to the compressor
         * by checking the variable "queued".  If I have all the
         * fields I need then don't send any more to the compressor.
         */ 

	if ((fieldsCaptured != _f->Nfields) && (queued < BUFS)) {

            /* Have any events arrived ???? 	*/
	    if (vlEventRecv(_svr, _path, &event) == DM_SUCCESS) {

		/* If it is a transfer event then we have another field .
		 * Use vlEventToDMBuffer to get the dmbuffer containing the
                 * new video field.
		 */
		if (event.reason == VLTransferComplete) {
		    VC( vlEventToDMBuffer(&event, &buf) );
		    
                    /* Check for a dropped field  */
		    newpair = dmBufferGetUSTMSCpair(buf);
		    if ( (fieldsCaptured != 0) && (newpair.msc != oldpair.msc + 1) ){
			fprintf(stderr, "Dropped an input %lld %lld \n", oldpair.msc, newpair.msc);
		    }
		    oldpair = newpair;

		    /* Send the new field to the compressor.	*/
    		    DC( dmICSend(_vice, buf, 0,(DMbuffer *)NULL) );
		    dmBufferFree(buf);
		    ++fieldsCaptured;
		    ++queued;

		    if (fieldsCaptured % 500 == 0) {
		        fprintf(stderr, ".... Captured %d fields\n", fieldsCaptured);
		    }
		}
	    } 
	}

	/* Any compressed fields to write???	*/

	if (dmICGetDstQueueFilled(_vice) != 0) {  /* Any output from the decompressor?*/

	    /* Get ptr to output data and write to file. */
	    DC( dmICReceive(_vice, &buf) );
	    compressedDataSize = dmBufferGetSize(buf);
	    dataPtr = (char *)dmBufferMapData(buf);

            /* Pad out field to "COMPRESSED_PAD" boundary with 0xff's        */
            tmp = compressedDataSize/COMPRESSED_PAD;
            if ((compressedDataSize % COMPRESSED_PAD) != 0) tmp++;
            pad = COMPRESSED_PAD * tmp - compressedDataSize;
    
            if (pad) {
                memcpy(dataPtr + compressedDataSize, _padBuffer, pad);
            }


            /* Inc. the compressed size by tha padded amount and write to disk. */
            compressedDataSize += pad;
	    OC (write(_f->fd, dataPtr, compressedDataSize));
	    dmBufferFree(buf);

            /* Update index table   */
            _indexBuffer[fieldsCompressed].offset = bytesCaptured;
            _indexBuffer[fieldsCompressed].pad = pad;

	    bytesCaptured += compressedDataSize;
	    ++fieldsCompressed;
	    queued--;
	}

        /* If there are onyl a few video fields buffered up (there is
         * room for 12) then take a break, you are working too hard.
         */ 
	if (vlGetFilledByNode(_svr, _path, _drn) < 2) {
	    sginap(1);   /* Nothing to do */
	}
    }

    /* Write the array of compressed field sizes at the end of the file.*/

    _indexBuffer[_f->Nfields].offset = bytesCaptured;
    _indexBuffer[_f->Nfields].pad = 0;

    _f->dataBytes = bytesCaptured;
    writeHeader(_f);
    writeIndexBuffer(_f, _indexBuffer);

    fsync(_f->fd);
    close(_f->fd);
    fprintf(stderr, "Done writing file\n");
}



void initVL()
{
    /* Initialize video for capture - set up a vid to mem path
     * and set the correct data formats on the memory node.
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }

    ERROR((devPtr == NULL), "No mvp device");

    /* Set up a capture _path from video-in to memory.	*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_VIDEO, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_MEM, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    /* Set attributes on the memory node.	*/
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _drn, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _drn, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC  (vlSetControl(_svr, _path, _drn, VL_CAP_TYPE, &val));

    /* Get the actual video field size 	*/
    vlGetControl(_svr, _path, _drn, VL_SIZE, &val);
    fprintf(stderr,"Video Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
    _f->uSize = val.xyVal.x;
    _f->vSize = val.xyVal.y;

    /* Make sure _f->vSize is a multiple of 8 - jpeg restriction. */
    _f->vSize = _f->vSize - (_f->vSize%8);

    _transferSize = vlGetTransferSize(_svr, _path);
}

void initCompression ( void)
{
    /* This routine initializes the compressor via the dmIC library.
     * It sets a variety of image parameters for the compressor.  In
     * addition, this routine  also sets up _inpool (where the live fields 
     * from video input go) and _outpool (where the compressed results go). 
     */
    DMparams* p;
    int numIC, i;
    float qf = (float)_f->qf;

    /* Find real-time jpeg compressor and try to open it */
    numIC = dmICGetNum();
    dmParamsCreate(&p);

    for (i = 0; i < numIC; i++) {

        /* Get description of next converter */
        if (dmICGetDescription(i, p) != DM_SUCCESS) {
            continue;
        }

        /* Look for the right combination */
        if ((dmParamsGetInt(p, DM_IC_ID) == 'jpeg') &&
            (dmParamsGetEnum(p, DM_IC_CODE_DIRECTION) == DM_IC_CODE_DIRECTION_ENCODE) &&
            (dmParamsGetEnum(p, DM_IC_SPEED) == DM_IC_SPEED_REALTIME)) {
	    break;
        }
    }
    dmParamsDestroy(p);

    if (dmICCreate(i, &_vice) != DM_SUCCESS) {
	fprintf(stderr,"Cannot open an appropriate real-time compressor\n");
	exit(1);
    }

    /* Set the params for dmIC source - the input in this case is
     * a live video input stream in YUV 4:2:2 format
     */
    dmParamsCreate(&p);
    DC(dmSetImageDefaults(p,_f->uSize,_f->vSize,DM_IMAGE_PACKING_CbYCrY));
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_UNCOMPRESSED);
    DC( dmICSetSrcParams(_vice, p) );

    /* Set the params for dmIC destination.  These have the same format
     * as the input except that the data is now compressed. 
     */

    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
    DC( dmICSetDstParams(_vice, p) );
    dmParamsDestroy(p);

    /*
     * Set the conversion parameters (q-factor)
     */
    dmParamsCreate(&p);
    dmParamsSetFloat(p, DM_IMAGE_QUALITY_SPATIAL, qf*DM_IMAGE_QUALITY_MAX/100.0);
    DC( dmICSetConvParams(_vice, p) );
    dmParamsDestroy(p);

    /* Create a dmbuffer pool for input to the dmIC.  This pool
     * is attached to the memory node of the vl capture path. 
     * So live video is loaded into this DMbufferpool (_inpool) and
     * sent directly to the compressor with no additional copying. 
     */

    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,BUFS,_transferSize,DM_FALSE,DM_TRUE) );
    DC( dmICGetSrcPoolParams(_vice, p) );		/* Load compressor buffering requirements */
    VC( vlDMPoolGetParams(_svr, _path, _drn, p) );	/* Load video buffering requirements */
    DC( dmBufferCreatePool(p, &_inpool) );
    VC( vlDMPoolRegister(_svr, _path, _drn, _inpool) ); /* Attach pool to vl memory node.	*/
    dmParamsDestroy(p);

    /* Create a dmbuffer pool for dmIC compressor output */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,BUFS,_transferSize,DM_FALSE,DM_TRUE));
    DC( dmICGetDstPoolParams(_vice, p) );
    DC( dmBufferCreatePool(p, &_outpool) );

    /* tell dmIC to use this for its output */
    DC( dmICSetDstPool(_vice, _outpool) );
    dmParamsDestroy(p);
} 
