/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_2stream_j = Plays two jpeg compressed video files from disk.  
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     file1                 - first input stream (file created with vgrab_j)
 *     file2                 - second input stream (file created with vgrab_j)
 *
 *
 * This program displays two streams of compressed video from disk to 
 * video-out.  The second stream is displayed as background, while the first
 * stream is displayed as a small PIP window in the lower left corner of the
 * background stream.  In order to keep from dropping fields, only every other
 * field from each file is used.  The combined result is sent to video out for
 * two field times (i.e. its 30 frames/sec. motion so things look a little 
 * choppy but play at the correct speed). 
 *
 * Depending on compression quality and other factors, this program may
 * have a hard time keeping up when both files are on the same disk.
 * 
 * DMbufferpools provide the application with special purpose memory that 
 * can be shared between the gfx, video, and compression devices.  A  
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of four DMbufferpools:
 *	_inpool - This pool contains "INBUFS" DMbuffers, used as input for the 
 *		  decompressor. Compressed fields are read into this pool.
 *     _outpool - This pool contains "OUTBUFS" DMbuffers, used to hold
 *		  output from the decompressor (YUV 422 video fields). 
 *     _vdmpool - This pool contains "GFXBUFS" DMbuffers, used as the 
 *		  the color buffers (i.e. frame buffers) for the pbuffer 
 *		  _vdmpbuffer.   These DMbuffers can also be passed directly 
 *		  to video-out (via vlDMBufferSend). 
 *     _texpool - This pool contains 1 DMbuffer, used as the color buffer
 *		  for the pbuffer _texpbuffer. 
 *
 * The pbuffers used by this program (_vdmpbuffer and _texpbuffer) use memory  
 * from a DMbufferpool for their color buffers.  In order to do this, these
 * pbuffers must be created as digital media pbuffers (passed as an attribute
 * to glXCreateGLXPbufferSGIX).  
 *
 * Being a digital media pbuffer allows _vdmpbuffer to cycle through the 
 * DMbuffers in _vdmpool, rendering into each of them.  These dmbuffers can 
 * then be sent to video out.  _vdmpbuffer uses a particular DMbuffer from
 * _vdmpool via a call to glXAssociateDMPbufferSGIX.  
 *
 * In addition, the fact that _texpbuffer is also a digital media pbuffer 
 * allows glCopyTexSubImage2DEXT to copy pixels from _texpbuffer to texture 
 * memory via a "copy by reference".  In this case, no real pixel copying 
 * occurs and _texpbuffer is really being used as the texure memory for 
 * _vdmpbuffer. 
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *	- After initializing everything, the routine doit() sprocs off
 *	  a reader thread (see ReadIt()) and a display thread (see PutField()).
 *	- The main thread then just hangs out.
 *
 *    Reader Thread: 
 *	- ReadIt() simply cycles through the the two files.
 *	- First, it reads two fields from the first file (and sends only the
 *	  second field to be decompressed), then it read two fields from the
 *	  the second file (again using only the second field).
 *	- So the input to the decompressor alternates between fields from the 
 *	  two *	  input files.
 *
 *    Display Thread:
 *	- PutField() uses select() to wait for valid output to arrive from the 
 *	  decompressor.  Valid output arrives in the form of a DMbuffer from 
 *	  _outpool. 
 *	- During each iteration of its play loop, PutField waits for a pair of
 *	  fields to arrive from the decompressor.  The first field (call it f1)
 *	  is always from the first input stream while the second field (call 
 *	  it f2) is always from the second input stream.
 *	- PutField() grabs the next available DMbuffer from _vdmpool (denote 
 *	  this nextDMbuf). 
 *	- glDrawPixels is used to draw f1 into the DMbuffer (_texbuf) associated
 *	  with _texpbuffer (so this glDrawPixels call color converts f1 to RGB 
 *	  and loads it into texture memory). 
 *	- f2 is drawn as background to nextDMbuf using glDrawPixels.
 *	- glXMakeCurrentReadSGI is then used to set up _texpbuffer as the read
 *        source for _vdmpbuffer (the color buffer for _vdmpbuffer is still 
 *	  nextDMbuf).
 *	- glCopyTexSubImage2DEXT is called.  This does a "copy-by-reference".  
 *	  The net effect is that _texpbuffer is now functioning as texture 
 *	  memory for _vdmpbuffer.
 *	- A texture mapped quad (where f1 is the texture) is drawn on top of 
 *	  the background (f2).  All this rendering is happening with _vdmpbuffer
 *	  and its associated DMbuffer, nextDMbuf.     
 *	- nextDMbuf is then sent twice to video-out via two successive calls 
 *	  to vlDMBufferSend(). 
 *	- PutField then loops back and waits for the next available output 
 *	  from the decompressor.
 *	
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <GL/gl.h>
#include <GL/glx.h>
#include <vl/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define INBUFS  12 
#define OUTBUFS 12 
#define GFXBUFS 12 
#define PRELOAD 8
#define FILL_CHECK 60
#define FILL_HIGH 9
#define FILL_LOW 4

#define VL_COMPRESSION_JPEG 2

/* All global variables in this file are preceded by an "_".  */

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

/* File related defines  */

int _Nfields;		/* Number of fields in the file.	*/
int _uSize;	        /* Width of field to decompress.	*/
int _vSize;		/* Heigth of field to decompress 	*/
int _CurField;	        /* Current field being read.	 	*/
jFieldInfo_t *_indexbuffer1; /* Array of compressed field sizes	*/
jFieldInfo_t *_indexbuffer2; /* Array of compressed field sizes	*/
fileHeader_t _f1;
fileHeader_t _f2;

/* Decompressor defines.  */

DMimageconverter _vice;	/* Handle to image decompressor.				*/
DMbufferpool   _inpool;	/* dmbuffer pool used as input to decompressor 			*/
DMbufferpool  _outpool;	/* dmbuffer pool used as output from decompressor 		*/
int          _inpoolfd;	/* fd to select on input to decompressor.	  		*/
int         _outpoolfd;	/* fd to select on _outpool.			  		*/
int              _icfd;	/* fd to select on output available fro decompressor 		*/


/* Defines for digital media pbuffer that drives fields to video out */ 

GLXPbufferSGIX _vdmpbuffer; /* dm-pbuffer that drives video. Associated with _vdmpool	*/
DMparams       *_vdmParams; /* dmParams associated with _vdmpbuffer.			*/
GLXContext     _vdmcontext; /* gfx context associated with _vdmpbuffer.			*/
DMbufferpool      _vdmpool; /* dmbuffer pool used as color buffer memory for _vdmpbuffer*/ 
		            /* These buffers are also sent to video out (vlDMBufferSend)*/	
int		_vdmpoolfd; /* fd to select on _vdmpool.				*/

/* Defines for digital media pbuffer that loads texture memory 	     */ 

GLXPbufferSGIX _texpbuffer; /* dm-pbuffer that serves as texture memory for _vdmpbuffer */ 
DMparams       *_texParams; /* dmParams associated with _texpbuffer.                    */
GLXContext    _textcontext; /* gfx context associated with _texpbuffer.			*/
DMbufferpool      _texpool; /* dmbuffer pool used as color buffer memory for _texpbuffer*/ 
DMbuffer           _texbuf; /* The one and only dmbuffer allocated from _texpool.	*/
int 	        _tex_width;
int	       _tex_height;

/* Misc defines. */
Display *_dpy;

/* Routine declarations. */
void initGL();
void doit();
void initVL();
void initDM4GLVL();
void initCompression();
void initDM4Tex();
void PutField();
void ReadIt();
void setscheduling();
void createTexture(unsigned int *, int, int, int, int);
GLXFBConfigSGIX getFBConfig(Display *, int);



main(int argc, char *argv[])
{
    int n;

    /* Get first file information.    */
    fileInit(argv[1], &_f1, FALSE);
    _indexbuffer1 = createIndexBuffer(_f1.Nfields);
    readIndexBuffer(&_f1, _indexbuffer1);
    ERROR ((_f1.packing != VL_COMPRESSION_JPEG), "File not a jpeg file");

    /* Get first file information.    */
    fileInit(argv[2], &_f2, FALSE);
    _indexbuffer2 = createIndexBuffer(_f2.Nfields);
    readIndexBuffer(&_f2, _indexbuffer2);
    ERROR ((_f2.packing != VL_COMPRESSION_JPEG), "File not a jpeg file");


    ERROR((_f1.uSize != _f2.uSize), "Dimensions (width) not equal");
    ERROR((_f1.vSize != _f2.vSize), "Dimensions (height) not equal");

    _uSize = _f1.uSize;
    _vSize = _f1.vSize;
    _Nfields = (_f1.Nfields < _f2.Nfields) ? _f1.Nfields : _f2.Nfields; 

    /* Set tetxure width and height based on field sizes from file.  */
    _tex_width = 1024;   		        /* Field width is 640, 768, or 720  */
    _tex_height = (_vSize == 240) ? 256 : 512;  /* 240 or 288			  */

    initVL();		/* Create mem to vid path. */
    initCompression();	/* Set up decompressor, including _inpool and _outpool */
    initDM4GLVL();	/* Set up _vdmpool	*/
    initDM4Tex();	/* Set up _texpool	*/
    initGL();		/* Create _vdmpbuffer, _texpbuffer and init. their contexts */

    fprintf(stderr, "Playing  %d fields of %dx%d \n", _Nfields, _uSize, _vSize);
    fprintf(stderr, "File 1 quality factor = %d File 2 quality factor = %d\n", _f1.qf, _f2.qf);

    doit();		/* sproc ReadIt() and sproc PutField()	*/
}

void initGL()
{
    /* This routine creates the two digital media pbuffers used by this
     * program (_vdmpbuffer and _texpbuffer).  
     */

    unsigned int texA;
    GLXFBConfigSGIX config;
    DMbuffer gfxbuf;
    int attrib[] = {GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE,
                    GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};

    _dpy = XOpenDisplay( ":0" ); 

    /* Get a 24 bit frame buffer configuration */
    config = getFBConfig(_dpy, 24);
    ERROR((config == NULL), "Can't get FB");

    /* Use this fb config. to set up _vdmpbuffer	*/ 
    GNC (_vdmpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _uSize, _vSize, attrib));
    GNC (_vdmcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));

    /* Allocate a color buffer from _vdmpool and associate it with _vdmpbuffer. 
     * There must be an associated DMBuffer in order to init the pbuffer context.
     */
    DC ( dmBufferAllocate(_vdmpool, &gfxbuf) );
    GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));

    /* Initialize context for _vdmpbuffer.  */
    GC (glXMakeCurrent(_dpy, _vdmpbuffer, _vdmcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_uSize,_vSize);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glPixelZoom (1.0, -1.0);     /* Need to flip the video when drawn as background */
    glRasterPos2f (-1.0, 1.0);

    /* Set up texture parameters. */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    createTexture(&texA, _tex_width, _tex_height, GL_RGBA, 0);
    BuildGfxLists(1, _uSize, _vSize, _tex_width, _tex_height,0);

    dmBufferFree(gfxbuf);  /* Done with buffer for now.	*/

    /* Now use config to set up _texpbuffer */ 
    GNC (_texpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _tex_width, _tex_height, attrib));
    GNC (_textcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    DC ( dmBufferAllocate(_texpool, &_texbuf) );

    /* Since there is only one DMbuffer required for texture memory, this
     * association is made only once in the beginning.
     */
    GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));

    /* Initialize context for _texpbuffer.	*/

    GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_tex_width, _tex_height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_tex_width, 0.0, (double)_tex_height, -1.0, 1.0);
    glRasterPos2i (0, 0);
}

void doit()
{
    /* Routne which sprocs the read and display thread. */

    int readPID, putPID;

    _CurField = 0;
    /* Surrender the gfx context.	*/
    GC (glXMakeCurrent(_dpy, None, NULL));

    /* Sproc off the read and display threads.	*/
    readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
    putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);

    while (1) sginap(100);  /*^C to exit	*/
}

void PutField()
{
    /* This routine is sproc'd off by doit.  It waits for a decompressed
     * field from the first stream and draws it as background. Then it waits
     * for a decompressed field from the second stream and draws it as a texture
     * mapped quad on top of the background.
     */
    fd_set fds;
    int i;
    int count = 0;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    double ust_per_msc;
    DMbuffer buf;
    DMbuffer gfxbuf;
    unsigned char *dataPtr;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    /* raise priority if root or setuid is set.	*/
    setscheduling();

    ust_per_msc = vlGetUSTPerMSC(_svr, _path, _src);

    /* The display loop.	*/
    while (1) {

        /* Select on _icfd, it returns when there is valid output
         * from the decompressor.
         */
	FD_ZERO(&fds);
	FD_SET(_icfd, &fds);
	OC (select(_icfd+1, &fds, NULL, NULL, NULL));

	/* Get ptr to output data  */
	DC( dmICReceive(_vice, &buf) );
        dataPtr = dmBufferMapData(buf);

        /* Load first stream into the texture pbuffer */
        GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));
        GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext));
        glDrawPixels(_uSize, _vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);
        dmBufferFree(buf);

        /* Get the next available DMbuffer from _vdmpool.	*/ 
	FD_ZERO(&fds);
	FD_SET(_vdmpoolfd, &fds);
        OC (select(_vdmpoolfd+1, NULL, &fds, NULL, NULL));
        DC (dmBufferAllocate(_vdmpool, &gfxbuf));

        /* Associate this new DMbuffer with _vdmpbuffer (i.e. it is now its
         * color buffer).  Then set up _texpbuffer as the read source for
         * the texture copy.
         */ 
        GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));
	GC (glXMakeCurrentReadSGI(_dpy, _vdmpbuffer, _texpbuffer, _vdmcontext));

        /* Wait for second stream and get a ptr to YUV data */
        FD_ZERO(&fds);
        FD_SET(_icfd, &fds);
        OC (select(_icfd+1, &fds, NULL, NULL, NULL));
        DC( dmICReceive(_vice, &buf) );
        dataPtr = dmBufferMapData(buf);

        /* Draw the second stream as background.		*/ 
        glDisable(GL_TEXTURE_2D); 
	glClear(GL_COLOR_BUFFER_BIT);
        glDrawPixels(_uSize, _vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

        /* Now call glCopyTexSubImage2DEXT to load the field from _texpbuffer
         * into texture memory.  This is actually a copy by reference, the net
         * effect being that _texpuffer (and its associated _texbuf) serves as
         * texture memory for _vdmpbuffer.
         */
        glEnable(GL_TEXTURE_2D); /* enable texturing */
        glCopyTexSubImage2DEXT( GL_TEXTURE_2D, 0, 0, 0, 0, 0, _tex_width, _tex_height);

        /* Draw the first stream as a texture mapped quad on top of the second stream */ 
        glPushMatrix();
        glScalef(0.4, 0.4, 0.4);
        glTranslatef(-1.4, -1.4, 0.0);
        glCallList(1);
        glPopMatrix();

        /* see if we are dropping any fields.  Check every eigth pass
         * through the loop. 
         */
	if (count % 8 == 0) {
            newmsc = vlGetFrontierMSC(_svr, _path, _src);
            if ((oldmsc != -1) && (newmsc != oldmsc + 8) && (count > FILL_CHECK)) {
                fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+8, newmsc);
            }
            oldmsc = newmsc;
	}

	/* Send the result to video out twice.  This must be done since we
         * only have the horsepower to decompress every other field from 
         * each stream.
         */ 
	VC ( vlDMBufferSend(_svr, _path, gfxbuf) );
	VC ( vlDMBufferSend(_svr, _path, gfxbuf) );
        dmBufferFree(buf);
        dmBufferFree(gfxbuf);
        count += 2;

	/* Don't start the transfer until we have some fields queued up. 	*/
        if (count == PRELOAD) {
            VC (vlBeginTransfer(_svr, _path, 0, NULL));
        }
#if 0
        /* Try to tune display process. */
        else if (count > FILL_CHECK) {
            i = vlGetFilledByNode(_svr, _path, _src);
            if (i > FILL_HIGH) sginap(1);
            if (i < FILL_LOW ) fprintf(stderr, "Filled down to %d\n", i);
        }
#endif

   } 
}

void ReadIt()
{
    /* This reader thread is sproc'd by doit.  It reads two fields
     * from the first stream, sending ONLY the second field to the decompressor.
     * It then reads two fields from the second streams, again sending only
     * the second to the decompressor.  So the decompressor alternates between
     * fields from the two streams.
     */
    int fd;
    int nextRead;
    int dataSize;
    int stream1 = 1;
    jFieldInfo_t *indexbuffer;
    unsigned char *dataPtr;
    fd_set fds;
    DMbuffer buf;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    lseek64 (_f1.fd, (off64_t)_f1.headerSize, SEEK_SET);
    lseek64 (_f2.fd, (off64_t)_f2.headerSize, SEEK_SET);
    setscheduling();

    while (1) {
        /* Wait for an input buffer to be available */ 
        FD_ZERO(&fds);
        FD_SET(_inpoolfd, &fds);
        OC (select(_inpoolfd+1, NULL, &fds, NULL, NULL)); 

        DC(dmBufferAllocate(_inpool, &buf));
        dataPtr = dmBufferMapData(buf);
    
        if (stream1) {
	    fd = _f1.fd;
	    indexbuffer = _indexbuffer1;
        }
        else {
	    fd = _f2.fd;
	    indexbuffer = _indexbuffer2;
        }

        /* Read two fields, but just send the second one of the pair.	*/ 
        nextRead = indexbuffer[_CurField+1].offset - indexbuffer[_CurField].offset;
        OC (read(fd, dataPtr, nextRead)); 

        nextRead = indexbuffer[_CurField+2].offset - indexbuffer[_CurField+1].offset;
        OC (read(fd, dataPtr, nextRead)); 

        /* Set the buffer size and send it to decompressor  */
	nextRead -= indexbuffer[_CurField+1].pad;
        dmBufferSetSize(buf, nextRead);

        /* Wait till there's room in _outpool */
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL)); 

        DC (dmICSend(_vice, buf, 0,(DMbuffer *)NULL) );
        dmBufferFree(buf);

        /* If we just finished reading from the second stream, then
         * move both streams onto the next frame.
         */
        if (stream1 == 0) {
            _CurField += 2;
            if (_CurField >= _Nfields) {
	        _CurField = 0;
	        lseek64 (_f1.fd, (off64_t)_f1.headerSize, SEEK_SET);
	        lseek64 (_f2.fd, (off64_t)_f2.headerSize, SEEK_SET);
		fprintf(stderr, "Done pass through files\n");
            }
	}
	stream1 = !stream1; /*Alternate between the two streams.	*/
   }
}

void initVL()
{
    /* Initialize the video stuff.  In particular, set
     * up a mem-> vid path.  The memory node for this play
     * path will be attached to the digital media pbuffers
     * where the textured video is being drawn.  Results can
     * then be sent directly from the pbuffer to video out.
     * See initDM4GLVL() for more details on this.
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR ((devPtr == NULL), "No mvp device");

    /* Create a mem to vid node.	*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE)); 

    SetVideoOutTiming(_svr, _path, VL_ANY, _uSize, _vSize);

    /* Set attributes on mem node.  Its rgb since it is going from gfx.	*/
    val.intVal = VL_FORMAT_RGB;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val)); 

    val.intVal = VL_PACKING_ABGR_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    /* This is new!  Why can't the video library figure this out */
    val.intVal = VL_LAYOUT_GRAPHICS;
    VC (vlSetControl(_svr, _path, _src, VL_LAYOUT, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
}


void initCompression ( void)
{
    /* This routine sets up the decomprssor - including its input
     * and output pools.
     */
    DMparams* p;
    int numIC, i;

    /* Find real-time jpeg compressor and try to open it */
    numIC = dmICGetNum();
    dmParamsCreate(&p);

    for (i = 0; i < numIC; i++) {

        /* Get description of next converter */
        if (dmICGetDescription(i, p) != DM_SUCCESS) {
            continue;
        }

        /* Look for the right combination */
        if ((dmParamsGetInt(p, DM_IC_ID) == 'jpeg') &&
            (dmParamsGetEnum(p, DM_IC_CODE_DIRECTION) == DM_IC_CODE_DIRECTION_DECODE) &&
            (dmParamsGetEnum(p, DM_IC_SPEED) == DM_IC_SPEED_REALTIME)) {
	    break;
        }
    }
    dmParamsDestroy(p);

    if (dmICCreate(i, &_vice) != DM_SUCCESS) {
	fprintf(stderr,"Cannot open an appropriate real-time compressor\n");
	exit(1);
    }

    /* set the params for dmIC source */
    dmParamsCreate(&p);
    DC(dmSetImageDefaults(p,_uSize,_vSize,DM_IMAGE_PACKING_CbYCrY));
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetSrcParams(_vice, p) );

    /* set the params for dmIC destionation.  These are the same as for 
     * dmIC source except for DM_IMAGE_COMPRESSION 
     */
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_UNCOMPRESSED);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetDstParams(_vice, p) );
    dmParamsDestroy(p);

    /* Create buffer pool for compressor input */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,INBUFS,_uSize*_vSize*2,DM_FALSE,DM_TRUE) );
    DC( dmICGetSrcPoolParams(_vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_inpool) );
    dmParamsDestroy(p);

    _inpoolfd = dmBufferGetPoolFD(_inpool);
    DC (dmBufferSetPoolSelectSize(_inpool, _uSize*_vSize*2) );

    /* Create buffer pool for dmIC compressor output */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,OUTBUFS,_uSize*_vSize*2,DM_FALSE,DM_TRUE));
    DC( dmICGetDstPoolParams(_vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_outpool) );

    /* tell dmIC to use this for its output */
    DC( dmICSetDstPool(_vice, _outpool) );
    dmParamsDestroy(p);

    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _uSize*_vSize*2) );
    _icfd = dmICGetDstQueueFD(_vice);
} 

void initDM4GLVL()
{
    /* This routine creates _vdmpool, the DMbufferpool that will
     * provide the color buffers (i.e. frame buffer memory) for the 
     * dm-pbuffer _vdmpbuffer.  This buffer pool is also attached to 
     * the vl memory node (_src) that is driving video out.  This allows
     * the application to render into DMbuffers from _vdmpool using OpenGL
     * AND to send the very same DMbuffers to video out (with no additional
     * copying.
     */
    DMparams *p;

    /* Set up image attributes for the buffer pool	*/
    dmParamsCreate(&_vdmParams);
    DC ( dmSetImageDefaults(_vdmParams, _uSize, _vSize, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_vdmParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_vdmParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Create parameter list p and load it with default pool information.	*/
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, GFXBUFS, _uSize*_vSize*4, DM_FALSE, DM_TRUE) );
  
    /* Now load p with specific image (GL) information from _vdmParams.	*/
    DC ( dmBufferGetGLPoolParams(_vdmParams, p) );

    /* Now load p with the required information to attache video to the buffer pool. */
    VC ( vlDMPoolGetParams(_svr, _path, _src, p) );

    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    
    /* Create the buffer pool based on all the info. loaded into p.	*/
    DC ( dmBufferCreatePool(p, &_vdmpool) );
    dmParamsDestroy(p);

    /* Get the fd we can use to select on the next available dmbuffer from
     * _vdmpool.  The set the select size for the pool as well.
     */
    _vdmpoolfd = dmBufferGetPoolFD(_vdmpool);
    DC (dmBufferSetPoolSelectSize(_vdmpool, _uSize*_vSize*4) );
}

void initDM4Tex()
{
    /* This routine creates the DMbufferpool that provides the DMBuffer
     * used as frame buffer memory for the dm-pbuffer _texpbuffer.  Only
     * one DMbuffer is alloacted in this pool (it is assigned to _texbuf inside
     * initGL().  _texbuf will end up serving as texture memory for _vdmpbuffer.
     */ 

    DMparams *p;

    /* Set up geometry attributes fro pool.	*/
    dmParamsCreate(&_texParams);
    DC ( dmSetImageDefaults(_texParams, _tex_width, _tex_height, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_texParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_texParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Load p with default info. + geometry info from _texparams.	*/
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, 1, _tex_width*_tex_height*4, DM_FALSE, DM_TRUE) );
    DC ( dmBufferGetGLPoolParams(_texParams, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);

    /* Create the buffer pool.	*/
    DC ( dmBufferCreatePool(p, &_texpool) );
    dmParamsDestroy(p);
}
