/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_j = Plays jpeg compressed video fields from disk to video out. 
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     fileName                 - File created with vgrab_j 
 *
 * DMbufferpools provide the application with special purpose memory that
 * can be shared between the gfx, video, and compression devices.  A
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of two DMbufferpools:
 *      _inpool - This pool contains "INBUFS" DMbuffers, used as input for the
 *                decompressor.  Compressed fields are read from disk directly 
 *		  into this pool. 
 *     _outpool - This pool contains "OUTBUFS" DMbuffers, used to hold
 *                output from the decompressor.  Fields from this DMbufferpool 
 *                then sent to video out. 
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *	- After doing initialization, the routine doit() is entered.
 *	- This routine sprocs off a reader thread (see ReadIt()).
 * 	- The main process then simply waits for output to become 
 * 	  available from the decompressor (via select) and sends it to video.
 *
 *    Reader Thread:
 *	- This thread reads compressed fields from disl into dmbuffers allocated
 *	  from _inpool.
 *	- After being read from disk, the fields are fed to the  decompressor.
 *	- The main thread (see doit()) then waits for output from the 
 *	  decompressor.
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <dmedia/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define INBUFS  12 
#define OUTBUFS 12
#define PRELOAD 4 
#define FILL_CHECK 60
#define VL_COMPRESSION_JPEG 2

/* All global variables are preceded by an _	*/

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

/* File related defines	*/
int _transferSize;		/* Bytes in decompressed field.		*/ 
int _CurField = 0;		/* Current field beong read off disk.	*/
jFieldInfo_t *_indexbuffer;	/* Array containing compressed filed sizes in bytes	*/
fileHeader_t _f;

/* dmIC related defines */
DMimageconverter _vice;		/* Handle to decompressor		*/
DMbufferpool     _inpool;	/* dmbuffer pool used as input to decompressor */
DMbufferpool     _outpool;	/* dmbuffer pool used as output from decompressor 
                                 * AND as input to VL_MEM node 		*/
int              _inpoolfd;	/* fd used to select on _inpool		*/
int              _outpoolfd;	/* fd used to select on _outpool	*/

/* Routine declarations.	*/
void doit();
void ReadIt();
void initVL();
void initCompression();

main(int argc, char *argv[])
{
    /* Get file information.    */
    fileInit(argv[1], &_f, FALSE);
    _indexbuffer = createIndexBuffer(_f.Nfields);
    readIndexBuffer(&_f, _indexbuffer);

    ERROR ((_f.packing != VL_COMPRESSION_JPEG), "File not a jpeg file");

    fprintf(stderr, "Playing %d fields with dimensions %dx%d with quality = %d\n",_f.Nfields,_f.uSize,_f.vSize,_f.qf);

    initVL();		/* Init video - set up up mem->vid path. */
    initCompression();  /* Set up _inpool and _outpool		 */
    doit();		/* Let it go.				 */
}

void doit( )
{
    /* This routine sprocs off the reader therad.  Then it just
     * loops waiting for decompressed fields to arrive from the
     * decompressor.  The uncompressed fields are returned in dmbuffers
     * from _outpool.  The buffers are then sent to video out.
     */
    int count = 0;
    int icfd;
    DMbuffer buf;
    fd_set fds;
    stamp_t oldmsc = -1;
    stamp_t newmsc;

    /* Get the fd for the destination queue of the decompressor.
     * Selecting on this fd allows the process to wait for available
     * output from the decompressor.
     */
    icfd = dmICGetDstQueueFD(_vice);

    setscheduling();	/* Raise priority if root or setuid */

    /* Start the reader thread.			*/
    sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);

    /* Now loop forever, getting output from the decompressor and
     * sending it to video out.
     */
    while (1) {
        /* Wait for output from decompressor */ 
        FD_ZERO(&fds);
        FD_SET(icfd, &fds);
        OC (select(icfd+1, &fds, NULL, NULL, NULL));

        /* Get the dmbuffer assoiated with the output */
        DC( dmICReceive(_vice, &buf) );
    
	/* Check for dropped fields - no need to check every time.	*/
	if (count % 4 == 0) {
            newmsc = vlGetFrontierMSC(_svr, _path, _src);
            if ((oldmsc != -1) && (newmsc != oldmsc + 4) && (count > FILL_CHECK)) {
		fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+4, newmsc);
            }
            oldmsc = newmsc;
	}

	/* Send to dmbuffer to video out */
        VC ( vlDMBufferSend(_svr, _path, buf) );
        dmBufferFree(buf);
	++count;

	/* get a few fields queued up before the transfer starts.*/
	if (count == PRELOAD) {
	    VC (vlBeginTransfer(_svr, _path, 0, NULL)); 
	}
    }
}

void ReadIt()
{
    /* This is the sproc'd routine that reads compressed fields off of 
     * disk into dmbuffers allocated from _inpool.  The fields are then 
     * sent to the decompressor.
     */

    fd_set fds;
    int nextRead;
    DMbuffer buf;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);

    while (1) {

        /* Wait for an input dmbuffers to be available in _inpool */ 
        FD_ZERO(&fds);
        FD_SET(_inpoolfd, &fds);
        OC (select(_inpoolfd+1, NULL, &fds, NULL, NULL));

        /* Read the fields into the input dmbuffers. */
        nextRead = _indexbuffer[_CurField+1].offset - _indexbuffer[_CurField].offset;
    	DC (dmBufferAllocate(_inpool, &buf));
        OC (read(_f.fd, dmBufferMapData(buf), nextRead));

        /* Set the buffer size and send it to decompressor  */
        nextRead -= _indexbuffer[_CurField].pad;
        dmBufferSetSize(buf, nextRead);

        /* Before actuallyu sending the buffer, make sure there is
         * a place to store the result in the _outpool.
         */ 
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

        DC (dmICSend(_vice, buf, 0,(DMbuffer *)NULL) );
        dmBufferFree(buf);

	/* Check for file wrap.		*/
        _CurField++;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass through file\n");
        }
    }
}


void initVL()
{
    /* Init. all the video.  In particualr, set up a memory to video
     * out path.  The memory node (_src) is attached to the DMbufferpool
     * _outpool in the routine initCompression().
     */
    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC(_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }

    ERROR((devPtr == NULL), "No mvp device");

    /* Create a mem to vid-out _path.	*/
    VC(_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC(_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC(_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC(vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set data attributes on the memory node 	*/
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Video Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);

    _transferSize = _f.uSize*_f.vSize*2;
    fprintf(stderr, "Transfer Size (bytes) = %d \n",_transferSize);
}

void initCompression ( void)
{
    /* This routine set up the real-time jpeg decomptressor via 
     * the dmIC library.  It creates the input (_inpool) and
     * output (_outpool) DMbufferpools used by the decompressor.
     * The output pool is attached to the memory node (_src) of
     * the mem->vid path.  This allows dmbuffers containing decompressed
     * fields to be sent directly to video out.
     */ 
    DMparams* p;
    int numIC, i;

    /* Find real-time jpeg compressor and try to open it */
    numIC = dmICGetNum();
    dmParamsCreate(&p);

    for (i = 0; i < numIC; i++) {

        /* Get description of next converter */
        if (dmICGetDescription(i, p) != DM_SUCCESS) {
            continue;
        }

        /* Look for the right combination */
        if ((dmParamsGetInt(p, DM_IC_ID) == 'jpeg') &&
            (dmParamsGetEnum(p, DM_IC_CODE_DIRECTION) == DM_IC_CODE_DIRECTION_DECODE) &&
            (dmParamsGetEnum(p, DM_IC_SPEED) == DM_IC_SPEED_REALTIME)) {
	    break;
        }
    }
    dmParamsDestroy(p);

    if (dmICCreate(i, &_vice) != DM_SUCCESS) {
	fprintf(stderr,"Cannot open an appropriate real-time compressor\n");
	exit(1);
    }

    /* set the params for dmIC source */
    dmParamsCreate(&p);
    DC(dmSetImageDefaults(p,_f.uSize,_f.vSize,DM_IMAGE_PACKING_CbYCrY));
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetSrcParams(_vice, p) );

    /* set the params for dmIC destionation.  These are the same as for 
     * dmIC source except for DM_IMAGE_COMPRESSION 
     */
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_UNCOMPRESSED);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetDstParams(_vice, p) );
    dmParamsDestroy(p);

    /* Create buffer pool for compressor input */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,INBUFS,_transferSize,DM_FALSE,DM_TRUE) );
    DC( dmICGetSrcPoolParams(_vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_inpool) );
    dmParamsDestroy(p);

    _inpoolfd = dmBufferGetPoolFD(_inpool);
    DC (dmBufferSetPoolSelectSize(_inpool, _transferSize) );

    /* Create buffer pool for dmIC compressor output.  The important
     * thing to note is that this pool will be shared with the memory
     * node of the vl mem->vid _path.  So, decompressed fields can be
     * sent directly to video out with no additional copying.
     */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,OUTBUFS,_transferSize,DM_FALSE,DM_TRUE));
    DC( dmICGetDstPoolParams(_vice, p) );	/* Get decompressor buffering requirements */
    VC( vlDMPoolGetParams(_svr, _path, _src, p) ); /* Get video buffering requirements	   */
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_outpool) );

    /* tell dmIC to use this for its output */
    DC( dmICSetDstPool(_vice, _outpool) );
    dmParamsDestroy(p);

    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _transferSize) );

} 
