/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_stack_j = Plays jpeg compressed video fields from disk to video out.
 *		   The field is sent through gfx, stacked on top of itself 
 *                 (oh, what a cool effect) and sent to video out.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     fileName                 - File created with vgrab_j
 *
 * DMbufferpools provide the application with special purpose memory that 
 * can be shared between the gfx, video, and compression devices.  A  
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of four DMbufferpools:
 *	_inpool - This pool contains "INBUFS" DMbuffers, used as input for the 
 *		  decompressor. Compressed fields are read into this pool.
 *     _outpool - This pool contains "OUTBUFS" DMbuffers, used to hold
 *		  output from the decompressor (YUV 422 video fields). 
 *     _vdmpool - This pool contains "GFXBUFS" DMbuffers, used as the 
 *		  the color buffers (i.e. frame buffers) for the pbuffer 
 *		  _vdmpbuffer.   These DMbuffers can also be passed directly 
 *		  to video-out (via vlDMBufferSend). 
 *     _texpool - This pool contains 1 DMbuffer, used as the color buffer
 *		  for the pbuffer _texpbuffer. 
 *
 * The pbuffers used by this program (_vdmpbuffer and _texpbuffer) use memory  
 * from a DMbufferpool for their color buffers.  In order to do this, these
 * pbuffers must be created as digital media pbuffers (passed in as an attribute
 * to glXCreateGLXPbufferSGIX).  
 *
 * Being a digital media pbuffer allows _vdmpbuffer to cycle through the 
 * DMbuffers in _vdmpool, rendering into each of them.  These dmbuffers can 
 * then be sent to video out.  _vdmpbuffer uses a particular DMbuffer from
 * _vdmpool via a call to glXAssociateDMPbufferSGIX.  
 *
 * In addition, the fact that _texpbuffer is also a digital media pbuffer allows
 * glCopyTexSubImage2DEXT to copy pixels from * _texpbuffer to texture memory 
 * via a "copy by reference".  In this case, no real pixel copying occurs and 
 * _texpbuffer is really being used as the texure memory for _vdmpbuffer. 
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *	- After initializing everything, the routine doit() sprocs off
 *	  a reader thread (see ReadIt()) and a display thread (see PutField()).
 *	- The main thread then just hangs out.
 *
 *    Reader Thread: 
 *	- ReadIt() simply cycles through the file.  It allocates DMbuffers from
 *	  _inpool and loads them with compressed fields read from disk. 
 *
 *    Display Thread:
 *	- PutField() uses select() to wait for valid output to arrive from the 
 *	  decompressor.  Valid output arrives in the form of a DMbuffer from 
 *	  _outpool. 
 *	- Once it has a valid decompressed field (in YCrCb format), PutField()
 *        grabs the next available DMbuffer from _vdmpool (call this nextDMbuf).
 *	- The field is drawn into nextDMbuf as the background via glDrawPixels.
 *	- glDrawPixels is called again to draw the field into the DMbuffer 
 *        (_texbuf) associated with _texpbuffer (so this glDrawPixels call color
 *	  converts the field to RGB and loads it into texture memory). 
 *	- glXMakeCurrentReadSGI is then used to set up _texpbuffer as the read
 *        source for _vdmpbuffer (the color buffer for _vdmpbuffer is still 
 *	  nextDMbuf).
 *	- glCopyTexSubImage2DEXT is called.  This does a "copy-by-reference".  
 *	  The net effect is that _texpbuffer is now functioning as texture 
 *	  memory for _vdmpbuffer.
 *	- Two texture mapped quads are drawn on top of the background.  
 *	- All this rendering is happening with _vdmpbuffer and its associated 
 *	  DMbuffer, nextDMbuf.     
 *	- nextDMbuf is then sent to video via vlDMBufferSend(). 
 *	- PutField then loops back and waits for the next available output 
 *	  from the decompressor.
 *	
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <GL/gl.h>
#include <GL/glx.h>
#include <vl/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define INBUFS  12 
#define OUTBUFS 12 
#define GFXBUFS 12 
#define PRELOAD 8
#define FILL_CHECK 60
#define FILL_HIGH 9 
#define FILL_LOW 4 
#define OUTBUF_HIGH 9

#define VL_COMPRESSION_JPEG 2

/* All global variables in this file are preceded by an "_".  */

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

/* File related defines  */
int _CurField;	                /* Current field being read. 	   */
jFieldInfo_t *_indexbuffer;     /* Array of compressed field sizes */
fileHeader_t _f;

/* Decompressor defines.  */
DMimageconverter vice;	/* Handle to image decompressor.				*/
DMbufferpool   _inpool;	/* dmbuffer pool used as input to decompressor 			*/
DMbufferpool  _outpool;	/* dmbuffer pool used as output from decompressor 		*/
int          _inpoolfd;	/* fd to select on input to decompressor.	  		*/
int         _outpoolfd;	/* fd to select on _outpool.			  		*/
int              _icfd;	/* fd to select on output available fro decompressor 		*/


/* Defines for digital media pbuffer that drives fields to video out */ 

GLXPbufferSGIX _vdmpbuffer; /* dm-pbuffer that drives video. Associated with _vdmpool	*/
DMparams       *_vdmParams; /* dmParams associated with _vdmpbuffer.			*/
GLXContext     _vdmcontext; /* gfx context associated with _vdmpbuffer.			*/
DMbufferpool      _vdmpool; /* dmbuffer pool used as color buffer memory for _vdmpbuffer*/ 
		            /* These buffers are also sent to video out (vlDMBufferSend)*/	
int             _vdmpoolfd; /* fd to select on _vdmpool.                                */

/* Defines for digital media pbuffer that loads texture memory 	     */ 

GLXPbufferSGIX _texpbuffer; /* dm-pbuffer that serves as texture memory for _vdmpbuffer */ 
DMparams       *_texParams; /* dmParams associated with _texpbuffer.                    */
GLXContext    _textcontext; /* gfx context associated with _texpbuffer.			*/
DMbufferpool      _texpool; /* dmbuffer pool used as color buffer memory for _texpbuffer*/ 
DMbuffer           _texbuf; /* The one and only dmbuffer allocated from _texpool.	*/
int 	        _tex_width;
int	       _tex_height;

/* Misc defines. */
Display *_dpy;

/* Routine declarations. */
void initGL();
void doit();
void initVL();
void initDM4GLVL();
void initCompression();
void initDM4Tex();
void PutField();
void ReadIt();
void setscheduling();
void createTexture(unsigned int *, int, int, int, int);
GLXFBConfigSGIX getFBConfig(Display *, int);



main(int argc, char *argv[])
{
    /* Get file information.    */
    fileInit(argv[1], &_f, FALSE);
    _indexbuffer = createIndexBuffer(_f.Nfields);
    readIndexBuffer(&_f, _indexbuffer);

    ERROR ((_f.packing != VL_COMPRESSION_JPEG), "File not a jpeg file");

    /* Set tetxure width and height based on field sizes from file.  */
    _tex_width = 1024;   		          /* Field width is 640, 768, or 720  */
    _tex_height = (_f.vSize == 240) ? 256 : 512;  /* 240 or 288		              */

    initVL();		/* Create mem to vid path. */
    initCompression();	/* Set up decompressor, including _inpool and _outpool */
    initDM4GLVL();	/* Set up _vdmpool	*/
    initDM4Tex();	/* Set up _texpool	*/
    initGL();		/* Create _vdmpbuffer, _texpbuffer and init. their contexts */

    fprintf(stderr, "Playing  %d fields of %dx%d \n", _f.Nfields, _f.uSize, _f.vSize);

    doit();		/* sproc ReadIt() and sproc PutField()	*/
}

void initGL()
{
    /* This routine creates the two digital media pbuffers used by this
     * program (_vdmpbuffer and _texpbuffer).  
     */

    unsigned int texA;
    GLXFBConfigSGIX config;
    DMbuffer gfxbuf;
    int attrib[] = {GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE,
                    GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};

    _dpy = XOpenDisplay( ":0" ); 

    /* Get a 24 bit frame buffer configuration */
    config = getFBConfig(_dpy, 24);
    ERROR((config == NULL), "Can't get FB");

    /* Use this fb config. to set up _vdmpbuffer	*/ 
    GNC (_vdmpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _f.uSize, _f.vSize, attrib));
    GNC (_vdmcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));

    /* Allocate a color buffer from _vdmpool and associate it with _vdmpbuffer. 
     * There must be an associated DMBuffer in order to init the pbuffer context.
     */
    DC ( dmBufferAllocate(_vdmpool, &gfxbuf) );
    GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));

    /* Initialize context for _vdmpbuffer.  */
    GC (glXMakeCurrent(_dpy, _vdmpbuffer, _vdmcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_f.uSize,_f.vSize);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glPixelZoom (1.0, -1.0);     /* Need to flip the video when drawn as background */
    glRasterPos2f (-1.0, 1.0);

    /* Set up texture parameters. */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    createTexture(&texA, _tex_width, _tex_height, GL_RGBA, 0);
    BuildGfxLists(1, _f.uSize, _f.vSize, _tex_width, _tex_height,0);

    dmBufferFree(gfxbuf);  /* Done with buffer for now.	*/

    /* Now use config to set up _texpbuffer */ 
    GNC (_texpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _tex_width, _tex_height, attrib));
    GNC (_textcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    DC ( dmBufferAllocate(_texpool, &_texbuf) );

    /* Since there is only one DMbuffer required for texture memory, this
     * association is made only once in the beginning.
     */
    GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));

    /* Initialize context for _texpbuffer.	*/

    GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_tex_width, _tex_height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_tex_width, 0.0, (double)_tex_height, -1.0, 1.0);
    glRasterPos2i (0, 0);
}

void doit()
{
    /* Sproc the read and display threads.	*/
    int readPID, putPID;

    _CurField = 0;
    /* Surrender the gfx context.	*/
    GC (glXMakeCurrent(_dpy, None, NULL));

    /* Sproc off the read and display threads.	*/
    readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
    putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);

    while (1) sginap(100);  /*^C to exit	*/
}

void PutField()
{
    /* This is the display thread sproc'd off by doit.  It waits for a 
     * decompressed field, draws the field as background, then draws two 
     * texture mapped quads (of the same field) on top of the background.
     */
    fd_set fds;
    int i;
    int count = 0;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    double ust_per_msc;
    DMbuffer buf;
    DMbuffer gfxbuf;
    unsigned char *dataPtr;
    float scale = 0.50;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    /* raise priority if root or setuid is set.	*/
    setscheduling();

    ust_per_msc = vlGetUSTPerMSC(_svr, _path, _src);

    /* The display loop.	*/
    while (1) {

        /* Select on _icfd, it returns when there is valid output
         * from the decompressor.
         */
	FD_ZERO(&fds);
	FD_SET(_icfd, &fds);
	OC (select(_icfd+1, &fds, NULL, NULL, NULL));

	/* Get ptr to output data  */
	DC( dmICReceive(vice, &buf) );
        dataPtr = dmBufferMapData(buf);

        /* Load the field into the texture pbuffer */
        GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));
        GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext));
        glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

        /* Get the next available DMbuffer from _vdmpool.	*/ 
        FD_ZERO(&fds);
        FD_SET(_vdmpoolfd, &fds);
        OC (select(_vdmpoolfd+1, NULL, &fds, NULL, NULL));
        DC (dmBufferAllocate(_vdmpool, &gfxbuf));

        /* Associate this new DMbuffer with _vdmpbuffer (i.e. it is now its
         * color buffer).  Then set up _texpbuffer as the read source for
         * the texture copy.
         */ 
        GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));
	GC (glXMakeCurrentReadSGI(_dpy, _vdmpbuffer, _texpbuffer, _vdmcontext));

        /* Draw the field (still pointed to by dataPtr) as backgroiund  */
        glDisable(GL_TEXTURE_2D); 
	glClear(GL_COLOR_BUFFER_BIT);
        glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

        /* Now call glCopyTexSubImage2DEXT to load the field from _texpbuffer
         * into texture memory.  This is actually a copy by reference, the net
         * effect being that _texpuffer (and its associated _texbuf) serves as
         * texture memory for _vdmpbuffer.
         */
        glEnable(GL_TEXTURE_2D); /* enable texturing */
        glCopyTexSubImage2DEXT( GL_TEXTURE_2D, 0, 0, 0, 0, 0, _tex_width, _tex_height);

        /* Draw the two texture mapped quads.	*/
        glPushMatrix();
        glScalef(scale, scale, scale);
        glCallList(1);
        glScalef(scale, scale, scale);
        glCallList(1);
        glPopMatrix();

        /* see if we are dropping any fields */
        if (count%4 == 0) {
            newmsc = vlGetFrontierMSC(_svr, _path, _src);
            if ((oldmsc != -1) && (newmsc != oldmsc + 4) && (count > FILL_CHECK)) {
                fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+4, newmsc);
            }
            oldmsc = newmsc;
	}

	/* Send to video out */
	VC ( vlDMBufferSend(_svr, _path, gfxbuf) );
        dmBufferFree(buf);
        dmBufferFree(gfxbuf);
        count++;

	/* Don't start the transfer until we have some fields queued up. 	*/
        if (count == PRELOAD) {
            VC (vlBeginTransfer(_svr, _path, 0, NULL));
        }
#if 0
        /* Try to tune display process.	*/
        else if (count > FILL_CHECK) {
            i = vlGetFilledByNode(_svr, _path, _src);
            if (i > FILL_HIGH) sginap(1); 
            if (i < FILL_LOW ) fprintf(stderr, "Filled down to %d\n", i);
        }
#endif
   } 
}

void ReadIt()
{
    /* This is the read thread, it grabs fields off of disk and
     * sends them to the decompressor.
     */
    int nextRead;
    int dataSize;
    fd_set fds;
    unsigned char *dataPtr;
    DMbuffer buf;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);

    setscheduling();

    while (1) {
        /* Wait for an input buffer to be available */ 
        FD_ZERO(&fds);
        FD_SET(_inpoolfd, &fds);
        OC (select(_inpoolfd+1, NULL, &fds, NULL, NULL)); 

	/* Grab the availbale input DMbuffer.	*/
        DC(dmBufferAllocate(_inpool, &buf));
        dataPtr = dmBufferMapData(buf);
    
        /* Read the compressed field */
        nextRead = _indexbuffer[_CurField+1].offset - _indexbuffer[_CurField].offset;
        OC (read(_f.fd, dataPtr, nextRead)); 
    
        /* Set the buffer size and send it to decompressor  */
        nextRead -= _indexbuffer[_CurField].pad;
        dmBufferSetSize(buf, nextRead);

        /* Wait till there's room in _outpool to store the result*/
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL)); 

	/* Since there is room to store the result, send the compressed field. */
        DC (dmICSend(vice, buf, 0,(DMbuffer *)NULL) );
        dmBufferFree(buf);

        _CurField += 1;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass through file\n");
        }

	/* If we have enough buffered up , take a break	*/
	if (dmICGetDstQueueFilled(vice) > OUTBUF_HIGH) sginap(1);
   }
}

void initVL()
{
    /* Initialize the video stuff.  In particular, set
     * up a mem-> vid path.  The memory node for this play
     * path will be attached to the digital media pbuffers
     * where the textured video is being drawn.  Results can
     * then be sent directly from the pbuffer to video out.
     * See initDM4GLVL() for more details on this.
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR ((devPtr == NULL), "No mvp device");

    /* Create a mem to vid node.	*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE)); 

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set attributes on mem node.  Its rgb since it is going from gfx.	*/
    val.intVal = VL_FORMAT_RGB;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val)); 

    val.intVal = VL_PACKING_ABGR_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    /* This is new!  Why can't the video library figure this out */
    val.intVal = VL_LAYOUT_GRAPHICS;
    VC (vlSetControl(_svr, _path, _src, VL_LAYOUT, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
}


void initCompression ( void)
{
    /* This routine sets up the decomprssor - including its input
     * and output pools.
     */
    DMparams* p;
    int numIC, i;

    /* Find real-time jpeg compressor and try to open it */
    numIC = dmICGetNum();
    dmParamsCreate(&p);

    for (i = 0; i < numIC; i++) {

        /* Get description of next converter */
        if (dmICGetDescription(i, p) != DM_SUCCESS) {
            continue;
        }

        /* Look for the right combination */
        if ((dmParamsGetInt(p, DM_IC_ID) == 'jpeg') &&
            (dmParamsGetEnum(p, DM_IC_CODE_DIRECTION) == DM_IC_CODE_DIRECTION_DECODE) &&
            (dmParamsGetEnum(p, DM_IC_SPEED) == DM_IC_SPEED_REALTIME)) {
	    break;
        }
    }
    dmParamsDestroy(p);

    if (dmICCreate(i, &vice) != DM_SUCCESS) {
	fprintf(stderr,"Cannot open an appropriate real-time compressor\n");
	exit(1);
    }

    /* set the params for dmIC source */
    dmParamsCreate(&p);
    DC(dmSetImageDefaults(p,_f.uSize,_f.vSize,DM_IMAGE_PACKING_CbYCrY));
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetSrcParams(vice, p) );

    /* set the params for dmIC destionation.  These are the same as for 
     * dmIC source except for DM_IMAGE_COMPRESSION 
     */
    dmParamsSetString(p, DM_IMAGE_COMPRESSION, DM_IMAGE_UNCOMPRESSED);
    dmParamsSetEnum(p,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
    DC( dmICSetDstParams(vice, p) );
    dmParamsDestroy(p);

    /* Create buffer pool for compressor input */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,INBUFS,_f.uSize*_f.vSize*2,DM_FALSE,DM_TRUE) );
    DC( dmICGetSrcPoolParams(vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_inpool) );
    dmParamsDestroy(p);

    _inpoolfd = dmBufferGetPoolFD(_inpool);
    DC (dmBufferSetPoolSelectSize(_inpool, _f.uSize*_f.vSize*2) );

    /* Create buffer pool for dmIC compressor output */
    dmParamsCreate(&p);
    DC(dmBufferSetPoolDefaults(p,OUTBUFS,_f.uSize*_f.vSize*2,DM_FALSE,DM_TRUE));
    DC( dmICGetDstPoolParams(vice, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_outpool) );

    /* tell dmIC to use this for its output */
    DC( dmICSetDstPool(vice, _outpool) );
    dmParamsDestroy(p);

    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _f.uSize*_f.vSize*2) );
    _icfd = dmICGetDstQueueFD(vice);
} 

void initDM4GLVL()
{
    /* This routine creates _vdmpool, the DMbufferpool that will
     * provide the color buffers (i.e. frame buffer memory) for the 
     * dm-pbuffer _vdmpbuffer.  This buffer pool is also attached to 
     * the vl memory node (_src) that is driving video out.  This allows
     * the application to render into DMbuffers from _vdmpool using OpenGL
     * AND to send the very same DMbuffers to video out (with no additional
     * copying.
     */
    DMparams *p;

    /* Set up image attributes for the buffer pool	*/
    dmParamsCreate(&_vdmParams);
    DC ( dmSetImageDefaults(_vdmParams, _f.uSize, _f.vSize, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_vdmParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_vdmParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Create parameter list p and load it with default pool information.	*/
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, GFXBUFS, _f.uSize*_f.vSize*4, DM_FALSE, DM_TRUE) );
  
    /* Now load p with specific image (GL) information from _vdmParams.	*/
    DC ( dmBufferGetGLPoolParams(_vdmParams, p) );

    /* Now load p with the required information to attache video to the buffer pool. */
    VC ( vlDMPoolGetParams(_svr, _path, _src, p) );

    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    
    /* Create the buffer pool based on all the info. loaded into p.	*/
    DC ( dmBufferCreatePool(p, &_vdmpool) );
    dmParamsDestroy(p);

    /* Get the fd we can use to select on the next available dmbuffer from
     * _vdmpool.  The set the select size for the pool as well.
     */
    _vdmpoolfd = dmBufferGetPoolFD(_vdmpool);
    DC (dmBufferSetPoolSelectSize(_vdmpool, _f.uSize*_f.vSize*4) );
}

void initDM4Tex()
{
    /* This routine creates the DMbufferpool that provides the DMBuffer
     * used as frame buffer memory for the dm-pbuffer _texpbuffer.  Only
     * one DMbuffer is alloacted in this pool (it is assigned to _texbuf inside
     * initGL().  _texbuf will end up serving as texture memory for _vdmpbuffer.
     */ 

    DMparams *p;

    /* Set up geometry attributes fro pool.	*/
    dmParamsCreate(&_texParams);
    DC ( dmSetImageDefaults(_texParams, _tex_width, _tex_height, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_texParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_texParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Load p with default info. + geometry info from _texparams.	*/
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, 1, _tex_width*_tex_height*4, DM_FALSE, DM_TRUE) );
    DC ( dmBufferGetGLPoolParams(_texParams, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);

    /* Create the buffer pool.	*/
    DC ( dmBufferCreatePool(p, &_texpool) );
    dmParamsDestroy(p);
}
