/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vidin2tex - This program takes live video input and displays it to a window on
 *             the gfx monitor.  The video is put into texture memory and mapped
 *	       onto a quad.  The video is put in texture by drawing it to a pbuffer
 *	       and then copying (by refernce) into texture memory. 
 *	   
 *	       Clicking the left, middle or right mouse button rotates the quad
 *	       (along with the video) around the x, y, or z axis respectively.
 *	       If things are rotating, then clickin ANY button will stop rotation.
 *	
 *	       The video fields are drawn into a digital media pbuffer and then
 *	       copied by reference into texture memory.  Texture memory is loaded
 *	       in this manner because:
 *		    1. This method allows video to come in in YUV format.
 *		    2. Loading tetxure memory directly (even for RGB) is much
 *		       slower than doing the copy by reference.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-gfx_frame      - Display a full frame by doubling scan lines.
 *			  (i.e a 640x240 field displayed as 640x480)
 *         OR
 *
 *      -gfx_half>      - Display a half height field.
 *			  (i.e a 640x240 field displayed as 320x240)
 *
 * If no argument is specified then a a field sized window is displayed. 
 *
 * The field size is determined by the "Timing" selected for "Default Input"
 * on the vcp.  More exactly:
 *
 *        Video Timing            Capture Dimensions
 *       ============            ===================
 *       Camera Video                640 x 240
 *       Square NTSC (525)           640 x 240
 *       CCIR NTSC (525)             720 x 240
 *       Square PAL (625)            768 x 288
 *       CCIR PAL (625)              720 x 288
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/signal.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <vl/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_image.h>
#include "common/handy.h"

/* Video related defines */
VLServer svr;
VLPath path;
VLNode src;
VLNode drn;
VLBuffer videoBuffer;

/* texture defines */
int tex_format = GL_RGBA;
int tex_width, tex_height;
unsigned int texA;
float rot[3];

/* Misc. defines */
int uSize;
int vSize;
int winx;
int winy;
int dbuffer;
int depth;
int rgb8 = 0;
int spinning = 0;

/* OpenGL  defines */
Display *dpy;
GLXContext context; 
GLXContext pbcontext;
GLXPbufferSGIX dmpbuffer;
DMbuffer dmBuf;
DMparams *dmParams;
Widget glw;

void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void initVL(int*, int*);
void initGL();
void initDM(void);
int BuildGfxLists(int start, int uSize, int vSize, int tex_width, int tex_height, int rhs);
void DrawVideo();
void createTexture(unsigned int *, int, int, int, int);
GLXFBConfigSGIX getFBConfig(Display *, int);


main(int argc, char *argv[])
{
    int i,n;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit",doit},{"doit",doit},{"doit",doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n<Btn3Down>:doit()\n";

    dbuffer = 1; winx =1; winy = 1;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-single") == 0) {
            dbuffer = 0;
        }
        else if (strcmp (argv[i], "-gfx_frame") == 0) {
	    winy = 2; winx = 1;
        }
        else if (strcmp (argv[i], "-gfx_half") == 0) {
	    winy = 1; winx = 2;
        }
        else if (strcmp (argv[i], "-rgb8") == 0) {
	    rgb8 = 1; 
        }
    }

    initVL(&uSize, &vSize);
    winx = uSize/winx;
    winy = vSize*winy;

    fprintf(stderr, "Field Size = %dx%d, Window Size = %dx%d\n", uSize, vSize, winx, winy); 


    toplevel = XtAppInitialize(&app_context, "4DgiftsGlx",
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv,
			       NULL,NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], GLwNdoublebuffer, dbuffer); n++;
    XtSetArg(args[n], XmNwidth, winx); n++;
    XtSetArg(args[n], XmNheight, winy);n++;
    glw = GLwCreateMDrawingArea(form, "glwidget", args, n);

    XtManageChild (glw);
    XtAddCallback(glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);
    XtOverrideTranslations(glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    initGL();
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    static int firstExpose = 1;

    if (firstExpose) {
	GC (glXMakeCurrent(dpy, XtWindow(glw), context));
	glClear(GL_COLOR_BUFFER_BIT);
	if (dbuffer) glXSwapBuffers(dpy,XtWindow(glw));
        GC (glXMakeCurrent(dpy, None, NULL));
        sproc((void (*)(void *))DrawVideo, PR_SADDR | PR_SFDS);
	firstExpose = 0;
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    switch(event->type) {
	case ButtonPress:
	    if (spinning) spinning = 0;
	    else {
		rot[0] = rot[1] = rot[2] =  0.0;
		rot[event->xbutton.button - 1] = 1.0;
		spinning = 1; 
	    }
	break;
	default:
	break;
    }
}

void DrawVideo()
{
    int val;
    int dospin;
    int count = 0;
    float scale = 1.0;
    float scaleInc = 0.001;
    float spin = 0.0;
    stamp_t oldmsc = -1;
    stamp_t newmsc;
    unsigned char *dataPtr;
    VLInfoPtr info;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    VC (vlBeginTransfer(svr, path, 0, NULL));

    val = (rgb8) ? GL_RGBA : GL_YCRCB_422_SGIX;

    while (1) {

        do {    /* Get the number and location of frames captured */
	    info = vlGetNextValid(svr, videoBuffer);
        } while (!info);

        dataPtr = vlGetActiveRegion(svr, videoBuffer, info);

        newmsc = vlGetFrontierMSC(svr, path, drn);
        if ((oldmsc != -1) && (newmsc != oldmsc + 1) && (count > 60)) {
	    fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+1, newmsc);
        }
        oldmsc = newmsc;

        GC (glXMakeCurrent(dpy, dmpbuffer, pbcontext));
        glDrawPixels(uSize, vSize, val, GL_UNSIGNED_BYTE, dataPtr);

        GC (glXMakeCurrentReadSGI(dpy, XtWindow(glw), dmpbuffer, context));
        glCopyTexSubImage2DEXT( GL_TEXTURE_2D, 0, 0, 0, 0, 0, tex_width, tex_height);

        dospin = spinning;

        if (dospin) {
            glPushMatrix();
            glScalef(scale, scale, scale);
            glRotatef(spin, rot[0], rot[1], rot[2]);
	}
        else {
	    spin = 0.0;
	    scale = 1.0;
	    scaleInc = 0.001;
	}

        if (dbuffer) glClear(GL_COLOR_BUFFER_BIT);
	glCallList(1);

        if (dospin) {
	    glPopMatrix();
	    spin += 0.5;
	    if (spin > 360.0) spin -= 360.0;

	    if ((scale <= .1) || (scale >= 1.0))  scaleInc = -scaleInc; 
	    scale += scaleInc;
	}

        if (dbuffer) glXSwapBuffers(dpy,XtWindow(glw));
	vlPutFree(svr, videoBuffer);
        ++count;
    }
}


void initVL(int *xdim, int *ydim)

{
    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC(svr = vlOpenVideo(""));
    vlGetDeviceList(svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR ((devPtr == NULL), "No mvp device");

    VC (src = vlGetNode(svr, VL_SRC, VL_VIDEO, VL_ANY));
    VC (drn = vlGetNode(svr, VL_DRN, VL_MEM, VL_ANY));
    VC (path = vlCreatePath(svr, devPtr->dev, src, drn)); 
    VC (vlSetupPaths(svr, (VLPathList)&path, 1, VL_SHARE, VL_SHARE));

    if (rgb8) val.intVal = VL_FORMAT_RGB;
    else val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(svr, path, drn, VL_FORMAT, &val));

    if (rgb8) val.intVal = VL_PACKING_ABGR_8;
    else val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(svr, path, drn, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(svr, path, drn, VL_CAP_TYPE, &val));

    vlGetControl(svr, path, drn, VL_SIZE, &val);
    *xdim = val.xyVal.x; 
    *ydim = val.xyVal.y;

    VNC(videoBuffer = vlCreateBuffer(svr, path, drn, 60));
    fprintf(stderr, "Transfer Size = %d \n",vlGetTransferSize(svr, path));
    VC (vlRegisterBuffer(svr, path, drn, videoBuffer));
}


/* Initialize GL. */
void
initGL()
{
    Arg args[18];
    XVisualInfo *vi;
    GLXFBConfigSGIX config;
    int val;
    int attrib[] = {GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE, 
                    GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};

    tex_width = 1024;                        /* Field width is 640, 768, or 720  */
    tex_height = (vSize <= 256) ? 256 : 512; /* 240 or 288                     */

    dpy = XtDisplay(glw);
    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(glw, args, 1);

    depth = vi->depth;
    GNC(context = glXCreateContext(dpy, vi, 0, GL_TRUE));
    GC (glXMakeCurrent(dpy, XtWindow(glw), context));

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    /* create texture A */
    createTexture(&texA, tex_width, tex_height, GL_RGBA, 0);

    /* init viewport */
    glViewport(0,0,winx,winy);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);

    /* enable texturing */
    glEnable(GL_TEXTURE_2D);
    glDisable(GL_DITHER);
    BuildGfxLists(1, uSize, vSize, tex_width, tex_height, 0);

    initDM();
    config = getFBConfig(dpy, depth);
    ERROR((config == NULL), "Can't get FB");

    GNC (dmpbuffer = glXCreateGLXPbufferSGIX(dpy, config, tex_width, tex_height, attrib));
    GNC (pbcontext = glXCreateContextWithConfigSGIX(dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    GNC (glXAssociateDMPbufferSGIX(dpy, dmpbuffer, dmParams, dmBuf));
    GC  (glXMakeCurrent(dpy, dmpbuffer, pbcontext));

    /* init viewport */
    glViewport(0,0,tex_width, tex_height);
    /* set up transformation matrix */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)tex_width, 0.0, (double)tex_height, -1.0, 1.0);

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glDisable(GL_TEXTURE_2D);
    glRasterPos2i (0, 0);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }
}

void initDM() 
{
    DMparams *poolParams;
    DMbufferpool pool;
    DMimagepacking dmpacking;

    DC (dmParamsCreate(&dmParams));
    dmpacking = (depth == 24) ? DM_IMAGE_PACKING_RGBA : DM_IMAGE_PACKING_XRGB1555;
    DC (dmSetImageDefaults(dmParams, tex_width, tex_height, dmpacking));
    DC (dmParamsSetEnum(dmParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS));
    DC (dmParamsCreate(&poolParams));
    DC (dmBufferSetPoolDefaults(poolParams, 1, tex_width*tex_height*4, DM_FALSE, DM_TRUE));
    DC (dmParamsSetEnum(poolParams, DM_POOL_SHARE, DM_POOL_INTERPROCESS));
    DC (dmBufferGetGLPoolParams(dmParams, poolParams));
    DC (dmBufferCreatePool(poolParams, &pool));
    DC (dmBufferAllocate(pool, &dmBuf));
}
