/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* 2streamfx_j- Play two compressed video streams from disk, perform an effect
 *	        and send the result to video-out and a screen node. 
 *
 *	        The two streams can be either lossy JPEG or lossless JPEG.
 *	        This program has been run on an Octane MXI with both a ProVideo 
 *	        and Octane compression board.  A lossless JPEG stream averages
 *	        about a 2:1 compression ratio (about 10 megabytes/sec).
 *
 *		This program made use of a Ciprico 6900 disk array plugged in
 *		to SCSI controller 1 of the Octane.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-gfx_half>      	- Display half height field (default is video
 *                                frame).
 *     file1			- file for first stream (created with vgrab_j). 
 *     <file2>			- file for 2nd stream   (created with vgrab_j). 
 *				  *** If a second file s not specified, then the 
 *				  first file is played out with no effect. 
 *
 * How to run the program:
 * =======================
 * When you run the program, a blank window appears.  Click once with the left 
 * mouse to start play, click again with the same button to stop play.  The 
 * initial effect performed is a blend.  Click at any time with the middle mouse
 * button to change the effect being played.  See the routine "loadIt" for 
 * effect choices. After entering the new effect, play resumes immediately.  
 * The effect calculation routines are inside common/fxcode.c 
 * 
 *
 * Overview of program flow: 
 * =========================
 * Each time you click the button and start play, two reader threads (one per
 * stream or file) and an effect thread are all sproc'd.  The reader threads
 * read compressed jpeg from disk and feed it to the compression board.
 * The effect thread retrieves the decompressed fields from memory (GL_ABGR_EXT
 * format), sends them to gfx (texture memory or frame buffer), reads back the
 * effect, and then sends the result to video-out and the screen node.  
 * Total system memory bandwidth looks like:
 *    Two streams @2:1 from disk to memory                         = 20 MB/sec.
 *    Two streams @2:1 from memory to compression card             = 20 MB/sec.
 *    Two RGBX streams from compression card to memory             = 80 MB/sec.
 *    Two RGBX streams from memory to gfx                          = 80 MB/sec.
 *    One RGB stream from gfx to memory (the effect)               = 30 MB/sec.
 *    One RGB stream from memory to video card                     = 30 MB/sec.
 *                                                         Total   =260 MB/sec.
 *                                                 Total to memory =130 MB/sec.
 *                                               Total from memory =130 MB/sec.
 * 
 * Here are the thread descriptions:
 * =================================
 *    1. UI thread - This threasd sits in the Xt main loop waiting for a
 *		     button press.  If a left button press, then the UI enters
 *		     the routine doit().  Inside doit() play is either stopped
 *		     or started depending upon the current play state.  If the
 *		     middle mouse button is pressed, then loadIt() is entered.
 *		     The user is prompted for a new effect, and play is resumed.
 *		     The UI thread directs the reader and effect threads to 
 *		     exit whenever play is stopped.
 *
 *    2. Readit() - One of these threads is sproc'd per stream.  The reader
 *		    thread reads compressed data from disk into the input
 *                  buffer of the compression codec associated with the stream 
 *		    (one codec per stream, see the stream_t structure for 
 *		    details. 
 *
 *		    The reader threads use direct I/O when reading from disk.
 *		    Each compressed field in an input file has been padded out 
 *		    with 0xffs so that its size is a multiple of 4k (this is
 *		    done while capturing).  The offset to the first field in
 *		    the file is also a multiple of 4k.  These two facts make it
 *		    easy to do direct I/O from anywhere in the file (you never 
 *		    need to switch back to buffered mode). 
 *
 *		    The reader threads run on cpu 0.
 *
 * 3. PlayEffect()- This is the thread which computes the effect and sends the
 *		    result to video-out. This thread waits for valid output
 *		    from each of the decompression codecs.  Once it has a valid
 *		    field for each stream it goes ahead and calculates the 
 *		    effect. 
 *
 *		    The effect is calculated in an off-screen pbuffer (see 
 *		    _pbuffer).  Usually one stream is drawn as backgroud via 
 *		    glDrawPixles and the other is texture mapeed.  The effect 
 *		    "depth" texture maps both streams.  All effect calculations 
 *		    are field based. The effect is read back from _pbuffer via 
 *		    glReadPixels and sent to video-out and a screen node (this
 *		    node is tied to the on-screen widget _glw).
 *
 *		    The effect thread runs on cpu 1.
 *
 * Notes:
 * ------
 * In single stream mode, this program still draws the field into a pbuffer and
 * then reads it back into a dmbuffer.  Since it is dealing with 4:2:2 data in this
 * case you draw/read half the number of lines (in GL_ABGR_EXT format).  This works
 * faster than bcopying, which would also require that the cl output buffers be
 * marked cacheable.
 *
 * For best results when drawing to the Octane Video screen node, make sure to do: 
 *              /usr/gfx/setmon -Fe 1280x1024_59 (NTSC).
 *			    OR
 *		/usr/gfx/setmon -Fe 1280x1024_49 (PAL).
 *
 * Sending video to a screen node means that the video hardware is inserting
 * video pixels onto the pixel bus just before the pixels arrive at the DACS
 * (which drive the monitor).  So there is no gfx control over the pixles going
 * to the monitor.
 */


#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ulocks.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/time.h>
#include <sys/cachectl.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/cl.h>
#include <vl/vl.h>
#include <vl/dev_mgv.h>
#include "common/handy.h"
#include "common/jinfo.h"
#include "common/fileutils.h"

#define CL_FIELD_BUFS 8 
#define CL_QUEUE_SIZE 16777216
#define CL_READ_SIZE 8388608
#define NUM_DMBUFFS 12
#define FILL_CHECK 32

typedef enum {FX_NONE, FX_BLEND, FX_WIPE, FX_PIP, FX_CROSS,
              FX_BOXBLEND, FX_SQUEEZE, FX_PERSPECTIVE, FX_DEPTH} FxType_t;

/* One stream structure per file */
typedef struct stream {
    int 	curField;
    char       *cbuffer;
    char       *fbuffer;
    CLhandle    clCodec;
    CLbufferHdl dataHandle;
    CLbufferHdl frameHandle;
    jFieldInfo_t *indexBuffer;
    fileHeader_t f;
} stream_t;

/* All globals preceded by an "_" */

/* Video globals 			*/
VLServer _svr;
VLPath _path;
VLNode _src;
VLNode _devNode;
VLNode _voutdrn;
VLNode _drn;
VLNode _csc;
int    _packing;

/* DMbuffer globals 			*/	
DMbufferpool _outpool;      /* dm buffer pool used to drive video output. */
int          _outpoolfd;    /* File descriptor for _outpool.              */

/* File definitions 			*/
int _uSize;
int _vSize;
int _winx;
int _winy;
int _clFieldSize;
int _vlFieldSize;
stream_t _s1;
stream_t _s2;

/* Play control variables */
int _IamPlaying = 0;

/* OpenGL defines. 			*/
int _dbuffer = 0;
int _tex_width;                          /* texture size */
int _tex_height;                         /* texture size */
int _tex_def = GL_RGBA;			 /* Can only decompress into RGBA or 4:2:2 */
int _tex_format = GL_ABGR_EXT;		 /* Can only decompress into ABGR format   */
int _readformat = GL_RGB;		 /* Always read back RGB to send to video  */
unsigned int _texA;                      /* texture name 			   */
FxType_t _fx = FX_BLEND;		 /* default effect type.		   */

Display *_dpy;
Widget _glw;				 /* Used to disply screen node.		   */
GLXContext _glw_context;

GLXPbufferSGIX _pbuffer;		 /* pbuffer where effect is calculated.	   */
GLXContext _pb_context;

/* Procedure definitions.	*/
GLXFBConfigSGIX getFBConfig(Display *, int);
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void motionCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void loadit(Widget , XEvent *, String *, Cardinal *);
void initVL();
void initCLMem();
void PlayEffect();
void ReadIt(stream_t *);
void GetInputFields(char **f1, char **f2);
void SendToVideo (int count);

/* Misc. external procedure definitions */
CLhandle clInitDecompressor(int u, int v, int wx, int wy, int qf, int rgb);
void createTexture(unsigned int *name, int width, int height, int format, int mip);
int GetVLFormat(int packing);
void SetVideoOutTiming(VLServer svr, VLPath path, VLNode node, int width, int height);
void setscheduling();
void setcpu(int);

main(int argc, char *argv[])
{
    char *f1 = NULL;
    char *f2 = NULL;
    int i,n;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"loadit", loadit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:loadit()\n";

    _winx = 1; _winy = 2;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-gfx_half") == 0) {
            _winx = 2; _winy = 1;
        }
        else if (f1 == NULL) {
	    f1 = argv[i];
        }
        else if (f2 == NULL) {
	    f2 = argv[i];
        }
    }

    /* If no second file specified, then no effects and change exit counters */
    if (f2 == NULL) _fx = FX_NONE;

    /* Open up first stream */
    fileInit(f1, &_s1.f, FALSE); 
    _s1.indexBuffer = createIndexBuffer(_s1.f.Nfields);
    readIndexBuffer(&_s1.f, _s1.indexBuffer);

    ERROR ((_s1.f.packing != VL_COMPRESSION_JPEG), "File 1 Not a jpeg file");

    /* Open up second stream */
    if (_fx != FX_NONE) {
        fileInit(f2, &_s2.f, FALSE); 
        _s2.indexBuffer = createIndexBuffer(_s2.f.Nfields);
        readIndexBuffer(&_s2.f, _s2.indexBuffer);

        ERROR ((_s2.f.packing != VL_COMPRESSION_JPEG), "File 2 Not a jpeg file");
        ERROR ((_s1.f.uSize != _s2.f.uSize), "Mismatch in file widths");
        ERROR ((_s1.f.vSize != _s2.f.vSize), "Mismatch in file heights");
    }

    /* Initialize miscellaneous globals */
    _s1.curField = _s2.curField = 0;
    _uSize = _s1.f.uSize;
    _vSize = _s1.f.vSize;
    _winx  = _uSize/_winx;
    _winy  = _vSize*_winy;

    if (_fx == FX_NONE) {
	_packing     = VL_PACKING_YVYU_422_8;
        _vlFieldSize = _uSize*_vSize*2; 
        _clFieldSize = _vlFieldSize; 
        _readformat  = GL_ABGR_EXT;
    }
    else {
	_packing     = VL_PACKING_BGR_8_P;
        _vlFieldSize = _uSize*_vSize*3; 
        _clFieldSize = _uSize*_vSize*4; 
        _readformat  = GL_RGB;
    }

    /* Initialize video hardware */
    initVL();

    /* Allocate input and output buffers for the compression library.  */ 
    initCLMem();

  
    /* Do all that nasty Xt stuff	*/
    toplevel = XtAppInitialize(&app_context, argv[0],
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv, NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    /* Create the on-screen widget that will be used for the screen
     * node.
     */
    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], GLwNdepthSize, 24); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight, _winy);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);

    XtAddEventHandler(toplevel, SubstructureNotifyMask, FALSE, (XtEventHandler)motionCB, toplevel);


    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the widget initialize call back for the visible
     * widget _glw.  This routine also sets up _pbuffer.
     */

    Arg args[18];
    XVisualInfo *vi;
    GLXFBConfigSGIX config;
    int val;
    int attrib[] = {GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};
    /* Get a 24 bit frame buffer configuration */
    _dpy = XOpenDisplay( ":0" );
    config = getFBConfig(_dpy, 24);
    ERROR((config == NULL), "Can't get FB");

    /* Use this fb config. to set up _pbuffer        */
    GNC (_pbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _uSize, _vSize, attrib));
    GNC (_pb_context = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    GC  (glXMakeCurrent(_dpy, _pbuffer, _pb_context));

    _tex_width = 1024; 
    _tex_height = (_vSize <= 256) ? 256 : 512; 

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_def, 0);

    /* init viewport */
    glViewport(0,0,_uSize,_vSize);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    
    BuildVideoLists(1, _uSize, _vSize, _tex_width, _tex_height);
    glDisable(GL_DITHER);

    /* Now set up the widget used for the screen node.	*/
    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(_glw, args, 1);

    /* create a visual context */
    GNC (_glw_context = glXCreateContext(_dpy, vi, 0, GL_TRUE));
    GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));

    glViewport(0,0,_winx,_winy);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.0,0.0,0.0,1.0);
    glColor4f(1.0, 1.0, 1.0, 1.0);

    glDisable(GL_DITHER);
    glRasterPos2f(-1.0, 1.0);
    glPixelZoom(1.0, -1.0);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }

}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    int x,y;
    Window dummyWin;
    VLControlValue val;
    static int first = 1;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    if (first) {

        /* If it is the very first expose, then tell the vl
         * that _glw is to be used fro the screen node.
         */
        GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));
        glClear(GL_COLOR_BUFFER_BIT);

        /* Declare window that gets data for screen node */
        val.intVal = XtWindow(_glw);
        VC(vlSetControl(_svr, _path, _drn, VL_WINDOW, &val));

        /* Set up scale factors.	*/
        val.fractVal.numerator = 1;
        val.fractVal.denominator =  _uSize/_winx;
        VC(vlSetControl(_svr, _path, _drn, VL_ZOOM, &val));

        val.xyVal.x = _winx;
        val.xyVal.y = _winy;
        VC(vlSetControl(_svr, _path, _drn, VL_SIZE, &val));

        val.xyVal.x = x;
        val.xyVal.y = y;
        VC(vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val));
	first = 0;

        fprintf(stderr, "\n");
        fprintf(stderr, "*** Click left mouse button to start/stop play\n");
        fprintf(stderr, "*** Click midle mouse button to change effect\n");
        fprintf(stderr, "\n");
    }
}

void motionCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Window dummyWin;
    int x,y;
    VLControlValue val;

    /* Need to tell video server where window has moved to in order
     * for screen node to look correct.
     */
    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    val.xyVal.x = x;
    val.xyVal.y = y;
    vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val);
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    int allDone;

    /* Routine to START/STOP play. 
     * This routine is entered directly when the left mouse button is clicked
     * or from loadit() when changing effects.
     */

    if (_IamPlaying) {
        /* Then stop playing.	*/
        _IamPlaying = 0;
        allDone = (_fx != FX_NONE) ? -3 : -2;
        
        /* Wait for both readers and the effect thread to exit. */
        while (_IamPlaying != allDone) sginap(1);
        _IamPlaying = 0;
 
        VC (vlEndTransfer(_svr, _path));
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));

        /* Destroy first codec */
        CC (clDestroyBuf(_s1.dataHandle));
        CC (clDestroyBuf(_s1.frameHandle));
        CC (clCloseCompressor(_s1.clCodec));

        /* Destroy second codec */
        if (_fx != FX_NONE) {
           CC (clDestroyBuf(_s2.dataHandle));
           CC (clDestroyBuf(_s2.frameHandle));
           CC (clCloseCompressor(_s2.clCodec));
	}

        fprintf(stderr, "Play stopped\n");
	glClear(GL_COLOR_BUFFER_BIT);
	if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));

    }
    else {
        /* Then start playing.	*/
        _IamPlaying = 1;
        if (_s1.curField %2 != 0) _s1.curField++;
        if (_s2.curField %2 != 0) _s2.curField++;
        fprintf(stderr, "STARTING AT FIELD %d %d\n", _s1.curField, _s2.curField);

        /* Surrender the graphics context.      */
        GC (glXMakeCurrent(_dpy, None, NULL));

        /* Set up first codec.		*/ 
        _s1.clCodec = clInitDecompressor(_uSize, _vSize, _uSize, _vSize, _s1.f.qf, (_fx != FX_NONE));
        CNC (_s1.dataHandle  = clCreateBuf(_s1.clCodec, CL_BUF_COMPRESSED, 1, CL_QUEUE_SIZE, (void **)&_s1.cbuffer));
        CNC (_s1.frameHandle = clCreateBuf(_s1.clCodec, CL_BUF_FRAME, CL_FIELD_BUFS, _clFieldSize, 
                                           (void **)&_s1.fbuffer));

        /* Set up second codec.		*/ 
        if (_fx != FX_NONE) {
            _s2.clCodec = clInitDecompressor(_uSize, _vSize, _uSize, _vSize, _s2.f.qf, (_fx != FX_NONE));
            CNC (_s2.dataHandle  = clCreateBuf(_s2.clCodec, CL_BUF_COMPRESSED, 1, CL_QUEUE_SIZE, (void **)&_s2.cbuffer));
            CNC (_s2.frameHandle = clCreateBuf(_s2.clCodec, CL_BUF_FRAME, CL_FIELD_BUFS, _clFieldSize, 
                                           (void **)&_s2.fbuffer));
	} 

        /* Start off the readers.	*/
        sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS, (void *)&_s1);
        if (_fx != FX_NONE) sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS, (void *)&_s2);

        /* Start each channel decompressing.	*/
        CC (clDecompress(_s1.clCodec, CL_CONTINUOUS_NONBLOCK, 0, NULL, NULL));
        if (_fx != FX_NONE) CC (clDecompress(_s2.clCodec, CL_CONTINUOUS_NONBLOCK, 0, NULL, NULL));

	/* Start the effect play back thread.	*/
        sproc((void (*)(void *))PlayEffect, PR_SADDR | PR_SFDS);
   }
}

void loadit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routiine is entered whenever a new effect is loaded.  This
     * is done by pushing the middle mouse button.
     */ 
    char etype[12];

    /* Stop play */
    if (_IamPlaying) doit(_glw, NULL, NULL, NULL);
    if (_fx == FX_NONE) return;

    fprintf(stderr, "Enter effect: ");
    gets(etype);

    if (strcmp (etype,      "blend") == 0)    _fx = FX_BLEND;
    else if (strcmp (etype, "pip") == 0)      _fx = FX_PIP;
    else if (strcmp (etype, "wipe") == 0)     _fx = FX_WIPE;
    else if (strcmp (etype, "cross") == 0)    _fx = FX_CROSS; 
    else if (strcmp (etype, "boxblend") == 0) _fx = FX_BOXBLEND; 
    else if (strcmp (etype, "squeeze") == 0)  _fx = FX_SQUEEZE; 
    else if (strcmp (etype, "persp") == 0)    _fx = FX_PERSPECTIVE; 
    else if (strcmp (etype, "depth") == 0)    _fx = FX_DEPTH; 
    else _fx = FX_BLEND;

    /* Resume play */
    doit(_glw, NULL, NULL, NULL);
}

void ReadIt(stream_t *s)
{
    /* This is the routine run by each of the read threads.  Those
     * thread are sproc'd from inside doit().
     */

    int avail;
    int nextRead;
    int wrap;
    int done;
    char *cptr;
    char *queuesEnd;
    off64_t bytesRead;
    off64_t tmp;
    off64_t dataBytes;

    setcpu(0);
    setscheduling();

    queuesEnd = s->cbuffer + CL_QUEUE_SIZE;
    tmp = s->f.headerSize + s->indexBuffer[s->curField].offset;
    bytesRead = s->indexBuffer[s->curField].offset;
    dataBytes = s->f.dataBytes;

    OC (lseek64 (s->f.fd, tmp, SEEK_SET));
    OC(fcntl(s->f.fd, F_SETFL, FDIRECT));

    /* Keep on reading until the UI thread says stop.  */
    while (_IamPlaying == 1) {

        tmp = dataBytes - bytesRead;

        nextRead = (tmp >= CL_READ_SIZE) ? CL_READ_SIZE : tmp;
		
	done = 0;
        while (!done) {
            CC (avail = clQueryFree(s->dataHandle, 0, (void **)&cptr, &wrap));

	    if (avail >= nextRead) {
		/* There is enough room for my next read.	*/
		done = 1; 
	    }
	    else if ((avail + cptr) == queuesEnd) {
		/* There is NOT enough room for my next read because
                 * I am hitting the end of the queue.
                 */
		done = 1;
		nextRead = avail;
	    }
	    else {
		/* Just ain't no room for now.	*/
	        sginap(1);
	    }
        } 
    
        OC (read(s->f.fd, cptr, nextRead));

        CC (clUpdateHead(s->dataHandle, nextRead));

        bytesRead += nextRead;

	if (bytesRead == dataBytes) {
	    fprintf(stderr, "Done pass through file\n");
	    lseek64 (s->f.fd, s->f.headerSize, SEEK_SET);
	    bytesRead = 0;
	}
    }

    CC(clDoneUpdatingHead(s->dataHandle));

    fprintf(stderr, "Done reading\n");
    --_IamPlaying;
    _exit(0);
}

void PlayEffect()
{
    /* This is the effect calculation thread sproc'd from doit()
     * It jumps off (for the entire current play duration) into
     * one of the effect routines.
     */

    int ec = (_fx != FX_NONE) ? -2 : -1;
    setcpu(1);

    GC  (glXMakeCurrent(_dpy, _pbuffer, _pb_context));

    switch (_fx) {
	case FX_NONE:        
	    do_single(_uSize, _vSize/2, _tex_width, _tex_height, 
		      _tex_format, _texA, &_IamPlaying,ec);      
	    break;
	case FX_BLEND:       
            do_blend(_uSize, _vSize, _tex_width, _tex_height, 
	             _tex_format, _texA, &_IamPlaying,ec);       
	    break;
	case FX_WIPE :       
	    do_wipe(_uSize, _vSize, _tex_width, _tex_height, 
                    _tex_format, _texA, &_IamPlaying,ec);        
            break;
	case FX_PIP  :       
            do_pip(_uSize, _vSize, _tex_width, _tex_height, 
                   _tex_format, _texA, &_IamPlaying,ec);         
            break;
	case FX_CROSS:       
            do_cross(_uSize, _vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
	case FX_BOXBLEND:    
            do_boxblend(_uSize, _vSize, _tex_width, _tex_height, 
                        _tex_format, _texA, &_IamPlaying,ec);    
            break;
	case FX_SQUEEZE:     
            do_squeeze(_uSize, _vSize, _tex_width, _tex_height, 
                       _tex_format, _texA, &_IamPlaying,ec);     
            break;
	case FX_PERSPECTIVE: 
            do_perspective(_uSize, _vSize, _tex_width, _tex_height, 
                           _tex_format, _texA, &_IamPlaying,ec); 
            break;
	case FX_DEPTH:       
            do_depth(_uSize, _vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
    }

    /* Surrender the graphics context.  */
    glXMakeCurrent(_dpy, None, NULL);

    fprintf(stderr, "Done drawing\n");
    --_IamPlaying;
    _exit(0);
}

void initVL()
{
    /* Set up video hardware.  The video path set up
     * by this routine has a VL_MEM  source node and both
     * a VL_VIDEO and VL_SCREEN for drain nodes.  Using the
     * screen node allows the app. to display things on the gfx 
     * monitor using no additional gfx horspower.
     *
     * Since the compression card only decompresses into GL_ABGR_EXT,
     * a VL_CSC node has to be placed between the source and drain nodes
     * to convert the RGB to YUV for video display.
     */ 

    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;
    int format;
    int memnode;
    int n;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact video device");

    format = GetVLFormat(_packing); 
    memnode = (_packing == VL_PACKING_YVYU_422_8) ? VL_ANY : VL_MGV_NODE_NUMBER_MEM_VGI_DL;

    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, memnode));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_SCREEN, VL_ANY));
    VC (_voutdrn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_devNode = vlGetNode(_svr, VL_DEVICE, 0, VL_ANY));

    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlAddNode(_svr, _path, _voutdrn));
    VC (vlAddNode(_svr, _path, _devNode));

    if (_packing != VL_PACKING_YVYU_422_8) {
        VC (_csc = vlGetNode(_svr, VL_INTERNAL, VL_CSC, VL_MGV_NODE_NUMBER_CSC));
        VC (vlAddNode(_svr, _path, _csc));
        fprintf(stderr, "Added CSC node\n");
    }

    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, _devNode, _uSize, _vSize);

    val.intVal = format;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = _packing;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    if (_packing != VL_PACKING_YVYU_422_8) {
	/* Set up input csc controls */
	val.intVal = format;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_RANGE, &val));
	val.intVal = _packing;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_PACKING, &val));

	/* Set up output csc controls */
	val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_RANGE, &val));
	val.intVal = VL_PACKING_YVYU_422_8;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_PACKING, &val));

	/* Turn on or off CSC constant hue algorithm */
	val.boolVal = 0;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_CONST_HUE, &val));

	/* Turn on or off alpha correction in the alpha channel */
	val.boolVal = 0;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_ALPHA_CORRECTION, &val));
    }

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "NUM_DMBUFFS" buffers
     * each containing a video field of bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,NUM_DMBUFFS, _vlFieldSize, DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.    */
    DC( dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS) );

    /* Create the dmbuffer pool used for video out.     */
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _vlFieldSize));

    dmParamsDestroy(p);
}

void GetInputFields (char **f1, char **f2)
{
    /* This routine is called by the effect calculation 
     * routines in common/fxcode.c.  It gives them pointers
     * to teh two decompressed fields.
     */
    int wrap;

    /* Wait until we get a valid field from each stream */
    CC (clQueryValid(_s1.frameHandle, 1, (void**)f1, &wrap));

    if (_fx != FX_NONE) {
	CC (clQueryValid(_s2.frameHandle, 1, (void**)f2, &wrap));
    }
}

void SendToVideo (int count)
{
    /* This routine is called by the effect calculation routines.
     * It reads back the effect from the frame buffer and sends the
     * result to video-out.
     */
    int y;
    fd_set fds;
    unsigned char *readBuffer;
    DMbuffer buffer;
    static stamp_t oldmsc;
    stamp_t newmsc;

    /* Tell cl we are done with the current set of fields. */
    CC (clUpdateTail(_s1.frameHandle, 1));
    
    if (_fx != FX_NONE) {
	CC (clUpdateTail(_s2.frameHandle, 1));
        y = _vSize;
    }
    else {
	y = _vSize/2;
    }

    /* Wait for a dmbuffer to send to video.		   */
    FD_ZERO(&fds);
    FD_SET(_outpoolfd, &fds);
    OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

    /* Read back the effect into the dmbuffer.		*/
    DC (dmBufferAllocate(_outpool, &buffer));
    readBuffer = dmBufferMapData(buffer);
    glReadPixels(0, 0, _uSize, y, _readformat, GL_UNSIGNED_BYTE, readBuffer);

    /* Send to video out.				*/
    VC (vlDMBufferPutValid(_svr, _path, _src, buffer) );
    dmBufferFree(buffer);

    if (count == 5) {
        VC (vlBeginTransfer(_svr, _path, 0, NULL));
    }

    /* Check and see if we are dropping.		*/
    newmsc = vlGetFrontierMSC(_svr, _path, _src);
    if ((count > FILL_CHECK) && (newmsc != oldmsc + 1)) {
	fprintf(stderr, "MSC drop old = %lld next = %lld\n", oldmsc, newmsc);
    }
    oldmsc = newmsc;

    _s1.curField++;
    if (_s1.curField == _s1.f.Nfields) _s1.curField = 0;
    
    _s2.curField++;
    if (_s2.curField == _s2.f.Nfields) _s2.curField = 0;
}


void initCLMem()
{

    int i, n;

    /* Set up compression buffers for both streams. .cbuffer refers to the compreesed
     * input buffer.  .fbuffer refers to the uncompressed output buffers.
     * Round the ouput buffer size to a multiple of the page size so we can mark
     * it non-cacheable.  The compressed buffers are already a multiple of the
     * page size.
     */
    i = CL_FIELD_BUFS*_clFieldSize;
    n = i/getpagesize();
    if ((i % getpagesize()) != 0) n++;
    n = n * getpagesize();

    /* Allocate aligned memory.	*/
    _s1.fbuffer = (char * )memalign(_s1.f.headerSize, n);
    _s1.cbuffer = (char * )memalign(_s1.f.headerSize, CL_QUEUE_SIZE);
    _s2.fbuffer = (char * )memalign(_s2.f.headerSize, n);
    _s2.cbuffer = (char * )memalign(_s2.f.headerSize, CL_QUEUE_SIZE);

    /* Mark frame buffers non-cacheable. The app. also issues
     * a no-access command on them in clInitDecompress.
     */
    i = 0;
    i |= cachectl  (_s1.fbuffer, n, UNCACHEABLE );
    i |= cachectl  (_s2.fbuffer, n, UNCACHEABLE );
    i |= cacheflush(_s1.fbuffer, n, DCACHE );
    i |= cacheflush(_s2.fbuffer, n, DCACHE );
    /*if (i < 0) fprintf(stderr, "cache manipulations on frame buffers failed\n"); */

    /* Mark compressed buffers as non-caheable	*/
    i = 0;
    i |= cachectl  (_s1.cbuffer, CL_QUEUE_SIZE, UNCACHEABLE );
    i |= cachectl  (_s2.cbuffer, CL_QUEUE_SIZE, UNCACHEABLE );
    i |= cacheflush(_s1.cbuffer, CL_QUEUE_SIZE, DCACHE );
    i |= cacheflush(_s2.cbuffer, CL_QUEUE_SIZE, DCACHE );
    /* if (i < 0) fprintf(stderr, "cache manipulations on compressed buffers failed\n");*/

    /* Pin the buffers in memory - this only works if you are root.	*/
    i = 0;
    i |= mpin(_s1.cbuffer, CL_QUEUE_SIZE);
    i |= mpin(_s2.cbuffer, CL_QUEUE_SIZE);
    i |= mpin(_s1.fbuffer, CL_FIELD_BUFS*_clFieldSize);
    i |= mpin(_s2.fbuffer, CL_FIELD_BUFS*_clFieldSize);
    if (i < 0) fprintf(stderr, "MPIN failed - need super user privilege\n");
}
