/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vgrab_j = Grabs live video in and writes to disk as compressed jpeg files. 
 *	     Uses Octane compression in conjunction with Octane ProVideo board.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-n fieldsToCapture>     - Number of fields to capture, default = 200.
 *                                This arg should be an even number.
 *     <-qf qualityFactor>      - Number between 1-101 that governs the quality 
 *				  of the compression.  A higher number implies
 *				  higher quality.  Default is 75.
 *				- A quality factor of 101 ==> lossless compression.
 *     fileName                 - File for results.
 *
 * While capturing, this program pads each compressed field out to a multiple
 * of 4k.  The padding is done with 0xff's.  Padding makes it very easy to
 * play back the file using direct IO.  For high quality factors, then fast 
 * disks will be required.  Lossless compression (-qf 101) requires about 10
 * megabytes/sec/stream.  This was done here using a Ciprico 6900 array hooked
 * up to controller 1 of the Octane.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>

#include <vl/vl.h>
#include <dmedia/cl.h>
#include <dmedia/cl_impactcomp.h>
#include "common/handy.h"
#include "common/jinfo.h"
#include "common/fileutils.h"

#define CL_QUEUE_SIZE 8388608
#define CL_WRITE_SIZE 6291456
#define COPY_BUFFER_SIZE 8388608
#define MAX_FIELDS 50000

/* All global variables are preceded by an "_"	*/

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

CLhandle _compHandle;
CLbufferHdl _dataHandle;
char *_cbuffer;
char *_copyBuffer;
char *_padBuffer; 
jFieldInfo_t *_indexBuffer;
fileHeader_t *_f;

/* Routine declarations */
void doit();
void initVLCL();
CLhandle clInitCompression(CLhandle handle, int u, int v, int qf);

main(int argc, char *argv[])
{
    int i;
    int n = 200;
    int qf = 101;
    char *Fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            n  = atoi(argv[i]);
            if (n > MAX_FIELDS) n = MAX_FIELDS; 
        }
        else if (strcmp (argv[i], "-qf") == 0) {
           ++i;
           qf = atoi(argv[i]);
        }
	else {
	    Fname = argv[i];
        }
    }

    _f = fileCreate(Fname, FALSE);
    _f->packing = VL_COMPRESSION_JPEG;
    _f->compressed = TRUE;
    _f->pad = COMPRESSED_PAD;
    _f->Nfields = n;
    _f->qf = qf;

    _indexBuffer = createIndexBuffer(_f->Nfields); 
    _cbuffer     = (char * )memalign(_f->headerSize, CL_QUEUE_SIZE);
    _copyBuffer  = (char * )memalign(_f->headerSize, COPY_BUFFER_SIZE);
    _padBuffer   = (char * )memalign(_f->headerSize, COMPRESSED_PAD);

    memset(_padBuffer, 0xff, COMPRESSED_PAD);

    initVLCL();		/* Initialize video and compression hardwre. */
    fprintf(stderr, "Capturing %d fields at %dx%d with quality = %d\n", _f->Nfields, _f->uSize, _f->vSize, _f->qf);
    doit();
}

void doit()
{
    /* The work routine.  It simply loops checking for the arrival of new video
     * fields (which are then sent to the decompressor) AND also checks for 
     * any compressed results coming back from the compressor (which are then 
     * written to disk).
     */

    CLimageInfo imageinfo;
    int prev_imagecount = 0;
    int fieldsCaptured = 0;
    int validData;
    int tmp;
    int wrap;
    long pad = 0;
    long copyBytes = 0;
    off64_t bytesCaptured = 0;
    void *dataPtr;
    char *copyPtr;

    setscheduling();  /* Raise priority if root or setuid */

    lseek64(_f->fd, _f->headerSize, SEEK_SET);
    OC(fcntl(_f->fd, F_SETFL, FDIRECT));

    VC (vlBeginTransfer(_svr, _path, 0, NULL)); 
    CC (clCompress(_compHandle, CL_CONTINUOUS_NONBLOCK, CL_EXTERNAL_DEVICE, NULL, NULL));

    /* Simply loop, grabbing video fields into dmbuffers, sending 
     * the fields to the compressor,  and then writing the compressed
     * result to disk.
     */

    copyPtr = _copyBuffer;

    while (fieldsCaptured != _f->Nfields) { 

        /* Wait for the next field */
	while (clGetNextImageInfo(_compHandle, &imageinfo, sizeof(CLimageInfo)) != SUCCESS) {
	    sginap(1);
	}

        if (imageinfo.imagecount != (prev_imagecount + 1)) {
	    fprintf(stderr,"Dropped field[s] had %d got %d\n", prev_imagecount, imageinfo.imagecount); 
        }
        prev_imagecount = imageinfo.imagecount;

        validData = clQueryValid(_dataHandle, 0, (void**)&dataPtr, &wrap);

        /* Write the field into the copy buffer - handle the wrap case.  */
        if (validData >=  imageinfo.size) {
            memcpy (copyPtr, dataPtr, imageinfo.size);
            copyPtr += imageinfo.size;
	    CC (clUpdateTail(_dataHandle, imageinfo.size));
        }
        else {  /* The data has wrapped around to the front of the queue */
            tmp = imageinfo.size - validData; 
            memcpy (copyPtr, dataPtr, validData);
            copyPtr += validData;
	    CC (clUpdateTail(_dataHandle, validData));

            validData = clQueryValid(_dataHandle, 0, (void**)&dataPtr, &wrap);
            memcpy (copyPtr, dataPtr, tmp);
            copyPtr += tmp;
	    CC (clUpdateTail(_dataHandle, tmp));
	}

        /* Pad out field to "COMPRESSED_PAD" boundary with 0xff's	*/ 
	tmp = imageinfo.size/COMPRESSED_PAD;
	if ((imageinfo.size % COMPRESSED_PAD) != 0) tmp++;
	pad = COMPRESSED_PAD * tmp - imageinfo.size;

        if (pad) {
	    memcpy(copyPtr, _padBuffer, pad);
	    copyPtr += pad;
	}

        /* Update index table	*/
        _indexBuffer[fieldsCaptured].offset = bytesCaptured;
        _indexBuffer[fieldsCaptured].pad = pad;

        tmp = imageinfo.size + pad;
        bytesCaptured += tmp; 
        copyBytes += tmp; 

        if (copyBytes > CL_WRITE_SIZE) {
	    OC( write(_f->fd, _copyBuffer, copyBytes));
            copyBytes = 0;
            copyPtr = _copyBuffer;
        } 

        fieldsCaptured++;
        if (fieldsCaptured % 200 == 0) fprintf(stderr, "Captured %d\n", fieldsCaptured);
    }

    if (copyBytes) {
        OC( write(_f->fd, _copyBuffer, copyBytes));
    }

    CC (clCloseDecompressor(_compHandle ));
    VC (vlEndTransfer(_svr, _path));

    _indexBuffer[_f->Nfields].offset = bytesCaptured;
    _indexBuffer[_f->Nfields].pad = COMPRESSED_PAD;
    _f->dataBytes = bytesCaptured;

    OC(fcntl(_f->fd, F_SETFL, 0x0));
    writeHeader(_f);
    writeIndexBuffer(_f, _indexBuffer);

    fsync(_f->fd);
    close(_f->fd);
}



void initVLCL()
{
    /* Initialize video for capture - set up a vid to mem path
     * and set the correct data formats on the memory node.
     */

    int i,n;
    int codecID;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact device");

    CC (clOpenCompressor(CL_JPEG_IMPACT, &_compHandle));
    CC (codecID = clGetParam( _compHandle, CL_IMPACT_VIDEO_INPUT_CONTROL));

    /* Set up a capture _path from video-in to memory.	*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_VIDEO, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_CODEC, codecID));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    /* Get the actual video field size 	*/
    VC (vlGetControl(_svr, _path, _drn, VL_SIZE, &val));
    _f->uSize = val.xyVal.x;
    _f->vSize = val.xyVal.y;

    /* Make sure _vSize is a multiple of 8 - jpeg restriction. */
    _f->vSize = _f->vSize - (_f->vSize%8);
    if (_f->vSize == 248) _f->vSize = 240;

    val.xyVal.x = _f->uSize; 
    val.xyVal.y = _f->vSize; 
    VC (vlSetControl(_svr, _path, _drn, VL_SIZE, &val));

    (void)clInitCompression(_compHandle, _f->uSize, _f->vSize, _f->qf); 
    CNC (_dataHandle = clCreateBuf(_compHandle, CL_BUF_COMPRESSED, 1, CL_QUEUE_SIZE, (void **)&_cbuffer));
}
