/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* gstreamfx_u -Plays two uncompressed streams from memory, performs an effect
 *	        in an on screen window which can be optionally mapped to video
 *		out using the Octane video screen node functionality. 
 *
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-vid>			- map gfx window to video out.
 *     file1			- file for first stream (created with vgrab_u). 
 *     file2			- file for 2nd stream   (created with vgrab_u). 
 *
 * Each stream is represented by the first sample from the corresponding input
 * file.  By limiting input to one sample per stream (in memory), this program 
 * can be used for prototyping video effects for uncompressed streams WITHOUT 
 * requiring storage capable of sustaining 60-80 MB/sec.
 *
 * The input files may contain GL_RGB (24 bits per pixel) or GL_RGBA (32 bits 
 * per pixel). The input files should contain video-sized frames (not fields) if
 * it is desired to use the -vid option (since a frame's worth of data is what 
 * gets mapped to video out).
 *
 * This program is similar to vstreamfx_u in the types of effects it
 * calculates but it uses the "other" approach to get things to video out.
 * vstreamfx_u calculates things in an off-screen pbuffer, reads the effect
 * back via glReadPixels and then sends the result to video-out and a screen
 * node.  As a result, vstreamfx_u can also handle yuv data AND it only has
 * to perform effects on field sized data when driving video out. gstreamfx_u
 * can only handle rgb(a) data and must render frame sized effects in order
 * to have things properly mapped to video out. 
 *
 * This program has been run on an Octane MXI with a ProVideo board.
 *
 * How to run the program:
 * =======================
 * When you run the program, a blank window appears.  Click once with the left 
 * mouse to start play, click again with the same button to stop play.  The 
 * initial effect performed is a blend.  Click at any time with the middle mouse
 * button to change the effect being played.  See the routine "loadIt" for 
 * effect choices. After entering the new effect, play resumes immediately.  
 * The effect calculation routines are inside common/fxcode.c 
 *
 * Overview of program flow: 
 * =========================
 * Each time you click the button and start play, the effect thread retrieves 
 * the video samples from memory and performs the effect in a gfx window.  If
 * -vid was specified, then the displayed result is automatically being mapped 
 * to video out using the screen node hardware.
 * 
 * Here are the thread descriptions:
 * =================================
 *    1. UI thread - This threasd sits in the Xt main loop waiting for a
 *		     button press.  If a left button press, then the UI enters
 *		     the routine doit().  Inside doit() play is either stopped
 *		     or started depending upon the current play state.  If the
 *		     middle mouse button is pressed, then loadIt() is entered.
 *		     The user is prompted for a new effect, and play is resumed.
 *		     The UI thread directs the effect thread to exit whenever 
 *		     play is stopped. 
 *
 * 2. PlayEffect()- This is the thread which computes the effect. The effect is 
 *		    calculated in an on-screen window, usually one stream is drawn 
 *		    as backgroud via glDrawPixles and the other is texture mapeed.  
 *		    The effect "depth" texture maps both streams.  If -vid
 *		    is specified then the result automatically goes to video out.
 *
 * Notes:
 * ------
 * For best results when driving video from an Octane screen node do: 
 *              /usr/gfx/setmon -Fe 1280x1024_59 (NTSC).
 *			    OR
 *		/usr/gfx/setmon -Fe 1280x1024_49 (PAL).
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ulocks.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/time.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <vl/vl.h>
#include "common/handy.h"
#include "common/fileutils.h"

typedef enum {FX_NONE, FX_BLEND, FX_WIPE, FX_PIP, FX_CROSS, FX_BOXBLEND,
              FX_SQUEEZE, FX_PERSPECTIVE, FX_DEPTH} FxType_t;

/* Video globals                        */
VLServer _svr;
VLPath _path;
VLNode _src;
VLNode _drn;
VLNode _devNode;
int    _vidout = 0;

/* File defines */
char *_sourceBuffer1; /* Pointer to first input frame. */
char *_sourceBuffer2; /* Pointer to second input frame. */
fileHeader_t _f1;     /* File header info for first stream. */
fileHeader_t _f2;     /* File header info for sewcond stream. */

/* texture defines */
unsigned int _texA;                      /* texture names */
int _tex_width;                          /* texture size */
int _tex_height;                         /* texture size */
int _tex_format  = GL_RGB;
FxType_t _fx = FX_BLEND;

/* OpenGL defines */
Display *_dpy;
Widget _glw;
GLXContext _glx_context;
int _dbuffer = 1;

/* Play control variables */
int _IamPlaying = 0;
int _count;

void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void motionCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void loadit(Widget , XEvent *, String *, Cardinal *);
void PlayEffect();
void GetInputFields(char **f1, char **f2);
void SendToVideo (int count);
void initVL(void);
int  BuildGfxLists(int start, int u, int v, int tw, int th, int rhs);


main(int argc, char *argv[])
{
    char *f1 = NULL;
    char *f2 = NULL;
    int i,n;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"loadit", loadit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:loadit()\n";

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-single") == 0) {
            _dbuffer = 0; 
        }
        if (strcmp (argv[i], "-vid") == 0) {
            _vidout = 1; 
        }
        else if (f1 == NULL) {
            f1 = argv[i];
        }
        else if (f2 == NULL) {
            f2 = argv[i];
        }
    }

    /* If no second file specified, then no effects and change exit counters */
    if (f2 == NULL) ERROR(1, "Need two input files");

    /* Open up first stream */
    fileInit(f1, &_f1, FALSE);

    /* Open up second stream */
    fileInit(f2, &_f2, FALSE);

    ERROR ((_f2.packing != _f1.packing), "Files not same packing\n");
    ERROR ((_f1.uSize != _f2.uSize), "Mismatch in file widths");
    ERROR ((_f1.vSize != _f2.vSize), "Mismatch in file heights");

    if (_f1.packing == VL_PACKING_ABGR_8)       _tex_format = GL_RGBA;
    else if (_f1.packing == VL_PACKING_BGR_8_P) _tex_format = GL_RGB;
    else ERROR(1,"Bad packing");

    _sourceBuffer1 = memalign(_f1.headerSize, _f1.readSize);
    _sourceBuffer2 = memalign(_f2.headerSize, _f2.readSize);

    lseek (_f1.fd, _f1.headerSize, SEEK_SET);
    OC ( read(_f1.fd, _sourceBuffer1, _f1.readSize));

    lseek (_f2.fd, _f2.headerSize, SEEK_SET);
    OC ( read(_f2.fd, _sourceBuffer2, _f2.readSize));

    if (_vidout) {
	initVL();
    }

    fprintf(stderr, "Playing %dx%d RGBA = %d\n", _f1.uSize, _f1.vSize, _tex_format == GL_RGBA); 
    
    toplevel = XtAppInitialize(&app_context, argv[0],
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv,
			       NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _f1.uSize); n++;
    XtSetArg(args[n], XmNheight, _f1.vSize);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);

    XtAddEventHandler(toplevel, SubstructureNotifyMask, FALSE, (XtEventHandler)motionCB, toplevel);

    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Arg args[18];
    XVisualInfo *vi;
    int val;

    _dpy = XtDisplay(_glw);

    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(_glw, args, 1);

    /* create a visual context */
    _glx_context = glXCreateContext(_dpy, vi, 0, GL_TRUE);

    GNC (_glx_context = glXCreateContext(_dpy, vi, 0, GL_TRUE));
    GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glx_context));

    _tex_width = 1024; 
    _tex_height = (_f1.vSize <= 256) ? 256 : 512; 

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_format, 0);

    /* init viewport */
    glViewport(0,0,_f1.uSize,_f1.vSize);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    
    (void)BuildGfxLists(1, _f1.uSize, _f1.vSize, _tex_width, _tex_height,1);

    glDisable(GL_DITHER);
    glReadBuffer(GL_FRONT);
    glRasterPos2f(-1.0, 1.0);
    glPixelZoom(1.0, -1.0);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }

}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    int x,y;
    Window dummyWin;
    VLControlValue val;
    static int first = 1;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    if (first) {
        GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glx_context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));
        glClear(GL_COLOR_BUFFER_BIT);

        if (_vidout) {
            /* Set up screen node mapping.	*/
	    val.xyVal.x = x;
	    val.xyVal.y = y;
	    vlSetControl(_svr, _path, _src, VL_ORIGIN, &val);

	    VC (vlBeginTransfer(_svr, _path, 0, NULL));
        }
	first = 0;
    }
}

void motionCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Window dummyWin;
    VLControlValue val;
    int x,y;

    if (_vidout) {
        XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

        val.xyVal.x = x;
        val.xyVal.y = y;
        vlSetControl(_svr, _path, _src, VL_ORIGIN, &val);
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    static unsigned long long then, now;
    double tdiff;
    int allDone;

    /* Routine to START/STOP play. 
     * This routine is entered directly when the left mouse button is clicked
     * or from loadit() when changing effects.
     */

    if (_IamPlaying) {
        /* Then stop playing.	*/
        _IamPlaying = 0;
        allDone = -1; 
	dmGetUST(&now);
        
        /* Wait for both readers and the effect thread to exit. */
        while (_IamPlaying != allDone) sginap(1);
        _IamPlaying = 0;
 
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _glx_context));

	tdiff = now - then;
	tdiff = tdiff/1000000000.0;
	fprintf(stderr, "Draw time = %f sec. OR fps = %f\n", tdiff, ((double)(_count))/tdiff);

    }
    else {
        /* Then start playing.	*/
        _IamPlaying = 1;

        /* Surrender the graphics context.      */
        GC (glXMakeCurrent(_dpy, None, NULL));

	/* Start the effect play back thread.	*/
        sproc((void (*)(void *))PlayEffect, PR_SADDR | PR_SFDS);
	dmGetUST(&then);
    }

}

void loadit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routiine is entered whenever a new effect is loaded.  This
     * is done by pushing the middle mouse button.
     */ 
    char etype[12];

    /* Stop play */
    if (_IamPlaying) doit(_glw, NULL, NULL, NULL);
    if (_fx == FX_NONE) return;

    fprintf(stderr, "Enter effect: ");
    gets(etype);

    if (strcmp (etype,      "blend") == 0)    _fx = FX_BLEND;
    else if (strcmp (etype, "pip") == 0)      _fx = FX_PIP;
    else if (strcmp (etype, "wipe") == 0)     _fx = FX_WIPE;
    else if (strcmp (etype, "cross") == 0)    _fx = FX_CROSS; 
    else if (strcmp (etype, "boxblend") == 0) _fx = FX_BOXBLEND; 
    else if (strcmp (etype, "squeeze") == 0)  _fx = FX_SQUEEZE; 
    else if (strcmp (etype, "persp") == 0)    _fx = FX_PERSPECTIVE; 
    else if (strcmp (etype, "depth") == 0)    _fx = FX_DEPTH; 
    else _fx = FX_BLEND;

    /* Resume play */
    doit(_glw, NULL, NULL, NULL);
}

void PlayEffect()
{
    /* This is the effect calculation thread sproc'd from doit()
     * It jumps off (for the entire current play duration) into
     * one of the effect routines.
     */

    int ec = 0; 

    GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glx_context));

    switch (_fx) {
	case FX_NONE:        
	    break;
	case FX_BLEND:       
            do_blend(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
	             _tex_format, _texA, &_IamPlaying,ec);       
	    break;
	case FX_WIPE :       
	    do_wipe(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                    _tex_format, _texA, &_IamPlaying,ec);        
            break;
	case FX_PIP  :       
            do_pip(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                   _tex_format, _texA, &_IamPlaying,ec);         
            break;
	case FX_CROSS:       
            do_cross(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
	case FX_BOXBLEND:    
            do_boxblend(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                        _tex_format, _texA, &_IamPlaying,ec);    
            break;
	case FX_SQUEEZE:     
            do_squeeze(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                       _tex_format, _texA, &_IamPlaying,ec);     
            break;
	case FX_PERSPECTIVE: 
            do_perspective(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                           _tex_format, _texA, &_IamPlaying,ec); 
            break;
	case FX_DEPTH:       
            do_depth(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
    }

    /* Surrender the graphics context.  */
    glXMakeCurrent(_dpy, None, NULL);

    fprintf(stderr, "Done drawing\n");
    --_IamPlaying;
    _exit(0);
}

void GetInputFields (char **f1, char **f2)
{
    /* The effect calculation code calls this routine to get pointers
     * to the input streams. See common/fxcode.c
     */
    *f1 = (char *)_sourceBuffer1;
    *f2 = (char *)_sourceBuffer2;
}

void SendToVideo (int count)
{
    /* This is the routine called by the effect code to send to
     * video out.  All it needs to do in this app. is swap 
     * buffers. See common/fxcode.c
     */
    if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));
    _count = count;
}


void initVL()
{
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    int n;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact video device");


    VC (_src = vlGetNode(_svr, VL_SRC, VL_SCREEN, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_devNode = vlGetNode(_svr, VL_DEVICE, 0, VL_ANY));

    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlAddNode(_svr, _path, _devNode));
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    n = _f1.vSize;
    if (n*2 > _f1.uSize) n = n/2;
    SetVideoOutTiming(_svr, _path, _devNode, _f1.uSize, n);
}
