/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* tplay_yuv_u = Plays uncompressed video fields (YUV 4:2:2) from disk
 *               to the gfx monitor.  The OpenGL color matirx functionality
 *		 is used to convert the data from YUV to RGB while loading
 *		 it into texture memory.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength> 	- Number of reads per readv call.
 *     <-gfx winx winy>     	- Size of gfx display window. 
 *     <-single>		- Force single buffering (default is double).
 *     fileName			- file created with vgrab_u. 
 *				  This file must contain fields of 4:2:2 data.
 *
 * When you run the program, a blank window appears.  Click once to start
 * play, click again to stop play, etc. This program has been run on an 
 * Octane MXI.
 *
 * There are three threads sprco'd off in this program:
 *	1. ReadIt()   - This thread uses readv to grab fields of YUV 4:2:2
 *		        data from disk.  In order to do direct I/O with readv(),
 *		        each field must be rounded to a page size (16k for R10k)
 *	                and loaded into page aligned buffers.
 *		       
 *		        This thread reads the fields into the "tail" (_rvtail)
 *		        of the _readvBufs queue.  The data then flows to:
 *					|
 *					|
 *				      \ | /
 *				       \|/
 *					V
 *	2. Expander() - This thread grabs fields from the "head" (_rvhead) of the
 *			_readvBufs queue.  Each field is expanded to YUVA, 8 bits
 *			per component.  The expanded result is palced at the "tail"
 *			(_cvtail) of the _converterBuf queue.  From here data flows to:
 *                                      |
 *                                      |       
 *                                    \ | /     
 *                                     \|/      
 *                                      V       
 *	3. PutField() - This thread grabs YUVA fields from the "head" (_cvhead) of the
 *			converterBufs queue and loads it into texture memory.  The color
 *			matrix is loaded, so loading the field to texture memory also
 *			converts it to RGBA.  Once loaded into texture meory, a texture map
 *			quad is drawn. 
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>
#include <sys/signal.h>
#include <sys/dmcommon.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <vl/vl.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16 
#define FILL_HIGH 16
#define FILL_LOW 4
#define BILL 1000000000.0


/* All global variables in this file are preceded by an "_".  */

/* Matrix stored in column major order.
 * Data is stored CbYCr where Cb = U, Cr = V.
 */
GLfloat _colorMat[16] = { 1.596027,   -0.812968,    0.000000,     0.0,
                          1.164383,    1.164384,    1.164384,     0.0,
                          0.000000,   -0.391762,    2.017232,     0.0,
                         -0.870787,    0.529591,   -1.081390,     0.0 };

/* Misc. declarations.	*/
int _winx = 0;		/* x dimension of gfx window.		*/
int _winy = 0;		/* y dimension of gfx window.		*/
int _CurField = 0;	/* Current field being processed.	*/
int _rgbSize;		
int _dbuffer;		/* Double buffering flag.		*/
int _iovcnt = 12;	/* Number of reads per readv call.	*/
int _IamPlaying = 0;    /* Am I playing or am I stopped?	*/
fileHeader_t _f;

/* Buffering declarations  */
char *_readvBufs;
char *_converterBufs;
int  _num_rvbufs; 
int  _num_cvbufs; 
volatile int _rvhead;	/* Head pointer for _readvBufs queue.	*/
volatile int _rvtail;	/* Tail pointer for _readvBufs queue.	*/
volatile int _cvhead;	/* Head pointer for _converterBufs  queue.	*/
volatile int _cvtail;	/* Tail pointer for _converterBufs queue.	*/

/* Gfx declarations  */
Display   *_dpy;
GLXContext _context;
Widget     _glw;

/* Texture declarations. */
int _tex_width;
int _tex_height;
int _tex_format = GL_RGBA;
int _texA;

/* Procdure declarations  */
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void initGL();
void PutField();
void ReadIt();
void Expander();

main(int argc, char *argv[])
{
    int i,n;
    int packing;
    char *Fname;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"doit", doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n";
    
    _dbuffer = 1;

    for (i=1; i < argc; i++) {
	if (strcmp (argv[i], "-single") == 0) {
	    _dbuffer = 0;
	}
        else if (strcmp (argv[i], "-gfx") == 0) {
            ++i;
            _winx = atoi(argv[i]);
            ++i;
            _winy = atoi(argv[i]);
        }
        else if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
	else {
	    Fname = argv[i];
	}
    }

    fileInit(Fname, &_f, TRUE);
    ERROR ((_f.packing != VL_PACKING_YVYU_422_8), "Must be 422 data\n");


    /* Cheat a liitle here, it makes the reading a lot simpler. */
    if (_f.Nfields < _iovcnt) {
	_iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }
	
    /* Set some state dependent on input from user and file attributes.*/
    if (_winx == 0) {
        _winx = _f.uSize;
        _winy = _f.vSize;
    }

    _num_rvbufs = 3*_iovcnt; 
    _num_cvbufs = _iovcnt;
    _rgbSize    = 2*_f.readSize;

    _readvBufs =     memalign(_f.headerSize, _num_rvbufs*_f.readSize);
    _converterBufs = memalign(_f.headerSize, _num_cvbufs*_rgbSize);

    fprintf(stderr, "Playing  %d fields of %dx%d using a read vector of length %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Padded read size (bytes) = %d \n", _f.readSize);
    fprintf(stderr, "Gfx window size = %dx%d\n", _winx, _winy);


    /* Do that nasty Motif stuff then head off to the Xt main loop.	*/

    toplevel = XtAppInitialize(&app_context, "4DgiftsGlx",
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv,
			       NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight, _winy);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);
    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);
}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the initialize call back routine for our widget.   This
     * is called automatically from the Xt loop.
     */

    Arg args[18];
    XVisualInfo *vi;
    int val;

    _dpy = XtDisplay(_glw);

    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(w, args, 1);

    /* create a context */
    _context = glXCreateContext(XtDisplay(_glw), vi, 0, GL_TRUE);
    GLwDrawingAreaMakeCurrent (_glw, _context);

    _tex_width = 1024;
    _tex_height = (_f.vSize <= 256) ? 256 : 512;

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_format, 0);

    /* init viewport */
    glViewport(0, 0, _winx, _winy);

    /* set up transformation matrix */
    glMatrixMode(GL_COLOR);
    glLoadMatrixf(_colorMat);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glColor4f(1.0, 1.0, 1.0, 1.0);
    glDisable(GL_DITHER);

    (void)BuildGfxLists(1, _f.uSize, _f.vSize, _tex_width, _tex_height,1);

    /* enable texturing */
    glEnable(GL_TEXTURE_2D);
    glBindTextureEXT(GL_TEXTURE_2D,_texA);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the expose callback routine, called whenever an expose happens.
     * Note, if real time play is happening then this routine cannot try to 
     * grab the gfx context.
     */

    static int firstExpose = 1;

    if (firstExpose) {
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy,XtWindow(_glw));
 	firstExpose = 0;
        
        fprintf(stderr, "\n");
	fprintf(stderr, "*****> Click in window to start play, click again to stop, etc.\n");
        fprintf(stderr, "\n");
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routine is called each time the left mouse button is clicked.
     * If things are stopped then the three play threads are sproc'd. 
     * If things are playing then the the threads are told to exit by 
     * setting _IamPlaying to 0.
     */

    int readPID, putPID, expPID;

    if (!_IamPlaying) {  /*Go and play */
	_IamPlaying = 1;
	_rvhead = _rvtail = 0; 
	_cvhead = _cvtail = 0; 
        GC (glXMakeCurrent(_dpy, None, NULL));  
        readPID = sproc((void (*)(void *))ReadIt,   PR_SADDR | PR_SFDS);
        expPID  = sproc((void (*)(void *))Expander, PR_SADDR | PR_SFDS);
        putPID  = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);
    }
    else { 
        _IamPlaying = 0;
        while (_IamPlaying != -3) sginap(1);  /* Wait for three threads */
        _IamPlaying = 0;
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));

        /* Make sure _CurField is pointing at an even field	*/
	if ((_CurField % 2) != 0) {
	    _CurField++;
	    if (_CurField >= _f.Nfields) _CurField = 0; 
        }
    }
}

void PutField()
{
    int tmp;
    int x,y;
    int count = 0;
    char *dataPtr;
    unsigned long long then, now;
    double tdiff;

    /* This thread exits second and sets _IamPlaying to -2.  */

    /* This thread takes YUVA fields from the head of _converterBufs
     * (put there by the thread Expander) and loads the result into
     * texture memory.  It then draws a quad with the video texture
     * mapped on top of it.
     */

    /* Exit if my parent dies */
    setcpu(1);
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);
    
    GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context)); /* Grab the context	*/

    x = _tex_width  - _f.uSize;
    y = _tex_height - _f.vSize;

    while (_IamPlaying != -1) {

        if (_cvhead == _cvtail) {
            /* fprintf(stderr, " _converterBufs queue is empty\n"); */
	    sginap(1);
	    continue;
        }

        /* Grab the next valid buffer from the head of the list	*/ 
        dataPtr = _converterBufs + _cvhead*_rgbSize;

        glTexSubImage2DEXT(GL_TEXTURE_2D, 0, x, y, _f.uSize, _f.vSize, _tex_format, 
                           GL_UNSIGNED_BYTE, dataPtr);
        glCallList(1);

        if (count == 0) dmGetUST(&then);
        count++;

	if (_dbuffer) {
	     glXSwapBuffers(_dpy,XtWindow(_glw));
	}

        tmp = _cvhead + 1;
        if (tmp == _num_cvbufs) tmp = 0;
	_cvhead = tmp;

    }  /* End of while (IamPlaying) */ 
    dmGetUST(&now);

    tdiff = now - then;
    tdiff = tdiff/BILL;
    fprintf(stderr, "Draw time = %f fps = %f\n", tdiff, ((double)(count))/tdiff);

    GC (glXMakeCurrent(_dpy, None, NULL));  /* Surrender gl context */
    _IamPlaying--;		            /* Tell doit() all is done */	
    _exit(0);
}

void ReadIt()
{
    /* This thread exits first and sets _IamPlaying to -1.  */

    /* This thread reads fields of YUV 4:2:2 data from disk
     * into the tail of _readvBufs.  The thread Expander()
     * will grab data from the head of _readvBufs and expand 
     * it to YUV 4:4:4. 
     */
    struct iovec iov[MAX_IOV];
    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;

    /* Exit if my parent dies */
    setcpu(0);
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    lseek64 (_f.fd, (off64_t)_f.headerSize+(_CurField*_f.readSize), SEEK_SET);

    while (_IamPlaying) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Now make sure there is room in _readvBufs to store 
         * "fieldsToRead" fields (which the consumer PutField() will later
         * display. 
         */
	do {
	    full = 0;
	    /* Check for the first full condition.  If the incremented
	     * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
	    tmphead = _rvhead;
	    tmptail = _rvtail + fieldsToRead;
            if (tmptail == _num_rvbufs) tmptail = 0;
            full = (tmptail == tmphead); 

	    /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
	    full |= ((tmphead > _rvtail) && (tmphead < _rvtail + fieldsToRead));

	    if (full) {
		/* fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail); */ 
		sginap(1);
	    }
	} while (full);

        /* Load the address of the dmbuffers into the readv vector "iov"	*/
        tmptail = _rvtail;
        for (i=0; i < fieldsToRead; i++) {
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = _readvBufs + tmptail*_f.readSize; 
	    tmptail++;
        }

        /* Read the fields.		*/
        OC (readv(_f.fd, iov, fieldsToRead));

        /* Inc. _rvtail, indicating more vaild data is in _readvBufs.	*/
        if (tmptail == _num_rvbufs) tmptail = 0;
        _rvtail = tmptail;
	
	/* Check for file wrap.		*/
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass through file\n");
        }

   }  /* end while(IamPLaying) */

   _IamPlaying--;
   _exit(0); 
}

void Expander()
{
    /* This thread exits last (third) and sets _IamPlaying to -3 */

    /* This routine takes fields from the head of _readvBufs (yuv 4:2:2) 
     * and expands the yuv field into yuv 4:4:4.  The expanded result is 
     * placed at the tail of the _converterBufs.  The PutField thread 
     * takes from the head of _converterBufs and draws the final image 
     * to the screen.
     */

    int tmphead;
    int tmptail;
    char *yuvPtr;
    char *rgbPtr;

    /* Exit if my parent dies */
    setcpu(0);
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);
    
    while (_IamPlaying != -2) {

        if (_rvhead == _rvtail) {
	    /* fprintf(stderr, "_readvBufs queue is empty.\n"); */
            sginap(1);
	    continue;
        }

        tmptail = _cvtail + 1;
        if (tmptail == _num_cvbufs) tmptail = 0; 
        if (tmptail == _cvhead) {       
	    /* fprintf(stderr, "No room in converter queue\n"); */
            sginap(1);
            continue;
        }

        /* Grab the next valid buffer from the head of the list	*/ 
        yuvPtr  = _readvBufs     + _rvhead*_f.readSize;
        rgbPtr  = _converterBufs + _cvtail*_rgbSize;

	expandyuv(_f.uSize*_f.vSize, yuvPtr, rgbPtr);

        tmphead = _rvhead + 1;
        if (tmphead == _num_rvbufs) tmphead = 0;
        _rvhead = tmphead;

        _cvtail = tmptail;

    }  /* End of while (IamPlaying) */ 

    _IamPlaying--;
    _exit(0);
}
