/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vstreamfx_u -Plays two uncompressed streams from memory, performs an effect
 *	        and sends the result to video-out and a screen node. 
 *
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-gfx_half>      	- Display half height field (default is video
 *                                frame).
 *     file1			- file for first stream (created with vgrab_u). 
 *     file2			- file for 2nd stream   (created with vgrab_u). 
 *				  Both files must contain fields, not frames. 
 *				  Field type capture is the default for vgrab_u. 
 *
 * Each stream is represented by the first field from the corresponding input
 * file.  By limiting input to one field per stream (in memory), this program 
 * can be used for prototyping video effects for uncompressed streams WITHOUT 
 * requiring storage capable of sustaining 60-80 MB/sec.
 *
 * The input files may contain packed rgb (24 bits per pixel), rgba (32 bits per
 * pixel) or packed yuv (24 bits per pixel).  Since all effects are calculated
 * in an off screen pbuffer, both yuv and rgb data can be handled.  The video 
 * screen node gurantees that the yuv data looks correct on the gfx monitor. 
 *
 * This program has been run on an Octane MXI with a ProVideo board.
 *
 * How to run the program:
 * =======================
 * When you run the program, a blank window appears.  Click once with the left 
 * mouse to start play, click again with the same button to stop play.  The 
 * initial effect performed is a blend.  Click at any time with the middle mouse
 * button to change the effect being played.  See the routine "loadIt" for 
 * effect choices. After entering the new effect, play resumes immediately.  
 * The effect calculation routines are inside common/fxcode.c 
 * 
 *
 * Overview of program flow: 
 * =========================
 * Each time you click the button and start play, the effect thread retrieves 
 * the video fields from memory, sends them to gfx (texture memory or 
 * frame buffer), reads back the effect, and then sends the result to video-out 
 * and the screen node. SO with the exception of disk IO, this program models
 * all the other aspects of performing uncompressed, real time video effects. 
 * 
 * Here are the thread descriptions:
 * =================================
 *    1. UI thread - This threasd sits in the Xt main loop waiting for a
 *		     button press.  If a left button press, then the UI enters
 *		     the routine doit().  Inside doit() play is either stopped
 *		     or started depending upon the current play state.  If the
 *		     middle mouse button is pressed, then loadIt() is entered.
 *		     The user is prompted for a new effect, and play is resumed.
 *		     The UI thread directs the reader and effect threads to 
 *		     exit whenever play is stopped.  Look inthe routine loadIt()
 *		     for effect choices.
 *
 *
 * 2. PlayEffect()- This is the thread which computes the effect and sends the
 *		    result to video-out.
 *
 *		    The effect is calculated in an off-screen pbuffer (see 
 *		    _pbuffer).  Usually one stream is drawn as backgroud via 
 *		    glDrawPixles and the other is texture mapeed.  The effect 
 *		    "depth" texture maps both streams.  All effect calculations 
 *		    are field based. The effect is read back from _pbuffer via 
 *		    glReadPixels and sent to video-out and a screen node (this
 *		    node is tied to the on-screen widget _glw).
 *
 *		    The effect calculation code is in common/fxcode.c 
 *
 * Notes:
 * ------
 * For best results when drawing to the Octane Video screen node, make sure to do: 
 *              /usr/gfx/setmon -Fe 1280x1024_59 (NTSC).
 *			    OR
 *		/usr/gfx/setmon -Fe 1280x1024_49 (PAL).
 *
 * Sending video to a screen node means that the video hardware is inserting
 * video pixels onto the pixel bus just before the pixels arrive at the DACS
 * (which drive the monitor).  So there is no gfx control over the pixles going
 * to the monitor.
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ulocks.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/time.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/cl.h>
#include <vl/vl.h>
#include <vl/dev_mgv.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define NUM_DMBUFFS 12
#define FILL_CHECK 32


typedef enum {FX_NONE, FX_BLEND, FX_WIPE, FX_PIP, FX_CROSS,
              FX_BOXBLEND, FX_SQUEEZE, FX_PERSPECTIVE, FX_DEPTH} FxType_t;

/* All globals preceded by an "_" */

/* Video globals 			*/
VLServer _svr;
VLPath _path;
VLNode _src;
VLNode _devNode;
VLNode _voutdrn;
VLNode _drn;
VLNode _csc;
int    _packing;

/* DMbuffer globals 			*/	
DMbufferpool _outpool;      /* dm buffer pool used to drive video output. */
int          _outpoolfd;    /* File descriptor for _outpool.              */

/* File definitions 			*/
int _winx;	/* Gfx window width */
int _winy;	/* Gfx window width */
int _vlFieldSize; /* Size in bytes of video field */

char *_sourceBuffer1;  /* Pointer to field for first stream */
char *_sourceBuffer2;  /* Pointer to field for second stream */
fileHeader_t _f1;      /* File info for first stream */
fileHeader_t _f2;      /* File info for second stream */

/* Play control variables */
int _IamPlaying = 0;

/* OpenGL defines. 			*/
int _dbuffer = 0;
int _tex_width;                          /* texture size */
int _tex_height;                         /* texture size */
int _tex_def    = GL_RGB;		 /* Can only decompress into RGBA or 4:2:2 */
int _tex_format = GL_RGB;		 /* Can only decompress into ABGR format   */
int _readformat = GL_RGB;		 /* Always read back RGB to send to video  */
unsigned int _texA;                      /* texture name 			   */
FxType_t _fx = FX_BLEND;		 /* default effect type.		   */

Display *_dpy;
Widget _glw;				 /* Used to disply screen node.		   */
GLXContext _glw_context;

GLXPbufferSGIX _pbuffer;		 /* pbuffer where effect is calculated.	   */
GLXContext _pb_context;

/* Procedure definitions.	*/
GLXFBConfigSGIX getFBConfig(Display *, int);
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void motionCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void loadit(Widget , XEvent *, String *, Cardinal *);
void initVL();
void PlayEffect();
void GetInputFields(char **f1, char **f2);
void SendToVideo (int count);

/* Misc. external procedure definitions */
int  BuildVideoLists(int start, int u, int v, int tw, int th);
CLhandle clInitDecompressor(int u, int v, int wx, int wy, int qf, int rgb);
void createTexture(unsigned int *name, int width, int height, int format, int mip);
int GetVLFormat(int packing);
void SetVideoOutTiming(VLServer svr, VLPath path, VLNode node, int width, int height);
void setscheduling();

main(int argc, char *argv[])
{
    char *f1 = NULL;
    char *f2 = NULL;
    int i,n;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"loadit", loadit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:loadit()\n";

    _winx = 1; _winy = 2;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-gfx_half") == 0) {
            _winx = 2; _winy = 1;
        }
        else if (f1 == NULL) {
	    f1 = argv[i];
        }
        else if (f2 == NULL) {
	    f2 = argv[i];
        }
    }

    /* If no second file specified, then no effects and change exit counters */
    if (f2 == NULL) ERROR(1, "Need two input files"); 

    /* Open up first stream */
    fileInit(f1, &_f1, TRUE); 

    /* Open up second stream */
    fileInit(f2, &_f2, TRUE); 

    ERROR ((_f2.packing != _f1.packing), "Files not same packing\n"); 
    ERROR ((_f1.uSize != _f2.uSize), "Mismatch in file widths");
    ERROR ((_f1.vSize != _f2.vSize), "Mismatch in file heights");

    if ((_f1.packing != VL_PACKING_ABGR_8)  && 
        (_f1.packing != VL_PACKING_BGR_8_P) && 
        (_f1.packing != VL_PACKING_UYV_8_P) ) {
    	ERROR(1,"Bad packing");
    }


    /* Initialize miscellaneous globals */
    _winx  = _f1.uSize/_winx;
    _winy  = _f1.vSize*_winy;
    _vlFieldSize = _f1.uSize*_f1.vSize*3; 
    _packing = _f1.packing; 

    if (_packing == VL_PACKING_ABGR_8) {
	_tex_format = GL_RGBA;
	_tex_def    = GL_RGBA;
        _packing    = VL_PACKING_BGR_8_P; /* always read GL_RGB */
    }
    _sourceBuffer1 = memalign(_f1.headerSize, _f1.readSize);
    _sourceBuffer2 = memalign(_f2.headerSize, _f2.readSize);

    /* Read the first field from each stream into memory */
    lseek (_f1.fd, _f1.headerSize, SEEK_SET);
    OC ( read(_f1.fd, _sourceBuffer1, _f1.readSize));

    /* Read the second field from each stream into memory */
    lseek (_f2.fd, _f2.headerSize, SEEK_SET);
    OC ( read(_f2.fd, _sourceBuffer2, _f2.readSize));

    /* Initialize video hardware */
    initVL();

    /* Do all that nasty Xt stuff	*/
    toplevel = XtAppInitialize(&app_context, argv[0],
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv, NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    /* Create the on-screen widget that will be used for the screen
     * node.
     */
    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], GLwNdepthSize, 24); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight, _winy);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);

    XtAddEventHandler(toplevel, SubstructureNotifyMask, FALSE, (XtEventHandler)motionCB, toplevel);


    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the widget initialize call back for the visible
     * widget _glw.  This routine also sets up _pbuffer.
     */

    Arg args[18];
    XVisualInfo *vi;
    GLXFBConfigSGIX config;
    int val;
    int attrib[] = {GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};
    /* Get a 24 bit frame buffer configuration */
    _dpy = XOpenDisplay( ":0" );
    config = getFBConfig(_dpy, 24);
    ERROR((config == NULL), "Can't get FB");

    /* Use this fb config. to set up _pbuffer        */
    GNC (_pbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _f1.uSize, _f1.vSize, attrib));
    GNC (_pb_context = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    GC  (glXMakeCurrent(_dpy, _pbuffer, _pb_context));

    _tex_width = 1024; 
    _tex_height = (_f1.vSize <= 256) ? 256 : 512; 

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    /* create texture A */
    createTexture(&_texA, _tex_width, _tex_height, _tex_def, 0);

    /* init viewport */
    glViewport(0,0,_f1.uSize,_f1.vSize);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    
    BuildVideoLists(1, _f1.uSize, _f1.vSize, _tex_width, _tex_height);
    glDisable(GL_DITHER);

    /* Now set up the widget used for the screen node.	*/
    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(_glw, args, 1);

    /* create a visual context */
    GNC (_glw_context = glXCreateContext(_dpy, vi, 0, GL_TRUE));
    GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));

    glViewport(0,0,_winx,_winy);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glClearColor(0.0,0.0,0.0,1.0);
    glColor4f(1.0, 1.0, 1.0, 1.0);

    glDisable(GL_DITHER);
    glRasterPos2f(-1.0, 1.0);
    glPixelZoom(1.0, -1.0);

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }

}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    int x,y;
    Window dummyWin;
    VLControlValue val;
    static int first = 1;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    if (first) {

        /* If it is the very first expose, then tell the vl
         * that _glw is to be used fro the screen node.
         */
        GC  (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));
        glClear(GL_COLOR_BUFFER_BIT);

        /* Declare window that gets data for screen node */
        val.intVal = XtWindow(_glw);
        VC(vlSetControl(_svr, _path, _drn, VL_WINDOW, &val));

        /* Set up scale factors.	*/
        val.fractVal.numerator = 1;
        val.fractVal.denominator =  _f1.uSize/_winx;
        VC(vlSetControl(_svr, _path, _drn, VL_ZOOM, &val));

        val.xyVal.x = _winx;
        val.xyVal.y = _winy;
        VC(vlSetControl(_svr, _path, _drn, VL_SIZE, &val));

        val.xyVal.x = x;
        val.xyVal.y = y;
        VC(vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val));

        fprintf(stderr, "\n");
        fprintf(stderr, "*** Click left mouse button to start/stop play\n");
        fprintf(stderr, "*** Click midle mouse button to change effect\n");
        fprintf(stderr, "\n");

	first = 0;
    }
}

void motionCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Window dummyWin;
    int x,y;
    VLControlValue val;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    val.xyVal.x = x;
    val.xyVal.y = y;
    vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val);
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    int allDone;

    /* Routine to START/STOP play. 
     * This routine is entered directly when the left mouse button is clicked
     * or from loadit() when changing effects.
     */

    if (_IamPlaying) {
        /* Then stop playing.	*/
        _IamPlaying = 0;
        allDone = -1; 
        
        /* Wait for both readers and the effect thread to exit. */
        while (_IamPlaying != allDone) sginap(1);
        _IamPlaying = 0;
 
        VC (vlEndTransfer(_svr, _path));
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _glw_context));

	glClear(GL_COLOR_BUFFER_BIT);
	if (_dbuffer) glXSwapBuffers(_dpy, XtWindow(_glw));

    }
    else {
        /* Then start playing.	*/
        _IamPlaying = 1;

        /* Surrender the graphics context.      */
        GC (glXMakeCurrent(_dpy, None, NULL));

	/* Start the effect play back thread.	*/
        sproc((void (*)(void *))PlayEffect, PR_SADDR | PR_SFDS);
   }
}

void loadit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routiine is entered whenever a new effect is loaded.  This
     * is done by pushing the middle mouse button.
     */ 
    char etype[12];

    /* Stop play */
    if (_IamPlaying) doit(_glw, NULL, NULL, NULL);
    if (_fx == FX_NONE) return;

    fprintf(stderr, "Enter effect: ");
    gets(etype);

    if (strcmp (etype,      "blend") == 0)    _fx = FX_BLEND;
    else if (strcmp (etype, "pip") == 0)      _fx = FX_PIP;
    else if (strcmp (etype, "wipe") == 0)     _fx = FX_WIPE;
    else if (strcmp (etype, "cross") == 0)    _fx = FX_CROSS; 
    else if (strcmp (etype, "boxblend") == 0) _fx = FX_BOXBLEND; 
    else if (strcmp (etype, "squeeze") == 0)  _fx = FX_SQUEEZE; 
    else if (strcmp (etype, "persp") == 0)    _fx = FX_PERSPECTIVE; 
    else if (strcmp (etype, "depth") == 0)    _fx = FX_DEPTH; 
    else _fx = FX_BLEND;

    /* Resume play */
    doit(_glw, NULL, NULL, NULL);
}

void PlayEffect()
{
    /* This is the effect calculation thread sproc'd from doit()
     * It jumps off (for the entire current play duration) into
     * one of the effect routines.
     */

    int ec = 0; 

    GC  (glXMakeCurrent(_dpy, _pbuffer, _pb_context));

    switch (_fx) {
	case FX_NONE:        
	    break;
	case FX_BLEND:       
            do_blend(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
	             _tex_format, _texA, &_IamPlaying,ec);       
	    break;
	case FX_WIPE :       
	    do_wipe(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                    _tex_format, _texA, &_IamPlaying,ec);        
            break;
	case FX_PIP  :       
            do_pip(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                   _tex_format, _texA, &_IamPlaying,ec);         
            break;
	case FX_CROSS:       
            do_cross(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
	case FX_BOXBLEND:    
            do_boxblend(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                        _tex_format, _texA, &_IamPlaying,ec);    
            break;
	case FX_SQUEEZE:     
            do_squeeze(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                       _tex_format, _texA, &_IamPlaying,ec);     
            break;
	case FX_PERSPECTIVE: 
            do_perspective(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                           _tex_format, _texA, &_IamPlaying,ec); 
            break;
	case FX_DEPTH:       
            do_depth(_f1.uSize, _f1.vSize, _tex_width, _tex_height, 
                     _tex_format, _texA, &_IamPlaying,ec);       
            break;
    }

    /* Surrender the graphics context.  */
    glXMakeCurrent(_dpy, None, NULL);

    --_IamPlaying;
    _exit(0);
}

void initVL()
{
    /* Set up video hardware.  The video path set up
     * by this routine has a VL_MEM  source node and both
     * a VL_VIDEO and VL_SCREEN for drain nodes.  Using the
     * screen node allows the app. to display things on the gfx 
     * monitor using no additional gfx horspower.
     *
     * Since the compression card only decompresses into GL_ABGR_EXT,
     * a VL_CSC node has to be placed between the source and drain nodes
     * to convert the RGB to YUV for video display.
     */ 

    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;
    int format;
    int memnode;
    int n;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact video device");

    format = GetVLFormat(_packing); 
    memnode = (_packing == VL_PACKING_YVYU_422_8) ? VL_ANY : VL_MGV_NODE_NUMBER_MEM_VGI_DL;

    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, memnode));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_SCREEN, VL_ANY));
    VC (_voutdrn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_devNode = vlGetNode(_svr, VL_DEVICE, 0, VL_ANY));

    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlAddNode(_svr, _path, _voutdrn));
    VC (vlAddNode(_svr, _path, _devNode));

    if (_packing != VL_PACKING_YVYU_422_8) {
        VC (_csc = vlGetNode(_svr, VL_INTERNAL, VL_CSC, VL_MGV_NODE_NUMBER_CSC));
        VC (vlAddNode(_svr, _path, _csc));
        fprintf(stderr, "Added CSC node\n");
    }

    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, _devNode, _f1.uSize, _f1.vSize);

    val.intVal = format;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = _packing;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    if (_packing != VL_PACKING_YVYU_422_8) {
	/* Set up input csc controls */
	val.intVal = format;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_RANGE, &val));
	val.intVal = _packing;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_PACKING, &val));

	/* Set up output csc controls */
	val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_RANGE, &val));
	val.intVal = VL_PACKING_YVYU_422_8;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_PACKING, &val));

	/* Turn on or off CSC constant hue algorithm */
	val.boolVal = 0;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_CONST_HUE, &val));

	/* Turn on or off alpha correction in the alpha channel */
	val.boolVal = 0;
	VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_ALPHA_CORRECTION, &val));
    }

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "NUM_DMBUFFS" buffers
     * each containing a video field of bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,NUM_DMBUFFS, _vlFieldSize, DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.    */
    DC( dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS) );

    /* Create the dmbuffer pool used for video out.     */
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _vlFieldSize));

    dmParamsDestroy(p);
}

void GetInputFields (char **f1, char **f2)
{
    /* This routine is called by the effect calculation routines in
     * common/fxcode.c.  It returns pointers to the next pair of input
     * fields to process.
     */

    *f1 = _sourceBuffer1;
    *f2 = _sourceBuffer2;
}

void SendToVideo (int count)
{
    /* This routine is called by the effect calculation routines.
     * It reads back the effect from the frame buffer and sends the
     * result to video-out.
     */

    unsigned char *readBuffer;
    DMbuffer buffer;
    static stamp_t oldmsc;
    stamp_t newmsc;
    fd_set fds;

    /* Wait for a dmbuffer to send to video.		   */
    FD_ZERO(&fds);
    FD_SET(_outpoolfd, &fds);
    OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

    /* Read back the effect into the dmbuffer.		*/
    DC (dmBufferAllocate(_outpool, &buffer));
    readBuffer = dmBufferMapData(buffer);
    glReadPixels(0, 0, _f1.uSize, _f1.vSize, _readformat, GL_UNSIGNED_BYTE, readBuffer);

    /* Send to video out.				*/
    VC (vlDMBufferPutValid(_svr, _path, _src, buffer) );
    dmBufferFree(buffer);

    if (count == 5) {
        VC (vlBeginTransfer(_svr, _path, 0, NULL));
    }

    /* Check and see if we are dropping.		*/
    newmsc = vlGetFrontierMSC(_svr, _path, _src);
    if ((count > FILL_CHECK) && (newmsc != oldmsc + 1)) {
	fprintf(stderr, "MSC drop old = %lld next = %lld\n", oldmsc, newmsc);
    }
    oldmsc = newmsc;
}
