/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <limits.h>
#include <GL/gl.h>
#include <GL/glx.h>
#include <vl/vl.h>

/* Create a texture and put its name into *texname. */
void
createTexture(unsigned int *texname, int tex_width, int tex_height, int tex_format, int mip)
{
	int i,texsize;
        int width, height;
	unsigned char *texbuf;

	/* determine size of LOD 0 */
	texsize = tex_width * tex_height * (tex_format == GL_RGBA ? 4 : 3);

	/* init texture buffer */
	texbuf = (unsigned char *)malloc(texsize);
	for(i = 0; i < texsize; i++) texbuf[i] = 0x00;

	/* generate texture name */
	glGenTexturesEXT(1,texname);

	/* bind to the texture */
	glBindTextureEXT(GL_TEXTURE_2D,*texname);

	/* set up filtering */
	if (mip) {
	    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	}
	else {
	    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
        }

	/* Always magnify linearly, mipmap good for shrinking only. */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);


	/* Now we have to actually initialize texture memory.  If doing
         * mip-mapping, then the impact video card sends 10 subimages
         * for use with filtering, (each subimage is called an LOD = level of 
         * detail).  If doing linear then only one LOD is required.
	 */

        /* create LOD's for this texture */
        if (mip) {
	    width = tex_width; height = tex_height;
	    for(i = 0; i < 10; i++) {
		glTexImage2D(GL_TEXTURE_2D,i,tex_format, width,height,0,tex_format, GL_UNSIGNED_BYTE,texbuf);
		width = width>>1; 
                height= height>>1;
		if (width == 0) width = 1; 
		if (height == 0) height = 1;
	    }
	}
        else {
	    /* create LOD 0 */
	    glTexImage2D(GL_TEXTURE_2D,0,tex_format,tex_width,tex_height, 0,tex_format,GL_UNSIGNED_BYTE,texbuf);
	}

	/* free up texture buffer */
	free(texbuf);
}


GLXFBConfigSGIX getFBConfig(Display *dpy, int d)
{
    GLXFBConfigSGIX config = NULL;
    GLXFBConfigSGIX* configs; 
    int configCount;
    int colorSize, alphaSize;
    int i;
    int *params;
    
    int params8[] = {
        GLX_RENDER_TYPE_SGIX,       GLX_RGBA_BIT_SGIX,
        GLX_DRAWABLE_TYPE_SGIX,     GLX_PBUFFER_BIT_SGIX,
        GLX_RED_SIZE,               8,
        GLX_GREEN_SIZE,             8,
        GLX_BLUE_SIZE,              8,
        GLX_ALPHA_SIZE,             8,
        None,
    };
    
    int params5[] = {
        GLX_RENDER_TYPE_SGIX,       GLX_RGBA_BIT_SGIX,
        GLX_DRAWABLE_TYPE_SGIX,     GLX_PBUFFER_BIT_SGIX,
        GLX_RED_SIZE,               5,
        GLX_GREEN_SIZE,             5,
        GLX_BLUE_SIZE,              5,
        GLX_ALPHA_SIZE,             1,
        None,
    };
    
    fprintf(stderr, "USING DEPTH %d\n", d);
    if (d == 24) {
        params = params8;
        colorSize = 8;
        alphaSize = 8; 
    }
    else {
        params = params5;
        colorSize = 5;
        alphaSize = 1; 
    }
    
    configCount = 0;
    configs = glXChooseFBConfigSGIX( dpy, DefaultScreen(dpy), params, &configCount);
    
    if (configCount == 0) {
        config = NULL;
        fprintf(stderr, "Could not find an FBConfig for depth %d\n", d);
    }
    
    for ( i = 0;  i < configCount;  i++ ) {
        int value;
        glXGetFBConfigAttribSGIX(dpy, configs[i], GLX_RED_SIZE, &value);
        if ( value != colorSize )  continue;
        glXGetFBConfigAttribSGIX(dpy, configs[i], GLX_GREEN_SIZE, &value);
        if ( value != colorSize )  continue;
        glXGetFBConfigAttribSGIX(dpy, configs[i], GLX_BLUE_SIZE, &value);        
        if ( value != colorSize )  continue;
        glXGetFBConfigAttribSGIX(dpy, configs[i], GLX_ALPHA_SIZE, &value);
        if ( value != alphaSize )  continue;
        break;
    }
    
    if ( i != configCount ) {
        config = configs[i];
    }
    return config;
}

int BuildVideoLists(int start, int uSize, int vSize, int tex_width, int tex_height)
{
    /* This routine builds the tiles used to texture map the
     * video.  
     */

    float t0[2],t1[2],t2[2],t3[3];          /* texture coordinates */
    float v0[2],v1[2],v2[2],v3[2];          /* vertice coordinates */
    float xoff, yoff;
    float midx, midy;

    /* The way video gets drawn to the screen is to do draw
     * a rectangle that corresponds to the size of the window.  With
     * texture mapping enabled, each pixel of the rectangle gets its
     * color value from the texture map (which just happens to contain
     * the field of video we want to draw).  What makes this all interesting
     * is that you can vary the size/orientation of the rectangle and the
     * texture mapping hardware filters the video image to fit the rectangle.
     */

    /* init vertice coordinates.  These coordinates match the default
     * projection space of -1 to 1 in both x and y.  In other words
     * these coordinates will exactly cover the entire window on the gfx 
     * monitor with a rectangle.
     */
    v0[0] = -1.0; v0[1] = -1.0;
    v1[0] =  1.0; v1[1] = -1.0;
    v2[0] =  1.0; v2[1] =  1.0;
    v3[0] = -1.0; v3[1] =  1.0;
    
    /* Texture space goes from (0,0) to (1,1).  So a 720x243
     * field is mapped onto this floating point range.  Each corner of
     * the rectangle needs to be assigned a location in texture space.
     *
     * Remember that textures are created in powers of 2, so you don't want to
     * take the entire texture map (i.e a 720x243 video field has a 1024x256 
     * texture space.  
     *
     * Due to Impact bug, we load texture so it abuts rhs of texture space.
     * glTexSubImage2DEXT stes the offsets to handle this.
     */ 
    
    xoff = (tex_width  - uSize)/(float)tex_width;
    yoff = (tex_height - vSize)/(float)tex_height; 
    midx = xoff + (1.0 - xoff)/2.0;
    midy = yoff + (1.0 - yoff)/2.0;

    t0[0] = xoff; t0[1] = yoff;
    t1[0] = 1.0;  t1[1] = yoff;
    t2[0] = 1.0;  t2[1] = 1.0;
    t3[0] = xoff; t3[1] = 1.0;

    glNewList(start, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* First tile */
    t0[0] = xoff; t0[1] = yoff;
    t1[0] = midx; t1[1] = yoff;
    t2[0] = midx; t2[1] = midy;
    t3[0] = xoff; t3[1] = midy;

    v0[0] = -1.0; v0[1] = -1.0;
    v1[0] =  0.0; v1[1] = -1.0;
    v2[0] =  0.0; v2[1] =  0.0;
    v3[0] = -1.0; v3[1] =  0.0;

    glNewList(start + 1, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Second tile */
    t0[0] = midx; t0[1] = yoff;
    t1[0] = 1.0;  t1[1] = yoff;
    t2[0] = 1.0;  t2[1] = midy;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  0.0; v0[1] = -1.0;
    v1[0] =  1.0; v1[1] = -1.0;
    v2[0] =  1.0; v2[1] =  0.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 2, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Third tile */
    t0[0] = 1.0; t0[1] = midy;
    t1[0] = 1.0; t1[1] = 1.0;
    t2[0] = midx; t2[1] = 1.0;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  1.0; v0[1] =  0.0;
    v1[0] =  1.0; v1[1] =  1.0;
    v2[0] =  0.0; v2[1] =  1.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 3, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Fourth tile */
    t0[0] = midx; t0[1] = 1.0;
    t1[0] = xoff; t1[1] = 1.0;
    t2[0] = xoff; t2[1] = midy;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  0.0; v0[1] =  1.0;
    v1[0] = -1.0; v1[1] =  1.0;
    v2[0] = -1.0; v2[1] =  0.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 4, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    return 5;
}

int BuildGfxLists(int start, int uSize, int vSize, int tex_width, int tex_height, int rhs)
{
    /* This last list flisp the texture coordinates if you display
     * to gfx instead of the screen node.
     */

    float t0[2],t1[2],t2[2],t3[3];          /* texture coordinates */
    float v0[2],v1[2],v2[2],v3[2];          /* vertice coordinates */
    float xoff, yoff;
    float xend, yend;
    float midx, midy;

    if (rhs) {
        xoff = (tex_width  - uSize)/(float)tex_width;
        yoff = (tex_height - vSize)/(float)tex_height; 
        xend = 1.0;
        yend = 1.0;
    }
    else {
        xoff = 0.0; 
        yoff = 0.0;
        xend = (float)uSize / (float)tex_width; 
        yend = (float)vSize / (float)tex_height; 
    }

    midx = xoff + (xend - xoff)/2.0;
    midy = yoff + (yend - yoff)/2.0;

    v0[0] = -1.0; v0[1] = -1.0;
    v1[0] =  1.0; v1[1] = -1.0;
    v2[0] =  1.0; v2[1] =  1.0;
    v3[0] = -1.0; v3[1] =  1.0;
    
    t0[0] = xoff; t0[1] = yend;
    t1[0] = xend; t1[1] = yend;
    t2[0] = xend; t2[1] = yoff;
    t3[0] = xoff; t3[1] = yoff;

    glNewList(start, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();


    /* First tile */
    t0[0] = xoff; t0[1] = yend;
    t1[0] = midx; t1[1] = yend;
    t2[0] = midx; t2[1] = midy;
    t3[0] = xoff; t3[1] = midy;

    v0[0] = -1.0; v0[1] = -1.0;
    v1[0] =  0.0; v1[1] = -1.0;
    v2[0] =  0.0; v2[1] =  0.0;
    v3[0] = -1.0; v3[1] =  0.0;

    glNewList(start + 1, GL_COMPILE);
        /* Draw the polygon. */
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Second tile */
    t0[0] = midx; t0[1] = yend;
    t1[0] = xend; t1[1] = yend;
    t2[0] = xend; t2[1] = midy;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  0.0; v0[1] = -1.0;
    v1[0] =  1.0; v1[1] = -1.0;
    v2[0] =  1.0; v2[1] =  0.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 2, GL_COMPILE);
        /* Draw the polygon. */
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Third tile */
    t0[0] = xend; t0[1] = midy;
    t1[0] = xend; t1[1] = yoff;
    t2[0] = midx; t2[1] = yoff;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  1.0; v0[1] =  0.0;
    v1[0] =  1.0; v1[1] =  1.0;
    v2[0] =  0.0; v2[1] =  1.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 3, GL_COMPILE);
        /* Draw the polygon. */
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    /* Fourth tile */
    t0[0] = midx; t0[1] = yoff;
    t1[0] = xoff; t1[1] = yoff;
    t2[0] = xoff; t2[1] = midy;
    t3[0] = midx; t3[1] = midy;

    v0[0] =  0.0; v0[1] =  1.0;
    v1[0] = -1.0; v1[1] =  1.0;
    v2[0] = -1.0; v2[1] =  0.0;
    v3[0] =  0.0; v3[1] =  0.0;

    glNewList(start + 4, GL_COMPILE);
        /* Draw the polygon. */
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();

    return 5;
}
