///////////////////////////////////////////////////////////////////////
//
// ScratchWindow.c++
//
//	A window for rendering OpenGL.  This class has no relation to the
//	dmu library.  It is just some sample code to show how to open
//	and set up a double buffering window for OpenGL rendering.
//	The following is an example of how this class is used.
//
//      ScratchWindow *window;
//      window = ScratchWindow::create(theDisplay, sizeX, sizeY, &argc, argv);
//      if (!window){
//         fprintf (stderr, "Failed to create scratch window!\n");
//         exit (-1);
//      }
//      window->setContext();
//
//      glViewport( 0, 0, sizeX, sizeY);
//      glMatrixMode(GL_PROJECTION);
//      glLoadIdentity();
//      glOrtho(0,sizeX,0,sizeY,0,1000);
//      . . . continue Open GL here . . 


// Copyright 1995, Silicon Graphics, Inc.
// ALL RIGHTS RESERVED
//
// UNPUBLISHED -- Rights reserved under the copyright laws of the United
// States.   Use of a copyright notice is precautionary only and does not
// imply publication or disclosure.
//
// U.S. GOVERNMENT RESTRICTED RIGHTS LEGEND:
// Use, duplication or disclosure by the Government is subject to restrictions
// as set forth in FAR 52.227.19(c)(2) or subparagraph (c)(1)(ii) of the Rights
// in Technical Data and Computer Software clause at DFARS 252.227-7013 and/or
// in similar or successor clauses in the FAR, or the DOD or NASA FAR
// Supplement.  Contractor/manufacturer is Silicon Graphics, Inc.,
// 2011 N. Shoreline Blvd. Mountain View, CA 94039-7311.
//
// THE CONTENT OF THIS WORK CONTAINS CONFIDENTIAL AND PROPRIETARY
// INFORMATION OF SILICON GRAPHICS, INC. ANY DUPLICATION, MODIFICATION,
// DISTRIBUTION, OR DISCLOSURE IN ANY FORM, IN WHOLE, OR IN PART, IS STRICTLY
// PROHIBITED WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SILICON
// GRAPHICS, INC.
//
////////////////////////////////////////////////////////////////////////

#include "ScratchWindow.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/GLw/GLwMDrawA.h>

#include <GL/glu.h>
#include <GL/glx.h>

#include <Xm/Frame.h>
#include <assert.h>
#include <stdio.h>		// fprintf()
#include <string.h>		// ffs()
#include <math.h>		// M_PI_2, tan()

#include <sys/types.h>
#include <sys/prctl.h>


#ifdef DEBUG
#define DEBUGONLY(x) x
#else
#define DEBUGONLY(x) // nothing
#endif

////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////

// commented out to fix warning from compiler
//static int appInit = 0;

static Display*		theDisplay 	= NULL;


static String   fallbackResources[] = {
    "*useSchemes: all", "*sgimode:True",
    "*glxwidget*width: 640", "*glxwidget*height: 480",
    "*frame*shadowType: SHADOW_IN",
    NULL};

static int      glAttribs[] = { GLX_RGBA, None,
                                GL_DOUBLEBUFFER, GL_TRUE};


////////////////////////////////////////////////////////////////////////
//
// scratchWindow implementation class
//
////////////////////////////////////////////////////////////////////////

class scratchWindow : public ScratchWindow {
private:
  GLboolean	windowValid;
  int		winWidth;
  int		winHeight;
  GLXContext	context;

  XtAppContext     appCtx;
  XVisualInfo*	   visInfo;
  Widget           topLevel;
  Widget           frame;
  Widget           glArea;
  
public:
  //
  // Constructors and destructors
  //
  
  scratchWindow( int width, int height, int *argc, String *argv );
  ~scratchWindow();
  
  //
  // ScratchWindow methods.
  //
  
  void   setContext();
  int    getWidth()		{ return winWidth; }
  int    getHeight()		{ return winHeight; }
  size_t getBufferSize();

  GLXContext getContext() {return context;}
  void   addGLCallback(String name, XtCallbackProc func, XtPointer data);
  void   eventLoop();

  void   addInput(int fd, XtInputCallbackProc func, XtPointer data);
  //
  // Methods used by create()
  //
  
  GLboolean isValid() { return windowValid; }
  
private:
  //
  // Private methods.
  //
  
  GLboolean findColormap();
  GLboolean findVisual();
  GLboolean openDisplay();
};

////////
//
// constructor
//
////////

scratchWindow::scratchWindow(
    int width,
    int height,
    int *argc, String* argv
    )
{
    assert( 0 < width );
    assert( 0 < height );

    winWidth  = width;
    winHeight = height;

    //
    // Make sure that we have a display open.
    //
    topLevel = XtOpenApplication(&appCtx,"libdmuScratchWindow", NULL, 0,
				 argc, argv,
				 fallbackResources, 
				 applicationShellWidgetClass, 
				 NULL, 0);


    theDisplay = XtDisplay(topLevel);

    frame = XmCreateFrame(topLevel, "frame", NULL, 0);
    XtManageChild(frame);
    
    visInfo = glXChooseVisual(theDisplay, DefaultScreen(theDisplay), glAttribs);
    if (!visInfo){
      fprintf (stderr, "Scratch window - can't get visual\n");
      windowValid = GL_FALSE;
      return;
    }

    glArea = XtVaCreateManagedWidget("glxwidget",
				       glwMDrawingAreaWidgetClass, 
				       frame,
				       GLwNrgba, True,
				       GLwNdepthSize, 8,
				       GLwNdoublebuffer, True,
				       XmNwidth, width,
				       XmNheight, height,
				       NULL);
    XtRealizeWidget(topLevel);
    
    XtVaGetValues(glArea, 
		  GLwNvisualInfo,&visInfo,NULL);
    
    
    context = glXCreateContext(theDisplay, visInfo, 0, GL_TRUE);

    // Set up the color map.
    //

    this->setContext();
    
    //
    // Clear the window.
    //

    glLoadIdentity();
    glClearColor( 0.0, 0.0, 0.0, 1.0 );
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glFlush();

    //
    // Make sure that the ABGR extension is supported.
    //

    if ( strstr( (const char*) glGetString(GL_EXTENSIONS),
		 "GL_EXT_abgr" ) == NULL ) {
	printf( "ABGR extension is not present.\n" );
	windowValid = GL_FALSE;
    }

    //
    // All done.
    //

    windowValid = GL_TRUE;
}

////////
//
// destructor
//
////////

scratchWindow::~scratchWindow()
{
    glXMakeCurrent( theDisplay, None, NULL );
    glXDestroyContext( theDisplay, context );
    XFlush( theDisplay );
}


////////
//
// getBufferSize
//
////////

size_t scratchWindow::getBufferSize(
    )
{
    // XXX This is hard-coded for RGBA
    return winWidth * winHeight * 4 * sizeof( GLubyte );
}

////////
//
// setContext
//
////////

void scratchWindow::setContext(
    )
{
  GLwDrawingAreaMakeCurrent(glArea, context);
}


void scratchWindow::addGLCallback(String name, 
				XtCallbackProc func, 
				XtPointer data){
  if (windowValid){
    XtAddCallback(glArea, name, func, data);
  }
}

void scratchWindow::addInput(int fd, XtInputCallbackProc func, XtPointer data){
  if (!windowValid)
    return;

  XtAppAddInput(appCtx, fd, (XtPointer) XtInputReadMask, func, data);
}

void scratchWindow::eventLoop(){
  
  if (!windowValid)
    return;

  XtAppMainLoop(appCtx);
}

////////////////////////////////////////////////////////////////////////
//
// ScratchWindow methods
//
////////////////////////////////////////////////////////////////////////

////////
//
// create
//
////////

ScratchWindow* ScratchWindow::create(
    Display* display,
    int width,
    int height,
    int *argc, 
    String *argv
    )
{
  
    theDisplay = display;

    scratchWindow* window = new scratchWindow( width, height, argc, argv );
    
    if ( window->isValid() ) {
	return window;
    }
    else {
	delete window;
	return NULL;
    }
}

////////
//
// Destructor
//
////////

ScratchWindow::~ScratchWindow(
    )
{
}


/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
