//
//  dmuChain.c++
//
//  Author: Ed Allard
//  Date:   Oct, 1996
//
//  Abstraction that allows you to containt a chain of objects
//  and work with them as a group.  The other main use of this
//  object is to contain a group of objects that you want
//  to build up together and hook up with  other stuff.  If you
//  inherit from this class, you can build up your own chain
//  inside this objects' setup method and then the outside
//  tree would not know what is in here.  For example, you
//  could make an chain object that would create either a
//  movie source or a video source depending on a file
//  name you pass.
//  NOTE: you must only give this object a chain. This means
//  that every object in the chain has only one child.
//
#include "dmuChain.h"


//****************************************************************
//
//  Public dmuChain methods
//
//****************************************************************


//----------------------------------------------------------
// dmuChain
//----------------------------------------------------------
//
// Constructor. Does nothing beyond initaliziation of
// member variables
//
dmuChain::dmuChain(){
  chainHead = NULL;
  chainTail = NULL;
}

//----------------------------------------------------------
// ~dmuChain
//----------------------------------------------------------
//
dmuChain::~dmuChain(){
}

//----------------------------------------------------------
// setup
//----------------------------------------------------------
//
// This method does not do anything on this objects but
// if you inherit from this class you might want to override
// this method to do some of your own stuff.
//
int dmuChain::setup(){
  return (DMU_UNIMPLEMENTED);
}

//----------------------------------------------------------
// addObject
//----------------------------------------------------------
//
// This method will add an object to the object on the tail
// end of the chain stored in the Chain object
//
int dmuChain::addObject(dmuObj *newObj){
  DEBUG2(fprintf (stderr, "adding object to chain tail\n"));
  DEBUG( if (!chainTail) return DMU_CHAIN_NO_CHAIN);

  return(chainTail->addObject(newObj));
}

//----------------------------------------------------------
// addSrcConstraints
//----------------------------------------------------------
//
// This will add the constraints of the object at the
// head of the chain
//
int dmuChain::addSrcConstraints(DMparams *params){
  DEBUG( if (!chainHead) return DMU_CHAIN_NO_CHAIN);

  return (chainHead->addSrcConstraints(params));
}

//----------------------------------------------------------
// addSrcConstraints
//----------------------------------------------------------
//
// This will add the constraints of the objects at the
// tail of the chain
//
int dmuChain::addDstConstraints(DMparams *params){
  DEBUG( if (!chainTail) return DMU_CHAIN_NO_CHAIN);

  return (chainTail->addDstConstraints(params));
}

//----------------------------------------------------------
// start
//----------------------------------------------------------
//
// This will call start on the object at the head of the
// chain.  Then the start method will propagate out of the
// chain if there is anything outside the chain
//
int dmuChain::start(int propagate){
  DEBUG( if (!chainHead) return DMU_CHAIN_NO_CHAIN );

  return (chainHead->start(propagate));
}

//----------------------------------------------------------
// getFD
//----------------------------------------------------------
//
// This method has no meaning in in this object.  There could
// be lots of fd depending on the chain.  You should call
// setFD instead.  This will add the fd for each object to
// the fdset automatically.
//
int dmuChain::getFD(){
  return (DMU_CHAIN_USE_SET_FD);
}

//----------------------------------------------------------
// setFD
//----------------------------------------------------------
//
// This takes an fdSet and add the fd for each object in the
// chain to the fdSet.  
//
int dmuChain::setFD(fd_set *fdSet,int *maxFD, int propagate ){
  DEBUG( if (!chainTail) return DMU_CHAIN_NO_CHAIN);
  dmuObj *obj = chainHead;
  dmuObjList *list = NULL;

  // we need to make sure that we add the fd for all the objects
  // in the chain, but don't propagate out if propagate is not
  // set, so we traverse the objects calling setFD on each one
  // until we reach the tail and then we propagate out if
  // propagate is set.
  while (obj){
    DEBUG2(fprintf (stderr,"chain calling setFD on %s\n",obj->getName()));

    // Call setFD on the current object
    obj->setFD(fdSet, maxFD, 0);

    // get the list of children of the current object
    list = obj->getObjectList();

    // if the list is empty, went through all of the objects
    if (!list){
      // this will exit the loop for us
      obj = NULL;
    }
    else {
      // set the current object to be the first object in the list of
      // children.  This is where the fan out case fails
      obj = list->obj;
      // we need to make sure that calling handle events on the
      // chain does not propagate outside of the Chain object
      if (obj == chainTail) {
	// call setFD on the tail and propagate if needed
	obj->setFD(fdSet, maxFD, propagate);
	// this will exit the loop for us
	obj = NULL;
      }
    }
  }

  return DMU_OK;
}

//----------------------------------------------------------
// receive
//----------------------------------------------------------
//
// calls receive on the head of the chain.
//
int dmuChain::receive(DMbuffer inBuffer){
  DEBUG( if (!chainHead) return DMU_CHAIN_NO_CHAIN);

  return (chainHead->receive(inBuffer));
}

//----------------------------------------------------------
// init
//----------------------------------------------------------
//
// calls init on the head of the chain.  This will propagate
// out of the chain if there is anything after the chain
//
int dmuChain::init(dmuParams *params ){
  DEBUG( if (!chainHead) return DMU_CHAIN_NO_CHAIN);

  return (chainHead->init(params));
}

//----------------------------------------------------------
// setChain
//----------------------------------------------------------
//
// pass the chain to contain into the object.  Once you have
// built up a chain and are done working with the objects
// individually, you can pass that chain into the chain object
// and use it as a wrapper around your chain from then on.
//
int  dmuChain::setChain(dmuObj *newHead){
  int chainLen = 1;
  dmuObjList *list;

  // Set the chain head to the new one
  if (!newHead) {
    return (DMU_CHAIN_ERROR);
  }
  chainHead = newHead;
  
  // Find the tail of the chain
  list = chainHead->getObjectList();
  if (!list){
    chainTail = chainHead;
    DEBUG2(fprintf (stderr, "Chain of length:%d set up\n",chainLen));
    DEBUG2(fprintf (stderr, "Head of chain:'%s', tail '%s'\n:%d set up\n",
		    chainHead->getName(),chainTail->getName(),chainLen));
    return (DMU_OK);
  }

  while (list->obj->getObjectList()){
    list = list->obj->getObjectList();
    chainLen++;
  }

  chainTail = list->obj;

  DEBUG2(fprintf (stderr, "Chain of length:%d set up\n",chainLen));
  DEBUG2(fprintf (stderr, "Head of chain:'%s', tail '%s'\n:%d set up\n",
		  chainHead->getName(),chainTail->getName(),chainLen));
  return (DMU_OK);
}

//----------------------------------------------------------
// handleEvents
//----------------------------------------------------------
//
// This method calls handleEvents individually on each of the objects
// in the chain.  This means that each objects should be sure
// that it only tries to do something if it's fd is set
//
int dmuChain::handleEvents(fd_set fdSet){
  dmuObj *obj = chainHead;
  dmuObjList *list = NULL;

  DEBUG2(fprintf (stderr,"\nchain handle events\n"));

  // iterate over all of the objects in the list and call handle events
  // NOTE:this assumes that the chain does not have any fan out anywhere.
  while (obj){
    DEBUG2(fprintf (stderr,"chain calling handle on %s\n",obj->getName()));

    // Call handle events on the current object
    obj->handleEvents(fdSet);

    // get the list of children of the current object
    list = obj->getObjectList();

    // if the list is empty, went through all of the objects
    if (!list){
      // this will exit the loop for us
      obj = NULL;
    }
    else {
      // set the current object to be the first object in the list of
      // children.  This is where the fan out case fails
      obj = list->obj;
      // we need to make sure that calling handle events on the
      // chain does not propagate outside of the Chain object
      if (obj == chainTail) {
	// call handle events on the tail
	obj->handleEvents(fdSet);
	// this will exit the loop for us
	obj = NULL;
      }
    }
  }

  DEBUG2(fprintf (stderr,"chain handle events done\n"));
  return (DMU_OK);
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
