//
//  dmuGLPBuffer.c++
//
//  Author: Demo Factory
//  Date:   Oct, 1996
//
//
// dmuGLPBuffer - This class is intended as a base class for any dmu
// object that wants to receive and manipulate DMbuffer data as an
// Open GL Pixel Buffer.  There are two key members to this class:
// pbuffer and newBuffer.
//
// pbuffer is the GL PBuffer which allows OpenGL operations to be
// performed off-screen to memory.  
//
// newBuffer is a handle to the latest DMBuffer which is passed to us
// from our parent object in the chain.  We will end up associating
// the DMBuffer (pointed to by newBuffer) with the pbuffer.  After
// this association has happened, the DMBuffer becomes the pixel area
// (i.e. the location in memory) where all off-screen rendering
// operations on the PBuffer will appear.
//
// To make sure there is always a valid pixel area associated with the
// pbuffer, either a user program or a subclass has to lock the
// pbuffer.  A locking semaphore, pbufferLock, is used around all
// operations that occur on pbuffer.  
//
// To make sure that newBuffer isn't changed or freed durring a
// PBuffer association, a semaphore lock is also placed around all
// operations that occur on newBuffer.  
//
// The basic methods (in order of use) for the class are:
//
// Construction - init's member variables
//
// Setup - Set's all the member variables we'll need to create the
// PBuffer.  This is also where the semaphores are created.
//
// Init - This is where the PBuffer is created.  Originally, there is
// no pixel area associated with pbuffer, so the buffer is locked.
//
// receive - When a dmbuffer is received from our parent, we obtain a
// lock on newBuffer.  If newBuffer was already pointing to something,
// then we dereference the old DMBuffer we were pointing to.  This
// dereference allows the DMBuffer to be destroyed if there are no
// other references to it in the program.  Next newBuffer is set to
// point to this new data, and a reference is added to the new buffer
// (to make sure it stays around, in case we need it).  This method is
// called internal to the library from the parent's handleEvents (or
// propagate) method.
//
// grab - This method is called when a program or a sub class want to
// use the PBuffer to perform rendering.  Here we obtain locks on both
// newBuffer and pbuffer.  We then associate newBuffer with pbuffer.
// This association does a few things: It dereferences the old
// dmbuffer that was associated with pbuffer.  It adds an implicit
// reference to newBuffer.  It also sets up the PBuffer to render in
// to, or out of.  Once the associate is finished, the lock on
// newBuffer is released, __but the lock on pbuffer is not__.  This
// lock ensures that the same data is valid and available to the
// pbuffer fromt the time of grab to the time of release.  note that
// grab is a blocking call, so you better release before you try to
// grab again.
//
#include "dmuGLPBuffer.h"


//****************************************************************
//
//  Public methods
//
//****************************************************************

//----------------------------------------------------------
// dmuGLPBuffer
//----------------------------------------------------------
//
// Constructor. Intializes some data members
//
dmuGLPBuffer::dmuGLPBuffer(){
  dpy = NULL;
  fbConfig = NULL;
  pbuffer  = 0;
  gfxParams   = NULL;
}

//----------------------------------------------------------
// ~dmuGLPBuffer
//----------------------------------------------------------
//
// Destructor - Destroy the PBuffer (if we've gotten far enough to
// create one)
//
dmuGLPBuffer::~dmuGLPBuffer(){
  if (pbuffer){
    DEBUG1(fprintf (stderr, "freeing up pbuffer\n"));
    glXDestroyGLXPbufferSGIX(dpy,pbuffer);
  }
}

//----------------------------------------------------------
// setup
//----------------------------------------------------------
//
// set all the member variables we'll need to create the
// pbuffer.  We need the X display, and the size of the data 
// that we'll be receiving before we can init() the object
// and create the PBuffer
//
int dmuGLPBuffer::setup(Display *_theDisplay, int _w, int _h){
  dmuLatch::setup();

  // Grab the display.. we'll need it for lots of stuff.. We also need
  // to know the size of the frames we'll be dealing with
  dpy = _theDisplay;
  //  bufferSizeX = w;
  //  bufferSizeY = h;
  imageSizeX = bufferSizeX  = _w;
  imageSizeY = bufferSizeY  = _h;
  int stat;

  // Setup a dmparam structure that describes what we've got here..
  stat = dmParamsCreate(&gfxParams);
  if (stat != DM_SUCCESS){
    return(DMU_PARAMS_CREATE_FAILED);
  }
  
  stat = dmSetImageDefaults(gfxParams, bufferSizeX, 
			    bufferSizeY, DMU_GLPBUFFER_PACKING);
  if (stat != DM_SUCCESS){
    return(DMU_POOL_DEFAULTS_FAILED);
  }

  stat = dmParamsSetEnum (gfxParams,DM_IMAGE_LAYOUT,DM_IMAGE_LAYOUT_GRAPHICS);
  if (stat != DM_SUCCESS){
    return(DMU_POOL_DEFAULTS_FAILED);
  }


  DEBUG1(fprintf (stderr, "\n====gfxParams for dmuGLPBuffer=====\n"));
  DEBUG1(printDMParams(gfxParams));
  DEBUG1(fprintf (stderr, "==================================\n"));

  return (DMU_OK);
}



//----------------------------------------------------------
// init
//----------------------------------------------------------
//
// create the pbuffer based on the parmeters set in setup()
// Once this is done, we are ready for frames to come our way
//
int dmuGLPBuffer::init (dmuParams *_params){
  int err = createPBuffer();
  if (err != DMU_OK) {
    return(err);
  }
  
  // Initialize our children
  return (dmuObj::init(_params));
}



//----------------------------------------------------------
// addSrcConstraints
//----------------------------------------------------------
//
// Add my graphics constraints to anyones params
// who wants to be my source based on my gfxParams (which are set in
// seutp()).
//
int dmuGLPBuffer::addSrcConstraints(DMparams *_params){
  int stat;

  stat = dmBufferGetGLPoolParams( gfxParams, _params );
  if (stat != DM_SUCCESS){
    return (DMU_BAD_CONSTRAINTS);
  }

  return (DMU_OK);
}




//----------------------------------------------------------
// addDstConstraints
//----------------------------------------------------------
//
// Add my graphics constraints to anyones params
// who wants to be my destination based on my gfxParams (which are set in
// seutp()).
//
int dmuGLPBuffer::addDstConstraints(DMparams *_params){
   int stat;

  stat = dmBufferGetGLPoolParams( gfxParams, _params );
  if (stat != DM_SUCCESS){
    return (DMU_BAD_CONSTRAINTS);
  }

  return (DMU_OK);
}


//----------------------------------------------------------
// grab
//----------------------------------------------------------
//
// This method grabs the most recent DMBuffer that we've
// receive()'d, locks it, and associates it with the pBuffer.  This
// prevents other processes from doing anything with the pbuffer or
// its associated dmbuffer until the the grab()'ing process has
// released it using release().  Note that the associate add's an
// implicit ref count to the dmbuffer, which stays there until
// associate is called again.
//
int dmuGLPBuffer::grab(){
  DEBUG2(fprintf (stderr, "grab\n"));
  Bool ok;

  dmuLatch::grab();
  
  ok = glXAssociateDMPbufferSGIX(dpy,pbuffer,gfxParams,latchBuffer);
  if (!ok){
    fprintf (stderr, "Error in associate");
  } else {
    //fprintf (stderr, "grabed\n");
  }

  return (ok?DMU_OK: DMU_GL_PBUFFER_ASSOC_FAILED);
}


//****************************************************************
//
//  Protected methods
//
//****************************************************************

//----------------------------------------------------------
// createPBuffer
//----------------------------------------------------------
//
int dmuGLPBuffer::createPBuffer() {
  int error;

  // Attributes we'll need to create the pbuffer
  int attrib[] = {
    GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
    GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE, 
    None
  };


  // Get a frame buffer configuration that matches our needs (see the 
  // dmuUtil.h header for info on this call)
  error = setupFBConfig(dpy,8,8,8,8,&fbConfig);

  if (error != DMU_OK){
    return (error);
  }

  DEBUG(fprintf (stderr, "FBConfig used to create pbuffer:\n"));
  DEBUG(printFBConfig(dpy,fbConfig));


  // Create the PBuffer
  pbuffer = glXCreateGLXPbufferSGIX(dpy,fbConfig,bufferSizeX,
				    bufferSizeY,attrib);
  if (pbuffer == None){
    return (DMU_GL_PBUFFER_ALLOC_FAILED);
  }

  fprintf(stderr, "dmparams (gfxparams) used for pbuffer:\n");
  printDMParams( gfxParams );

  return(DMU_OK);
}









/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
