//
//  dmuIC.c++
//
//  Author: Tim Davison
//  Date:   Oct, 1996
//
//  This file is the basis of the image conversion object of the framework
//  Currently, the inheritance hierachy looks like this:
//  dmuObj -> dmuIC ---> dmuICjencoder
//                  |
//                   --> dmuICjdecoder
//  There are a bunch of things that are the same about the decoder and
//  encoder objects, so this object wraps up the similarities.
//
#include "dmuIC.h"


//****************************************************************
//
//  Public methods
//
//****************************************************************


//----------------------------------------------------------
// dmuIC 
//----------------------------------------------------------
//
// Constructor. Does nothing beyond initaliziation of
// member variables
//
dmuIC::dmuIC() {
  fd = -1;
  speed = DM_IC_SPEED_REALTIME;
  direction =  DM_IC_CODE_DIRECTION_UNDEFINED;
  srcPacking = DM_IMAGE_PACKING_CbYCrY;
  dstPacking = DM_IMAGE_PACKING_CbYCrY;
  srcCompression = DM_IMAGE_UNCOMPRESSED;
  dstCompression = DM_IMAGE_UNCOMPRESSED;
  imgConvCtx = NULL;
}

//----------------------------------------------------------
// ~dmuIC
//----------------------------------------------------------
//
// destroy the image conversion context if one has been created
//
dmuIC::~dmuIC() {
  if (imgConvCtx) {
    dmICDestroy(imgConvCtx);
  }
}

//----------------------------------------------------------
// setup
//----------------------------------------------------------
//
// This method creates the image conversion context needed by the
// dmIC library to know how to convert images.  This method also
// sets parameters within the image conversion context.  
//
int dmuIC::setup() {
  DMstatus stat;
  DEBUG2( fprintf(stderr, "%s:setup\n", getName()));

  // we need two sets of params: src and dst
  DMparams *srcParams;
  DMparams *dstParams;
  DMparams *convParams;

  // create a parameter structure 
  stat = dmParamsCreate(&srcParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED );

  stat = dmParamsCreate(&dstParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED );

  stat = dmParamsCreate(&convParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED );


  // Setup the DMIC source Params
  //
  // set up the defaults for the image
  // These are the parameters of the source image
  stat = dmSetImageDefaults(srcParams, imageSizeX, imageSizeY, srcPacking);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_SET_IMAGE_DEFAULTS_FAILED);

  // Set the parameter to say that the source image is uncompressed
  stat = dmParamsSetString(srcParams, DM_IMAGE_COMPRESSION, srcCompression);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_PARAMS_SET_STRING_FAILED);

  // Set the parameter to tell it which direction is up
  stat =
    dmParamsSetEnum(srcParams, DM_IMAGE_ORIENTATION, DM_IMAGE_TOP_TO_BOTTOM);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_PARAMS_SET_ENUM_FAILED);

  
  // Setup the DMIC destination Params
  //
  // set up the defaults for the image source
  // These are the parameters of the source image
  stat = dmSetImageDefaults(dstParams, imageSizeX, imageSizeY, dstPacking);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_SET_IMAGE_DEFAULTS_FAILED);
  	 
  // Change the compression type in the param structure so that the
  // destination is told to be a jpeg compressed image
  stat = dmParamsSetString(dstParams, DM_IMAGE_COMPRESSION, dstCompression);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_PARAMS_SET_STRING_FAILED);

  // Set the parameter to tell it which direction is up
  stat =
    dmParamsSetEnum(dstParams, DM_IMAGE_ORIENTATION, DM_IMAGE_TOP_TO_BOTTOM);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_PARAMS_SET_ENUM_FAILED);

  
  // Setup the DMIC conversion Params
  //
  // we need to set the speed of the conversion
  stat = dmParamsSetEnum(convParams, DM_IC_SPEED, speed);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_PARAMS_SET_ENUM_FAILED);
		  

  // DEBUG: print out the src, dst, and conversion  param structures
  DEBUG3( fprintf(stderr, "%s::setup:printing src dmic params\n", getName()));
  DEBUG3( printDMParams(srcParams));
  DEBUG3( fprintf(stderr, "%s::setup:printing dst dmic params\n", getName()));
  DEBUG3( printDMParams(dstParams));
  DEBUG3( fprintf(stderr, "%s::setup:printing conv dmic params\n", getName()));
  DEBUG3( printDMParams(convParams));


  // Setup the Image Converter
  //
  // try to find an image converter that matches the parameters set above
  int imgConvID = dmICChooseConverter(srcParams, dstParams, convParams);
  DMUIC_ERR( (imgConvID < 0),           DMU_DMIC_FOUND_NO_USABLE_DMIC );

  DEBUG3( fprintf(stderr, "%s::setup: creating imgconvctx\n", getName()));
  DEBUG3( fprintf(stderr, "%s::setup:   using id:%d\n", getName(), imgConvID));

  // create a context with the index that we found above and put it into
  // to object's private member imgConvCtx
  stat = dmICCreate(imgConvID, &imgConvCtx); 
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_CREATE_FAILED );
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );

  // get the max frame size and store it
  if (direction == DM_IC_CODE_DIRECTION_ENCODE) {
    bufferSize = dmImageFrameSize(srcParams);
  } else if (direction == DM_IC_CODE_DIRECTION_DECODE) {
    bufferSize = dmImageFrameSize(dstParams);
  }
  DEBUG3( fprintf(stderr, "%s::setup: buffersize:%d\n",getName(), buffersize));
  
  DEBUG3( fprintf(stderr, "%s::setup: setting src params\n", getName()));
  // set the source params for the img conversion context
  stat = dmICSetSrcParams(imgConvCtx, srcParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_SRC_PARAMS_FAILED);

  DEBUG3( fprintf(stderr, "%s::setup: setting dst params\n", getName()));
  // set the destination params for the img conversion context
  stat = dmICSetDstParams(imgConvCtx, dstParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_DST_PARAMS_FAILED);

  DEBUG3( fprintf(stderr, "%s::setup: setting conv params\n", getName()));
  // set the conversion params for the img conversion context
  stat = dmICSetConvParams(imgConvCtx, convParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_CONV_PARAMS_FAILED);

  // clean up the params structure
  dmParamsDestroy(srcParams);
  dmParamsDestroy(dstParams);
  dmParamsDestroy(convParams);
  
  // return no error
  return DMU_OK;
}


//----------------------------------------------------------
// init
//----------------------------------------------------------
//
// This method creates the dmbuffer pool that the image converter
// dumps into.  The dmbuffer it creates is dependent on
// all of it's children also, so you must have the chain of
// objects put together before you call this method.  This
// method must also be propagated down to all the objects
// in the chain below it.
// 
int dmuIC::init(dmuParams *params) {
  DMstatus   stat;
  int        err;
  
  DEBUG2( fprintf(stderr, "%s: init\n", getName()));

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );

  DMparams *dmicParams;
  // create a parameter structure
  stat = dmParamsCreate(&dmicParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED );
  // set up the defaults for the image
  // These are the parameters of the source image
  stat = dmBufferSetPoolDefaults(dmicParams, numBuffers, bufferSize,
				 DMU_NOT_CACHED, DMU_NOT_MAPPED);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_POOL_DEFAULTS_FAILED );
	 
  // Add my own constraints to the parameter structure
  err = addDstConstraints(dmicParams);
  DMUIC_ERR( (err != DMU_OK),           DMU_BAD_CONSTRAINTS );
  
  // add the constraints of the children to the parameter structure
  // The addChildConstraints method is an inherited method from dmuObj
  // which calls addSrcConstraints on each of the children.  
  err = addChildConstraints(dmicParams);
  DMUIC_ERR( (err != DMU_OK),           DMU_BAD_CONSTRAINTS );
  
  DEBUG( fprintf(stderr, "%s::init: pool creation params\n", getName()));
  DEBUG( printDMParams(dmicParams));

  // Create the pool with the given parameters
  stat = dmBufferCreatePool(dmicParams, &bufferPool);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_POOL_CREATE_FAILED );

  // Tell the img Conversion context that the pool you just created is
  // the pool to dump data into.
  stat = dmICSetDstPool(imgConvCtx, bufferPool);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_DST_POOL_FAILED );
  
  // clean up the params structure
  dmParamsDestroy(dmicParams);  

  // return the error from propagating this method on to the rest of the
  // chain.  The base method calls init on all of the children of an
  // object
  return(dmuObj::init());
}


//----------------------------------------------------------
// addSrcConstraints
//----------------------------------------------------------
//
//  This method adds the constraints it imposes on it's source pool.
//  This means that the parent of this object will invoke this method
//  to add the constraints of the parents destination pool
//
int dmuIC::addSrcConstraints(DMparams *params) {
  DMstatus stat;

  DEBUG2( fprintf(stderr, "%s: addSrcConstraints\n", getName()));

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );
  DMUIC_ERR( (params == NULL),          DMU_BAD_INPUT);
  
  // add the source constraints to the parameter structure passed in
  stat = dmICGetSrcPoolParams(imgConvCtx, params);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_SRC_POOL_PARAMS_FAILED);

  return DMU_OK;
}


//----------------------------------------------------------
// addDstConstraints
//----------------------------------------------------------
//
int dmuIC::addDstConstraints(DMparams *params) {
  DMstatus stat;

  DEBUG2( fprintf(stderr, "%s: addDstConstraints\n", getName()));

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );
  DMUIC_ERR( (params == NULL),          DMU_BAD_INPUT);
  
  // add the Destination constraints to the parameter structure passed in
  stat = dmICGetDstPoolParams(imgConvCtx, params);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_DST_POOL_PARAMS_FAILED);

  return DMU_OK;
}


//----------------------------------------------------------
// getFD
//----------------------------------------------------------
//
// This method will return the file descriptor that corresponds to
// the destination buffer receiving new data.  Then this file descriptor
// is set then the object has data to pass on to it's children
//
int dmuIC::getFD() {
  DEBUG2( fprintf(stderr, "%s: getFD\n", getName()));
  
  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );

  // if we haven't grapped the fd yet, grab it and store it for future
  // use
  if (fd == -1) {
    fd = dmICGetDstQueueFD(imgConvCtx);
  }
  return(fd);
}


//----------------------------------------------------------
// handleEvents
//----------------------------------------------------------
//
// This method checks a fd_set to see if the object's FD is set.  If
// it is that means that there is waiting data from VICE. If there is
// waiting data, the method propagates the data to it's children
// and free's the buffer.  This function is most likely used in the
// event loop after a select call.
//
int dmuIC::handleEvents(fd_set fdSet) {
  DMstatus stat;
  int      err;

  DEBUG2( fprintf(stderr, "%s: handleEvents\n", getName()));

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );

  // check to see if this object FD is set
  if (isStarted() && FD_ISSET(fd, &fdSet)) {
    DEBUG2( fprintf(stderr, "%s:   - really got an event\n", getName()));
    DMbuffer buf;

    // get a pointer to the buffer containing the converted data
    stat = dmICReceive(imgConvCtx, &buf);
    DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_RECEIVE_FAILED);

    // Take the converted buffer and propogate it to the object's children
    // This method is inherited from dmuObj and is not overridden
    err = propagate(buf);
    DMUIC_ERR( (err != DMU_OK),           DMU_GENERIC_ERROR );
    
    // we are done with the buffer, the children should have add a ref by
    // now, so we should free the buffer
    stat = dmBufferFree(buf);
    DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_BUFFER_FREE_FAILED);

    // everything seems fine, return OK
    return DMU_OK;
  }

  // Everything is fine, but I didn't do anything
  return DMU_NOT_MY_EVENT;
}


//----------------------------------------------------------
// receive
//----------------------------------------------------------
//
// The is the method called on this object when it's parent
// propagates data down to the parent's children
// In this object, this method needs only pass the data
// on to ICE using the image conversion context
//
int dmuIC::receive(DMbuffer inBuffer) {
  DMstatus stat;

  DEBUG2( fprintf(stderr, "%s: recieve\n", getName()));

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );
  DMUIC_ERR( (inBuffer == NULL),        DMU_BAD_INPUT);

  if (isStarted()){
    // pass the data to the VICE queue
    stat = dmICSend(imgConvCtx, inBuffer, 0, 0);
    DMUIC_ERR( (stat != DM_SUCCESS),      DMU_DMIC_SEND_FAILED);
  }

  // everything is fine, return OK
  return DMU_OK;
}


//****************************************************************
//
//  Debugging methods
//
//****************************************************************

//----------------------------------------------------------
// setPrintReturnErr
//----------------------------------------------------------
//
// Debugging function to tell the object to print an error to stderr
// if it is returning an error.
int dmuIC::setPrintReturnErr(int _printReturnErr) {
  printReturnErr = _printReturnErr;

  return(DMU_OK);
}




/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
