//
//  dmuICjencoder.c++
//
//  Author: Tim Davison
//  Date:   Oct, 1996
//
// This file implements a jpeg encoder.  It is based on the dmuIC object.
// The dmuIC superclass of this object implements all of the steady state
// functionality of this object.  This object only overrides the initialization
// methods on the dmuIC object.  There is also a method provided for setting
// the quality of the compression. 
// The Hierachy looks like this:
//  dmuObj -> dmuIC ---> dmuICjencoder
//                  |
//                   --> dmuICjdecoder
//
#include "dmuICjencoder.h"

//****************************************************************
//
//  Public dmuICjencoder methods
//
//****************************************************************

//----------------------------------------------------------
// dmuICjencoder
//----------------------------------------------------------
//
dmuICjencoder::dmuICjencoder() {
  quality = (float)DMU_DMIC_JENCODE_DEFAULT_QUALITY/100.0;
  direction = DM_IC_CODE_DIRECTION_ENCODE;
  srcCompression = DM_IMAGE_UNCOMPRESSED;
  dstCompression = DM_IMAGE_JPEG;
}


//----------------------------------------------------------
// dmuICjencoder 
//----------------------------------------------------------
//
// constructor that takes a default height and width of the
// frames it is recieving
//
dmuICjencoder::dmuICjencoder(int _imageSizeX, int _imageSizeY) {
  quality = (float)DMU_DMIC_JENCODE_DEFAULT_QUALITY/100.0;
  direction = DM_IC_CODE_DIRECTION_ENCODE;
  srcCompression = DM_IMAGE_UNCOMPRESSED;
  dstCompression = DM_IMAGE_JPEG;
  imageSizeX = _imageSizeX;
  imageSizeY = _imageSizeY;
}


//----------------------------------------------------------
// ~dmuICjencoder
//----------------------------------------------------------
//
dmuICjencoder::~dmuICjencoder() {
}

//----------------------------------------------------------
// setup
//----------------------------------------------------------
//
// This method overrides the base class setup method.  It sets
// some more default information after it has called the base
// class method.
//
int dmuICjencoder::setup() {
  int err;

  // setup the image conversion context by calling the base classes
  //  setup method
  err = dmuIC::setup();
  DMUIC_ERR( (err != DMU_OK),           DMU_GENERIC_ERROR );

  // set the initial quality 
  err = setQuality((int)(quality*100));
  DMUIC_ERR( (err != DMU_OK),           DMU_GENERIC_ERROR );
  
  return (DMU_OK);
}


//----------------------------------------------------------
// setQuality
//----------------------------------------------------------
//
//  This method sets the quality of the image conversion prosses.
//  It translates to the Q factor in jpeg compression.  The range
//  of values that this method takes is:
//   DMU_DMIC_JENCODE_MIN_QUALITY < quality < DMU_DMIC_JENCODE_MAX_QUALITY
//  or in numerical terms:
//   0 < quality < 100
//  
//  This method then converts the number into a float between 0 and 1
//  to pass to the image conversion context.  
//
int dmuICjencoder::setQuality(int _quality) {
  DMstatus stat;

  DEBUG2( fprintf(stderr, "%s:setting quality:%d\n", getName(), newQuality) );

  // Check initial conditions for this method
  DMUIC_ERR( (imgConvCtx == NULL),      DMU_DMIC_NO_IMG_CTX_DEFINED );
  DMUIC_ERR( (_quality > DMU_DMIC_JENCODE_MAX_QUALITY),
	     DMU_DMIC_INVALID_QUALITY);
  DMUIC_ERR( (_quality < DMU_DMIC_JENCODE_MIN_QUALITY),
	     DMU_DMIC_INVALID_QUALITY);

  // set up the quality member variable.  
  quality = (float)_quality/100.0;

  DMparams *convParams;

  // create a parameter structure
  stat = dmParamsCreate(&convParams);
  DMUIC_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED );

  // set a float value within the parameter structure.  In this case, set
  // the DM_IMAGER_QUALITY_SPATIAL member variable
  stat = dmParamsSetFloat(convParams, DM_IMAGE_QUALITY_SPATIAL, quality);
  DMUIC_ERR( (stat != DM_SUCCESS),     DMU_DMIC_PARAMS_SET_STRING_FAILED);

  // Add or change this parameter in the image conversion context
  stat = dmICSetConvParams(imgConvCtx, convParams);
  DMUIC_ERR( (stat != DM_SUCCESS),     DMU_DMIC_SET_CONV_FAILED );

  // clean up the params structure
  dmParamsDestroy(convParams);

  // no errors
  return DMU_OK;
}

/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
