//
//  dmuLatch.c++
//
//  Author: Demo Factory
//  Date:   Oct, 1996
//
//
#include "dmuLatch.h"

//****************************************************************
//
//  Public dmuLatch methods
//
//****************************************************************

//----------------------------------------------------------
// dmuLatch
//----------------------------------------------------------
//
// Constructor. Initializes some data members and create a semaphore
// to use to lock the data
//
dmuLatch::dmuLatch(){
  firstBufferIn = 0;
  newBuffer = NULL;
  latchBuffer = NULL;
  newBufferLock = NULL;
  latchBufferLock = NULL;
  haveNewBuffer = DMU_FALSE;
}

//----------------------------------------------------------
// ~dmuLatch
//----------------------------------------------------------
//
// Destructor.  Frees the semaphores.
//
dmuLatch::~dmuLatch(){
  if (newBufferLock)
    usfreesema(newBufferLock,dmuArena);
    
  if (latchBufferLock)
      usfreesema(latchBufferLock,dmuArena);
}

//----------------------------------------------------------
// setup
//----------------------------------------------------------
//
// create two semaphores for later use in locking the buffers
// past to this object
//
int dmuLatch::setup(){

  // Create a lock for the latch.  Initialize to 0, since we 
  // don't want to be able to 'grab()' the dmbuffer until we 
  // have something in it.
  latchBufferLock  = usnewsema(dmuArena, 0);

  // Create a lock for the newBuffer. Initialize to 1, since we 
  // want our first receive to be able to set the newBuffer
  newBufferLock = usnewsema(dmuArena, 1);

  return (DMU_OK);
}

//----------------------------------------------------------
// init
//----------------------------------------------------------
//
// Call init on all of the children
//
int dmuLatch::init (dmuParams *_params){
  // Initialize our children
  return (dmuObj::init(_params));
}


//----------------------------------------------------------
// grab
//----------------------------------------------------------
//
// This method grabs the most recent DMBuffer that we've
// receive()'d, locks it.  This
//
int dmuLatch::grab(){
  // Lock the PBuffer - this stays locked until release() is called.
  lockLatchBuffer();

  // Lock the newBuffer to make sure that receive doesn't get called
  // while we try to associate it
  lockNewBuffer();
  
  // free the latched buffer
  dmBufferFree(latchBuffer);
  // switch the pointer so we keep track of the new buffer
  latchBuffer = newBuffer;
  // add a ref to the buffer so it stays around
  dmBufferAttach(latchBuffer);

  // unlock the new buffer
  unlockNewBuffer();
  haveNewBuffer = DMU_FALSE;
  return (DMU_OK);
}


//----------------------------------------------------------
// release
//----------------------------------------------------------
//
// This method unlocks the pbuffer, so we can assocate new
// data with it.
//
int dmuLatch::release(){
  DEBUG2(fprintf (stderr, "release\n"));
  if (latchBuffer){
    dmuObj::propagate(latchBuffer);
  }
  unlockLatchBuffer();
  return (DMU_OK);
}


//----------------------------------------------------------
// addSrcConstraints
//----------------------------------------------------------
//
// my source constraints are the constraints of my children
// so return the contstraints of the children
//
int dmuLatch::addSrcConstraints(DMparams *_params){
  return (addChildConstraints(_params));
}

//----------------------------------------------------------
// addDstConstraints
//----------------------------------------------------------
//
// my destination constraints are the constraints of my parent
// destination- so return the contstraints of my parent
//
int dmuLatch::addDstConstraints(DMparams *_params){
  if (parent)
    return (parent->addDstConstraints(_params));
  else 
    return (DMU_OK);
}



//----------------------------------------------------------
// receive
//----------------------------------------------------------
//
// When we receive a DMbuffer, we keep a pointer, newBuffer,
// which points to it.  newBuffer allways points to the latest data
// that's come down the pipe.  We leave it up to grab() to grab and
// lock this newBuffer, and to glAssociate it with the pbuffer.
//
int dmuLatch::receive(DMbuffer _inBuffer){

  DEBUG(fprintf (stderr,"dmuLatch(%s) Received video frame\n",getName()));
  DEBUG(fprintf (stderr,"lock new buf\n"));

  // Lockt newBuffer, so nobody tries to use it while we switch the
  // pointer.
  lockNewBuffer();

  if (firstBufferIn){
    DEBUG(fprintf (stderr,"de-ref prev buf\n"));
    dmBufferFree(newBuffer);
  }

  newBuffer = _inBuffer;
  dmBufferAttach (newBuffer);

  if (!firstBufferIn){
    DEBUG(fprintf (stderr,"first buf in\n"));
    firstBufferIn = 1;

    latchBuffer = newBuffer;
    dmBufferAttach(latchBuffer);

    unlockLatchBuffer();
  }
  
  // Set the newBuffer handle to point to the new buffer we're
  // received.  dmBufferAttach add's a reference to it, which we'll
  // later de-ref when we're done with it, by calling dmBufferFree 
  DEBUG2(fprintf (stderr,"set pointer and attach\n"));

  // Now the newBuffer is OK to use, so we can unlock it
  DEBUG2(fprintf (stderr,"unlock new buf\n"));
  unlockNewBuffer();
  
  haveNewBuffer = DMU_TRUE;
  return (DMU_OK);
}

//----------------------------------------------------------
// mapBuffer
//----------------------------------------------------------
//
// returns a pointer to the data within the buffer
//
unsigned char* dmuLatch::mapBuffer(){
//   assert(!(ustestsema(latchBufferLock)));
  return ((unsigned char *)(dmBufferMapData(latchBuffer)));
}

//----------------------------------------------------------
// handleEvents
//----------------------------------------------------------
//
// This method allows you to pass buffer right through the
// latch.
//
int dmuLatch::handleEvents(fd_set _fdSet){
  if (!haveNewBuffer){
    return (DMU_NOT_MY_EVENT);
  } else {
    grab();
    release();
  }
  return(DMU_OK);
}


//****************************************************************
//
//  Protected dmuLatch methods
//
//****************************************************************

// lockNewBuffer - Blocking attempt to get the semaphore on the newBuffer
//
void dmuLatch::lockNewBuffer(){
  int err;
  err = uspsema (newBufferLock);
  if (err < 0){
    fprintf (stderr,"dmuLatch::lockNewBuffer failed !\n");
  }
}

// unlockNewBuffer - Free up a locked newBuffer
//
void dmuLatch::unlockNewBuffer(){
  int err = usvsema (newBufferLock);
  if (err < 0){
    fprintf (stderr, "dmuLatch::unlockNewBuffer failed.\n");
  }
}

// lockLATCHBuffer - Blocking attempt to get the semaphore on the LatchBuffer
//
void dmuLatch::lockLatchBuffer(){
  int err;
  err = uspsema (latchBufferLock);
  if (err < 0){
    fprintf (stderr,"dmuLatch::lockLatchBuffer failed !\n");
  }
}

// unlockLatchBuffer - Free up a locked LatchBuffer
//
void dmuLatch::unlockLatchBuffer(){
  int err = usvsema (latchBufferLock);
  if (err < 0){
    fprintf (stderr, "dmuLatchBuffer::unlockLatchBuffer failed.\n");
  }
}


/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
