#include "dmuMovieIn.h"
#include <stropts.h>
#include <poll.h>

#include <sys/types.h>
#include <signal.h>
#include <wait.h>

// dmuMovieIn - This object implements all the movie file setup
// required for reading a movie off disk, and passing it down the dmu
// chain.  It also implements a separate read thread that is invoked
// by the start method.  The read thread is resposible for continually
// getting the latest data from the movie object into a dmBuffer, so
// we can propagate it down the chain.  The only behavior that a 
// subclass needs to implement is init() and getImageFrame.
//
// Setup() opens the movie file and gets all the parameters we'll need
// to set up a DMBuffer pool.
//
// Init() is handled by inheriting classes
//
// start() does a number of things:
//
// First, it creates 2 semaphores.  The currentLock semaphore is for
// internal use.  This locks the currentBuffer DMBuffer handle.  We
// need to lock this handle, because the readLoop, which is in a
// separate thread, is always trying to update the buffer pointed to
// by the current handle, but when handleEvents is called, we 
// want to make sure that the readLoop doesn't change the handle
// on us before we can pass the latest data down the chain.  This 
// semaphore is initialized to the unlocked state.
//
// The canRead semaphore is a pollable semaphore.  This is used to
// notify the user program how many DMBuffers there are available for
// reading from this object.  The user program should select or poll
// on the FD associated with canRead.  When the select falls through,
// then we have a valid movie frame to read, and handleEvents should
// be called to propagate the latest data down the chain.  Once this
// has happened, the semaphore will be decremented, so the same frame
// isn't read twice.  Note that the read thread is continually reading
// frames, so there may be another frame available right away
// (i.e. the semaphore may go above 1), and the next select on the FD
// will fall through immediately as well.  This semaphore is
// initialized to the locked (no resources available) state until a
// frame has been read.
//
// Once the semaphores are created, a thread is sproc'd off to run the
// read loop.  Killing this thread is handled in the objects
// destructor.  
//
// The read loop continually tries to read new frames into the
// DMBufferPool of the object. (this is created in the subclass's
// init() method).  The class has a DMBufferPool with room for 2
// frames in it.  The first thing the read loop does is try to
// allocate a new frame from this buffer pool.  If both frames are
// allocated and busy, then the loop blocks until there is enough room
// to allocate a buffer.
//
// There are 2 buffer handles used in the event loop: readBuffer is
// used by the class to point to the data that's currently being read
// and/or rendered from disk.  This is the buffer that we have just
// allocated from the pool. Once the readBuffer is filled (via
// getImageFrame), a lock is obtained on currentBuffer, and then the
// current currentBuffer DMbuffer is dereferenced, and then pointed at
// the readBuffer.  This means that currentBuffer is always pointing
// at the most recent valid data.  After this has happened, the
// pollable semaphore is incremented to notify the user process that
// there's a new buffer to be handled.
//
// The actual reading of the movie frame, and getting the DMBuffer to
// look the way we want it to is done in the getImageFrame method.
// This method should be implemented in subclasses of dmuMovieIn to do
// the correct reading and possibly decompression of the movie.



// Function prototypes (for non-method functions used by sproc and signal)
static void startThread(void *movieInObject);
static void stopReadLoop(int);

// constructor - initialize the fd's and other data members
//
dmuMovieIn::dmuMovieIn(){
  readThreadFD = -1;
  fileName = NULL;
  movieID    = 0;
  imageTrack = 0;
  durationTime = 0;
  firstFrameHappened = 0;
  handledCurrent = 1;
  lastFrameIndex = -1;
  bufferPool = NULL;
  currentFrameIndex = 0;
  isPlaying = 0;
  isLooping = 1;
}

// destructor - Kill the readThread if it's running.  Once this is
// dead, destry the buffer pool that we've been reading into, close up
// the movie, and free up the semaphore  that we've been using
// for the polable semaphore.
//
dmuMovieIn::~dmuMovieIn(){
  
  if (readThreadPID != -1){
    DEBUG1(fprintf (stderr, "sending kill to read thread\n"));
    kill(readThreadPID, SIGTERM);
    waitpid(readThreadPID, NULL, 0);
    readThreadPID = -1;
  }
  //if (bufferPool){
  //dmBufferDestroyPool(bufferPool);
  //}

  // free up memory resources
  delete(fileName);
  if (movieID)
    mvClose(movieID);

}

// setup - open the movie file and set all the size parameters based
// on the image track of the file.
//
int dmuMovieIn::setup(char *newFileName){
  DMstatus stat;

  if (newFileName == NULL){
    fprintf (stderr, "dmuMovieIn::setup - invalid file name\n");
    return (DMU_MOVIE_FILE_ERROR);
  }

  delete(fileName);
  fileName = strdup(newFileName);

  stat = mvOpenFile( fileName, O_RDONLY, &movieID);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "dmuMovieIn::setup - can't open file '%s'",fileName);
    return (DMU_MOVIE_FILE_ERROR);
  }
  
  stat = mvFindTrackByMedium(movieID,DM_IMAGE,&imageTrack);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "Unable to find image track in movie!\n");
    mvClose(movieID);
    return (DMU_MOVIE_FILE_ERROR);
  }

  numFrames = mvGetTrackLength(imageTrack) - 1;
  frameRate = mvGetImageRate(imageTrack);

  double tmp = (1.0/frameRate) * numFrames * 1000000000.0;
  durationTime = (long long)tmp;

  DEBUG(fprintf (stderr, "duration of the movie is:%lld\n",durationTime));
  DEBUG(fprintf (stderr, "duration of the movie is:%lf\n",tmp));
  DEBUG(fprintf (stderr, "duration of the movie is:%lf\n",
		  (double)1.0/frameRate* numFrames));

  
  imageSizeX  = mvGetImageWidth(imageTrack);
  imageSizeY = mvGetImageHeight(imageTrack);
  
  bufferSizeX = roundToPower(imageSizeX);
  bufferSizeY = roundToPower(imageSizeY);
  
  texCoordScaleXD = (double) imageSizeX  / bufferSizeX;
  texCoordScaleYD = (double) imageSizeY / bufferSizeY;
  
  texCoordScaleXF = (float) texCoordScaleXD;
  texCoordScaleYF = (float) texCoordScaleYD;

  interlacing = mvGetImageInterlacing(imageTrack);

  return (DMU_OK);
}

// init - This method is virtual and set up the the dmBuffer Pool and
// possibly a PBuffer so we can read and/or render frames from the
// movie into dmBuffers.  This depends on what kind of frames and what
// form the resulting data should take
//
int dmuMovieIn::init(dmuParams *params){
  return (DMU_VIRTUAL_CALLED);
}


// handleEvents - should be called when the main process is done
// select-ing or polling on the readThreadFD.  This will pass the
// latest movie frame down the dmu graph, and free up the buffer, so
// we can fill it again.  This will decrement the canRead semaphore,
// so we don't try to process the same frame twice.
//
int dmuMovieIn::handleEvents(fd_set fdSet){
  int err;

  if (FD_ISSET(readThreadFD, &fdSet)) {


    lockCurrentBuffer();

    // Pass on the latest buffer.
    propagate(currentBuffer);

    // Now we're done with it so de-ref it
    dmBufferFree(currentBuffer);

    // Let the other thread know that we freed it so we only free it once
    handledCurrent = 1;

    unlockCurrentBuffer();

    // Decrement the count of available dmbuffers, 'cause we just freed
    // one up 
    err = uspsema(canRead);
    
    // Debug
    DEBUG2( int test = ustestsema(canRead);
	    fprintf (stderr, "semaphore value:%d\n",test) );
    
    if (err < 0){
      return (DMU_UNIMPLEMENTED);
    }
    return (DMU_OK);
  }
  else {
    return (DMU_NOT_MY_EVENT);
  }
}


// start - Create the canRead pollable semaphore, and start the read
// loop.
//
int dmuMovieIn::start(int propagate){
  int err;

  // Create the pollable semaphore that user programs will select on 
  // to get notification of new frames
  DEBUG1(fprintf(stderr, "Creating semaphore\n"));
  canRead = usnewpollsema(dmuArena,0);
  if (!canRead){
    fprintf (stderr, "Unable to create semaphore for movie read thread\n");
    return (DMU_MOVIE_SEMA_ERROR);
  }

  readThreadFD = usopenpollsema(canRead, 0777);
  if (readThreadFD < 0){
    fprintf (stderr, "Unable to attach FD to semaphore\n");
    return (DMU_MOVIE_SEMA_ERROR);    
  }

  // Intialize the semaphore so the first select or poll will block until
  // we have good data to give them
  err = uspsema(canRead);
  if (err < 0) {
    fprintf (stderr, "uspsema error\n");
    return (DMU_MOVIE_SEMA_ERROR);    
  }
  DEBUG1( int test = ustestsema(canRead);
	  fprintf (stderr, "semaphore value:%d\n",test) );


  isPlaying = 1;

  // Create a semaphore to lock the current buffer pointer, so we 
  // know it's always safe to grab it.  Originally it is safe, since
  // we haven't read anything in.  
  currentLock = usnewsema(dmuArena, 1);

  // Sproc off the read thread - make sure that the file descriptor
  // tables stay in sync so we can poll on the same FD for the canRead
  // semaphore
  readThreadPID = sproc(startThread,PR_SADDR|PR_SFDS,this);
  DEBUG1(fprintf (stderr, "readThreadPID: %d\n", readThreadPID));
  if (readThreadPID < 0){
    fprintf (stderr, "Unable to start read thread\n");
    return (DMU_MOVIE_THREAD_ERROR);
  }

  return (dmuObj::start(propagate));
}


// Helper function that signal uses to make the readThread exit
static void stopReadLoop(int){
  fprintf (stderr, "Exiting movie read loop\n");
  exit (0);
}

// readLoop - spin and fill the buffer pool as fast as we can with
// movie frames
//
void dmuMovieIn::readLoop(){
  int err;
  DMstatus stat;
  unsigned long long nowTime;
  fd_set writeFDSet;

  // We want to die when we catch sigterm from our parent too.
  signal (SIGTERM,stopReadLoop);

  // Get the time right now.. This is our reference time for where
  // we should be playing in the movie.
  dmGetUST(&startTime);
  DEBUG1(fprintf (stderr, "STARTTIME: %d\n",startTime));
  
  // prctl so that we die when our parent object dies
  prctl (PR_TERMCHILD, NULL);

  DEBUG(fprintf (stderr, "Starting movieIn readLoop thread\n"));
  for (int i = 0; ;i++){
    

    // Wait untill there's space available for us to allocate a buffer
    DEBUG2(fprintf(stderr, "S"));
    FD_ZERO(&writeFDSet);
    FD_SET(poolFD, &writeFDSet);
    select(poolFD+1,NULL,&writeFDSet,NULL,NULL);
    
    //DEBUG(fprintf(stderr, "A"));
    // Allocate a buffer big enough for a frame
    stat = dmBufferAllocate(bufferPool, &readBuffer);
    if (stat != DM_SUCCESS){
      fprintf (stderr, "dmuMovieIn: DMBuffer allocate Failed\n");
      continue;
    }
    
    if (isPlaying){
      // Get an image frame from a movie.  This is different in the
      // different subclasses, but every subclass has to implement the
      // getImageFrame method, 'cause the read loop calls it here.
      dmGetUST(&nowTime);

      unsigned long long difTime = nowTime - startTime;

      if (!isLooping && (difTime > durationTime)){
	err = getImageFrameTime(durationTime);
	pause();
      }
      else {
	err = getImageFrameTime((nowTime - startTime)%durationTime);
      }

      while (err == DMU_MOVIE_FRAME_REPEAT){
	DEBUG2(fprintf(stderr, "repeat frame\n"));
	dmGetUST(&nowTime);
	err = getImageFrameTime((nowTime - startTime)%durationTime);
	sginap(2);
      }
    }
    else {
      err = getImageFrameIndex(currentFrameIndex);
      if (err == DMU_MOVIE_FRAME_REPEAT){
	dmBufferFree(readBuffer);
	sginap(1);
	continue;
      }
    }
    
    // Here we free up the old data that currentBuffer pointed to and
    // point it at the data we just read in.  We lock currentBuffer,
    // so nobody (i.e. the read thread) tries to use the old data
    // before we can swap the pointer and free the memory
    lockCurrentBuffer();

    if (!firstFrameHappened){
      firstFrameHappened = 1;
    }
    else{
      if (!handledCurrent){
	// Decrement the count of available dmbuffers, because we're
	// freeing up the old currentBuffer before setting the new
	// pointer.
	err = uspsema(canRead);
	dmBufferFree(currentBuffer);
      }
    }
    handledCurrent = 0;

    // Set the currentBuffer pointer to the latest frame
    currentBuffer = readBuffer;

    // Now currentBuffer is safe.
    unlockCurrentBuffer();

    // Increment the count of available buffers, so anyone blocking
    // will know there's a frame ready
    DEBUG2(fprintf(stderr, "V"));
    err = usvsema(canRead); //increment count of available buffers

    // Debug stuff to print out the state of the pollable semaphore
    if (err < 0){
      fprintf (stderr, "Unable to increment semaphore\n");
    }
    DEBUG2(fprintf(stderr, "."));
  }

}

// startThread - helper function to start the readLoop method in a
// dmuMovieIn object
//
static void startThread(void *movieInObject){
  ((dmuMovieIn *)movieInObject)->readLoop();
}

// getFD - return the file descriptor of the canRead semaphore.
//
int dmuMovieIn::getFD(){
  return (readThreadFD);
}

void dmuMovieIn::pause(){
  isPlaying = 0;
}

void dmuMovieIn::gotoFrame(int frameIndex){
    unsigned long long nowTime;

  if (frameIndex >= numFrames){
    currentFrameIndex = numFrames -1;
  }
  else {
    currentFrameIndex = frameIndex;
  }

  // back calculate the startTime to keep it in sync
  dmGetUST(&nowTime);
  long long movieTime = (long long) ((double)currentFrameIndex/frameRate * 1000000000);
  startTime = nowTime - movieTime;
}

void dmuMovieIn::resume(){
  isPlaying = 1;
}



// Wrapper functions around the locking of the currentBuffer pointer.
//
int dmuMovieIn::lockCurrentBuffer(){
  uspsema (currentLock);
  return (DMU_OK);
}

int dmuMovieIn::unlockCurrentBuffer(){
  usvsema (currentLock);
  return (DMU_OK);
}



int dmuMovieIn::setBufferDimensions(int newSizeX,int newSizeY){
  if ((newSizeX < imageSizeX) || (newSizeY < imageSizeY)){
    return (DMU_MOVIE_BAD_BUFFER_SIZE);
  }
  else {
    bufferSizeX = newSizeX;
    bufferSizeY = newSizeY;

    texCoordScaleXD = (double) imageSizeX  / bufferSizeX;
    texCoordScaleYD = (double) imageSizeY / bufferSizeY;
    
    texCoordScaleXF = (float) texCoordScaleXD;
    texCoordScaleYF = (float) texCoordScaleYD;

    return (DMU_OK);
  }
}

/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
