#include "dmuMovieInCompressed.h"

// dmuMovieInCompressed implements the functionality needed to read,
// raw, compressed JPEG frames from a movie file into DMbuffers.  To
// do this, the object implements the getImageFrame method to use the
// movie library to perform an mvGetCompressedImageSize(3dm) operation
// into an already allocated DMBuffer.

// The getImageFrame() method figures out which frame to play.  It
// then gets a pointer to the data portion of the readBuffer DMBuffer,
// and tells the movie library to read the frame into the readBuffer.
// Since this is not using the Rendering capabilities of the movie
// library, no OpenGL PBuffers or contexts are needed (as they are in
// dmuRenderMovieIn).  Note that DMBuffer allocation is handled in
// dmuMovieIn during the readLoop().

dmuMovieInCompressed::dmuMovieInCompressed(){
}

dmuMovieInCompressed::~dmuMovieInCompressed(){

}

// init - set up a dmBuffer pool with parameters for a compressed,
// top-to-bottom movie of the correct size for the movie that we
// opened in setup().  We don't create a pBuffer here, like we did in
// dmuRenderMovie.c++, because we don't do any GL rendering -  We just
// read the bits off the disk.
int dmuMovieInCompressed::init(dmuParams *params){
  DMstatus stat;
  int err;

  DEBUG(fprintf(stderr, "Enter MovieIn Init\n"));

  // Set up the parameter structure - right now we're assuming
  // jpeg compressed frames or fields
  stat = dmParamsCreate(&poolParams);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  //  stat = dmSetImageDefaults(poolParams,bufferSizeX,
  //			    bufferSizeY,DM_IMAGE_PACKING_RGBA);
  stat = dmSetImageDefaults(poolParams,bufferSizeX,
			    bufferSizeY,DM_IMAGE_PACKING_CbYCrY);
  dmParamsSetString(poolParams, DM_IMAGE_COMPRESSION, DM_IMAGE_JPEG);
  dmParamsSetEnum(poolParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM);
  int bufSize = dmImageFrameSize(poolParams);

  // clean up the params structure
  dmParamsDestroy(poolParams);

  stat = dmParamsCreate(&poolParams);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  stat = dmBufferSetPoolDefaults(poolParams, 2, bufSize, DM_FALSE, DM_FALSE);
  if (stat != DM_SUCCESS){
    return (DMU_POOL_DEFAULTS_FAILED);
  }
  
  // Constrain based our our childrens' source constraints
  DEBUG(fprintf(stderr, "Add Child Constraints \n"));
  err = addChildConstraints(poolParams);
  if (err != DMU_OK){
    return (DMU_BAD_CONSTRAINTS);
  }

  stat = dmParamsSetInt( poolParams, DM_BUFFER_COUNT, 5);
  if (stat != DMU_OK){
    return (DMU_BAD_CONSTRAINTS);
  }

  // Create the DMBufferPool
  DEBUG(fprintf (stderr, "Create Pool\n"));
  DEBUG(fprintf (stderr, "Parameters:\n"));
  DEBUG(printDMParams(poolParams));
  stat = dmBufferCreatePool (poolParams, &bufferPool);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "error:%d",stat);
    return (DMU_POOL_CREATE_FAILED);
  }


  // Let the pool know how much room we need to read a frame from disk
  // into a dmBuffer.  Be conservative, and assume uncompressed frame
  // sizes
 

  // Allocate a buffer big enough for a frame so we can get the byte count
  stat = dmBufferAllocate(bufferPool, &readBuffer);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "dmbuffer allocate Failed\n");
    return  (DMU_POOL_CREATE_FAILED);
  }
  
  int bytes = dmBufferGetSize(readBuffer);
  
  dmBufferFree(readBuffer);
  
  DEBUG(fprintf (stderr, "Set Select size\n"));

  stat = dmBufferSetPoolSelectSize(bufferPool,bytes);
  if (stat != DM_SUCCESS){
    dmBufferDestroyPool(bufferPool);
    return (DMU_POOL_CREATE_FAILED);
  }

  // Get the file descriptor associated with our buffer pool - This
  // lets us select and block until there's room before we try to
  // allocate a DMBuffer.
  DEBUG(fprintf (stderr, "Get FD\n"));
  poolFD = dmBufferGetPoolFD(bufferPool);
  if (poolFD < 0){
    dmBufferDestroyPool(bufferPool);
    return (DMU_POOL_CREATE_FAILED);
  }
  //--End: Create dmbuffer pool
  
  return (dmuObj::init(params));

}

// getImageFrame - Determines what frame corresponds to the time
// passed in as movieTime. The DMbuffer has already been allocated by
// the readLoop, and is pointed to by readBuffer, so we map the buffer
// so that we have a memory pointer.  Then just read into the memory
// pointed to by pointer.
int dmuMovieInCompressed::getImageFrameTime(long long movieTime){
  currentFrameIndex = (int) (frameRate * (double)movieTime/1000000000);

  return (getImageFrameIndex(currentFrameIndex));
}


int dmuMovieInCompressed::getImageFrameIndex(int frameIndex){
  int bufSize;
  char *bufPtr;

  if (frameIndex == lastFrameIndex){
    DEBUG2(fprintf(stderr, "getImageFrame(frame:%d):%d\n",frameIndex,bufSize));
    return (DMU_MOVIE_FRAME_REPEAT);
  }
  
  bufSize = mvGetCompressedImageSize(imageTrack, frameIndex);
  DEBUG2(fprintf(stderr, "getImageFrame(frame:%d):%d\n",frameIndex,bufSize));

  bufPtr = (char *)dmBufferMapData(readBuffer);
  mvReadCompressedImage(imageTrack,frameIndex,bufSize,bufPtr);
  dmBufferSetSize(readBuffer,bufSize);

  lastFrameIndex = frameIndex;
  return (DMU_OK);

}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
