#include "dmuMovieInRender.h"

// dmuMovieInRender implements the functionality needed to read,
// decompress and render a movie into an off-screen rendering area.
// To do this, the object implements the getImageFrame method to use
// the movie library to perform an mvRenderMovieToOpenGL (3dm)
// operation into an off screen rendering area.

// The init() method handles setting up both a DMBuffer Pool, and an
// OpenGL PBuffer.  The DMBuffer is where the memory comes from to
// read our frames into. The PBuffer sets up an off-screen rendering
// area that can be associated with a DMBuffer to allow Open GL to
// perform rendering operations into the DMBuffer.

// The getImageFrame() method figures out which frame to play, and
// tells OpenGL to render the frame into the PBuffer (which is
// associated with the most recently allocated DMBuffer (called
// readBuffer).  Note that DMBuffer allocation is handled in
// dmuMovieIn during the readLoop().


dmuMovieInRender::dmuMovieInRender(){
  dpy = NULL;
  pbufParams = NULL;
  fbConfig = NULL;
  offsetX = offsetY = 0;
  zoomX = zoomY = 1.0;
}

dmuMovieInRender::~dmuMovieInRender(){

}

// init - Set up a GL PBuffer, and a dmBuffer Pool.  When we render a
// frame, we'll allocate a buffer from the pool, associate the GL
// PBuffer with it, Make the PBuffer the current drawable, and render
// into the PBuffer (and hence the DMBuffer).  This means that 
// the PBuffer and DMbuffer pool have to have consistent constraints
// between them.
int dmuMovieInRender::init(dmuParams *params){
  DMstatus stat;
  int err;

  DEBUG1(fprintf(stderr, "Enter MovieIn Init\n"));

  int attrib[] = {
    GLX_PRESERVED_CONTENTS_SGIX,    GL_TRUE,
    GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE,
    (int) None
  };

  // Create 2 dmparam structures - one for the pbuffer and one for the
  // pool.  Any overlapping values that are set must be consistent 
  // between the two
  stat = dmParamsCreate(&pbufParams);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  stat = dmSetImageDefaults(pbufParams,bufferSizeX,
			    bufferSizeY,DM_IMAGE_PACKING_RGBA);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }
  
  stat = dmParamsSetEnum(pbufParams,DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  // Constrain the parameters based on our children 
  DEBUG1(fprintf(stderr, "Add Child Constraints \n"));
  err = addChildConstraints(pbufParams);
  if (err != DMU_OK){
    return (DMU_BAD_CONSTRAINTS);
  }
  DEBUG1(fprintf(stderr, "Done with Add Child Constraints \n"));


  //--Start: Create PBuffer

  // Open the local display 
  dpy = XOpenDisplay(":0");
  fprintf (stderr, "xdisplay:%x\n",dpy);

  //  Get a FBconfig that we can use
  err = setupFBConfig(dpy,8,8,8,8,&fbConfig);
  if (stat != DMU_OK){
    fprintf (stderr, "dmuMovieIn::init - Unable to setup fb config\n");
    return (stat);
  }
  
  // Create the PBuffer
  pbuffer = glXCreateGLXPbufferSGIX(dpy,
				    fbConfig,
				    bufferSizeX,
				    bufferSizeY,
				    attrib);
  if (pbuffer == None){
    return (DMU_GL_PBUFFER_ALLOC_FAILED);
  }

  // Create a context for the PBuffer, so we can draw to it.
  context = glXCreateContextWithConfigSGIX(dpy,fbConfig,
					   GLX_RGBA_TYPE_SGIX,
					   NULL, GL_TRUE);

  if (!context){
    fprintf (stderr, "Unable to create context \n");
    exit (-1);
  }
  //--End: Create PBuffer


  //--Start: Create dmbuffer pool

  // Setup buffer pool parameters
  stat = dmParamsCreate(&poolParams);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  stat = dmBufferSetPoolDefaults(poolParams, 0, 0,
				 DM_FALSE, DM_FALSE);
  if (stat != DM_SUCCESS){
    return (DMU_POOL_DEFAULTS_FAILED);
  }

  stat = dmBufferGetGLPoolParams(pbufParams,poolParams);
  if (stat != DM_SUCCESS){
    return (DMU_POOL_CREATE_FAILED);
  }

  stat = dmParamsSetInt( poolParams, DM_BUFFER_COUNT, 25);
  if (stat != DMU_OK){
    return (DMU_BAD_CONSTRAINTS);
  }


  // Create the buffer pool
  DEBUG1(fprintf (stderr, "Create Pool\n"));
  DEBUG1(fprintf (stderr, "Parameters:\n"));
  DEBUG1(printDMParams(poolParams));
  stat = dmBufferCreatePool (poolParams, &bufferPool);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "error:%d",stat);
    return (DMU_POOL_CREATE_FAILED);
  }

  // Allocate a buffer big enough for a frame so we can get the byte count
  stat = dmBufferAllocate(bufferPool, &readBuffer);
  if (stat != DM_SUCCESS){
    fprintf (stderr, "dmbuffer allocate Failed\n");
    return  (DMU_POOL_CREATE_FAILED);
  }

  int bytes = dmBufferGetSize(readBuffer);
  
  dmBufferFree(readBuffer);

  // Tell the pool how much room we'll need to allocate for a frame
  DEBUG1(fprintf (stderr, "Set Select size:%d\n",bytes));
  stat = dmBufferSetPoolSelectSize(bufferPool,(bufferSizeX * bufferSizeY));
  if (stat != DM_SUCCESS){
    dmBufferDestroyPool(bufferPool);
    return (DMU_POOL_CREATE_FAILED);
  }

  // Get the file descriptor associated with the pool so we can 
  // select on it to make sure that there's space to allocate
  // a frame-size DMBuffer.
  poolFD = dmBufferGetPoolFD(bufferPool);
  if (poolFD < 0){
    dmBufferDestroyPool(bufferPool);
    return (DMU_POOL_CREATE_FAILED);
  }
  //--End: Create dmbuffer pool

  // Initialize our children
  return (dmuObj::init(params));
}

// getImageFrame - Determines what frame corresponds to the time
// passed in as movieTime.  Make our pbuffer the current context, and
// render the frame the DMbuffer.  The DMbuffer has already been
// allocated by the readLoop, and is pointed to by readBuffer.

int dmuMovieInRender::getImageFrameIndex (int frameIndex){
  currentFrameIndex = frameIndex;
  double tmp = (((double)1.0/frameRate * frameIndex) * 1000000000.0);
  
  DEBUG2(fprintf(stderr, "getImageFrameIndex(frame:%ld)\n",currentFrameIndex)); 
  DEBUG2(fprintf(stderr, "getImageFrameIndex(tmp:%lf)\n",tmp));


  return getImageFrameTime((long long) tmp);
}

int dmuMovieInRender::getImageFrameTime (long long movieTime){
  int err;

  
  currentFrameIndex = (int) (frameRate * (double)movieTime/1000000000);
  DEBUG2(fprintf(stderr, "getImageFrame(time:%lld)\n",movieTime));
  DEBUG2(fprintf(stderr, "getImageFrame(frame:%d)\n",currentFrameIndex));

  if (currentFrameIndex == lastFrameIndex){
    //fprintf (stderr, "repeat\n");
    return (DMU_MOVIE_FRAME_REPEAT);
  }

  // Point the pbuffer to the frame
  err = glXAssociateDMPbufferSGIX(dpy,pbuffer,pbufParams,readBuffer);
  if (!err){
    fprintf (stderr, "\nglxAssociate Failed\n");
    dmBufferFree(readBuffer);
    return (DMU_GL_PBUFFER_ASSOC_FAILED);
  }
  

  // Render the frame into the pbuffer
  err = glXMakeCurrent(dpy,pbuffer,context);

  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT);

  //glTranslatef(1,1,0);
  DEBUG2(fprintf(stderr, "R"));

  mvRenderMovieToOpenGL(movieID,movieTime,1000000000);

  lastFrameIndex = currentFrameIndex;

  DEBUG2(fprintf (stderr, "exit getImageFrame\n"));
  return (DMU_OK);
}
 

void dmuMovieInRender::setImageOffset(float x, float y){
  MVrect rect;

  offsetX = x;
  offsetY = y;
  
  rect.left = (int)offsetX;
  rect.right = (int)(offsetX + (imageSizeX * zoomX));
  rect.bottom = (int)offsetY;
  rect.top = (int)(offsetY + (imageSizeY * zoomY));

  mvSetMovieRect(movieID,rect);
}

void dmuMovieInRender::setImageZoom(float x, float y){
  MVrect rect;

  zoomX = x;
  zoomY = y;
  
  rect.left = (int)offsetX;
  rect.right = (int)(offsetX + (imageSizeX * zoomX));
  rect.bottom = (int)offsetY;
  rect.top = (int)(offsetY + (imageSizeY * zoomY));

  mvSetMovieRect(movieID,rect);  
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
