//
//  dmuNetworkRecv.c++
//
//  Author: Demo Factory
//  Date:   Oct, 1996
//
//
//  This object read data from across the network and put
//  the data into dmbuffers to pass down the chain.  It uses
//  a special read thread to get the data from the network.
//  Using this extra read thread allows this object to be
//  asychronous.  
//
#include "dmuNetworkRecv.h"

// system header files needed for the networking stuff
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>

// prototypes for some helper functions
static void sprocReadLoop(void *dmuNetworkRecv);
static void stopReadLoop(int);

//****************************************************************
//
//  Public dmuNetworkRecv methods
//
//****************************************************************

//----------------------------------------------------------
// dmuNetworkRecv
//----------------------------------------------------------
//
// Constructor. Does nothing beyond initaliziation of
// member variables
//
dmuNetworkRecv::dmuNetworkRecv() {
  imageSizeX = 0;
  imageSizeY = 0;
  packing = DM_IMAGE_PACKING_CbYCrY;

  readLoopPID = -1;
  firstFrameHappened = 0;
  poolFD = -1;
  handledCurrent = 1;
  canRead = NULL;
  currentLock = NULL;
  currentBuffer = NULL;
  readBuffer = NULL;
  bufferPool = NULL;
  lastHeaderVal = 0;
}

//----------------------------------------------------------
// dmuNetworkRecv
//----------------------------------------------------------
//
// Constructor.
// This constructor will also setup the correct image size
// to work with
//
dmuNetworkRecv::dmuNetworkRecv(int _imageSizeX, int _imageSizeY) {
  imageSizeX = _imageSizeX;
  imageSizeY = _imageSizeY;
  packing = DM_IMAGE_PACKING_CbYCrY;
}

//----------------------------------------------------------
// ~dmuNetworkRecv
//----------------------------------------------------------
//
// Destructor
//
dmuNetworkRecv::~dmuNetworkRecv() {
}

//----------------------------------------------------------
// init
//----------------------------------------------------------
//
// This method creates a pool of dmbuffer.  These buffers are
// used to hold the data from the network.  Then these buffers
// are passed down the line to the children.
//
int dmuNetworkRecv::init(dmuParams *) {
  DMstatus stat;
  DMparams *params;
  int err;
  int bufferSize = 0;
  
  DEBUG2( fprintf(stderr, "%s: init\n", getName()));

  DMparams *tmpParams;
  // create a parameter structure
  dmParamsCreate(&tmpParams);
  // set up the defaults for the buffers
  dmSetImageDefaults(tmpParams, imageSizeX, imageSizeY, packing);
  // get the size of each buffer 
  bufferSize = dmImageFrameSize(tmpParams);
  
  // create a parameter structure
  stat = dmParamsCreate(&params); 
  DMUNET_ERR( (stat != DM_SUCCESS),      DMU_PARAMS_CREATE_FAILED);

  // setup some parameters to create the buffer pool with
  stat = dmBufferSetPoolDefaults(params, 2, bufferSize,
				 DMU_NOT_CACHED, DMU_NOT_MAPPED);
  DMUNET_ERR( (stat != DM_SUCCESS),      DMU_POOL_DEFAULTS_FAILED );

  // add the constraints of the children of this object
  err = addChildConstraints(params);
  DMUNET_ERR( (err != DMU_OK),           DMU_BAD_CONSTRAINTS );

  DEBUG2( printDMParams(params) );

  // create the buffer pool
  stat = dmBufferCreatePool(params, &bufferPool);
  DMUNET_ERR( (stat != DM_SUCCESS),      DMU_POOL_CREATE_FAILED);

  // clean up the params structure
  dmParamsDestroy(params);    

  // grab an fd associated w/ the pool
  poolFD = dmBufferGetPoolFD(bufferPool);
  if (poolFD < 0){
    dmBufferDestroyPool(bufferPool);
    return (DMU_POOL_CREATE_FAILED);
  }

  // seutp a pollable semaphore to lock the 
  DEBUG1(fprintf(stderr, "Creating semaphore\n"));
  canRead = usnewpollsema(dmuArena,0);
  if (!canRead){
    fprintf (stderr, "Unable to create semaphore for movie read thread\n");
    return (DMU_MOVIE_SEMA_ERROR);
  }

  // attach an fd to the pollable semaphore we just created
  canReadFD = usopenpollsema(canRead, 0777);
  if (canReadFD < 0){
    fprintf (stderr, "Unable to attach FD to semaphore\n");
    return (DMU_MOVIE_SEMA_ERROR);    
  }

  // Intialize the semaphore so the first select or poll will block until
  // we have good data to give them
  err = uspsema(canRead);
  if (err < 0) {
    fprintf (stderr, "uspsema error\n");
    return (DMU_MOVIE_SEMA_ERROR);    
  }

  // Create a semaphore to lock the current buffer pointer, so we 
  // know it's always safe to grab it.  Originally it is safe, since
  // we haven't read anything in.  
  currentLock = usnewsema(dmuArena, 1);

  return(dmuObj::init());
}

//----------------------------------------------------------
// start
//----------------------------------------------------------
//
// This method will start a thread that will read buffers from
// the network.  
// 
int dmuNetworkRecv::start(int _propagate) {
  // if there is no data object, we have an error condition
  if (!dataObj){
    fprintf (stderr,"dmuNetworkRecv::Start occured before setting data obj\n");
    return (DMU_NETWORK_ERROR);
  }

  // If a read thread exists, kill it
  if (readLoopPID != -1){
    kill (readLoopPID, SIGTERM);
    readLoopPID = -1;
  }

  // set the started state.  This will be set later again, but
  // the read loop needs this.
  started = DMU_TRUE;
  
  // sproc off the read thread.  
  readLoopPID = sproc(sprocReadLoop, PR_SADDR|PR_SFDS,this);
  DEBUG(fprintf (stderr, "netRecv:readThreadPID: %d\n", readLoopPID));

  return(dmuNetwork::start(_propagate));
}

//----------------------------------------------------------
// receive
//----------------------------------------------------------
//
// This method should never be called on this object.
// This object can never be anything but the head of a chain
// so nothing should pass buffers to this object.
//
int dmuNetworkRecv::receive(DMbuffer _inBuffer) {
  return(DMU_OK);
}

//----------------------------------------------------------
// handleEvents
//----------------------------------------------------------
//
// This method will deal with a new buffer of data just recieved
// from the network.  
// 
int dmuNetworkRecv::handleEvents(fd_set _fdSet) {
  int err;
  char *buf = NULL;

  // if the object is in the started state, do something with the
  // data
  if (isStarted() && FD_ISSET(canReadFD, &_fdSet)){

    lockCurrentBuffer();

    // Pass on the latest buffer.
    propagate(currentBuffer);

    // Now we're done with it so de-ref it
    dmBufferFree(currentBuffer);

    handledCurrent = 1;

    unlockCurrentBuffer();

    // Decrement the count of available dmbuffers, 'cause we just freed
    // one up 
    err = uspsema(canRead);

    // everything is fine so return
    return(DMU_OK);
  }
  else {
    return (DMU_NOT_MY_EVENT);
  }
}

 
//----------------------------------------------------------
// stop
//----------------------------------------------------------
//
// This method stops the read loop inside the object
//
int dmuNetworkRecv::stop(){

  lockCurrentBuffer();

  int test = ustestsema(canRead);
  DEBUG( fprintf (stderr, "(stop) : semaphore value:%d\n",test) );

  if (test == 0){
    dmBufferFree(currentBuffer);
    handledCurrent = 1;
    uspsema(canRead);
  }

  if (readLoopPID != -1){
    kill (readLoopPID, SIGTERM);
    readLoopPID = -1;
    DEBUG( fprintf (stderr, "waiting for child to die\n") );
    waitpid(readLoopPID, &test, NULL);
    DEBUG( fprintf (stderr, "he's dead\n") );
  }

  int err = dmuNetwork::stop();
  unlockCurrentBuffer();
  return (err);
}


//----------------------------------------------------------
// readLoop
//----------------------------------------------------------
//
//
void dmuNetworkRecv::readLoop(){
  int dataSize;
  int err;
  
  // We want to die when we catch sigterm from our parent too.
  signal (SIGTERM,stopReadLoop);

  // prctl so that we die when our parent object dies
  prctl (PR_TERMCHILD, NULL);

  while (1){
    fd_set writeFDSet;
    fd_set readFDSet;
    DMstatus stat;
    int err;

    // Wait untill there's space available for us to allocate a buffer
    //DEBUG(fprintf(stderr, "S"));
    FD_ZERO(&writeFDSet);
    FD_SET(poolFD, &writeFDSet);
    select(poolFD+1,NULL,&writeFDSet,NULL,NULL);

    //DEBUG(fprintf(stderr, "A"));
    // Allocate a buffer big enough for a frame
    stat = dmBufferAllocate(bufferPool, &readBuffer);
    if (stat != DM_SUCCESS){
      fprintf (stderr, "dmuNetworkRecv: DMBuffer allocate Failed\n");
      continue;
    }

    FD_ZERO(&readFDSet);
    int dataFD;
    dataObj->getFD(&dataFD);
    FD_SET(dataFD,&readFDSet);

    select(dataFD + 1, &readFDSet, NULL, NULL, NULL);

    lockCurrentBuffer();

    if (isStarted()){
      if (!dataObj){
	readLoopPID = -1;
	fprintf (stderr, "no data obj.. bye!\n");
	unlockCurrentBuffer();
	exit(0);
      }
      
      char *buf = (char *) dmBufferMapData(readBuffer);
      DEBUG1(fprintf(stderr, "%s: reading the header\n", getName()));
      err = dataObj->getHeader(&lastHeaderVal);
      if (err != PENT_OK){
	fprintf (stderr, "no data header\n");
	readLoopPID = -1;
	unlockCurrentBuffer();
	exit(0);
      }
      
      DEBUG1(fprintf(stderr, "%s: reading the dataSize\n", getName()));
      err = dataObj->getSize(&dataSize);
      if (err != PENT_OK || dataSize == 0){
	fprintf (stderr, "no data size\n");
	readLoopPID = -1;
	unlockCurrentBuffer();
	exit(0);
      }
      DEBUG1(fprintf(stderr, "%s: dataSize:%d\n", getName(), dataSize));
      
      DEBUG1(fprintf(stderr, "%s: reading the data\n", getName()));
      err = dataObj->getData(buf);
      if (err != PENT_OK){
	fprintf (stderr, "no data \n");
	readLoopPID = -1;
	unlockCurrentBuffer();
	exit(0);
      }
      
      dmBufferSetSize(readBuffer, dataSize);
      
      if (!firstFrameHappened){
	firstFrameHappened = 1;
      }
      else{
	if (!handledCurrent){
	  // Decrement the count of available dmbuffers, because we're
	  // freeing up the old currentBuffer before setting the new
	  // pointer.
	  err = uspsema(canRead);
	  dmBufferFree(currentBuffer);
	}
      }
      handledCurrent = 0;
      
      // Set the currentBuffer pointer to the latest frame
      currentBuffer = readBuffer;
      
    }
    else {
      fprintf (stderr, "I should stop \n");
      readLoopPID = -1;
      unlockCurrentBuffer();
      exit(0);
    }

    // Now currentBuffer is safe.
    unlockCurrentBuffer();


    // Increment the count of available buffers, so anyone blocking
    // will know there's a frame ready
    err = usvsema(canRead); //increment count of available buffers
    
    DEBUG1( int test = ustestsema(canRead);
	   fprintf (stderr, "semaphore value:%d\n",test) );

  }
}

//----------------------------------------------------------
// getFD
//----------------------------------------------------------
//
//
int dmuNetworkRecv::getFD(){
  return (canReadFD);
}

//****************************************************************
//
//  Static dmuNetworkRecv functions
//
//****************************************************************

//----------------------------------------------------------
// stopReadLoop
//----------------------------------------------------------
//
static void stopReadLoop(int){
  fprintf (stderr, "stopping netRecv read loop\n");
  exit (0);
}

//----------------------------------------------------------
// sprocReadLoop
//----------------------------------------------------------
//
// helper method to call the objects read thread
//
static void sprocReadLoop(void *_recvObj){
  ((dmuNetworkRecv *)_recvObj)->readLoop();
}



/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
