#include "dmuVideo.h"

// The dmuVideo object handles all of the functionality that is common
// between videoIn and videoOut objects.  Theis includes openning and
// closing the video server, destroying the path on destruction, and
// setting up the control values for the video path.  
//


// Static members that help us make sure there's only one video server
// connection open at a time.
VLServer dmuVideo::server = NULL;
int dmuVideo::videoOpen   = DMU_FALSE;
int dmuVideo::nObjects = 0;

// Constructor - Initialize member variables.  If this is the first
// object to be created,  then open a connection to the video server
// which all the dmuVideo objects will use.
dmuVideo::dmuVideo(){
  drainNode   = 0;
  sourceNode  = 0;
  path        = 0;
  transferStarted = DMU_FALSE;
  
  // Keep track of the number of dmuVideo objects
  nObjects += 1;

  // Start up the video server if we have to 
  if (videoOpen == DMU_FALSE){
    int err = dmuVideo::openVideo();
    if (err == DMU_OK){
      videoOpen = DMU_TRUE;
    }
    else {
      videoOpen = DMU_FALSE;
    }
  }
}

// Destructor - Destroy our video path, and shut down the video
// server, if we need to.
dmuVideo::~dmuVideo(){
  nObjects -= 1;

  // Destroy the video path 
  if (path) {
    DEBUG1(fprintf (stderr, "destroying video path\n"));
    vlSetupPaths(server, &path, 1, VL_DONE_USING, VL_DONE_USING);
    vlEndTransfer (server,path);
    vlDestroyPath(server,path);
  }

  // Shut down the video server if we're the last dmuVideo object
  if (nObjects <= 0){
    nObjects = 0;
    if (videoOpen == DMU_TRUE){
      dmuVideo::closeVideo();
      videoOpen = DMU_FALSE;
    }
  }
}


// addSrcConstraints - Need to implement in subclasses.  There are 
// no constraints common to all video objects
int dmuVideo::addSrcConstraints(DMparams *params){
  return (DMU_UNIMPLEMENTED);
}

// addDstConstraints - Constrain the params structure based on our
// video path. This is used to constrain our the parameters for a
// the DMbufferPool that the video path will be dumping into (usually
// used internally)
int dmuVideo::addDstConstraints(DMparams *params){
  // Constrain params based on our video path
  int stat;
  int bufCount,prevBufCount;
  
  // Remember how many buffers the params used to have, 
  prevBufCount = dmParamsGetInt( params, DM_BUFFER_COUNT );
  dmParamsSetInt( params, DM_BUFFER_COUNT, 0 );

  // Get the params from VL so we know how to create a dmbuffer pool
  stat = vlDMPoolGetParams(server, path, drainNode, params);
  if (stat != 0){
    char errorBuf[DM_MAX_ERROR_DETAIL];
    fprintf(stderr, "dmuVideo - Error description:%s\n", 
	    dmGetError(NULL, errorBuf));
    fprintf(stderr, "   detail:%s\n", errorBuf);
    return (DMU_BAD_CONSTRAINTS);
  }
  
  // Since the last call nukes all the buffer count constraint, we'll 
  // add the old buffer count back int.
  bufCount = dmParamsGetInt( params, DM_BUFFER_COUNT );
  dmParamsSetInt( params, DM_BUFFER_COUNT, prevBufCount + bufCount );

  return (DMU_OK);
}

int dmuVideo::setup(){
  return (DMU_VIRTUAL_CALLED);
}

int dmuVideo::init(dmuParams *params){
  return (DMU_VIRTUAL_CALLED);
}

int dmuVideo::handleEvents(fd_set fdSet){
  return (DMU_VIRTUAL_CALLED);
}

int dmuVideo::receive(DMbuffer inBuffer){
  return (DMU_VIRTUAL_CALLED);
}


//
// Static methods
//

// openVideo - open a connection to the video server
//
int dmuVideo::openVideo(){
  DEBUG(fprintf (stderr, "Opening video server\n"));

  server = vlOpenVideo("");
  if (!server){
    fprintf (stderr, "Unable to open video server\n");
    vlPerror("Video Error in openVideo: ");
    return (DMU_VIDEO_ERROR);
  }

  return (DMU_OK);
}

// closeVideo - close the connection to the video server
//
int dmuVideo::closeVideo(){
  DEBUG(fprintf (stderr, "Closing video server connection\n"));
  vlCloseVideo(server);
  return (DMU_OK);
}


// start - begin the transfer, regardless of direction.  Also propagate
//         the start message down to my children
int dmuVideo::start(int propagate){
  int stat;
  
  if (transferStarted == DMU_TRUE){
    return (DMU_VIDEO_STARTED);
  }

  // Start up the video transfer
  stat = vlBeginTransfer(server, path, 0, NULL);
  if (stat != 0){
    fprintf (stderr, "\n\nFAILED TO START VIDEO TRANSFER\n");
    return (DMU_VIDEO_START_FAILED);
  }

  DEBUG1(fprintf (stderr, "\nStarted Video Transfer\n"));
  
  transferStarted = DMU_TRUE;

  // Start up my children
  return (dmuObj::start(propagate));
}


// Return the file descriptor associated with the video path
//
int dmuVideo::getFD(){
  int stat;
  int fd;

  stat = vlPathGetFD (server, path, &fd);
  if (stat != 0){
    return (-1);
  }
  else{
    return (fd);
  }
}


// Helper functions for setting and getting VL control values
//
int dmuVideo::setIntControl(VLPath path, VLNode node,int control, int value){
  VLControlValue val;
  int stat;
  
  val.intVal = value;
  stat = vlSetControl (server, path, node, control, &val);

  if (stat){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }
  else {
    return (DMU_OK);
  }
}

int dmuVideo::setFractControl (VLPath path, VLNode node, 
			       int control, int num, int denom){
  VLControlValue val;
  int stat;
  
  val.fractVal.numerator   = num;
  val.fractVal.denominator = denom;
  stat = vlSetControl (server, path, node, control, &val);

  if (stat){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }
  else {
    return (DMU_OK);
  }
}

int dmuVideo::setXYControl (VLPath path, VLNode node, 
			       int control, int x, int y){
  VLControlValue val;
  int stat;
  
  val.xyVal.x   = x;
  val.xyVal.y   = y;
  stat = vlSetControl (server, path, node, control, &val);


  if (stat){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }
  else {
    return (DMU_OK);
  }
}


int dmuVideo::getXYControl (VLPath path, VLNode node,int control, 
			    int *returnX, int *returnY){
  VLControlValue val;
  int stat;
  
  stat = vlGetControl( server, path, node, control, &val );
  if ( stat != 0){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }
  else {
    *returnX = val.xyVal.x;
    *returnY = val.xyVal.y;
    return (DMU_OK);
  }
 
}


// getBufferBytes - return the size, in bytes, of the video transfer
// associated with the current path. (in bytes)
//
int dmuVideo::getBufferBytes(int *returnSize){
  int val;

  if (path <= 0){
    return (DMU_VIDEO_NO_PATH);
  }
  else{
    val = vlGetTransferSize(server, path);
    if (val < 0){
      return (DMU_VIDEO_BAD_CONTROL_VALUE);
    }
    *returnSize = val;
    return (DMU_OK);
  }
}

int dmuVideo::getTypeByName(const char *deviceName){
  VLDevList devlist;
  VLNode    source;
  vlGetDeviceList(server, &devlist);
  
  VLNodeInfo *videoInput = NULL;
  for(int i=0;i<devlist.numDevices;i++){
    
    for (int j=0;j<devlist.devices[i].numNodes;j++){
      VLNodeInfo &node = devlist.devices[i].nodes[j];
      if ( node.kind == VL_VIDEO && strstr(node.name,deviceName) != NULL )
	videoInput = &node;
    }
  }
  
  if (videoInput == NULL) {
    fprintf (stderr, "Unaable to locate requested video source type '%s'\n", 
	     deviceName);
    return (-1);
  }
  
  return (videoInput->number);
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
