#include "dmuVideoIn.h"
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_image.h>
#include <dmedia/vl.h>
#include <dmedia/vl_mvp.h>


// dmuVideoIn - This object creates a VIDEO->MEMORY path using VL.  
// See dmuVideoIn.h for more details on how this works.

dmuVideoIn::dmuVideoIn(){

}

dmuVideoIn::~dmuVideoIn(){
}


// setup - setup the video path from default video source -> mem node
//         no dm buffer pool creation or constraint happens here..
dmuVideoIn::setup(int source, int srcClass){
  int stat;


  // Standard VL path setup
  sourceNode = vlGetNode(server, VL_SRC, srcClass, source);
  if (sourceNode < 0) {
    return (DMU_VIDEO_GETNODE_FAILED);
  }

  drainNode = vlGetNode(server, VL_DRN, VL_MEM, VL_ANY);
  if (drainNode < 0) { 
    return (DMU_VIDEO_GETNODE_FAILED);
  }

  path = vlCreatePath(server, VL_ANY, sourceNode, drainNode);
  if (path < 0){
    return (DMU_VIDEO_CREATEPATH_FAILED);
  }

  DEBUG1(fprintf(stderr, "Trying to setup video path\n"));

  stat = vlSetupPaths(server, &path, 1, VL_SHARE, VL_SHARE);

  if (stat != 0){
    return (DMU_VIDEO_SETUPPATH_FAILED);
  }
  DEBUG1(fprintf(stderr, "**Returned:%d\n",stat));
  

  // For debug puposes, we'll listen to all the events..  This list
  // will get trimmed significantly when things are more stable.
  vlSelectEvents(server, path, 
		 VLTransferCompleteMask | 
		 VLStreamBusyMask |
		 VLStreamPreemptedMask |
		 VLAdvanceMissedMask |
		 VLStreamAvailableMask |
		 VLSyncLostMask |
		 VLStreamStartedMask |
		 VLStreamStoppedMask |
		 VLSequenceLostMask |
		 VLControlChangedMask |
		 VLTransferCompleteMask |
		 VLTransferFailedMask |
		 VLEvenVerticalRetraceMask |
		 VLOddVerticalRetraceMask |
		 VLFrameVerticalRetraceMask |
		 VLDeviceEventMask |
		 VLDefaultSourceMask |
		 VLControlRangeChangedMask |
		 VLControlPreemptedMask |
		 VLControlAvailableMask |
		 VLDefaultDrainMask |
		 VLStreamChangedMask |
		 VLTransferError);

  return (setDefaultControls());
}

// setDefaultControls() - Setup all the default controls on the video
// path.  This assumes that the video nodes and path have been created
// and set up
//
int dmuVideoIn::setDefaultControls(){
  int err = 0;

  err = this->setTiming(VIDEOIN_DEFAULT_TIMING);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_TIMING"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }


  err = this->setPacking(VIDEOIN_DEFAULT_PACKING);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_PACKING"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }

  err = this->setCaptureType(VIDEOIN_DEFAULT_CAP_TYPE);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_CAP_TYPE"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }

  /*err = this->setRate(VIDEOIN_DEFAULT_RATE,1);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_RATE"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }*/

  err = this->setFormat(VIDEOIN_DEFAULT_FORMAT);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_FORMAT"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }

  err = this->setLayout(VIDEOIN_DEFAULT_LAYOUT);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_LAYOUT"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }

  err = this->setZoom(VIDEOIN_DEFAULT_ZOOM_NUM, VIDEOIN_DEFAULT_ZOOM_DENOM);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_ZOOM"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }

  /*err = this->setOffset(VIDEOIN_DEFAULT_OFFSET_X, VIDEOIN_DEFAULT_OFFSET_Y);
  DEBUG1(fprintf (stderr, "Setting: %s \n","VIDEOIN_DEFAULT_OFFSET"));
  if (err < -1){
    return (DMU_VIDEO_BAD_CONTROL_VALUE);
  }*/

  return (err);
}

// init - create and associate a dmbuffer pool with the video path.
// The constraints of the buffer pool are based on the video path as
// well as on my children's source constraints.
//
int dmuVideoIn::init(dmuParams *params){
  int stat;

  int width;
  int height;
  getXYControl(path, drainNode, VL_SIZE, &width, &height);
  DEBUG1(fprintf (stderr, "SIZE: %d x %d\n", width, height));

  imageSizeX = bufferSizeX = width;
  imageSizeY = bufferSizeY = height;

  // Create a parameter structure for the buffer pool
  stat = dmParamsCreate(&poolParams);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }

  stat = dmBufferSetPoolDefaults(poolParams, 0, 0, DM_FALSE, DM_FALSE);
  if (stat != DM_SUCCESS){
    return (DMU_PARAMS_CREATE_FAILED);
  }
  
  // Constrain it based on our video path
  stat = addDstConstraints(poolParams);
  if (stat != DMU_OK){
    return (DMU_BAD_CONSTRAINTS);
  }

  // Constrain based on all of our children's constraints
  stat = addChildConstraints(poolParams);
  if (stat != DMU_OK){
    return (stat);
  }

  // Make sure we have room for at least 4 buffers.
  dmParamsSetInt( poolParams, DM_BUFFER_SIZE,
		  dmParamsGetInt( poolParams, DM_BUFFER_SIZE ) );
  dmParamsSetInt( poolParams, DM_BUFFER_COUNT, 6);

  
  DEBUG1(fprintf (stderr, "Pool parameters before creating buffer\n"));
  DEBUG1(printDMParams( poolParams ));
  DEBUG1(fprintf (stderr, "-\n"));


  // Create a buffer pool with our currently compounded constraints
  stat = dmBufferCreatePool(poolParams, &bufferPool);
  if (stat != DM_SUCCESS){
    return (DMU_POOL_CREATE_FAILED);
  }

  // Register the pool with our VL path
  stat = vlDMPoolRegister( server, path, drainNode, bufferPool);
  if (stat != 0){
    return (DMU_VIDEO_BUF_REGISTER_FAILED);
  }

  fprintf (stderr, "parameters for video:\n");
  printDMParams(poolParams);

  // Initialize children
  return (dmuObj::init(params));
}

// handleEvents - receive a video event from VL If the video event is
// a TransferComplete event, we have a new frame - convert it to a
// dmbuffer and propagate a reference to this new dmBuffer to our
// children.
//
// The current implementation (for debug purposes) returns either
// VIDEOIN_TRANSFER_COMPLETE, or VIDEOIN_RANDOM_EVENT, depending on
// what kind of event was received.
int dmuVideoIn::handleEvents(fd_set fdSet){
  DEBUG1( fprintf(stderr, " dmuVideoIn::handleEvents:%s\n", getName()));

  int stat;
  VLEvent  event;
  DMbuffer buffer;

  // Make sure that my FD is set in the fdSet
  if (FD_ISSET(getFD(), &fdSet)){
    //Grab an event off the vl event queue
    stat = vlEventRecv(server, path, &event);
    while ( stat == -1 && vlGetErrno() == VLAgain ){
	sginap( 1 );
	stat = vlEventRecv( server, path, &event );
    }

    if (stat < 0){
      fprintf (stderr, "VideoIn : couldn't get an event\n");
      return (DMU_VIDEO_NO_EVENT);
    }
  }
  else {
    // I shouldn't look for an event - I'm not the one they're looking
    // for.
    return (DMU_NOT_MY_EVENT);
  }
  
  DEBUG2(fprintf (stderr, "!"));
  switch (event.reason){
  case VLStreamBusy:
    DEBUG2(fprintf (stderr, "event: VLStreamBusy:\n"));
    break;
  case VLStreamPreempted:
    DEBUG2(fprintf (stderr, "event: VLStreamPreempted:\n"));
    break;
  case VLAdvanceMissed:
    DEBUG2(fprintf (stderr, "event: VLAdvanceMissed:\n"));
    break;
  case VLStreamAvailable:
    DEBUG2(fprintf (stderr, "event: VLStreamAvailable:\n"));
    break;
  case VLSyncLost:
    DEBUG2(fprintf (stderr, "event: VLSyncLost:\n"));
    break;
  case VLStreamStarted:
    DEBUG2(fprintf (stderr, "event: VLStreamStarted:\n"));
    break;
  case VLStreamStopped:
    DEBUG2(fprintf (stderr, "event: VLStreamStopped:\n"));
    break;
  case VLSequenceLost:
    DEBUG2(fprintf (stderr, "event: VLSequenceLost:\n"));
    break;
  case VLControlChanged:
    DEBUG2(fprintf (stderr, "event: VLControlChanged:\n"));
    break;
  case VLTransferFailed:
    DEBUG2(fprintf (stderr, "event: VLTransferFailed:\n"));
    break;
  case VLFrameVerticalRetrace:
    DEBUG2(fprintf (stderr, "event: VLFrameVerticalRetrace:\n"));
    break;
  case VLDeviceEvent:
    DEBUG(fprintf (stderr, "event: VLDeviceEvent:\n"));
    return (VIDEOIN_SHUTTER_EVENT);
    //    break;
  case VLDefaultSource:
    DEBUG2(fprintf (stderr, "event: VLDefaultSource:\n"));
    break;
  case VLControlRangeChanged:
    DEBUG2(fprintf (stderr, "event: VLControlRangeChanged:\n"));
    break;
  case VLControlPreempted:
    DEBUG2(fprintf (stderr, "event: VLControlPreempted:\n"));
    break;
  case VLControlAvailable:
    DEBUG2(fprintf (stderr, "event: VLControlAvailable:\n"));
    break;
  case VLDefaultDrain:
    DEBUG2(fprintf (stderr, "event: VLDefaultDrain:\n"));
    break;
  case VLStreamChanged:
    DEBUG2(fprintf (stderr, "event: VLStreamChanged:\n"));
    break;
  case VLTransferError:
    DEBUG2(fprintf (stderr, "event: VLTransferErr:\n"));
    break; 
  case VLEvenVerticalRetrace:
    DEBUG2(fprintf (stderr, "event: VLEvenVerticalRetrace:\n"));
    break;
  case VLOddVerticalRetrace:
    DEBUG2(fprintf (stderr, "event: VLOddVerticalRetrace:\n"));
    break;
  case VLTransferComplete:
    //DEBUG(fprintf (stderr, "event:%s\n"," VLTransferCompleteMask:\n"));

    stat = vlEventToDMBuffer(&event, &buffer);
    if (stat < 0){
      fprintf (stderr, "Unable to convert video to dmbuffer\n");
      return (DMU_VIDEO_TO_DMBUFFER_FAILED);
    }
    
    propagate(buffer);
    // For now, free it right away, 'cause we don't have anything
    // to do with it!  Normally we propagate here..
    dmBufferFree(buffer);

    return (VIDEOIN_TRANSFER_COMPLETE);
    // uncomment this line if we get rid of the return above
    // break;
  default:
    fprintf (stderr, "Unknown event:%x",event.reason);
  }
  return (VIDEOIN_RANDOM_EVENT);
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
