//
//  chainTest.h
//
//  Author: Tim Davison <td@pobox.com>
//  Date:   Oct, 1996
//
// <dmu>
// This sample program shows how to use the chain object.  In this
// case we are taking the compressionTest and putting part of it
// in a chain.  Specifically, we are wrapping up the encoder
// and jpeg decoder into a chain.  See the CompressionTest for
// more details about exactly what this test does.
// In this case, the chain encapsulates a processing step.  If you
// wanted to change the processing you would only have to change
// the chain you put into the chain object.  The Event loop would
// stay the same.
// </dmu>
//
#include "dmuVideoIn.h"
#include "dmuGLDisplay.h"
#include "dmuICjencoder.h"
#include "dmuICjdecoder.h"
#include "dmuChain.h"

#define dmuCheck(dmFunc) { \
    int dmuErrID; \
    if ((dmuErrID = (dmFunc)) != DMU_OK) {\
       fprintf(stderr,"*Error:%d:%s\n", dmuErrID,getErrorString(dmuErrID)); \
    }\
    else { \
       fprintf(stderr,"No Error\n"); \
    } \
 }

int main(int argc, char *argv[]){
  dmuVideoIn     *videoIn;
  dmuGLDisplay   *rawDisplay;
  dmuGLDisplay   *compressedDisplay;
  dmuICjencoder  *encoder;
  dmuICjdecoder  *decoder;
  dmuChain       *chain;

  int sizeX,sizeY;
  int compressionQuality = 75;

  if (argc > 1) {
    compressionQuality = atoi(argv[1]);
  }
  
  // -- Construction --
  // create a video object
  videoIn = new dmuVideoIn();
  videoIn->setName("Video1");

  // create a jpeg encoder object
  encoder = new dmuICjencoder();
  encoder->setName("encoder");

  // create a jpeg decoder object
  decoder = new dmuICjdecoder();
  decoder->setName("decoder");

  // create a display object
  rawDisplay = new dmuGLDisplay();
  rawDisplay->setName("rawDisplay");

  // create a display object
  compressedDisplay = new dmuGLDisplay();
  compressedDisplay->setName("compressedDisplay");

  // create a chain object
  chain = new dmuChain();
  chain->setName("chain");

  // -- Wire the Chain --
  // Wire the things together that will be in the chain
  encoder->addObject(decoder);
  chain->setChain(encoder);
  
  // -- Wire --
  // Wire the video to a display and then to the chain
  // then wire the chain to the output display
  videoIn->addObject(rawDisplay);
  videoIn->addObject(chain);
  chain->addObject(compressedDisplay);

   // Set up our objects.  We'll print out what we're doing as we go

  // -- Setup --

  //   -- Begin Video Object Setup ----
  //  see VideoTest.c++ for more detailed comments about this video stuff
 
  fprintf (stderr, "Setting up video path:");
  dmuCheck(videoIn->setup(VL_ANY,VL_VIDEO));
  fprintf (stderr, "Setting up packing: ");
  dmuCheck(videoIn->setPacking(VL_PACKING_YVYU_422_8));
  videoIn->getImageDimensions(&sizeX,&sizeY);
  fprintf (stderr, "Video Image Size: %d, %d\n",sizeX,sizeY);

  //   -- End Video Setup ---
  //   -- Begin Encoder Object Setup ----
  
  // tell the encoder what size of image to expect
  encoder->setImageDimensions(sizeX, sizeY);
  // go set it up
  fprintf(stderr, "setting up the encoder:");
  dmuCheck( encoder->setup() );
  // tell the encoder what quality to use
  // you must call setup before you call this method.  
  fprintf(stderr, "setting encoder quality:");
  dmuCheck( encoder->setQuality(compressionQuality) );

  //   -- End Encoder Setup ---
  //   -- Begin Decoder Object Setup ----

  // tell the decoder what size of image to expect
  decoder->setImageDimensions(sizeX, sizeY);
  // go set it up
  fprintf(stderr, "setting up the decoder:");
  dmuCheck( decoder->setup() );
  
  //   -- End Decoder Setup ---
  //   -- Begin Raw Display Setup --
  
  rawDisplay->setPacking( GL_YCRCB_422_SGIX);
  rawDisplay->setImageDimensions(sizeX, sizeY);
  rawDisplay->setup();

  //   -- End Raw Display Setup ---
  //   -- Begin Compressed Display Setup --

  compressedDisplay->setPacking( GL_YCRCB_422_SGIX);
  compressedDisplay->setImageDimensions(sizeX, sizeY);
  compressedDisplay->setup();

  //   -- End compressed Display Setup ---
  
  // -- Initialize --

  // Init only has to be called at the head of the chain.
  fprintf (stderr, "Initializing: ");
  dmuCheck( videoIn->init()  );

   // -- Start --

  // Start only has to be called at the head of the chain.  The video
  // transfer has now started.
  fprintf (stderr, "Fire it up!: ");
  dmuCheck( videoIn->start() );

  // -- Begin Handle events loop --
  fd_set readFDSet;
  int maxFD;
  for(int i = 0; i < 5000; i ++) {
    // Setup our file descriptor set
    FD_ZERO(&readFDSet);
    videoIn->setFD(&readFDSet, &maxFD, 1);

    // block until one of the fd's has some data on it
    select (maxFD+1, &readFDSet, NULL, NULL, NULL);

    // let the objects handle their events.  each object checks to see
    // if their own fd is set in the FDset and if so, they act on it.  
    chain->handleEvents(readFDSet);
    videoIn->handleEvents(readFDSet);
  }
  // -- End Handle events loop --

  // destroy everything on exit
  delete(decoder);
  delete(encoder);
  delete(videoIn);
  delete(rawDisplay);
  delete(compressedDisplay);

}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
