#include "dmu.h"
#include "dmuMovieInRender.h"
#include "dmuVideoOut.h"
#include "dmuGLCtx.h"
#include "dmuGLTex2d.h"
#include <dmedia/vl_mvp.h>
#include <dmedia/vl.h>

// compositeTest.c++ - <dmu>Renders a movie into a dmbuffer, and
// passes that on to a dmuGLCtx.  Text, and a second movie texture
// mapped onto a cube are rendered into in the dmuGLCtx (using Open
// GL) and then the composite is sent out to the video output (as
// selected by vcp). Demonstrates the use of the context object object
// for compositing operations and the use of the video output object
// using graphics layout, rgba data.  Also demostrates an event loop
// which is using multiple chains (a la multiTexTest). </dmu>


// Simple macro for checking the result of a dmu method and printing
// the results if there's ana error.
#define dmuCheck(dmFunc) { \
    int dmuErrID; \
    if ((dmuErrID = (dmFunc)) != DMU_OK) {\
       fprintf(stderr,"*Error:%d:%s\n", dmuErrID, \
	       getErrorString(dmuErrID) ); \
	       exit(-1); \
    }\
    else { \
	     /*fprintf(stderr,"No Error\n");*/ \
    } \
 }

void initText();
void render(int frameNumber);
void renderString(float x, float y, char *string);
void drawCube(void);
void drawFace(Bool borders,float z);


// Globals - makes the render function easer to deal with.  In a real
// app you'd just make 'em menber variables in a class.
int firstFrameRendered = 0;
int sizeX,sizeY;
Display *dpy;
dmuMovieInRender *movieIn2;

int main(int argc, char *argv[]){
  // dmu objects we'll use
  dmuMovieInRender *movieIn;
  dmuGLCtx         *context;
  dmuVideoOut      *videoOut;

  dmuGLTex2d       *texture;

  // other variables we'll need
  fd_set readFDSet;
  int err;
  int texSize = 512;

  if (argc < 3){
    fprintf (stderr, "Usage: compositeTest <movie file name> <movie file 2>\n");
    exit (-1);
  }

  // -- Construction --
  // Create our objects.
  movieIn  = new dmuMovieInRender();
  videoOut = new dmuVideoOut();
  context  = new dmuGLCtx();

  movieIn2 = new dmuMovieInRender;
  texture  = new dmuGLTex2d;

  // -- Wire --
  // Wire up the path (video->context->videoOut).
  movieIn->addObject(context);
  context->addObject(videoOut);

  movieIn2->addObject(texture);

  //   -- Begin Video Object Setup ----

  dmuCheck(videoOut->setup(VL_ANY));
  dmuCheck(videoOut->setPacking(VL_PACKING_ABGR_8));
  //dmuCheck(videoOut->setSize(640,480));

  // Textures and off-screen renders are in "graphics mode"
  dmuCheck(videoOut->setLayout(VL_LAYOUT_GRAPHICS));

  // Get the size of a single frame of video 
  videoOut->getImageDimensions(&sizeX,&sizeY);
  fprintf (stderr, "Video Image Size: %d, %d\n\n",sizeX,sizeY);


  
  //   -- Begin movie setup
  dmuCheck(movieIn->setup(argv[1]));

  // Set the buffer size to video output size, so we can send the
  // movie directly to video output.
  dmuCheck(movieIn->setBufferDimensions(sizeX,sizeY));

  // We can set the offset of the movie rendering, so that it's in the
  // middle of the video output.
  movieIn->setImageOffset( (sizeX - movieIn->getImageSizeX())/2,
			   (sizeY - movieIn->getImageSizeY())/2);
    


  // Setup the second movie as in movieTexTest.c++
  dmuCheck(movieIn2->setup(argv[2]));
  dmuCheck(movieIn2->setBufferDimensions(texSize,texSize));

  //   -- Begin context setup

  dpy = XOpenDisplay( ":0" );
  context->setup(dpy,sizeX,sizeY);

  // -- Texture setup
  dmuCheck(texture->setup(dpy,texSize,texSize));

  // -- Init --

  // Init only has to be called at the head of the chains
  fprintf (stderr, "Initializing: \n");
  dmuCheck(movieIn->init());
  dmuCheck(movieIn2->init());

  // -- Start --

  // Start only has to be called at the head of the chain.  The video
  // transfer has now started.
  fprintf (stderr, "Fire it up!: \n");
  dmuCheck(movieIn->start());
  dmuCheck(movieIn2->start());

  int maxFD;

  // -- Begin Handle events loop --
  for (int i = 0; ;i++){
    // Setup our file descriptor set
    maxFD = 0;
    FD_ZERO(&readFDSet);

    // Set the file descriptors we're interested in.
    movieIn->setFD(&readFDSet, &maxFD);
    movieIn2->setFD(&readFDSet, &maxFD);

    // Wait for something to happen on the video path
    select (maxFD + 1,&readFDSet, NULL, NULL, NULL);

    err = movieIn2->handleEvents(readFDSet);

    err = movieIn->handleEvents(readFDSet);

    if (err == DMU_OK){
      // If we've sent the last frame out OK, then we should prepare a
      // new frame (which is passed to video out via release)
      err = context->grab();
      if (err == DMU_OK){
	context->makeCurrent();

	// Setup texturing if its safe
	glEnable (GL_TEXTURE_2D);
	texture->defineTex2D();

	render(i);
      }
      context->release();
    }

    err = videoOut->handleEvents(readFDSet);
  }
  // -- End Handle events loop --

  // On exit, we should destroy the objects in the chain (except we'll 
  // never get here, because we were in an infite loop above :)
  //  delete (movieIn);
  //  delete (context);
  //  delete (videoOut);
}



// Perform rendering on top of the frame 
void render(int frameNumber){
  char tmpBuf[100];

  if (!firstFrameRendered){
    glEnable(GL_DEPTH_TEST);
    // OpenGL setup for the window
    glViewport( 0, 0, sizeX, sizeY);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,sizeX,0,sizeY,-100,1000);
    
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    initText();
  }

  glEnable (GL_BLEND);
  glBlendFunc(GL_ONE, GL_ONE_MINUS_CONSTANT_COLOR_EXT);
  //glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
  glBlendColorEXT(0.10,0.10,0.10,0.1);

  glEnable(GL_DEPTH_TEST);
  glClearColor(0,0,0,0);
  glClear(GL_DEPTH_BUFFER_BIT);

  glMatrixMode( GL_MODELVIEW );

  glPushMatrix();
  glTranslatef(sizeX/2,sizeY/2,0);
  glRotatef(frameNumber/2,0.0,1.0,0.0);
  glTranslatef(200,0,0);
  glRotatef(frameNumber/4,1.0,0.0,0.0);
  
  glScalef(0.2,0.2,0.2);
  drawCube();
  glPopMatrix();

  glPushMatrix();
  glTranslatef(sizeX/2 - 160,sizeY - 100,0);
  glRotatef(-frameNumber,0.0,0.0,1.0);
  glColor3f(1.0,0.0,0.0);

  glDisable (GL_TEXTURE_2D);

  glBegin(GL_QUADS);
  glVertex3d(0.0,0.0,0.0);
  glVertex3d(-4,25,0.0);
  glVertex3d(0.0,40.0,0.0);
  glVertex3d(4,25,0.0);
  glEnd();

  glColor3f(1.0,0.0,0.0);

  glPopMatrix();

  sprintf (tmpBuf, "Frame: %d",frameNumber);
  renderString(sizeX/2 - (100),sizeY - 100,tmpBuf);
}


//Bases for display lists for alphabet and numeric fonts.
#define ALPHABETBASE 0x1000

void initText(){
  XFontStruct *fontStruct;
  
  fontStruct = XLoadQueryFont(dpy,"-*-courier-*-r-*-*-30-*-*-*-*-*-*-*");
  
  if (!fontStruct) {
    fprintf (stderr, "Unable to load font\n");
    return;
  }

  glXUseXFont(fontStruct->fid, '!', '~' - '!',  ALPHABETBASE);
}


// renderString - Render a string to screen at raster coords (x,y)
//
void renderString(float x, float y, char *string){
  int len = strlen(string);

  for (int i = 0; i < len; i++){
    glRasterPos2f(x + (i*20),y);
    glCallList(ALPHABETBASE + string[i] - '!');
  }
}


void drawCube(void) {
    glColor4f( 0.5, 0.5, 0.5, 1.0 );
    
    /* front */
    drawFace( False, 160 );
    
    /* right */
    glPushMatrix();
    glRotatef( 90.0, 0.0, 1.0, 0.0 );
    drawFace( False, 160 );
    glPopMatrix();
    
    /* left */
    glPushMatrix();
    glRotatef( -90.0, 0.0, 1.0, 0.0 );
    drawFace( False, 160 );
    glPopMatrix();
    
    /* back */
    glPushMatrix();
    glRotatef( 180.0, 0.0, 1.0, 0.0 );
    drawFace( False, 160 );
    glPopMatrix();
    
    /* top */
    glPushMatrix();
    glRotatef( -90.0, 1.0, 0.0, 0.0 );
    drawFace( False, 160 );
    glPopMatrix();
    
    /* bottom */
    glPushMatrix();
    glRotatef( 90.0, 1.0, 0.0, 0.0 );
    drawFace( False, 160);
    glPopMatrix();
}


void drawFace(Bool borders,float z) {

  glEnable( GL_TEXTURE_2D );
  glBegin( GL_QUADS );

  movieIn2->movieTexCoord2f( 0.0, 0.0 ); glVertex3i( -160, -160, z );
  movieIn2->movieTexCoord2f( 0.0, 1.0 ); glVertex3i( -160,  160, z );
  movieIn2->movieTexCoord2f( 1.0, 1.0 ); glVertex3i(  160,  160, z );
  movieIn2->movieTexCoord2f( 1.0, 0.0 ); glVertex3i(  160, -160, z );

  glEnd();
    
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
