#include "dmu.h"
#include "dmuMovieInRender.h"
#include "dmuVideoOut.h"
#include "dmuGLCtx.h"
#include <dmedia/vl_mvp.h>
#include <dmedia/vl.h>

// ctxTest.c++ - <dmu>Renders a movie into a dmbuffer, and passes that
// on to a dmuGLCtx.  Text is rendered on top of the movie in the
// dmuGLCtx (usingOpen GL) and then the composit is sent out to the
// video output (as selected by vcp). Demonstrates the use of the
// context object object for compositing operations and the use of the
// video output object using graphics layout, rgba data.  </dmu>


// Simple macro for checking the result of a dmu method and printing
// the results if there's ana error.
#define dmuCheck(dmFunc) { \
    int dmuErrID; \
    if ((dmuErrID = (dmFunc)) != DMU_OK) {\
       fprintf(stderr,"*Error:%d:%s\n", dmuErrID, \
	       getErrorString(dmuErrID) ); \
	       exit(-1); \
    }\
    else { \
	     /*fprintf(stderr,"No Error\n");*/ \
    } \
 }

void initText();
void render(int frameNumber);
void renderString(float x, float y, char *string);


// Globals - makes the render function easer to deal with.  In a real
// app you'd just make 'em menber variables in a class.
int firstFrameRendered = 0;
int sizeX,sizeY;
Display *dpy;

int main(int argc, char *argv[]){
  // dmu objects we'll use
  dmuMovieInRender *movieIn;
  dmuGLCtx         *context;
  dmuVideoOut      *videoOut;

  // other variables we'll need
  fd_set readFDSet;
  int err;

  if (argc < 2){
    fprintf (stderr, "Usage: movieToVideo <movie file name>\n");
    exit (-1);
  }

  // -- Construction --
  // Create our objects.
  movieIn = new dmuMovieInRender();
  movieIn->setName("Movie input");

  videoOut = new dmuVideoOut();
  videoOut->setName("Video Output");

  context = new dmuGLCtx();
  context->setName("Context");

  // -- Wire --
  // Wire up the path (video->context->videoOut).
  movieIn->addObject(context);
  context->addObject(videoOut);

  // -- Setup --

  //   -- Begin Video Object Setup ----

  dmuCheck(videoOut->setup(VL_ANY));
  dmuCheck(videoOut->setPacking(VL_PACKING_ABGR_8));
  dmuCheck(videoOut->setFormat(VL_FORMAT_RGB));
  dmuCheck(videoOut->setLayout(VL_LAYOUT_GRAPHICS));

  // Get the size of a single frame of video 
  videoOut->getImageDimensions(&sizeX,&sizeY);
  fprintf (stderr, "Video Image Size: %d, %d\n\n",sizeX,sizeY);

  dmuCheck(movieIn->setup(argv[1]));


  //   -- Begin movie setup
  // Set the buffer size to video size, so we can send the movie 
  // directly to video.
  dmuCheck(movieIn->setBufferDimensions(sizeX,sizeY));

  // We can set the offset of the movie rendering, so that it's in the
  // middle of the video output.
  movieIn->setImageOffset( (sizeX - movieIn->getImageSizeX())/2,
			   (sizeY - movieIn->getImageSizeY())/2);
    

  //   -- Begin context setup

  dpy = XOpenDisplay( ":0" );
  context->setup(dpy,sizeX,sizeY);
  

  //   -- End  Setup ---
		 

  // -- Init --

  // Init only has to be called at the head of the chain.
  fprintf (stderr, "Initializing: \n");
  dmuCheck(movieIn->init());

  // -- Start --

  // Start only has to be called at the head of the chain.  The video
  // transfer has now started.
  fprintf (stderr, "Fire it up!: \n");
  dmuCheck(movieIn->start());

  // -- Begin Handle events loop --
  for (int i = 0; ; i++){
    int maxFD = 0;

    // Setup our file descriptor set
    FD_ZERO(&readFDSet);
    movieIn->setFD(&readFDSet, &maxFD);
    videoOut->setFD(&readFDSet, &maxFD);

    // Wait for something to happen on the video path
    select (maxFD + 1,&readFDSet, NULL, NULL, NULL);
    
    err = movieIn->handleEvents(readFDSet);
    if (err == DMU_OK){
      // If we've send the last one out OK, then we should 
      // prepare and pass the video out object another frame.
      err = context->grab();
      if (err == DMU_OK){
	context->makeCurrent();
	render(i);
      }
      context->release();
    }

    err = videoOut->handleEvents(readFDSet);

  }
  // -- End Handle events loop --

  // On exit, we should destroy the objects in the chain (except we'll 
  // never get here, because we were in an infite loop above :)
  //  delete (movieIn);
  //  delete (context);
  //  delete (videoOut);
}



// Perform rendering on top of the frame 
void render(int frameNumber){
  char tmpBuf[100];

  if (!firstFrameRendered){
    // OpenGL setup for the window
    glViewport( 0, 0, sizeX, sizeY);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0,sizeX,0,sizeY,0,1000);
    
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    initText();
  }

  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glTranslatef(sizeX/2 - 160,sizeY - 100,0);

    
  //glDisable (GL_TEXTURE_2D);

  glColor3f(0,0.2,1.0);
  glBegin(GL_QUADS);
  glVertex3d(-45,-45,0);
  glVertex3d(-45,45,0);
  glVertex3d(45,45,0);
  glVertex3d(45,-45,0);
  glEnd();

  glRotatef(-frameNumber,0,0,1.0);

  glColor3f(1.0,0.0,0.0);

  glBegin(GL_QUADS);
  glVertex3d(0,0,0);
  glVertex3d(-4,25,0);
  glVertex3d(0,40,0);
  glVertex3d(4,25,0);
  glEnd();

  glColor3f(1.0,0.0,0.0);

  glPopMatrix();

  sprintf (tmpBuf, "Frame: %d",frameNumber);
  renderString(sizeX/2 - (100),sizeY - 100,tmpBuf);
}


//Bases for display lists for alphabet and numeric fonts.
#define ALPHABETBASE 0x1000

void initText(){
  XFontStruct *fontStruct;
  
  fontStruct = XLoadQueryFont(dpy,"-*-courier-*-r-*-*-30-*-*-*-*-*-*-*");
  
  if (!fontStruct) {
    fprintf (stderr, "Unable to load font\n");
    return;
  }

  glXUseXFont(fontStruct->fid, '!', '~' - '!',  ALPHABETBASE);
}


// renderString - Render a string to screen at raster coords (x,y)
//
void renderString(float x, float y, char *string){
  int len = strlen(string);

  for (int i = 0; i < len; i++){
    glRasterPos2f(x + (i*20),y);
    glCallList(ALPHABETBASE + string[i] - '!');
  }
}
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
