#include "dmu.h"
#include "dmuMovieInRender.h"
#include "dmuGLTex2d.h"
#include <dmedia/vl_mvp.h>
#include "ScratchWindow.h"
#include <unistd.h>

// movieTexTest.c++ - <dmu> Sample program displaying a movie file as
// a texture map on a rotating polygon using the dmuMovieInRender
// object.  </dmu>This uses libdmu to create a simple chain:
// 
// dmuMovieInRender -> dmuGLTex2d
//
// <dmu>The textured polygon is rendered to a double-buffered Open GL
//window created through the ScratchWindow class. </dmu> Notice how
//defineTex2d is used in the handleEvents loop to define and use new
//textures for OpenGL rendering

void render(dmuMovieInRender *movieIn, 
	    Display *theDisplay, 
	    int sizeX, 
	    int sizeY);

#define dmuCheck(dmFunc) { \
    int dmuErrID; \
    if ((dmuErrID = (dmFunc)) != DMU_OK) {\
       fprintf(stderr,"*Error:%d:%s\n", dmuErrID, \
	       getErrorString(dmuErrID) ); \
       exit(-1);\
    }\
    else { \
	     /*fprintf(stderr,"No Error\n");*/ \
    } \
 }


int main(int argc, char *argv[]){
  dmuMovieInRender *movieIn;
  dmuGLTex2d *texture;

  fd_set readFDSet;
  int maxFD;
  int imageSizeX,imageSizeY;
  int bufferSizeX,bufferSizeY;

  Display *theDisplay;
  ScratchWindow *window;

  // Make sure we have a command line arg to point us to a movie
  if (argc < 2){
    fprintf (stderr, "USAGE: movieTexTest <movie file name>\n");
    exit (-1);
  }

  // Open up the X Display - We need it for the texture and scratch
  // window.
  theDisplay = XOpenDisplay( ":0" );

  // -- Construction --
  movieIn = new dmuMovieInRender;
  movieIn->setName("MovieIn ");

  texture = new dmuGLTex2d;
  texture->setName("Texture");


  // -- Wire --
  movieIn->addObject(texture);
  

  // -- Setup --

  // Point the movieIn object at the movie file 
  dmuCheck(movieIn->setup(argv[1]));

  // We need to get the buffer dimensions of the movie, which is
  // rounded up to the next power of 2.  This lets us set 
  // the texture size
  movieIn->getBufferDimensions(&bufferSizeX,&bufferSizeY);
  fprintf (stderr,"Movie buffer dimensions:%d,%d\n\n",bufferSizeX,bufferSizeY);

  // We get the image dimensions so we know how big to make the window
  // and the polygon we're texture mapping on to.
  movieIn->getImageDimensions(&imageSizeX,&imageSizeY);
  fprintf (stderr, "Movie image dimensions: %d, %d\n\n",imageSizeX,imageSizeY);

  // Set up the texture object to have the same size as the buffers in
  // the movie object
  dmuCheck(texture->setup(theDisplay, bufferSizeX, bufferSizeY));

  // -- Init --
  fprintf (stderr, "Initializing: \n\n");
  dmuCheck(movieIn->init());

  // -- Start --
  fprintf (stderr, "Fire it up!\n");
  dmuCheck(movieIn->start());


  // -- Other setup -- 

  // Setup an Open GL window to render into - this is not a dmuObj
  window = ScratchWindow::create(theDisplay, imageSizeX,imageSizeY,&argc,argv);
  if (!window){
    fprintf (stderr, "Failed to create scratch window!\n");
    exit (-1);
  }
  
  // OpenGL setup for the window
  glViewport( 0, 0, imageSizeX, imageSizeY);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,imageSizeX,0,imageSizeY,0,1000);

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  glEnable(GL_TEXTURE_2D);
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE_EXT);


  // -- Handle Events loop --
  for (int i = 0; i < 5000; i++){
    
    // Setup our File descriptors
    FD_ZERO(&readFDSet);
    movieIn->setFD(&readFDSet,&maxFD);

    // Wait for something to happen to the movie object (i.e. we get a frame)
    select (maxFD+1,&readFDSet, NULL, NULL, NULL);

    int error = movieIn->handleEvents(readFDSet);

    // Tell the texture to setup Open GL to use the latest
    // movie frame as a texture map.
    dmuCheck(texture->defineTex2D());
    

    // Rotate and render the texture map onto a polygon
    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glTranslatef(imageSizeX/2,imageSizeY/2,0);
    glRotatef(i,0,0,1.0);
    render(movieIn,theDisplay,imageSizeX,imageSizeY);
    glPopMatrix();

    // Swap buffers in the window
    glXSwapBuffers(theDisplay, glXGetCurrentDrawable());
  }


  
  delete (texture);
  delete (movieIn);
}


// Handles the clearing of the window and rendering of the polygon
void render(dmuMovieInRender *movieIn,
	    Display *theDisplay, 
	    int imageSizeX, 
	    int imageSizeY)
{
  // Clear the window
  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT);
  
  glColor3f(1.0,1.0,1.0);

  glBegin(GL_QUADS);

  // Here we use the convenience methods on the movie object to get
  // the proper texture coordinates, since our texture is smaller than
  // the DMBuffer it lives in
  //glTexCoord2d(0,0);
  movieIn->movieTexCoord2d(0.0,0.0);
  glVertex3d(-imageSizeX/2,-imageSizeY/2,0.0);

  //glTexCoord2d(1,0);
  movieIn->movieTexCoord2d(1.0,0.0);
  glVertex3d(imageSizeX/2,-imageSizeY/2,0.0);

  //glTexCoord2d(1,1);
  movieIn->movieTexCoord2d(1.0,1.0);
  glVertex3d(imageSizeX/2,imageSizeY/2,0.0);

  //glTexCoord2d(0,1);
  movieIn->movieTexCoord2d(0.0,1.0);
  glVertex3d(-imageSizeX/2,imageSizeY/2,0.0);

  glEnd();
  
}  
/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
