#include "dmu.h"
#include "dmuObj.h"
#include "dmuVideoIn.h"
#include "dmuMovieInRender.h"
#include "dmuGLTex2d.h"
#include <dmedia/vl_mvp.h>
#include "ScratchWindow.h"
#include <unistd.h>

// multiTexTest.c++ - <dmu> Sample program demonstrating the use of 2
// dmu chains in the same program to display a movie texture map and a
// video texture map simultaneously using the same handleEvents loop
// </dmu>
// 
// dmuVideoIn -> dmuGLTex2d ------------|
//                                      |---> openGL rendering
// dmuMovieInRender-> dmuGLTex2d    ----|
//

void render(int frame);

#define dmuCheck(dmFunc) { \
    int dmuErrID; \
    if ((dmuErrID = (dmFunc)) != DMU_OK) {\
       fprintf(stderr,"*Error:%d:%s\n", dmuErrID, \
	       getErrorString(dmuErrID) ); \
					     /*exit(-1);*/\
    }\
    else { \
	     /*fprintf(stderr,"No Error\n");*/ \
    } \
 }

// Functions to set up the different chains.  We're going to make the
// dmu object pointers, and some of the variables we get from them,
// global so we don't have to pass the pointers around to all the
// functions in the sample program.  Alternatively, all these
// functions and pointers could be encapsulated in a class, so all the
// class methods would have access to the object pointers.

dmuVideoIn *videoIn;
dmuGLTex2d *videoTex;
dmuMovieIn *movieIn;
dmuGLTex2d *movieTex;

Display *theDisplay;
ScratchWindow *window;

int videoImageSizeX;
int videoImageSizeY;

int movieImageSizeX;
int movieImageSizeY;

int movieBufferSizeX;
int movieBufferSizeY;


int setupVideo();
int setupMovie(char *fileName);

int setupVideo(){

  // -- Construction --
  videoIn = new dmuVideoIn;
  videoIn->setName("Video Input");

  videoTex = new dmuGLTex2d;
  videoTex->setName("Video Texture");

  // -- Wire --
  videoIn->addObject(videoTex);

  // -- Setup --
  // Use current vcp setting for video source
  dmuCheck(videoIn->setup(VL_ANY));

  // Use ABGR_8 packing and formatting texture mapping
  dmuCheck(videoIn->setPacking(VL_PACKING_ABGR_8));
  dmuCheck(videoIn->setFormat(VL_FORMAT_RGB));
  dmuCheck(videoIn->setCaptureType(VL_CAPTURE_INTERLEAVED));

  // Use graphics layout for texture maps 
  dmuCheck(videoIn->setLayout(VL_LAYOUT_GRAPHICS));

  // Texture maps need to be powers of 2. Setting the zoom
  // to 4/5 and then cropping to 512/256 gets us a pretty
  // good area
  //dmuCheck(videoIn->setZoom(4,5));
  dmuCheck(videoIn->setSize(512,512));
  dmuCheck(videoIn->setOrigin(0,32));

  videoIn->getImageDimensions(&videoImageSizeX,&videoImageSizeY);

  // Set up the texture object to have the same size
  dmuCheck(videoTex->setup(theDisplay, videoImageSizeX, videoImageSizeY));

  // Intalize the video chain
  dmuCheck(videoIn->init());
  
  return (DMU_OK);
}

int setupMovie(char *fileName){
  // -- Construction --
  movieIn = new dmuMovieInRender;
  movieIn->setName("Movie Input (Rendering)");

  movieTex = new dmuGLTex2d;
  movieTex->setName("Movie Texture");

  // -- Wire --
  movieIn->addObject(movieTex);

  // -- Setup --

  // Point the movieIn object at the movie file 
  dmuCheck(movieIn->setup(fileName));

  movieIn->getBufferDimensions(&movieBufferSizeX,&movieBufferSizeY);
  movieIn->getImageDimensions(&movieImageSizeX,&movieImageSizeY);

  // Set up the texture object to have the same size as movie buffers
  dmuCheck(movieTex->setup(theDisplay, movieBufferSizeX, movieBufferSizeY));

  // -- Init --
  dmuCheck(movieIn->init());

  return (DMU_OK);
}

int main(int argc, char *argv[]){
  fd_set readFDSet;
  int maxFD;

  // Since defineTex2d is a blocking call, we need to make sure
  // we've got valid texture data before attempting to render 
  // using the video and movie textures.
  int gotFirstVideoFrame = 0;
  int gotFirstMovieFrame = 0;

  if (argc < 2){
    fprintf (stderr, "USAGE: multiTexTest <movie file name>\n");
    exit (-1);
  }

  // Open up the X Display - We need it for the texture and scratch
  // window.
  theDisplay = XOpenDisplay( ":0" );

  setupVideo();
  setupMovie(argv[1]);

  // -- Start --
  fprintf (stderr, "Fire 'em up!\n");
  dmuCheck(videoIn->start());
  dmuCheck(movieIn->start());

  // -- Other GL setup -- 

  // Setup an Open GL window to render into - this is not a dmuObj
  window = ScratchWindow::create(theDisplay, videoImageSizeX*2, 
				 videoImageSizeY * 2, &argc, argv);
  if (!window){
    fprintf (stderr, "Failed to create scratch window!\n");
    exit (-1);
  }
  
  // OpenGL setup for the window
  glViewport( 0, 0, videoImageSizeX*2, videoImageSizeY*2);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,videoImageSizeX*2,0,videoImageSizeY*2,-1000,1000);
  //glFrustum (0,videoImageSizeX*2,0,videoImageSizeY*2,-1000,1000);

  glEnable(GL_DEPTH_TEST);

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  glEnable(GL_TEXTURE_2D);


  // -- Handle Events loop --
  int eventReturn;
  int shouldRender;

  for (int i = 0; i < 5000; i++){
    shouldRender = 0;
    
    // Setup our File descriptors
    FD_ZERO(&readFDSet);
    videoIn->setFD(&readFDSet,&maxFD);
    movieIn->setFD(&readFDSet,&maxFD);

    // Wait for something to happen to the video object
    select (maxFD+1,&readFDSet, NULL, NULL, NULL);

    eventReturn = videoIn->handleEvents(readFDSet);

    if (eventReturn == VIDEOIN_TRANSFER_COMPLETE){
      if (gotFirstVideoFrame != 2){
	gotFirstVideoFrame++;
      }
      shouldRender = 1;
    }

    eventReturn = movieIn->handleEvents(readFDSet);
    if (eventReturn != DMU_NOT_MY_EVENT){
      if (gotFirstMovieFrame != 2){
	gotFirstMovieFrame++;
      }
      //shouldRender = 1;
    }

    if (!shouldRender || (gotFirstMovieFrame != 2) || (gotFirstVideoFrame != 2)){
    }
    //window->setContext();
    else {
      render(i);
    }

    // Swap buffers in the window
  }


  delete (movieIn);
  delete (movieTex);
  delete (videoTex);
  delete (videoIn);
}


// Handles the clearing of the window and rendering of the polygon
void render(int frame){
  frame /=2;

  window->setContext();

  glClearColor(0.0,0.0,0.0,0.0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  glColor3f(1.0,1.0,1.0);

  glMatrixMode( GL_MODELVIEW );
  
  dmuCheck(videoTex->defineTex2D());

  glPushMatrix();
  glTranslatef(videoImageSizeX,videoImageSizeY,0);
  glRotatef(frame,0,1.0,0.0);
  glTranslatef(0,0,300);
  glRotatef(-2*frame,0,1.0,0.0);

  glBegin(GL_QUADS);
  glTexCoord2d(0.0,0.0);glVertex3d(-videoImageSizeX/2,-videoImageSizeY/2,0.0);
  glTexCoord2d(1.0,0.0);glVertex3d(videoImageSizeX/2,-videoImageSizeY/2,0.0);
  glTexCoord2d(1.0,1.0);glVertex3d(videoImageSizeX/2,videoImageSizeY/2,0.0);
  glTexCoord2d(0.0,1.0);glVertex3d(-videoImageSizeX/2,videoImageSizeY/2,0.0);
  glEnd();

  glPopMatrix();
  
  glFlush();

  dmuCheck(movieTex->defineTex2D());

  glPushMatrix();
  glTranslatef(videoImageSizeX,videoImageSizeY,0);
  glRotatef(frame,0,1,0.0);
  glTranslatef(0.0,0.0,-300);
  glRotatef(-2*frame,0.0,1,0.0);

  glBegin(GL_QUADS);

  movieIn->movieTexCoord2d(0.0,0.0);
  glVertex3d(-movieImageSizeX/2,-movieImageSizeY/2,10.0);

  movieIn->movieTexCoord2d(1.0,0.0);
  glVertex3d(movieImageSizeX/2,-movieImageSizeY/2,10.0);

  movieIn->movieTexCoord2d(1.0,1.0);
  glVertex3d(movieImageSizeX/2,movieImageSizeY/2,10.0);

  movieIn->movieTexCoord2d(0.0,1.0);
  glVertex3d(-movieImageSizeX/2,movieImageSizeY/2,10.0);

  glEnd();
  glPopMatrix();

  glFinish();

  glXSwapBuffers(theDisplay, glXGetCurrentDrawable());
}  



/****************************************************************************
  *                                                                          *
  *               Copyright (c) 1996-97, Silicon Graphics, Inc.              *
  *                                                                          *
  *   Permission to use,  copy, modify, distribute, and sell this software   *
  *   and its documentation for any purpose is hereby granted without fee,   *
  *   provided that the name of  Silicon  Graphics  may not be used in any   *
  *   advertising  or  publicity  relating  to  the  software  without the   *
  *   specific, prior written permission of Silicon Graphics.                *
  *                                                                          *
  *   THIS SOFTWARE IS PROVIDED "AS-IS"  AND WITHOUT WARRANTY OF ANY KIND,   *
  *   EXPRESS,  IMPLIED OR OTHERWISE,  INCLUDING  WITHOUT LIMITATION,  ANY   *
  *   WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.       *
  *                                                                          *
  *   IN  NO  EVENT  SHALL  SILICON  GRAPHICS  BE  LIABLE FOR ANY SPECIAL,   *
  *   INCIDENTAL,  INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,  OR  ANY   *
  *   DAMAGES  WHATSOEVER  RESULTING  FROM  LOSS OF USE,  DATA OR PROFITS,   *
  *   WHETHER OR NOT ADVISED OF  THE  POSSIBILITY  OF  DAMAGE,  AND ON ANY   *
  *   THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR   *
  *   PERFORMANCE OF THIS SOFTWARE.                                          *
  *                                                                          *
  ****************************************************************************/
