// videoSnap -- This application uses libdmu to grab a video frame
// every five seconds, map the bytes into a character array, and
// display them to the screen.. This is an example of how to put the
// dmu event loop into a second thread, and to use the latch object,
// to have new video data available on demand.  Note that there is
// no byte copying going on here.  The video hardware performs
// dma's into main memory, and when the latch object is asked to 
// 'grab' the latest data, a pointer to that memory location is
// returned.  This means that the second dmu thread is a low-cost
// thread.  Note also, that the bytes could be brought in as 
// any video library supported format.
#include "dmuVideoIn.h"
#include "dmuLatch.h"
#include "dmu.h"
#include "ScratchWindow.h"


// Variables made global for simplest sample code
dmuVideoIn *videoIn;
dmuLatch   *videoLatch;
ScratchWindow *window;
int sizeX, sizeY;
pid_t      eventLoopPID;

// Function prototypes
int setupVideo();
int setupWindow(int argc, char *argv[]);
void eventLoop(void */*no params required*/);

// Entry point.
int main(int argc, char *argv[]){
  int err;
  unsigned char *bufBytes;

  err = setupVideo();
  if (err < 0){
    fprintf (stderr, "Unable to setup video.  Aborting\n");
    exit (1);
  }

  err = setupWindow(argc, argv);
  if (err < 0){
    fprintf (stderr, "Unable to setup window.  Aborting\n");
    exit (1);
  }

  for (int i = 0; i < 1000; i++){
    fprintf (stderr, "sleeping...\n");
    sleep (5);
    fprintf (stderr, "i woke up!\n");
    
    fprintf (stderr, "grab!\n");
    videoLatch->grab();

    // *** This is where you get a pointer to your data ****
    fprintf (stderr, "map!\n");
    bufBytes = videoLatch->mapBuffer();
    
    fprintf (stderr, "display!\n");

    glDrawBuffer(GL_FRONT);
    glDrawPixels(sizeX, sizeY, 
		 GL_YCRCB_422_SGIX, 
		 GL_UNSIGNED_BYTE, bufBytes);
    

    videoLatch->release();
  }
}


int setupVideo(){
  int err;

  // Create Objects.
  fprintf (stderr, "Creating dmu objects:\n");
  videoIn = new dmuVideoIn();
  videoLatch = new dmuLatch();

  fprintf (stderr, "Creating dmu chain:\n");
  videoIn->addObject(videoLatch);
  videoIn->setup(VL_ANY, VL_VIDEO);
  
  // Tell's the video object that we want YUV pixels
  fprintf (stderr, "Setting up Video packing: ");
  videoIn->setPacking(VL_PACKING_YVYU_422_8);

  videoIn->getImageDimensions(&sizeX,&sizeY);
  fprintf (stderr, "Video Image Size: %d, %d\n\n",sizeX,sizeY);
  
  videoLatch->setup();
  

  fprintf (stderr, "Initializing dmu chain:\n");
  err = videoIn->init();
  if (err != DMU_OK){
    fprintf (stderr, "Unable to initialize dmu chain\n");
    return (-1);
  }

  fprintf (stderr, "Starting dmu chain:\n");
  videoIn->start();

  fprintf (stderr, "Starting dmu handleEvents thread\n");
  eventLoopPID = sproc(eventLoop, PR_SADDR | PR_SFDS, NULL);
  return (0);
}


void eventLoop(void */*no params required*/){
  fd_set readFDSet;
  while (1){
    int maxFD;

    // Setup our file descriptor set
    FD_ZERO(&readFDSet);
    videoIn->setFD(&readFDSet, &maxFD);

    // Wait for something to happen on the video path
    select (maxFD + 1,&readFDSet, NULL, NULL, NULL);

    // Handle the event - this will cause the latest frame to 
    // be passed to the latch and held on to until it's grabbed 
    // or another frame comes down
    videoIn->handleEvents(readFDSet);
  }
}


int setupWindow(int argc, char *argv[]){

  Display *theDisplay = XOpenDisplay( ":0" );

  // Setup an Open GL window to render into - this is not a dmuObj
  window = ScratchWindow::create(theDisplay, sizeX, sizeY, 
				 &argc, argv);
  
  window->setContext();
  if (!window){
    fprintf (stderr, "Failed to create scratch window!\n");
    return (-1);
  }

  // OpenGL setup for the window
  glViewport( 0, 0, sizeX, sizeY);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,sizeX,0,sizeY,0,1000);

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_DEPTH_TEST);
  return (0);
}
