/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* gvplay_u = Plays uncompressed video fields from disk to 
 *	      video-out and to the gfx monitor.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength> 	- Number of reads per readv call.
 *     <-gfx_frd or     	- Display a full frame by doubling scan lines.
 *      -gfx_fri or     	- Display a full frame by interlacing fields.
 *      -gfx_half>      	- Display half height field.
 *     <-single>		- Force single buffering (default is double).
 *     fileName			- file created with vgrab_u. 
 *
 *
 * This program plays fields of uncompresed video from disk to video-out and gfx
 * in real time.  The video is stored in YUV 4:2:2 (16 bits per pixel) format.  
 * Video was stored using a Ciprico 6900 ultra-scsi raid array.
 *
 * The readv system call is used (along with direct I/O) for getting
 * the data from disk.  In order to use readv with direct I/O each field
 * must be rounded up to a page boundary.  The fields are stored this
 * way on disk.  Fortunately, each vl buffer element handed back by the vl
 * is also padded out to a 64k boundary.  Exploiting this "feature" is
 * what makes it all work.  Fields are therefore rounded to 16k (most
 * systems have either 4k or 16k page sizes).
 *
 * The optional flag "-iov" allows the user to specify the number of reads for
 * readv to perform.  Each read corresponds to a video field padded out
 * to a 16k boundary (the default is 12 reads per readv - smaller values like
 * 6 and 8 are also good and use less memory).
 *
 * By default, the gfx window is the size of a field of video (i.e. 720x243 for
 * CCIR 525).  -gfx_half forces the gfx window to be half sized (360x243 for
 * CCIR 525).  -gfx_frd and -gfx_fri cause the window to display an entire frame
 * (720x486 for CCIR 525).  For -gfx_fri, single buffering is always forced
 * on and the SGI interlace extension is used.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After starting the program, a blank window appears.
 *      - Click with the left mouse button to start play.  
 *	- Click again to stop play then again to start again. Each button click
 *	  causes the main thread to enter the routine doit().
 *      - Each time play is started,  a read thread (see ReadIt) and display
 *        thread (see PutField) are sproc'd from inside the routine doit().
 *      - While things are playing, the main thread is in  XtAppMainLoop(), 
 *	  waiting for another button click.
 *
 *    Read Thread: 
 *      - ReadIt() simply cycles through the file.  It allocates DMbuffers from
 *        _outpool and loads them with uncompressed fields read from disk.
 *      - This thread uses the readv system call to read _iovcnt fields at a 
 *	  time into allocated dmbuffers.
 *      - After the read has finished,   ReadIt() marks these dmbuffers as 
 *	  valid for output by placing them in the array _readvBufs starting 
 *	  at index _dmtail.
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	  by _f.readSize).
 *	   
 *    Display Thread:
 *	- The display thread (see PutField) continually checks _readvBufs for 
 *	  valid data to send to video-out.
 *	- When a dmbuffer is sent to video-out (using vlDMBufferSend) its 
 *	  video-out display time is calcualted usng vlGetFrontierMSC() and 
 *	  saved away. 
 *	- The display thread then draws the field to the gfx monitor (using 
 *	  glDrawPixels) when the corresponding video-out display time has been 
 *	  reached (this keeps video-out and gfx in sync.). 
 *      - **** WHEN RUNNING IN DOUBLE BUFFERED MODE, this seems to work best 
 *	  when the gfx timing is set to 75 Hz (see xscreen).
 * 
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16 
#define MAX_DMBUFS 48
#define FILL_CHECK 60
#define FILL_HIGH 16
#define FILL_LOW 4

/* All global variables in this file are preceded by an "_".  */

/* This structure contains dmbuffer/usr pairs used to sync. play between
 * the gfx monitor and video out.
 */

typedef struct {
    DMbuffer buf;
    stamp_t ust;
} GFXBuf; 

/* Video related declarations */

VLServer _svr;
VLPath   _path;		/* App. uses a mem to vid path.	*/
VLNode   _src;		/* A VL_MEM node.		*/
VLNode   _drn;		/* A VL_VIDEO node.		*/

/* Misc. declarations.	*/
int _winx;		/* x dimension of gfx window.		*/
int _winy;		/* y dimension of gfx window.		*/
int _CurField = 0;	/* Current field being processed.	*/
int _dbuffer;		/* Double buffering flag.		*/
int _inter;		/* Interlacing flag.			*/
int _iovcnt = 12;	/* Number of reads per readv call.	*/
volatile int _IamPlaying = 0;  /* Am I playing or am I stopped?	*/
fileHeader_t _f;

/* DM buffer declarations  */

DMbufferpool _outpool;	/* dm buffer pool used to drive video and gfx. */
DMbuffer    *_readvBufs;/* Queue of dmbuffers with valid data read from disk */
volatile int _dmhead;	/* Head pointer for _readvBufs queue.	*/
volatile int _dmtail;	/* Tail pointer for _readvBufs queue.	*/
int          _DMBUFS;	/* Number of dm buffers in _outpool	*/ 
int          _outpoolfd;/* File descriptor for _outpool.	*/

/* Gfx declarations  */

Display   *_dpy;
GLXContext _context;
Widget     _glw;

/* Procdure declarations  */
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void initVL();
void initGL();
void PutField();
void ReadIt();
int sginap (long );

main(int argc, char *argv[])
{
    int i,n;
    char *Fname;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"doit", doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n";
    
    _inter = 0; _dbuffer = 1; _winx =1; _winy = 1;

    for (i=1; i < argc; i++) {
	if (strcmp (argv[i], "-single") == 0) {
	    _dbuffer = 0;
	}
        else if (strcmp (argv[i], "-gfx_fri") == 0) {
            _inter = 1; _dbuffer = 0; _winy = 2; _winx = 1;
        }
        else if (strcmp (argv[i], "-gfx_frd") == 0) {
            _inter = 0; 	      _winy = 2; _winx = 1;
	}
        else if (strcmp (argv[i], "-gfx_half") == 0) {
            _inter = 0;               _winy = 1; _winx = 2;
        }
        else if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
	else {
	    Fname = argv[i];
	}
    }

    fileInit(Fname, &_f, TRUE);
    fprintf(stderr, "Playing %d of %dx%d using vector %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Field read size = %d File header size = %d\n", _f.readSize, _f.headerSize);

    /* Set some state dependent on input from user and file attributes.*/
    _winx  = _f.uSize/_winx;
    _winy  = _f.vSize*_winy;
    _DMBUFS = 3*_iovcnt;
    _readvBufs = (DMbuffer *)malloc(_iovcnt*sizeof(DMbuffer));


    /* Cheat a liitle here, it makes the reading a lot simpler. */
    if (_f.Nfields < _iovcnt) {
	_iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }
	
    /* Set up video hardware  */
    initVL();

    /* Do that nasty Motif stuff then head off to the Xt main loop.	*/

    toplevel = XtAppInitialize(&app_context, "4DgiftsGlx",
			       (XrmOptionDescList)NULL , 0,
			       (int*)&argc,
			       (String*)argv,
			       NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight, _winy);n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);
    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);
}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the initialize call back routine for our widget.   This
     * is called automatically from the Xt loop.
     */

    float xdecim, ydecim;
    Arg args[18];
    XVisualInfo *vi;

    _dpy = XtDisplay(_glw);

    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(w, args, 1);

    /* create a context */
    _context = glXCreateContext(XtDisplay(_glw), vi, 0, GL_TRUE);
    GLwDrawingAreaMakeCurrent (_glw, _context);

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glColor4f(0.0, 0.0, 0.0, 1.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_winx, 0.0, (double)_winy, -1.0, 1.0);

    /* Thje orientation of the video field is top to bottom.  In order
     * to make it look right on the gfx monitor we need to do a negative
     * y for pixel zoom and a glRatserPos to the upper left corner.
     */
    if (_inter) {
        glEnable(GL_INTERLACE_SGIX);
        glPixelZoom (1.0, -1.0);
    }
    else {
	xdecim = ((float)(_winx))/_f.uSize;
	ydecim = ((float)(_winy))/_f.vSize;
	glPixelZoom (xdecim, -ydecim);
    }
    glRasterPos2i (0, _winy);
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the expose callback routine, called whenever an expose happens.
     * Note, if real time play is happening then this routine cannot try to 
     * grab the gfx context.
     */

    static int firstExpose = 1;

    if (firstExpose) {
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy,XtWindow(_glw));
 	firstExpose = 0;
	fprintf(stderr, "*****> Click in window to start play, click again to stop, etc.\n");
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    /* This routine is called each time the left mouse button is clicked.
     * If things are stopped then the read and display threads are sproc'd.
     * If things are playing then the read and display thread are stopped
     * by setting _IamPlaying to 0.
     */

    int readPID, putPID;

    if (!_IamPlaying) {  /*Go and play */
	_IamPlaying = 1;
	_dmhead = _dmtail = 0; 
        GC (glXMakeCurrent(_dpy, None, NULL));  
        readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
        putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);
    }
    else { 
        /* Setting IamPlaying to 0 will stop the ReadIt().  Upon exiting,
         * ReadIt() sets IamPlaying to -1, this will stop PutField().  Upon
         * exiting, PutField sets IamPLaying to -2.
         */
        _IamPlaying = 0;
        while (_IamPlaying != -2) sginap(1);  /* Wait for both threads */
        _IamPlaying = 0;
	VC  (vlEndTransfer(_svr, _path));
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));

        /* Make sure _CurField is pointing at an even field	*/
	if ((_CurField % 2) != 0) {
	    _CurField++;
	    if (_CurField >= _f.Nfields) _CurField = 0; 
        }
    }
}

void PutField()
{
    /* This routine has two main tasks.  It checks _readvBufs[_dmhead] for
     * a valid dmbuffer to send to video out.  After sending the buffer to video
     * out, it saves it in go[tail].buf along with the ust for when this field 
     * will appear at video out.  So the "go" queue contains dmbuffers which 
     * have been sent to video out and are scheduled to go to gfx.  Each time 
     * through the loop go[head].ust is tested to see if it is time to draw the
     *  next field to gfx.  This keeps video and gfx play in sync. 
     *
     * The ReadIt() thread is constantly reading fields off of disk and 
     * placing them at the end (i.e. _dmtail) of _readvBufs.  This thread
     * consumes from the front (i.e. _dmhead) of _readvBufs.
     */

    GFXBuf go[MAX_DMBUFS];  /* (dmbuf,ust) pairs that need to go to gfx */
    int head = 0;           /* The head of the go queue	- the next field for gfx output */ 
    int tail = 0;           /* The tail of the go queue - the field just sent to video  */ 
    int count = 0;
    int i;
    int scan;
    stamp_t cur_ust;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    USTMSCpair pair;
    double ust_per_msc;
    void *dataPtr;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);
    
    GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context)); /* Grab the context	*/
    ust_per_msc = vlGetUSTPerMSC(_svr, _path, _src);

    /* If interlacing, the first field starts on a different line
     * depending if 525 or 625 timing.
     */
    scan = (_f.vSize == 288) ? 0 : 1;

    while (_IamPlaying != -1) {

        if (head != tail) {
            /* Gfx output queue is NOT empty.			*/
	    dmGetUST((unsigned long long *)&cur_ust);	/* Sample the current ust.	*/

            /* If the current ust has passed the video-out ust for the head 
             * of the go list then draw the head of the go list to gfx.
             */
	    if (cur_ust > go[head].ust) {
	        dataPtr = dmBufferMapData(go[head].buf);

	        if (_inter) {
	            glRasterPos2i (0, _winy - scan);
	            scan = !scan; 
	        }

	        glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

                if (_dbuffer) {
                     if (!_inter)  glRecti(0, _winy, _winx, _winy-1);
                     glXSwapBuffers(_dpy,XtWindow(_glw));
                }

	        dmBufferFree(go[head].buf); /* This dmbuufer can now be freed. */
                head++;			    /* Point head at next entry in go list */
                if (head == _DMBUFS) head = 0; 
	    }
	    /* Otherwise, if there is some extra time until next gfx update,
	     * then take a break. 
	     */
	    else if ((go[head].ust - cur_ust) > 13000000) { /* in nanoseconds */
	        sginap(1);
	    }
	}   /* End if (head != tail) */

        /* The rest of the loop deals with removing dmbuffers from the head
         * of the _readvBufs queue.  The thread ReadIt() adds to the tail of 
         * the _readvBufs queue.
         */
	i = tail + 1;
        if (i == _DMBUFS) i = 0;

        if (i == head) {	/* No room in go to queue up next entry */
	    /* fprintf(stderr, "FULL GFX OUTPUT QUEUE\n"); */
	    continue; 
        }

        if (_dmhead == _dmtail) {	/* _readvBufs queue is empty.	*/
	    /* fprintf(stderr, "EMPYT READ QUEUE\n"); */
	    continue;
        }
        else {
            /* Grab the head of _readvBufs and place in tail of go queue. */
	    go[tail].buf = _readvBufs[_dmhead];
	    i = _dmhead + 1;
	    if (i == _DMBUFS) i = 0;
            _dmhead = i;
        }

        if (count == 0) {
	    VC (vlBeginTransfer(_svr, _path, 0, NULL));
	    vlGetUSTMSCPair(_svr, _path, _drn, VL_ANY, _src, &pair);
        }

        /* Get MSC for the sample and check for drops	*/
        newmsc = vlGetFrontierMSC(_svr, _path, _src);
        if ((oldmsc != -1) && (newmsc != oldmsc + 1) && (count > FILL_CHECK)) {
            fprintf(stderr, "MSC drop old = %lld, new = %lld\n", oldmsc, newmsc);
        }
        oldmsc = newmsc;

        /* Send the dmbufer to video out and calculate the video-out ust for
         * the field.  This is used later to draw the field to gfx.
         */
	VC ( vlDMBufferSend(_svr, _path, go[tail].buf) );
        go[tail].ust = pair.ust + (newmsc - pair.msc)*ust_per_msc;

        /* Point the tail at the next free location in go queue */
        tail++;	
        if (tail == _DMBUFS) tail = 0;
        count++;

    }  /* End of while (IamPlaying) */ 

    /* Seems to help field swap bug.	*/
    if ((count%2) != 0) {
        VC ( vlDMBufferSend(_svr, _path, go[head].buf) );
    }

    /* We have to free any dmbuffer in use.  The go queue
     * contains entries that have been sent to video but not yet
     * drawn to gfx.  _readvBufs contains entries read offof disk
     * but not sent anywhere.
     */

    while (head != tail) {
	dmBufferFree(go[head].buf);
	head++;
	if (head == _DMBUFS) head = 0;
    }

    while (_dmhead != _dmtail) {
	dmBufferFree(_readvBufs[_dmhead]);
	_dmhead++;
	if (_dmhead == _DMBUFS) _dmhead = 0;
    }

    GC (glXMakeCurrent(_dpy, None, NULL));  /* Surrender gl context */
    _IamPlaying--;		            /* Tell doit() all is done */	
    _exit(0);
}

void ReadIt()
{
    /* This routine uses readv to read uncompressed fields off of
     * disk into dmbuffers allocated from _outpool.  After a dmbuffer
     * is filled with valid data it is added to _readvBufs[_dmtail].
     * The display thread (see PutField()) consumes fields from 
     * _readvBufs[_dmhead].
     *
     * This routine exits before PutField(), it decrements _IamPlaying to -1
     * upon exit which causes PutField to exit.
     */

    fd_set fds;
    struct iovec iov[MAX_IOV];
    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    lseek64 (_f.fd, (off64_t)(_f.headerSize + (_CurField*_f.readSize)), SEEK_SET);

    while (_IamPlaying) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Wait until there are enough bytes in the dmbuffer pool
         * to do a readv of "fieldsToRead" fields.
         */ 
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        DC (dmBufferSetPoolSelectSize(_outpool, fieldsToRead*_f.readSize) );
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

        /* Now make sure there is room in _readvBufs to store 
         * "fieldsToRead" dmbuffers (which the consumer PutField() will later
         * display. 
         */
	do {
	    full = 0;
	    /* Check for the first full condition.  If the incremented
	     * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
	    tmphead = _dmhead;
	    tmptail = _dmtail + fieldsToRead;
            if (tmptail == _DMBUFS) tmptail = 0;
            full = (tmptail == tmphead); 

	    /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
	    full |= ((tmphead > _dmtail) && (tmphead < _dmtail + fieldsToRead));

	    if (full) {
		/* fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail); */
		sginap(1);
	    }
	} while (full);

        /* Load the address of the dmbuffers into the readv vector "iov"	*/
        tmptail = _dmtail;
        for (i=0; i < fieldsToRead; i++) {
	    while (dmBufferAllocate(_outpool, &(_readvBufs[tmptail])) == DM_FAILURE)sginap(1);
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[tmptail]);
	    tmptail++;
        }

        /* Read the fields.		*/
        OC (readv(_f.fd, iov, fieldsToRead));

        /* Inc. _dmtail, indicating more vaild data is in _readvBufs.	*/
        if (tmptail == _DMBUFS) tmptail = 0;
        _dmtail = tmptail;
	
	/* Check for file wrap.		*/
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass through file\n");
        }

   }  /* end while(IamPLaying) */

   _IamPlaying--;
   _exit(0); 
}


void initVL()
{
    /* Initialize the video stuff.  In particular, set up
     * a VL_MEM to VL_VIDEO path.  The VL_MEM node is attached 
     * to the DMbufferpool _outpool, these dmbuffers are filled
     * with data by ReadIt().  Once filled they are sen to video-out
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No mvp device");

    /* Set up a mem to vid path.		*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set three data attributes on the memory node.	*/
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "_DMBUFS" buffers
     * each containing _f.readSize bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f.readSize,DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and 
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMPoolGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.	*/
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);

    /* Create the dmbuffer pool used for video out.	*/
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _iovcnt*_f.readSize) );
    dmParamsDestroy(p);

} 
