/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vgrab_u = Grabs uncompressed video to disk.
 * 
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of writes per writev call.
 *			          Default is 12, 8 and 16 are also good values.
 *     <-n fieldsToCapture> 	- Number of fields to capture, default = 200.
 *				  This arg should be an even number.
 *     fileName                 - File for results. 
 *
 * This is a real simple program to capture fields of uncompresed video
 * to disk in real time.  The video is recorded in YUV 4:2:2 (16 bits
 * per pixel) format.  Video was captured using a Ciprico 6900 ultra-
 * scsi raid array.
 *
 * The writev system call is used (along with direct I/O) for getting 
 * the data to disk.  In order to use writev with direct I/O each field
 * must be rounded up to a page aligned boundary when being wriiten to disk.  
 * Fortunately, each buffer element handed back by the vl is padded to a 
 * 64k page-aligned boundary. Exploiting this "feature" is what makes it work.
 *
 * The flag "-iov" allows the user to specify the number of writes for
 * writev to perform.  Each write corresponds to a video field padded out
 * to a 16k boundary (most SGI machines are 4k or 16k page sizes).. 
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() is called to
 *        perform the real work.
 *	- doit() sits in a tight loop grabbing pointers to fields accumulating
 *	  in the VLBuffer.  
 *	- Once "_iovcnt" fields have been accumulated, they are written to disk
 *	  using writev().
 *	- After capturing the desired number of fields, the program exits.
 */


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <vl/vl.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16

/* All global variables in this file are preceded by an "_".  */

/* File globals */
fileHeader_t *_f;
int  _iovcnt = 12;		/* # of writes for writev.		*/

/* Video globals  */
VLServer _svr;			/* Connection to server.		*/ 
VLNode   _Csrc;			/* Source for capture (video node).     */
VLNode   _Cdrn;			/* Drain for capture (memory node).	*/
VLPath   _Cpath;		/* Capture path.			*/


/* Procudure declarations		*/
void doit();
void initVL();

main(int argc, char *argv[])
{
    int i,n;
    char * fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            n = atoi(argv[i]);
	}
        else if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
	else {
	    fname = argv[i];
	}
    }

    _f = fileCreate(fname, TRUE);
    _f->packing = VL_PACKING_YVYU_422_8; 
    _f->compressed = FALSE;
    _f->pad = _f->headerSize;
    _f->Nfields = n;

    fprintf(stderr, "Capturing %d fields,  writev vector length = %d\n", _f->Nfields,_iovcnt);
    initVL();
    doit();
}


void doit()
{
    /* The routine which does all the work capturing and writing 
     * fields to disk.
     */
    struct iovec iov[MAX_IOV];
    VLInfoPtr info;
    VLBuffer videoBuffer;
    stamp_t oldmsc = -1;
    stamp_t newmsc;
    int i, toGet;
    int fieldsCaptured = 0;

    setscheduling();  /* Raise priority if root or setuid */

    /* Register the buffer and begin data flowing  into the buffer.
     */
    VNC (videoBuffer = vlCreateBuffer(_svr, _Cpath, _Cdrn, _iovcnt*3));
    VC (vlRegisterBuffer(_svr, _Cpath, _Cdrn, videoBuffer));
    VC (vlBeginTransfer(_svr, _Cpath, 0, NULL));

    /* Now loop and capture the fields. */
    while (fieldsCaptured < _f->Nfields) {

	toGet = ((fieldsCaptured + _iovcnt) <= _f->Nfields) ?
                         _iovcnt : (_f->Nfields - fieldsCaptured);

        for (i= 0; i < toGet; i++) {
            do {    /* Get pointer to next valid field */ 
                info = vlGetNextValid(_svr, videoBuffer);
            } while (!info);

	    /* Load the write size and write address for the vector */
            iov[i].iov_len = _f->readSize;
            iov[i].iov_base = vlGetActiveRegion(_svr, videoBuffer, info);

	    /* Check the MSC of the sample to see if we are dropping
             * any fields.
             */
            newmsc = vlGetFrontierMSC(_svr, _Cpath, _Cdrn);
            if ((oldmsc != -1) && (newmsc != oldmsc + 1)) {
                fprintf(stderr, "MSC drop old = %lld, new = %lld\n", oldmsc, newmsc);
            }
            oldmsc = newmsc;
	}

	OC (writev(_f->fd, iov, toGet));  /* write the fields to disk. 	*/

	/* Indicate that the buffer entries are free for re-use.	*/
        for (i=0; i < toGet; i++) {
	    vlPutFree(_svr, videoBuffer);
	}
	fieldsCaptured += toGet;
       
        if (fieldsCaptured % 600 == 0) {
	    fprintf(stderr, "...Captured %d fields\n", fieldsCaptured);
	}
    }
    fprintf(stderr, "\n");

    VC (vlEndTransfer(_svr, _Cpath));
    VC (vlDeregisterBuffer(_svr, _Cpath, _Cdrn, videoBuffer));
    VC (vlDestroyBuffer(_svr, videoBuffer));

    /* Write out the correct header info.       */
    writeHeader(_f);
    fsync(_f->fd);
    close(_f->fd);
    fprintf(stderr, "Done writing %d fields to file\n", fieldsCaptured);
}

void initVL()
{
    /* Set up the video hardware for capturing data in YUV 4:2:2
     * format.  This includes creating a VL_VIDEO to VL_MEM path.
     */
    int i, n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
	devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices; 
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No mvp video device");

    /* Grab a video node and a memory node, then create a
     * path from video-in to memory.
     */
    VC (_Csrc = vlGetNode(_svr, VL_SRC, VL_VIDEO, VL_ANY));
    VC (_Cdrn = vlGetNode(_svr, VL_DRN, VL_MEM, VL_ANY));
    VC (_Cpath = vlCreatePath(_svr, devPtr->dev, _Csrc, _Cdrn));
    VC (vlSetupPaths(_svr, (VLPathList)&_Cpath, 1, VL_SHARE, VL_SHARE));

    /* Set the attributes for the incoming video.  This is
     * done on the memory drain.
     */
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_CAP_TYPE, &val));

    vlGetControl(_svr, _Cpath, _Cdrn, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
    _f->uSize = val.xyVal.x;
    _f->vSize = val.xyVal.y;

    _f->readSize = vlGetTransferSize(_svr, _Cpath);
    val.intVal = _f->readSize;

    i = _f->readSize/_f->headerSize;
    if ((_f->readSize % _f->headerSize) != 0) ++i;
    _f->readSize = _f->headerSize*i;

    fprintf(stderr, "Transfer size  = %d Write size  = %d\n", val.intVal, _f->readSize);
}
