/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_stack_u = Plays uncompressed video fields from disk to video out.
 *                 The field is sent through gfx, stacked on top of itself
 *                 (oh, what a cool effect) and sent to video out.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of reads per readv call.
 *     fileName                 - file created with vgrab_u.
 *
 * The readv system call is used (along with direct I/O) for getting
 * the data from disk.  In order to use readv with direct I/O each field
 * must be rounded up to a page boundary.  The fields are stored this
 * way on disk.  Fortunately, each vl buffer element handed back by the vl
 * is also padded out to a 64k boundary.  Exploiting this "feature" is
 * what makes it all work.
 *
 * The optional flag "-iov" allows the user to specify the number of reads for
 * readv to perform.  Each read corresponds to a video field padded out
 * to a 16k boundary (the default is 12 reads per readv - smaller values like
 * 6 and 8 are also good and use less memory).
 *
 * DMbufferpools provide the application with special purpose memory that
 * can be shared between the gfx, video, and compression devices.  A
 * DMbufferpool is divided into equal sized compartments called DMbuffers.
 * The application allocates one of these compartments by calling 
 * dmBufferAllocate.
 *
 * This particular program makes use of three DMbufferpools:
 *    _readpool - This pool contains "_DMBUF" DMbuffers, used to hold 
 *	  	  uncompressed YUV 4:2:2 fields read from disk. 
 *     _vdmpool - This pool contains "GFXBUFS" DMbuffers, used as the
 *                the color buffers (i.e. frame buffers) for the pbuffer
 *                _vdmpbuffer.   These DMbuffers can also be passed directly
 *                to video-out (via vlDMBufferSend).
 *     _texpool - This pool contains 1 DMbuffer, used as the color buffer
 *                for the pbuffer _texpbuffer.
 *
 * The pbuffers used by this program (_vdmpbuffer & _texpbuffer) are able to use
 * memory from a DMbufferpool for their color buffers.  To do this, these
 * pbuffers must be created as digital media pbuffers (this is passed in as 
 * an attribute to glXCreateGLXPbufferSGIX).  Being a digital media pbuffer 
 * allows _vdmpbuffer to cycle through the DMbuffers in _vdmpool, rendering 
 * into each of them.  These dmbuffers can then be sent to video out.  
 *
 * _vdmpbuffer uses a particular DMbuffer from _vdmpool via a call to 
 * glXAssociateDMPbufferSGIX.  In addition, the fact that _texpbuffer is also
 * a digital media pbuffer allows glCopyTexSubImage2DEXT to copy pixels from
 * _texpbuffer to texture memory via a "copy by reference".  In this case, 
 * no real pixel copying occurs and _texpbuffer is really being used as the 
 * texure memory for _vdmpbuffer.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() sprocs off
 *        a reader thread (see ReadIt()) and a display thread (see PutField()).
 *      - The main thread then just hangs out.
 *
 *    Reader Thread:
 *      - ReadIt() simply cycles through the file.  It allocates DMbuffers from
 *        _readpool and loads them with uncompressed fields read from disk.
 *	- This thread uses the readv system call to read _iovcnt fields 
 *	  into allocated dmbuffers.
 *	- After the read has finished,   ReadIt() marks these dmbuffers as valid
 *	  for output by placing them in the array _readvBufs starting at 
 *	  index _dmtail. 
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	  by _f.readSize).
 *
 *    Display Thread:
 *      - PutField() waits for a field to be placed in _readvBufs[_dmhead]. 
 *      - It then grabs the next available DMbuffer from _vdmpool (denote 
 *	  this by nextDMbuf).
 *      - The field is drawn into nextDMbuf as the background by calling 
 *	  glDrawPixels.
 *      - glDrawPixels is called again to draw the field into the DMbuffer 
 *	  (_texbuf) associated with _texpbuffer (so this glDrawPixels call 
 *	  color converts the field to RGB and loads it into texture memory).
 *      - glXMakeCurrentReadSGI is then used to set up _texpbuffer as the read
 *        source for _vdmpbuffer (the color buffer for _vdmpbuffer is still 
 *	  nextDMbuf).
 *      - glCopyTexSubImage2DEXT is called.  This does a "copy-by-reference".
 *        The net effect is that _texpbuffer is now functioning as texture 
 *	  memory for _vdmpbuffer.
 *      - Two texture mapped quads are drawn on top of the background.  
 *	  All this rendering is happening with _vdmpbuffer and its associated 
 *	  DMbuffer, nextDMbuf.
 *      - nextDMbuf is then sent to video via vlDMBufferSend().
 *      - PutField then loops back and waits for the next available field.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ulocks.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>

#include <GL/gl.h>
#include <GL/glx.h>
#include <vl/vl.h>
#include <dmedia/dm_buffer.h>
#include <dmedia/dm_imageconvert.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16 
#define GFXBUFS 16 
#define PRELOAD 8
#define FILL_CHECK 60
#define FILL_HIGH 12
#define FILL_LOW 4

/* All global variables in this file are preceded by an "_".  */

/* Video related defines */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _drn;

/* File related defines  */
fileHeader_t _f;
int _CurField;          /* Current field being read.            */
int _iovcnt = 12;	/* Number of read per readv call.	*/

/* Defines for digital media pbuffer that drives fields to video out */

GLXPbufferSGIX _vdmpbuffer; /* dm-pbuffer that drives video. Associated with _vdmpool   */
DMparams       *_vdmParams; /* dmParams associated with _vdmpbuffer.                    */
GLXContext     _vdmcontext; /* gfx context associated with _vdmpbuffer.                 */
DMbufferpool      _vdmpool; /* dmbuffer pool used as color buffer memory for _vdmpbuffer*/
                            /* These buffers are also sent to video out (vlDMBufferSend)*/
int             _vdmpoolfd; /* fd to select on _vdmpool.                                */

/* Defines for digital media pbuffer that loads texture memory       */

GLXPbufferSGIX _texpbuffer; /* dm-pbuffer that serves as texture memory for _vdmpbuffer */
DMparams       *_texParams; /* dmParams associated with _texpbuffer.                    */
GLXContext    _textcontext; /* gfx context associated with _texpbuffer.                 */
DMbufferpool      _texpool; /* dmbuffer pool used as color buffer memory for _texpbuffer*/
DMbuffer           _texbuf; /* The one and only dmbuffer allocated from _texpool.       */
int             _tex_width;
int            _tex_height;

/* DMbuffer pool that fields are read into from disk.		     */

DMbufferpool _readpool;         /* dm buffer pool used to hold fields read from disk.	*/ 
DMbuffer    *_readvBufs;        /* Queue of dmbuffers with valid data read from disk 	*/
volatile int _dmhead, _dmtail;  /* Head and tail pointers for _readvBufs queue. 	*/
int          _DMBUFS;           /* Number of dm buffers in _readpool             	*/
int          _readpoolfd;       /* File descriptor used to select on _readpool.		*/ 

Display *_dpy;

/* Routine declarations.	*/
void initCB();
void doit();
void initVL();
void initGL();
void initDM4GLVL();
void initDM4Tex();
void initDM4Read();
void PutField();
void ReadIt();
int sginap (long );
GLXFBConfigSGIX getFBConfig(Display *, int);

main(int argc, char *argv[])
{
    int i,n;
    char *Fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
        else {
            Fname = argv[i];
        }
    }

    fileInit(Fname, &_f, TRUE);
    fprintf(stderr, "Playing %d of %dx%d using vector %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Field read size = %d File header size = %d\n", _f.readSize, _f.headerSize);

    /* Cheat a liitle here, it makes the reading a lot simpler. */
    if (_f.Nfields < _iovcnt) {
        _iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }

    /* Set tetxure width and height based on field sizes from file.  */
    _tex_width = 1024;                        /* Field width is 640, 768, or 720  */
    _tex_height = (_f.vSize == 240) ? 256 : 512;  /* 240 or 288                     */

    _DMBUFS = 3*_iovcnt;
    _readvBufs = (DMbuffer *)malloc(_DMBUFS*sizeof(DMbuffer));

    initVL();		/* Create mem to vid path. */
    initDM4Read();	/* Set up _readpool.	   */
    initDM4GLVL();	/* Set up _vdmpool         */
    initDM4Tex();	/* Set up _texpool         */
    initGL();		/* Create _vdmpbuffer, _texpbuffer and init. their contexts */

    doit();	/* sproc ReadIt() and sproc PutField()  */
}


void initGL()
{
    /* This routine creates the two digital media pbuffers used by this
     * program (_vdmpbuffer and _texpbuffer).  
     */

    unsigned int texA;
    GLXFBConfigSGIX config;
    DMbuffer gfxbuf;
    int attrib[] = {GLX_DIGITAL_MEDIA_PBUFFER_SGIX, GL_TRUE,
                    GLX_PRESERVED_CONTENTS_SGIX, GL_TRUE,
                    None};

    _dpy = XOpenDisplay( ":0" ); 

    /* Get a 24 bit frame buffer configuration */
    config = getFBConfig(_dpy, 24);
    ERROR((config == NULL), "Can't get FB");

    /* Use this fb config. to set up _vdmpbuffer	*/ 
    GNC (_vdmpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _f.uSize, _f.vSize, attrib));
    GNC (_vdmcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));

    /* Allocate a color buffer from _vdmpool and associate it with _vdmpbuffer. 
     * There must be an associated DMBuffer in order to init the pbuffer context.
     */
    DC ( dmBufferAllocate(_vdmpool, &gfxbuf) );
    GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));

    /* Initialize context for _vdmpbuffer.  */
    GC (glXMakeCurrent(_dpy, _vdmpbuffer, _vdmcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_f.uSize,_f.vSize);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glPixelZoom (1.0, -1.0);     /* Need to flip the video when drawn as background */
    glRasterPos2f (-1.0, 1.0);

    /* Set up texture parameters. */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
    createTexture(&texA, _tex_width, _tex_height, GL_RGBA, 0);
    BuildGfxLists(1, _f.uSize, _f.vSize, _tex_width, _tex_height,0);

    dmBufferFree(gfxbuf);  /* Done with buffer for now.	*/

    /* Now use config to set up _texpbuffer */ 
    GNC (_texpbuffer = glXCreateGLXPbufferSGIX(_dpy, config, _tex_width, _tex_height, attrib));
    GNC (_textcontext = glXCreateContextWithConfigSGIX(_dpy, config, GLX_RGBA_TYPE_SGIX, NULL, True));
    DC ( dmBufferAllocate(_texpool, &_texbuf) );

    /* Since there is only one DMbuffer required for texture memory, this
     * association is made only once in the beginning.
     */
    GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));

    /* Initialize context for _texpbuffer.	*/

    GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext)); 
    glClearColor(0.0,0.0,0.0,1.0);
    glDisable(GL_DITHER);
    glViewport(0,0,_tex_width, _tex_height);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_tex_width, 0.0, (double)_tex_height, -1.0, 1.0);
    glRasterPos2i (0, 0);
}

void doit()
{
    /* The main work routine - it sprocs off a read thread (ReadIt)
     * and a disply thread (PutField) to do the real work.
     */
    int readPID, putPID;

    _CurField = 0;
    _dmhead = _dmtail = 0; 

    /* Surrender the gfx context.       */
    GC (glXMakeCurrent(_dpy, None, NULL));

    /* Sproc off the read and display threads.  */
    readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
    putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);

    while (1) sginap(100);  /*^C to exit        */

}

void PutField()
{
    /* This routine grabs DMbuffers with valid datd from _readvBufs[_dmhead].
     * Is is sprco'd off inside doit() along with the read thread (see ReadIt).
     * The field is drawn as background and then texture mapped onto two
     * quads (which are then drawn over the backgorund image).
     *
     * The ReadIt() thread is constantly reading fields off of disk and
     * placing them at the end (i.e. _dmtail) of _readvBufs.  This thread
     * consumes from the front (i.e. _dmhead) of _readvBufs.
     */

    fd_set fds;
    int i;
    int count = 0;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    void *dataPtr;
    DMbuffer gfxbuf;
    float scale = 0.50;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();

    while (1) {

	/* If no fields available, try again	*/
        if (_dmhead == _dmtail) {
	    continue;
        }

        /* Load the field into the texture pbuffer */
        GC (glXAssociateDMPbufferSGIX(_dpy, _texpbuffer, _texParams, _texbuf));
        GC (glXMakeCurrent(_dpy, _texpbuffer, _textcontext));
        dataPtr = dmBufferMapData(_readvBufs[_dmhead]);
        glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);

        /* Get the next available DMbuffer from _vdmpool.       */
        FD_ZERO(&fds);
        FD_SET(_vdmpoolfd, &fds);
        OC (select(_vdmpoolfd+1, NULL, &fds, NULL, NULL));
        DC (dmBufferAllocate(_vdmpool, &gfxbuf));

        /* Associate this new DMbuffer with _vdmpbuffer (i.e. it is now its
         * color buffer).  Then set up _texpbuffer as the read source for
         * the texture copy.
         */
        GC (glXAssociateDMPbufferSGIX(_dpy, _vdmpbuffer, _vdmParams, gfxbuf));
        GC (glXMakeCurrentReadSGI(_dpy, _vdmpbuffer, _texpbuffer, _vdmcontext));

        /* Draw the field (still pointed to by dataPtr) as backgroiund  */
        glDisable(GL_TEXTURE_2D);
        glClear(GL_COLOR_BUFFER_BIT);
        glDrawPixels(_f.uSize, _f.vSize, GL_YCRCB_422_SGIX,GL_UNSIGNED_BYTE,dataPtr);


        /* Now call glCopyTexSubImage2DEXT to load the field from _texpbuffer
         * into texture memory.  This is actually a copy by reference, the net
         * effect being that _texpuffer (and its associated _texbuf) serves as
         * texture memory for _vdmpbuffer.
         */
        glEnable(GL_TEXTURE_2D); /* enable texturing */
        glCopyTexSubImage2DEXT( GL_TEXTURE_2D, 0, 0, 0, 0, 0, _tex_width, _tex_height);

        /* Draw the two texture mapped quads.   */
        glPushMatrix();
        glScalef(scale, scale, scale);
        glCallList(1);
        glScalef(scale, scale, scale);
        glCallList(1);
        glPopMatrix();

        /* See if we are dropping any fields.  There seems to be a
         * a bug with vlGetFrontierMSC() where it sometimes reports
         * bogus results if called too often, so check every fourth put.
         */
        if (count % 4 == 0) {
	    newmsc = vlGetFrontierMSC(_svr, _path, _src);
	    if ((oldmsc != -1) && (newmsc != oldmsc + 4) && (count > FILL_CHECK)) {
	        fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+4, newmsc);
	    }
	    oldmsc = newmsc;
        }

        /* Send to video out */
	VC ( vlDMBufferSend(_svr, _path, gfxbuf) );
	DC ( dmBufferFree(_readvBufs[_dmhead]) );
	DC ( dmBufferFree(gfxbuf) );
        count++;

	/* Point head pointer at next valid entry.		*/
        i = _dmhead + 1;
        if (i == _DMBUFS) i = 0;
        _dmhead = i;

        /* Don't start the transfer until we have some fields queued up.        */
        if (count == PRELOAD) {
            VC (vlBeginTransfer(_svr, _path, 0, NULL));
        }
        else if (count > FILL_CHECK) {
	    i = vlGetFilledByNode(_svr, _path, _src);
	    if (i > FILL_HIGH) sginap(1);
            if (i < FILL_LOW ) fprintf(stderr, "Filled down to %d\n", i);
	}
    } 
}

void ReadIt()
{
    /* This routine uses readv to read uncompressed fields off of
     * disk into dmbuffers allocated from _readpool.  After a dmbuffer
     * is filled with valid data it is added to _readvBufs[_dmtail].
     * The display thread (see PutField()) consumes fields from 
     * _readvBufs[_dmhead].
     */

    fd_set fds;
    int i, fieldsToRead;
    int tmphead, tmptail;
    int full;
    struct iovec iov[MAX_IOV];

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();
    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);

    while (1) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Wait till there's room in _readpool */
        FD_ZERO(&fds);
        FD_SET(_readpoolfd, &fds);
        DC (dmBufferSetPoolSelectSize(_readpool, fieldsToRead*_f.readSize) );
        OC (select(_readpoolfd+1, NULL, &fds, NULL, NULL));

	do {
	    full = 0;
	    /* Check for the first full condition.  If the incremented
	     * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
	    tmphead = _dmhead;
	    tmptail = _dmtail + fieldsToRead;
            if (tmptail == _DMBUFS) tmptail = 0;
            full = (tmptail == tmphead); 

	    /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
	    full |= ((tmphead > _dmtail) && (tmphead < _dmtail + fieldsToRead));

	    if (full) {
		/* fprintf(stderr, "Full read buffer %d %d\n", tmphead, _dmtail); */
		sginap(1);
	    }
	} while (full);

        /* Load the address of the dmbuffers into the readv vector "iov"        */
        tmptail = _dmtail;
        for (i=0; i < fieldsToRead; i++) {
	    while (dmBufferAllocate(_readpool, &(_readvBufs[tmptail])) == DM_FAILURE);
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[tmptail]);
	    tmptail++;
        }

        /* Read the fields.             */
        OC (readv(_f.fd, iov, fieldsToRead));

        /* Inc. _dmtail, indicating more vaild data is in _readvBufs.   */
        if (tmptail == _DMBUFS) tmptail = 0;
        _dmtail = tmptail;
	
	/* Check for file wrap.         */
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass throughfile\n");
        }
   }
}


void initVL()
{
    /* Initialize the video stuff.  In particular, set
     * up a mem-> vid path.  The memory node for this play
     * path will be attached to the digital media pbuffers
     * where the textured video is being drawn.  Results can
     * then be sent directly from the pbuffer to video out.
     * See initDM4GLVL() for more details on this.
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No mvp device");

    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set the data attributes on the memory node.    */
    val.intVal = VL_FORMAT_RGB;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = VL_PACKING_ABGR_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    /* Need this we are going direcly fro gfx to video out.	*/
    val.intVal = VL_LAYOUT_GRAPHICS;
    VC (vlSetControl(_svr, _path, _src, VL_LAYOUT, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
}

void initDM4Read()
{
    DMparams *p;

    /* Create dmbuffer pool to read fields into */ 
    dmParamsCreate(&p);

    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f.readSize,DM_FALSE,DM_TRUE));
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC( dmBufferCreatePool(p, &_readpool) );

    _readpoolfd = dmBufferGetPoolFD(_readpool);
    DC (dmBufferSetPoolSelectSize(_readpool, _iovcnt*_f.readSize) );
    dmParamsDestroy(p);
} 

void initDM4GLVL()
{
    /* This routine creates _vdmpool, the DMbufferpool that will
     * provide the color buffers (i.e. frame buffer memory) for the
     * dm-pbuffer _vdmpbuffer.  This buffer pool is also attached to
     * the vl memory node (_src) that is driving video out.  This allows
     * the application to render into DMbuffers from _vdmpool using OpenGL
     * AND to send the very same DMbuffers to video out (with no additional
     * copying.
     */

    DMparams *p;

    /* Set up image attributes for the buffer pool      */
    dmParamsCreate(&_vdmParams);
    DC ( dmSetImageDefaults(_vdmParams, _f.uSize, _f.vSize, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_vdmParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_vdmParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Create parameter list p and load it with default pool information.       */
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, GFXBUFS, _f.uSize*_f.vSize*4, DM_FALSE, DM_TRUE) );

    /* Now load p with specific image (GL) information from _vdmParams. */
    DC ( dmBufferGetGLPoolParams(_vdmParams, p) );

    /* Now load p with the required information to attache video to the buffer pool. */
    VC ( vlDMPoolGetParams(_svr, _path, _src, p) );

    /* Create the buffer pool based on all the info. loaded into p.     */
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);
    DC ( dmBufferCreatePool(p, &_vdmpool) );
    dmParamsDestroy(p);

    /* Get the fd we can use to select on the next available dmbuffer from
     * _vdmpool.  The set the select size for the pool as well.
     */
    _vdmpoolfd = dmBufferGetPoolFD(_vdmpool);
    DC (dmBufferSetPoolSelectSize(_vdmpool, _f.uSize*_f.vSize*4) );
}

void initDM4Tex()
{
    /* This routine creates the DMbufferpool that provides the DMBuffer
     * used as frame buffer memory for the dm-pbuffer _texpbuffer.  Only
     * one DMbuffer is alloacted in this pool (it is assigned to _texbuf inside
     * initGL().  _texbuf will end up serving as texture memory for _vdmpbuffer.
     */

    DMparams *p;

    /* Set up geometry attributes for pool.     */
    dmParamsCreate(&_texParams);
    DC ( dmSetImageDefaults(_texParams, _tex_width, _tex_height, DM_IMAGE_PACKING_RGBA) );
    DC ( dmParamsSetEnum(_texParams,DM_IMAGE_ORIENTATION,DM_IMAGE_TOP_TO_BOTTOM) );
    DC ( dmParamsSetEnum(_texParams, DM_IMAGE_LAYOUT, DM_IMAGE_LAYOUT_GRAPHICS) );

    /* Load p with default info. + geometry info from _texparams.       */
    dmParamsCreate(&p);
    DC ( dmBufferSetPoolDefaults(p, 1, _tex_width*_tex_height*4, DM_FALSE, DM_TRUE) );
    DC ( dmBufferGetGLPoolParams(_texParams, p) );
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);

    /* Create the buffer pool.  */
    DC ( dmBufferCreatePool(p, &_texpool) );
    dmParamsDestroy(p);
}
