/*
 * Copyright (c) 1996 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vplay_u = Plays uncompressed video fields from disk to video out
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of reads for per readv call.
 *                                Default is 12, 8 and 16 are also good values.
 *     fileName                 - File created by vgrab_u
 *
 * The readv system call is used (along with direct I/O) for getting
 * the data from disk.  In order to use readv with direct I/O each field
 * must be rounded up to a page boundary.  The fields are stored this
 * way on disk.  Fortunately, each vl buffer element handed back by the vl
 * is also padded out to a 64k boundary.  Exploiting this "feature" is
 * what makes it all work.
 *
 * The optional flag "-iov" allows the user to specify the number of reads for
 * readv to perform.  Each read corresponds to a video field padded out
 * to a 16k boundary (the default is 12 reads per readv - smaller values like
 * 6 and 8 are also good and use less memory).
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After doing initialization, the routine doit() is entered.
 *      - Inside doit(), the program sprocs a read thread (see ReadIt) and
 *        a display thread (PutField()). 
 *      - The main thread just hangs out. 
 *
 *    Read Thread: 
 *      - ReadIt() simply cycles through the file.  It allocates DMbuffers from
 *        _outpool and loads them with uncompressed fields read from disk.
 *      - This thread uses the readv system call to read _iovcnt fields.
 *        into allocated dmbuffers.
 *      - After the read has finished,   ReadIt() marks these dmbuffers as valid
 *        for output by placing them in the array _readvBufs starting at 
 *	  index _dmtail.
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	  by _readSize).
 *	   
 *    Display Thread:
 *	- The display thread (see PutField) continually checks 
 *	  _readvBufs[_dmhead] for valid data to send to video-out 
 *	  (via vlDMBufferSend()).
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/uio.h>
#include <vl/vl.h>
#include <dmedia/dm_buffer.h>
#include "common/handy.h"
#include "common/fileutils.h"


#define MAX_IOV 16 
#define FILL_CHECK 60
#define FILL_HIGH 12
#define FILL_LOW 4
#define PRELOAD 8

/* All global variables in this file are preceded by an "_".  */

/* Video related declarations */

VLServer _svr;
VLPath   _path;		/* App. uses a mem to vid path.	*/
VLNode   _src;		/* A VL_MEM node.		*/
VLNode   _drn;		/* A VL_VIDEO node.		*/

/* Misc. declarations.	*/
fileHeader_t _f;
int _iovcnt = 12;	/* Number of reads per readv call.	*/
int _CurField = 0;	/* Current field being processed.	*/

/* DM buffer declarations  */

DMbufferpool _outpool;		/* dm buffer pool used to drive video and gfx. */
DMbuffer    *_readvBufs;	/* Queue of dmbuffers with valid data read from disk */
volatile int _dmhead, _dmtail;	/* Head and tail pointers for _readvBufs queue.	*/
int          _DMBUFS;		/* Number of dm buffers in _outpool		*/ 
int          _outpoolfd;	/* File descriptor for _outpool.		*/

/*Procedure declarations */
void doit();
void initVL();
void PutField();
void ReadIt();

main(int argc, char *argv[])
{
    int i,n;
    char *Fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
	else {
	    Fname = argv[i];
	}
    }

    fileInit(Fname, &_f, TRUE);

    /* Cheat a liitle here, it makes the reading a lot simpler. */
    if (_f.Nfields < _iovcnt) {
	_iovcnt = _f.Nfields;
    }
    else if ((_f.Nfields % _iovcnt) != 0) {
        _f.Nfields -= (_f.Nfields % _iovcnt);
    }
	
    /* Set some state dependent on input from user and file attributes.*/
    _DMBUFS = 3*_iovcnt; 
    _readvBufs = (DMbuffer *)malloc(_DMBUFS*sizeof(DMbuffer));

    /* Set up video hardware  */
    initVL();

    fprintf(stderr, "Playing  %d fields of %dx%d using a read vector of length %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Padded read size (bytes) = %d \n", _f.readSize);

    doit();
}

void doit()
{
    /* sproc the read and display threads */
    int readPID, putPID;
    _dmhead = _dmtail = 0; 
    readPID = sproc((void (*)(void *))ReadIt, PR_SADDR | PR_SFDS);
    putPID = sproc((void (*)(void *))PutField, PR_SADDR | PR_SFDS);
    while (1) sginap(100);
}

void PutField()
{
    /* This routine constanylt _readvBufs[_dmhead] for a valid dmbuffer to send 
     * to video out.  It also uses vlGetFrontierMSC to determine if any
     * fields are being dropped. 
     *
     * The ReadIt() thread is constantly reading fields off of disk and 
     * placing them at the end (i.e. _dmtail) of _readvBufs.  This thread
     * consumes from the front (i.e. _dmhead) of _readvBufs.
     */

    int i;
    int count = 0;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    DMbuffer buf;
    void *dataPtr;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();

    while (1) {

        if (_dmhead == _dmtail) {	/* _readvBufs queue is empty.	*/
	    sginap(1);
	    continue;
        }

        /* Grab the head of _readvBufs and place in tail of go queue. */
        buf = _readvBufs[_dmhead];
        ++_dmhead;
        if (_dmhead == _DMBUFS) _dmhead = 0;

        /* See if we are dropping any fields.  There seems to be a
         * a bug with vlGetFrontierMSC() where it sometimes reports
         * bogus results if called too often, so check every fourth put.
         */
        if (count % 4 == 0) {
            newmsc = vlGetFrontierMSC(_svr, _path, _src);
            if ((oldmsc != -1) && (newmsc != oldmsc + 4) && (count > FILL_CHECK)) {
                fprintf(stderr, "MSC drop expected %lld got %lld\n", oldmsc+4, newmsc);
            }
            oldmsc = newmsc;
        }

        /* Send the dmbufer to video out and calculate the video-out ust for
         * the field.
         */
	VC ( vlDMBufferSend(_svr, _path, buf) );
	DC ( dmBufferFree(buf) );

        count++;

        /* Don't start the transfer until we have some fields queued up.        */
        if (count == PRELOAD) {
            VC (vlBeginTransfer(_svr, _path, 0, NULL));
        }
        else if (count > FILL_CHECK) {
            i = vlGetFilledByNode(_svr, _path, _src);
            if (i > FILL_HIGH) sginap(1);
            if (i < FILL_LOW ) fprintf(stderr, "Filled down to %d\n", i);
        }
    } 
}

void ReadIt()
{
    /* This routine uses readv to read uncompressed fields off of
     * disk into dmbuffers allocated from _outpool.  After a dmbuffer
     * is filled with valid data it is added to _readvBufs[_dmtail].
     * The display thread (see PutField()) consumes fields from 
     * _readvBufs[_dmhead].
     */

    fd_set fds;
    struct iovec iov[MAX_IOV];
    int i, fieldsToRead;
    int tmphead, tmptail;
    unsigned char full;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    setscheduling();

    lseek64 (_f.fd, _f.headerSize + ((off64_t)_CurField*_f.readSize), SEEK_SET);

    while (1) {
        fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
                         _iovcnt : (_f.Nfields - _CurField);

        /* Wait until there are enough bytes in the dmbuffer pool
         * to do a readv of "fieldsToRead" fields.
         */ 
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        DC (dmBufferSetPoolSelectSize(_outpool, fieldsToRead*_f.readSize) );
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

        /* Now make sure there is room in _readvBufs to store 
         * "fieldsToRead" dmbuffers (which the consumer PutField() will later
         * display. 
         */
	do {
	    full = 0;
	    /* Check for the first full condition.  If the incremented
	     * tail equals the head then wait, otherwise the consumer
             * will see tail == head and think the queue is empty
             */
	    tmphead = _dmhead;
	    tmptail = _dmtail + fieldsToRead;
            if (tmptail == _DMBUFS) tmptail = 0;
            full = (tmptail == tmphead); 

	    /* The other full condition is that the head is between tail and
             * (tail + fieldsPerRead).
             */
	    full |= ((tmphead > _dmtail) && (tmphead < _dmtail + fieldsToRead));

	    if (full) {
		/* fprintf(stderr, "Full readv buffer %d %d\n", tmphead, tmptail); */
		sginap(1);
	    }
	} while (full);

        /* Load the address of the dmbuffers into the readv vector "iov"	*/
        tmptail = _dmtail;
        for (i=0; i < fieldsToRead; i++) {
	    while (dmBufferAllocate(_outpool, &(_readvBufs[tmptail])) == DM_FAILURE);
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[tmptail]);
	    tmptail++;
        }

        /* Read the fields.		*/
        OC (readv(_f.fd, iov, fieldsToRead));

        /* Inc. _dmtail, indicating more vaild data is in _readvBufs.	*/
        _dmtail = tmptail;
        if (_dmtail == _DMBUFS) _dmtail = 0;
	
	/* Check for file wrap.		*/
        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
	    lseek64 (_f.fd, (off64_t)_f.headerSize, SEEK_SET);
	    fprintf(stderr, "Done pass through file\n");
        }

   } 

}


void initVL()
{
    /* Initialize the video stuff.  In particular, set up
     * a VL_MEM to VL_VIDEO path.  The VL_MEM node is attached 
     * to the DMbufferpool _outpool, these dmbuffers are filled
     * with data by ReadIt().  Once filled they are sent to video-out
     */

    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;

    VNC (_svr = vlOpenVideo(""));

    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "mvp") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No mvp device");

    /* Set up a mem to vid path.		*/
    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, VL_ANY, _f.uSize, _f.vSize);

    /* Set three data attributes on the memory node.	*/
    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    vlGetControl(_svr, _path, _src, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "_DMBUFS" buffers
     * each containing _f.readSize bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f.readSize,DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and 
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMPoolGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.	*/
    dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS);

    /* Create the dmbuffer pool used for video out.	*/
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    DC (dmBufferSetPoolSelectSize(_outpool, _iovcnt*_f.readSize) );
    dmParamsDestroy(p);
} 
