/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* gvplay_u - plays uncompressed video fields (4:2:2 or 24 bit yuv/rgb)
 *            from disk and sends the result to both video-out and to a 
 *	      screen node for display on the gfx monitor - OCTANE VIDEO ONLY.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of reads per readv call.
 *     <-rate frameRate>	- play at a slower rate.
 *     <-gfx_half>              - Display half height field.
 *     fileName                 - A file created with vgrab_u.
 *				  The file must contain fields (not frames).
 *				  Fields are the default for vgrab_u.
 *
 *
 * Sending video to a screen node means that the video hardware is inserting
 * video pixels onto the pixel bus just before the pixels arrive at the DACS 
 * (which drive the monitor).  So there is no gfx control over the pixles going 
 * to the monitor.
 *
 * The readv system call is used (along with direct I/O) for getting
 * the data from disk.  In order to use readv with direct I/O each field
 * must be rounded up to a page (16k) boundary.  The fields are stored this
 * way on disk.  Fortunately, each vl buffer element handed back by the vl
 * is also padded out to a 16k page boundary.  Exploiting this "feature" is
 * what makes it all work. A Ciprico 6900 disk array was uesd to store the
 * media.
 *
 * The optional flag "-iov" allows the user to specify the number of reads for
 * readv to perform.  Each read corresponds to a video field padded out
 * to a 16k boundary (the default is 12 reads per readv).
 *
 * The flag "-rate" allows the user to play things at a slower speed (for
 * example -rate 15 would play NTSC material at half speed).  In this case
 * only the first field from each pair of fields in the media file is sent
 * the video hardware.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After starting the program, a blank window appears.
 *      - Click with the left mouse button to start play.  
 *	- Click again to stop play then again to start again. Each button click
 *	  causes the main thread to enter the routine doit().
 *      - Each time play is started,  a play thread (see PlayIt) 
 *        is sproc'd from inside the routine doit().
 *      - While things are playing, the main thread is in  XtAppMainLoop(), 
 *	  waiting for another button click.
 *
 *    Play Thread: 
 *      - PlayIt() simply cycles through the file.  It allocates buffers from
 *        _outpool and loads them with uncompressed fields read from disk.
 *      - This thread uses the readv system call to read _iovcnt fields at a 
 *	  time into allocated dmbuffers.
 *      - After the read has finished,   PlayIt() marks these buffers as 
 *	  valid for output. 
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	  by _readSize).
 *
 * For best results when drawing to the Octane Video screen node, be sure to do:
 *              /usr/gfx/setmon -Fe 1280x1024_59 (NTSC).
 *                          OR
 *              /usr/gfx/setmon -Fe 1280x1024_49 (PAL).
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/signal.h>
#include <sys/uio.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/vl.h>
#include <dmedia/vl_mgv.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16
#define FILL_LOW 15
#define FILL_CHECK 60

/* All globals are preceded by "_" */

/* vl related globals */
VLServer _svr;
VLPath   _path;
VLNode   _src;
VLNode   _devNode;
VLNode   _voutdrn;
VLNode   _drn;
VLNode   _csc;
int      _IamPlaying = 0;
int      _rate = 0;

/* DMbuffer globals */
DMbufferpool _outpool;  /* dm buffer pool used to drive video and gfx. */
DMbuffer    *_readvBufs;/* Queue of dmbuffers with valid data read from disk */
int          _DMBUFS;   /* Number of dm buffers in _outpool     */
int          _outpoolfd;/* File descriptor for _outpool.        */
int          _iovcnt = 12;

/* Gfx declarations  */
Display   *_dpy;
GLXContext _context;
Widget     _glw;

/* Misc. definitions */
int _winx;
int _winy;
int _CurField = 0;
fileHeader_t _f;

void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void motionCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void SetVideoOutTiming(VLServer _svr, VLPath _path, VLNode node, int width, int height);
int  GetVLFormat(int packing);
void PlayIt();
void PutField(int i);
void initVL();

main(int argc, char *argv[])
{
    char *Fname;
    int i,n;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app__context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"doit", doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n";

    _winx = 1; _winy = 2;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-gfx_half") == 0) {
	    _winx = 2; _winy = 1;
        }
        else if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
        }
        else if (strcmp (argv[i], "-rate") == 0) {
            ++i;
            _rate = atoi(argv[i]);
        }
        else  {
	    Fname = argv[i];
        }
    }

    fileInit(Fname, &_f, TRUE);
    fprintf(stderr, "Playing %d of %dx%d using vector %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Field read size = %d File header size = %d\n", _f.readSize, _f.headerSize); 

    _winx  = _f.uSize/_winx;
    _winy  = _f.vSize*_winy;
    _DMBUFS = 3*_iovcnt;
    _readvBufs = (DMbuffer *)malloc(_iovcnt*sizeof(DMbuffer));

    initVL();

    toplevel = XtAppInitialize(&app__context, "4DgiftsGlx",
                               (XrmOptionDescList)NULL , 0,
                               (int*)&argc,
                               (String*)argv,
                               NULL, NULL, 0);

    XtAppAddActions(app__context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, FALSE); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight,_winy); n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);

    XtAddEventHandler(toplevel, SubstructureNotifyMask, FALSE, (XtEventHandler)motionCB, toplevel);

    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app__context);
}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    /* This is the initialize call back routine for our widget.   This
     * is called automatically from the Xt loop.
     */

    Arg args[18];
    XVisualInfo *vi;
    float xdecim, ydecim;

    _dpy = XtDisplay(_glw);
    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(w, args, 1);

    /* create a _context */
    _context = glXCreateContext(XtDisplay(_glw), vi, 0, GL_TRUE);
    GLwDrawingAreaMakeCurrent (_glw, _context);

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glColor4f(0.0, 0.0, 0.0, 1.0);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0.0, (double)_winx, 0.0, (double)_winy, -1.0, 1.0);

    xdecim = ((float)(_winx))/_f.uSize;
    ydecim = ((float)(_winy))/_f.vSize;
    glPixelZoom (xdecim, -ydecim);
    glRasterPos2i (0, _winy);
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    int x,y;
    Window dummyWin;
    VLControlValue val;
    static int first = 1;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);
    if (first) {
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);

        /* Declare window that gets data for screen node */
        val.intVal = XtWindow(_glw);
        VC(vlSetControl(_svr, _path, _drn, VL_WINDOW, &val));

        /* Set up scale factors.	*/
        val.fractVal.numerator = 1;
        val.fractVal.denominator =  _f.uSize/_winx;
        VC(vlSetControl(_svr, _path, _drn, VL_ZOOM, &val));

        val.xyVal.x = _winx;
        val.xyVal.y = _winy;
        VC(vlSetControl(_svr, _path, _drn, VL_SIZE, &val));

        val.xyVal.x = x;
        val.xyVal.y = y;
        VC(vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val));
        first = 0;

        fprintf(stderr, "\n");
        fprintf(stderr, "*** Click left mouse button to start/stop play\n");
        fprintf(stderr, "\n");
    }
}

void motionCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Window dummyWin;
    int x,y;
    VLControlValue val;

    XTranslateCoordinates(_dpy, XtWindow(_glw),
                          RootWindow(_dpy, DefaultScreen(_dpy)),
                          0, 0,
                          &x, &y,
                          &dummyWin);

    val.xyVal.x = x;
    val.xyVal.y = y;
    vlSetControl(_svr, _path, _drn, VL_ORIGIN, &val);
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    int i;
    VLControlValue val;

    if (_IamPlaying) {
        
        /* Then stop play, wait for "PlayIt" thread to return.  */
	_IamPlaying = 0;
        while (_IamPlaying == 0)  sginap(1);
	_IamPlaying = 0;

        VC (vlEndTransfer(_svr, _path));
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);
    }
    else {
        /* Then start play, sproc "PlayIt" thread. */
        _IamPlaying = 1;

	if (_CurField %2 != 0) _CurField++;

        lseek64 (_f.fd, (off64_t)(_f.headerSize + _CurField*_f.readSize), SEEK_SET);

        /* Preload some video 	*/
        for (i=0; i < _iovcnt; i++) {
            PutField(i);
        }

        /* Sproc the thread to read samples and send them to video */
        sproc((void (*)(void *))PlayIt, PR_SADDR | PR_SFDS);
    }
}

void PlayIt ()
{
    /* This thread is sproc'd from doit() and reads video fields from
     * disk.  It deposits them in dmBuffers, where the hardware then
     * sends them to video out and the screen node (gfx monitor).
     */
    int i;
    int count = 0;
    int skip;
    int fieldsToRead;
    fd_set fds;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    struct iovec iov[MAX_IOV];

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    VC (vlBeginTransfer(_svr, _path, 0, NULL));

    while (_IamPlaying) {

	fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
		         _iovcnt : (_f.Nfields - _CurField);	
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        DC (dmBufferSetPoolSelectSize(_outpool, fieldsToRead*_f.readSize) );
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

        for (i=0; i < fieldsToRead; i++) {
            DC( dmBufferAllocate(_outpool, &(_readvBufs[i])) );
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[i]);
        }

        /* Read the data */
        OC (readv(_f.fd, iov, fieldsToRead));
	
        for (i=0; i < fieldsToRead; i++) {

            /* If VL_RATE was set, then send only the first field
             * from each pair.
             */
            skip = _rate * (count%2 != 0);

            if (skip == 0) {
	        newmsc = vlGetFrontierMSC(_svr, _path, _src);
		if ((oldmsc != -1) && (newmsc != oldmsc + 1) && (count > FILL_CHECK)){
	            fprintf(stderr, "MSC drop old = %lld next = %lld\n", oldmsc, newmsc);
	        }
	        oldmsc = newmsc;

		VC (vlDMBufferPutValid(_svr, _path, _src, _readvBufs[i]));
	    }

            dmBufferFree(_readvBufs[i]);
	    count++;
	}

#if 0
        if (count % 100 == 0) {
	    i = vlGetFilledByNode(_svr, _path, _src);
	    if ( i < FILL_LOW) fprintf(stderr, "Filled down to %d\n", i);
	} 
#endif

        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
            lseek64 (_f.fd, (off64_t)(_f.headerSize), SEEK_SET);
        }
    }

    _IamPlaying--;
    _exit(0);
}

void PutField(int n)
{
    int skip;
    int size;
    DMbuffer buffer;

    while (dmBufferAllocate(_outpool, &buffer) == DM_FAILURE)sginap(1);

    OC (size = read(_f.fd, dmBufferMapData(buffer), _f.readSize) );

    /* If VL_RATE was set, then send only the first field
     * from each pair.
     */

    skip = _rate * (n%2 != 0);

    if (skip == 0) {
	 VC (vlDMBufferPutValid(_svr, _path, _src, buffer) );
    }

    dmBufferFree(buffer);

    _CurField += 1;
    if (_CurField >= _f.Nfields) {
        _CurField = 0;
        lseek64 (_f.fd, (off64_t)(_f.headerSize), SEEK_SET);
    }
}

void initVL()
{
    /* Set up video hardware.	*/
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;
    int format;
    int memnode;
    int n;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact video device");

    format = GetVLFormat(_f.packing);

    memnode = (_f.packing == VL_PACKING_YVYU_422_8) ? VL_ANY : VL_MGV_NODE_NUMBER_MEM_VGI_DL;

    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, memnode));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_SCREEN, VL_ANY));
    VC (_voutdrn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_devNode = vlGetNode(_svr, VL_DEVICE, 0, VL_ANY));

    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlAddNode(_svr, _path, _voutdrn));
    VC (vlAddNode(_svr, _path, _devNode));

    if (_f.packing != VL_PACKING_YVYU_422_8) {
        VC (_csc = vlGetNode(_svr, VL_INTERNAL, VL_CSC, VL_MGV_NODE_NUMBER_CSC));
        VC (vlAddNode(_svr, _path, _csc));
        fprintf(stderr, "Added CSC node\n");
    }

    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, _devNode, _f.uSize, _f.vSize);

    val.intVal = format;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));

    val.intVal = _f.packing;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));

    if (_rate) {
        fprintf(stderr, "Setting rate to %d\n", _rate);
	val.intVal = VL_CAPTURE_FIELDS; /*VL_CAPTURE_ODD_FIELDS; */
	VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

	val.fractVal.numerator   = _rate;
	val.fractVal.denominator = 1;
	VC (vlSetControl(_svr, _path, _src,VL_RATE, &val));
    }
    else {
	val.intVal = VL_CAPTURE_NONINTERLEAVED;
	VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));
    }

    if (_f.packing != VL_PACKING_YVYU_422_8) {
        /* Set up input csc controls */
        val.intVal = format;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_RANGE, &val));
        val.intVal = _f.packing;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_IN_PACKING, &val));

        /* Set up output csc controls */
        val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_RANGE, &val));
        val.intVal = VL_PACKING_YVYU_422_8;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_OUT_PACKING, &val));

        /* Turn on or off CSC constant hue algorithm */
        val.boolVal = 0;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_CONST_HUE, &val));

        /* Turn on or off alpha correction in the alpha channel */
        val.boolVal = 0;
        VC (vlSetControl(_svr, _path, _csc, VL_MGV_CSC_ALPHA_CORRECTION, &val));
    }

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "_DMBUFS" buffers
     * each containing _f.readSize bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f.readSize,DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.    */
    DC( dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS) );

    /* Create the dmbuffer pool used for video out.     */
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    dmParamsDestroy(p);

}
