/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* playtotex_u - reads uncompressed video fields (YUV 4:2:2 only) from disk
 *	       and sends them to both video-out and the texture drain node on 
 *	       Octane video card for real time display.
 *
 * Command line arguments, optional choices are enclosed in <>:
 *     <-single>                - Single buffer the gfx window. 
 *     <-gfx winx winy>		- Size of window on gfx monitor.
 *     <-mip>			- Use mip-maps when loading fields into
 *				  texture memory from the video card.
 *     fileName                 - A file created with vgrab_u.
 *				  The file must contain fields (not frames).
 *				  Fields are the default for vgrab_u.
 *
 *
 * The readv system call is used (along with direct I/O) for getting
 * the data from disk.  In order to use readv with direct I/O each field
 * must be rounded up to a page (16k) boundary.  The fields are stored this
 * way on disk.  Fortunately, each vl buffer element handed back by the vl
 * is also padded out to a 16k page boundary.  Exploiting this "feature" is
 * what makes it all work.
 *
 * This program has been run on an Octane MXI with a ProVideo board.
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After starting the program, a blank window appears.
 *      - Click with the left mouse button to start play.  
 *	- Click again to stop play then again to start again. Each button click
 *	  causes the main thread to enter the routine doit().
 *      - Each time play is started,  a play thread (see PlayIt) 
 *        is sproc'd from inside the routine doit().
 *      - While things are playing, the main thread is in  XtAppMainLoop(), 
 *	  waiting for another button click.
 *
 *    Play Thread: 
 *      - PlayIt() simply cycles through the file.  It allocates buffers from
 *        _outpool and loads them with uncompressed fields read from disk.
 *      - This thread uses the readv system call to read _iovcnt fields at a 
 *	  time into allocated dmbuffers.
 *      - After the read has finished,   PlayIt() marks these buffers as 
 *	  valid for output. 
 *      - Each field is padded out to 16k on disk (this value is represented 
 *	  by _readSize).
 *
 *    FrameAdvance:
 *      - This thread's only job is to move fields from the video card
 *	  into texture memory.
 *	- This is done by calling glCopyTexSubImage2DEXT with the texture
 *        drain node on the video board established as a read source for this
 *	  operation (see _glxvidsrc variable).
 *      - Once in tetxure memory, the video field is drawn to the gfx monitor 
 *	  by drawing a rectangle where each vertex is  assgined a texture coordinate 
 *	  correspondingg to a corner of the video field in texture space.  The size
 *        of the rectangle is arbitrary and this is what allows arbitrary sizing. 
 *
 * Notes:
 * ------
 * For best results make sure and do: 
 *              /usr/gfx/setmon -Fe 1280x1024_59 (NTSC).
 *                          OR
 *              /usr/gfx/setmon -Fe 1280x1024_49 (PAL).
 *
 * The app. could try to send RGB into the video texture node, but then a
 * a color space converter node would have to be added to the path.  The data
 * would gotr from RGB, to YUV into the texture node, then back to RGB for input
 * into texture memory - why bother.
 *
 */
 
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/signal.h>
#include <sys/uio.h>

#include <X11/Intrinsic.h>
#include <Xm/Form.h>
#include <GL/GLwMDrawA.h>
#include <GL/gl.h>
#include <dmedia/vl.h>
#include <dmedia/vl_mgv.h>
#include "common/handy.h"
#include "common/fileutils.h"


#define MAX_IOV 16 
#define FILL_LOW 10
#define FILL_HIGH 20 
#define FILL_CHECK 60

/* All global variables are preced by an "_" */


/* Video related defines */
VLServer _svr;
VLPath _path;
VLNode _src;
VLNode _vout_drn;
VLNode _drn;
VLNode _devNode;

/* DMbuffer globals */
DMbufferpool _outpool;  /* dm buffer pool used to drive video and gfx. */
DMbuffer    *_readvBufs;/* Queue of dmbuffers with valid data read from disk */
int          _DMBUFS;   /* Number of dm buffers in _outpool     */
int          _outpoolfd;/* File descriptor for _outpool.        */

/* texture defines */
int _mipmapmode = 0;
int _tex_width;               		/* texture size */
int _tex_height;               		/* texture size */
int _tex_hoff;                          /* horizontal offset for video load */
unsigned int _texA,_texB;               /* texture names */

fileHeader_t _f;
int _dbuffer = 1;
int _iovcnt = 12;
int _winx = 0;
int _winy = 0; 
int _CurField = 0;

int _IamPlaying = 0;
int _waitingForEnd;
int _waitingCounting;

/* OpenGL defines */
Display *_dpy;
GLXContext _context;             
GLXVideoSourceSGIX _glxvidsrc; 
Widget _glw;

/* Proceduere declarations	*/
void createTexture(unsigned int *texname, int width, int height, int format, int mip);
void initCB( Widget w, caddr_t client_data, caddr_t call_data);
void exposeCB( Widget w, caddr_t client_data, caddr_t call_data);
void doit(Widget , XEvent *, String *, Cardinal *);
void initVL();
void BuildList();
void PlayIt();
void FrameAdvance();
void PutField();

main(int argc, char *argv[])
{
    char *Fname;
    int i,n;
    int packing;
    Arg args[20];
    XtTranslations transTable;
    XtAppContext app_context;
    Widget toplevel, form;
    XtActionsRec actionsTable[] = {{"doit", doit},{"doit", doit},};
    char defaultTranslations[] = "<Btn1Down>:doit()\n<Btn2Down>:doit()\n";

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-gfx") == 0) {
            ++i;
            _winx = atoi(argv[i]);
            ++i;
            _winy = atoi(argv[i]);
        }
        else if (strcmp (argv[i], "-mip") == 0) {
            fprintf(stderr, "USING MIPMAPS\n");
	    _mipmapmode = 1;
	}
	else if (strcmp (argv[i], "-single") == 0) {
	    _dbuffer = 0;
	}
        else  {
	    Fname = argv[i];
        }
    }

    fileInit(Fname, &_f, TRUE);
    ERROR ((_f.packing != VL_PACKING_YVYU_422_8), "Must be 422 data\n");

    fprintf(stderr, "Playing %d of %dx%d using vector %d\n", _f.Nfields, _f.uSize, _f.vSize, _iovcnt);
    fprintf(stderr, "Field read size = %d File header size = %d\n", _f.readSize, _f.headerSize);

    if (_winx == 0) {
	_winx = _f.uSize;
	_winy = _f.vSize;
    }

    _DMBUFS = 3*_iovcnt;
    _readvBufs = (DMbuffer *)malloc(_iovcnt*sizeof(DMbuffer));
    initVL();

    toplevel = XtAppInitialize(&app_context, "4DgiftsGlx",
                               (XrmOptionDescList)NULL , 0,
                               (int*)&argc,
                               (String*)argv, NULL, NULL, 0);

    XtAppAddActions(app_context, actionsTable, XtNumber(actionsTable));
    transTable = XtParseTranslationTable(defaultTranslations);
    form = XmCreateForm(toplevel, "form", (ArgList)NULL, 0);
    XtManageChild(form);

    n = 0;
    XtSetArg(args[n], GLwNdoublebuffer, _dbuffer); n++;
    XtSetArg(args[n], GLwNdepthSize, 24); n++;
    XtSetArg(args[n], GLwNrgba, TRUE); n++;
    XtSetArg(args[n], XmNwidth, _winx); n++;
    XtSetArg(args[n], XmNheight,_winy); n++;
    _glw = GLwCreateMDrawingArea(form, "_glwidget", args, n);

    XtManageChild (_glw);
    XtAddCallback(_glw, GLwNexposeCallback, (XtCallbackProc)exposeCB, 0);
    XtAddCallback(_glw, GLwNginitCallback, (XtCallbackProc)initCB, 0);
    XtOverrideTranslations(_glw, transTable);
    XtRealizeWidget(toplevel);
    XtAppMainLoop(app_context);

}

void initCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    Arg args[18];
    XVisualInfo *vi;
    int _tex_format = GL_RGB;
    int val;

    _dpy = XtDisplay(_glw);

    XtSetArg(args[0], GLwNvisualInfo, &vi);
    XtGetValues(_glw, args, 1);

    /* create a visual context */
    _context = glXCreateContext(_dpy, vi, 0, GL_TRUE);

    GNC (_context = glXCreateContext(_dpy, vi, 0, GL_TRUE));
    GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));

    if (_mipmapmode) {  /* Then limited to 512 x 256 */
        _tex_width = 512;
        _tex_height = 256;
    }
    else {
	_tex_width = 1024; 
        _tex_height = (_f.vSize == 243) ? 256 : 512; 
    }

    /* there is a TRAM load restriction that forces us to load
     * our video field such that its right hand edge must be on
     * a 32 texel boundary. We will set the horizontal offset
     * for our sub-loads so that this restiction is satisfied
     */
    if (_mipmapmode) _tex_hoff = 0;
    else _tex_hoff = ((_f.uSize & 0x1f) ? (0x20 - (_f.uSize & 0x1f)) : 0x00);

    /* set texture environment */
    glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);

    /* Tell video that gfx has double buffered textures set up.  */
    glHint(GL_TEXTURE_MULTI_BUFFER_HINT_SGIX,GL_FASTEST);

    /* create textures A & B */
    createTexture(&_texA, _tex_width, _tex_height, _tex_format, _mipmapmode);
    createTexture(&_texB, _tex_width, _tex_height, _tex_format, _mipmapmode);

    /* init viewport */
    glViewport(0,0,_winx,_winy);

    /* set up transformation matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    /* set clear color */
    glClearColor(0.0,0.0,0.0,1.0);
    
    /* determine color of polygons */
    glColor4f(1.0, 1.0, 1.0, 1.0);
    
    /* enable texturing */
    glEnable(GL_TEXTURE_2D);

    /* create read drawable for video. When glCopyTexSubImage2DEXT()
     * is called in FrameAdvance(). It transfers from this video source
     * to gfx texture memory.
     */  
    _glxvidsrc = glXCreateGLXVideoSourceSGIX(_dpy,DefaultScreen(_dpy),_svr,
					    _path,VL_TEXTURE,_drn);
    BuildList();

    /* check for GL errors */
    while ((val = glGetError()) != GL_NO_ERROR) {
        fprintf(stderr,"******* GL ERROR ******* 0x%x\n",val);
    }
}

void exposeCB( Widget w, caddr_t client_data, caddr_t call_data)
{
    static int first = 1;

    if (first) {
        GC (glXMakeCurrent(_dpy, XtWindow(_glw), _context));
        glClear(GL_COLOR_BUFFER_BIT);
        if (_dbuffer) glXSwapBuffers(_dpy,XtWindow(_glw));
        first = 0;

        fprintf(stderr, "\n");
        fprintf(stderr, "*** Click left mouse button to start/stop play\n");
        fprintf(stderr, "\n");
    }
}

void doit( Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    int i;
    VLControlValue val;
    VLTransferDescriptor xferDesc;
    VLEvent ev;

    if (_IamPlaying) {

        /* Then stop play, wait for "PlayIt" thread to return.  */
        _IamPlaying = 0;
        while (_IamPlaying != -2)  sginap(1);
        _IamPlaying = 0;

        VC (vlEndTransfer(_svr, _path));
    }
    else {
        _IamPlaying = 1;

        if (_CurField % 2 != 0) _CurField += 1;
        lseek (_f.fd, _f.headerSize + _CurField*_f.readSize, SEEK_SET);

        for (i=0; i < 2*_iovcnt; i++) {
            PutField();
        }
	glXMakeCurrent(_dpy, None, NULL);
        sproc((void (*)(void *))FrameAdvance, PR_SADDR | PR_SFDS);
        sproc((void (*)(void *))PlayIt, PR_SADDR | PR_SFDS);
        VC (vlBeginTransfer(_svr, _path, 0, NULL));

    }
}

void PlayIt ()
{
    /* This thread is sproc'd from doit() and reads video fields from
     * disk.  It deposits them in dmBuffers, where the hardware then
     * sends them to video out and the screen node (gfx monitor).
     */
    int i;
    int count = 0;
    int fieldsToRead;
    fd_set fds;
    stamp_t oldmsc = -1;
    stamp_t newmsc = -1;
    struct iovec iov[MAX_IOV];

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    while (_IamPlaying)  {

	fieldsToRead = ((_CurField + _iovcnt) <= _f.Nfields) ?
		         _iovcnt : (_f.Nfields - _CurField);	
        FD_ZERO(&fds);
        FD_SET(_outpoolfd, &fds);
        DC (dmBufferSetPoolSelectSize(_outpool, fieldsToRead*_f.readSize) );
        OC (select(_outpoolfd+1, NULL, &fds, NULL, NULL));

        for (i=0; i < fieldsToRead; i++) {
            DC( dmBufferAllocate(_outpool, &(_readvBufs[i])) );
            iov[i].iov_len = _f.readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[i]);
        }

        /* Read the data */
        OC (readv(_f.fd, iov, fieldsToRead));
	
        for (i=0; i < fieldsToRead; i++) {

	    newmsc = vlGetFrontierMSC(_svr, _path, _src);
	    if ((oldmsc != -1) && (newmsc != oldmsc + 1) && (count > FILL_CHECK)){
	        fprintf(stderr, "MSC drop old = %lld next = %lld\n", oldmsc, newmsc);
	    }
	    oldmsc = newmsc;

            VC (vlDMBufferPutValid(_svr, _path, _src, _readvBufs[i]));
            dmBufferFree(_readvBufs[i]);
	    count++;
	}

#if 0
        if (count % 100 == 0) {
	    i = vlGetFilledByNode(_svr, _path, _src);
	    if ( i < FILL_LOW) fprintf(stderr, "Filled down to %d\n", i);
	} 
#endif

        _CurField += fieldsToRead;
        if (_CurField >= _f.Nfields) {
	    _CurField = 0;
            lseek (_f.fd, _f.headerSize, SEEK_SET);
            fprintf(stderr, "Done pass through file\n");
        }
    }

    _IamPlaying--;
    _exit(0);
}

void PutField()
{
    int size;
    DMbuffer buffer;

    while (dmBufferAllocate(_outpool, &buffer) == DM_FAILURE)sginap(1);

    OC (size = read(_f.fd, dmBufferMapData(buffer), _f.readSize) );

    VC (vlDMBufferPutValid(_svr, _path, _src, buffer) );
    dmBufferFree(buffer);

    _CurField += 1;
    if (_CurField >= _f.Nfields) {
        _CurField = 0;
        lseek (_f.fd, _f.headerSize, SEEK_SET);
        fprintf(stderr, "Done pass through file\n");
    }
}

void FrameAdvance()
{
    /* This routine gets sproc'd off by doit().  Its only job is
     * to transfer video from Imapct video to Impact gfx texture
     * memory.  Things have already been preloaded into the vlBuffer
     * and vlBeginTransfer() has been called.  So at this point video
     * is streaming into the field buffer that sits between the tmi
     * on the video card and gfx texture memory.
     *
     * glCopyTexSubImage2DEXT() is called to move the video field from the 
     * video card to gfx texture memory.  If it is not called quickly
     * enough then the field buffer on the video card will get overwritten
     * and you will miss sending a field to texture memory.
     *
     * glCopyTexSubImage2DEXT does not block, it starts the transfer from
     * the tmi to texture memory and then returns to the caller.   The
     * glHint() call in this program sets up doubled buffered texture memory.
     * Two textures are created (_texA and _texB) and you can be loading one
     * from the tmi while drawing out of the other.  This overlap is required
     * in order to hit real time rates.
     */

    int twidth, theight;

    /* Exit if my parent dies */
    sigset(SIGHUP, SIG_DFL);
    (void)prctl(PR_TERMCHILD);

    if (_mipmapmode) {
	twidth = 512; theight = 256;
    }
    else {
        twidth = _f.uSize; theight = _f.vSize;
    }

    /* Grab the read and draw contexts, which the main thread has
     * surrendered.
     */
    GC (glXMakeCurrentReadSGI(_dpy, XtWindow(_glw), _glxvidsrc, _context));

    /* load A */
    glBindTextureEXT(GL_TEXTURE_2D,_texA);
    glCopyTexSubImage2DEXT(GL_TEXTURE_2D,0,_tex_hoff,0,0,0, twidth, theight);
    
    /* load B */
    glBindTextureEXT(GL_TEXTURE_2D,_texB);
    glCopyTexSubImage2DEXT(GL_TEXTURE_2D,0,_tex_hoff,0,0,0, twidth, theight);

    while (_IamPlaying != -1) {
	/* draw A  while B is loading*/
        glBindTextureEXT(GL_TEXTURE_2D,_texA);
        glCallList(1);
        if (_dbuffer) glXSwapBuffers(_dpy,XtWindow(_glw));

	/* start the next load of A */
        glCopyTexSubImage2DEXT(GL_TEXTURE_2D,0,_tex_hoff,0,0,0, twidth, theight);

	/* draw B while A is loading */
        glBindTextureEXT(GL_TEXTURE_2D,_texB);
        glCallList(1);
        if (_dbuffer) glXSwapBuffers(_dpy,XtWindow(_glw));

	/* start the next load of B */
        glCopyTexSubImage2DEXT(GL_TEXTURE_2D,0,_tex_hoff,0,0,0, twidth, theight);
    }

    /* Surrender the read and draw contexts so the main thread
     * can use them again.
     */
    GC (glXMakeCurrentReadSGI(_dpy, None, None, NULL));
    _IamPlaying--;
    _exit(0);
}

void initVL()
{
    int i,n;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
        devPtr = &(devlist.devices[n]);
        if (strcmp (devPtr->name, "impact") == 0) {
            n = devlist.numDevices;
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No impact video device");

    VC (_src = vlGetNode(_svr, VL_SRC, VL_MEM, VL_ANY));
    VC (_drn = vlGetNode(_svr, VL_DRN, VL_TEXTURE, VL_MGV_NODE_NUMBER_TEXTURE));
    VC (_vout_drn = vlGetNode(_svr, VL_DRN, VL_VIDEO, VL_ANY));
    VC (_devNode = vlGetNode(_svr, VL_DEVICE, 0, VL_ANY));

    VC (_path = vlCreatePath(_svr, devPtr->dev, _src, _drn)); 
    VC (vlAddNode(_svr, _path, _vout_drn));
    VC (vlAddNode(_svr, _path, _devNode));
    VC (vlSetupPaths(_svr, (VLPathList)&_path, 1, VL_SHARE, VL_SHARE));

    SetVideoOutTiming(_svr, _path, _devNode, _f.uSize, _f.vSize);

    if (_mipmapmode) {
        /* Enable mip-mapping */
        val.intVal = VL_MGV_TEXTURE_MIPMAP_ON;
        VC (vlSetControl(_svr,_path,_drn,VL_MGV_TEXTURE_MIPMAP_MODE,&val));
    
        /* Set the H and V scale sizes.  This tells the tmi hardware
         * to scale down the outgoing texture (the one being sent to the 
         * gfx card).  This is required because in mip-map mode 
         * the texture in gfx memory can be no bigger than 512 x 256
         * while the video is either 720x243 or 720x288).  
	 *
	 * When the corresponding rectangle is drawn on the screen, the
         * 512x256 texture map is filtered either up or down to fit the
         * gfx window.
	 * 
  	 */
        if (_f.uSize > 512) {
	    val.fractVal.numerator = 512; val.fractVal.denominator = _f.uSize;
	    VC (vlSetControl(_svr,_path,_drn,VL_MGV_HASPECT,&val));
	}

        if (_f.vSize > 256) {
            val.fractVal.numerator = 256; val.fractVal.denominator = _f.vSize;
            VC (vlSetControl(_svr,_path,_drn,VL_MGV_VASPECT,&val));
	}
    }

    else {
        val.intVal = VL_MGV_TEXTURE_MIPMAP_OFF;
        VC (vlSetControl(_svr,_path,_drn,VL_MGV_TEXTURE_MIPMAP_MODE,&val));
    }

    val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
    VC (vlSetControl(_svr, _path, _src, VL_FORMAT, &val));
    val.intVal = VL_PACKING_YVYU_422_8;
    VC (vlSetControl(_svr, _path, _src, VL_PACKING, &val));
    val.intVal = VL_CAPTURE_NONINTERLEAVED;
    VC (vlSetControl(_svr, _path, _src, VL_CAP_TYPE, &val));

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "_DMBUFS" buffers
     * each containing _f.readSize bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f.readSize,DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMGetParams(_svr, _path, _src, p) );

    /* Flag pool for use by several threads.    */
    DC( dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS) );

    /* Create the dmbuffer pool used for video out.     */
    DC( dmBufferCreatePool(p, &_outpool) );
    _outpoolfd = dmBufferGetPoolFD(_outpool);
    dmParamsDestroy(p);
}

void BuildList()
{
    float t0[2],t1[2],t2[2],t3[3];          /* texture coordinates */
    float v0[2],v1[2],v2[2],v3[2];          /* vertice coordinates */

    /* The way video gets drawn to the screen is to do draw
     * a rectangle that corresponds to the size of the window.  With
     * texture mapping enabled, each pixel of the rectangle gets its
     * color value from the texture map (which just happens to contain
     * the field of video we want to draw).  What makes this all interesting
     * is that you can vary the size/orientation of the rectangle and the
     * texture mapping hardware filters the video image to fit the rectangle.
     */

    /* init vertice coordinates.  These coordinates match the default
     * projection space of -1 to 1 in both x and y.  In other words
     * these coordinates will exactly cover the entire window on the gfx 
     * monitor with a rectangle.
     */
    v0[0] = -1.0; v0[1] = -1.0;
    v1[0] =  1.0; v1[1] = -1.0;
    v2[0] =  1.0; v2[1] =  1.0;
    v3[0] = -1.0; v3[1] =  1.0;
    
    /* Texture space goes from (0,0) to (1,1).  So a 720x243
     * field is mapped onto this floating point range.  Each corner of
     * the rectangle needs to be assigned a location in texture space.
     *
     * Remember that textures are created in powers of 2, so you don't want to
     * take the entire texture map (i.e a 720x243 video field has a 1024x256 
     * texture space.  So for example t1[0] (which pairs with v1[0]) equals 
     * _f.uSize/_tex_width = 720/1024 (don't worry about _tex_hoff hack, you 
     * get the idea).
     */ 
    
    /* init texture coordinates */
    t0[0] = (float)_tex_hoff / (float)_tex_width; 
    t0[1] = 0.0;
    t1[0] = (float)(_tex_hoff + _f.uSize) / (float)_tex_width;
    t1[1] = 0.0;
    t2[0] = (float)(_tex_hoff + _f.uSize) / (float)_tex_width;
    t2[1] = (float)_f.vSize / (float)_tex_height;
    t3[0] = (float)_tex_hoff / (float)_tex_width; 
    t3[1] = (float)_f.vSize / (float)_tex_height;
    
    if (_mipmapmode) {
        /* Then texture width =  512 so take it all. */ 
        t1[0] = t2[0] = 1.0;  
        /* Then texture height = 256 so take it all. */
        if (_f.vSize == 288) t2[1] = t3[1] = 1.0;
    }

    glNewList(1, GL_COMPILE);
        /* Draw the polygon. */ 
        glBegin(GL_QUADS);
        glTexCoord2f(t0[0],t0[1]);
        glVertex2f(v0[0],v0[1]);
        glTexCoord2f(t1[0],t1[1]);
        glVertex2f(v1[0],v1[1]);
        glTexCoord2f(t2[0],t2[1]);
        glVertex2f(v2[0],v2[1]);
        glTexCoord2f(t3[0],t3[1]);
        glVertex2f(v3[0],v3[1]);
        glEnd();
    glEndList();
}
