/*
 * Copyright (c) 1997 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the name of Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER
 * RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT ADVISED OF THE
 * POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* vgrab_u = Grabs uncompressed video to disk using Octane ProVideo
 *	     board using DMbuffers.
 * 
 * Command line arguments, optional choices are enclosed in <>:
 *     <-iov vectorLength>      - Number of writes per writev call.
 *			          Default is 12, 8 and 16 are also good values.
 *     <-n fieldsToCapture> 	- Number of fields to capture, default = 200.
 *				  This arg should be an even number.
 *     <-rgba OR		- Capture RGBA fields to disk (40 MB/sec!!)
 *      -rgbp OR 		- Capture packed RGB fields to disk (30 MB/sec).
 *      -yuvp OR		- Capture packed YUV fields to disk (30 MB/sec).
 *	-yuv422> 	        - Capture YUV 4:2:2 to disk (20 MB/sec.). 
 *
 *     <-frames> 		- Capture frames instead of fields. 
 *
 *     fileName                 - File for results. 
 *
 * This is a real simple program to capture fields of uncompresed video
 * to disk in real time.  Video was captured using a Ciprico 6900 ultra-
 * scsi raid array.  For some formats, multiple arrays are required (like 
 * -rgba or -rgbp).  However, some of the other example programs require files
 * with only a couple of fields (see vstreamfx or gstreamfx).  This program can 
 * capture 2 fields of rgba, rgbp, yuvp or yuv422  to even an ordinary drive. 
 *
 * The writev system call is used (along with direct I/O) for getting 
 * the data to disk.  In order to use writev with direct I/O each field
 * must be rounded up to a page boundary when being wriiten to disk.  
 * Fortunately, each buffer element handed back by the vl is also padded 
 * out to a page boundary. Exploiting this "feature" is what makes it work.
 *
 * The flag "-iov" allows the user to specify the number of writes for
 * writev to perform.  Each write corresponds to a video field padded out
 * to a 16k boundary. 
 *
 * So here is the program flow from 50,000 feet.
 *
 *    Main Thread:
 *      - After initializing everything, the routine doit() is called to
 *        perform the real work.
 *	- doit() sits in a tight loop grabbing pointers to fields accumulating
 *	  in the DMBuffer.  
 *	- Once "_iovcnt" fields have been accumulated, they are written to disk
 *	  using writev().
 *	- After capturing the desired number of fields, the program exits.
 */


#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <dmedia/vl.h>
#include <dmedia/vl_mgv.h>
#include "common/handy.h"
#include "common/fileutils.h"

#define MAX_IOV 16

/* All global variables in this file are preceded by an "_".  */

/* File globals */
int  _iovcnt = 12;		/* # of writes for writev.		*/
fileHeader_t *_f;

/* Video globals  */
VLServer _svr;			/* Connection to server.		*/ 
VLNode   _Csrc;			/* Source for capture (video node).     */
VLNode   _Cdrn;			/* Drain for capture (memory node).	*/
VLPath   _Cpath;		/* Capture path.			*/
VLNode   _csc;
int      _capType = VL_CAPTURE_NONINTERLEAVED;

/* DMbuffer globals */
DMbufferpool _inpool;  /* dm buffer pool used to drive video and gfx. */
DMbuffer    *_readvBufs;/* Queue of dmbuffers with valid data read from disk */
int          _DMBUFS;   /* Number of dm buffers in _outpool     */


/* Procudure declarations		*/
void doit();
void initVL();
int GetVLFormat(int packing);

main(int argc, char *argv[])
{
    int i;
    int  packing = VL_PACKING_YVYU_422_8;
    int  imageCount = 200;
    char * fname;

    for (i=1; i < argc; i++) {
        if (strcmp (argv[i], "-n") == 0) {
            ++i;
            imageCount = atoi(argv[i]);
	}
        else if (strcmp (argv[i], "-iov") == 0) {
            ++i;
            _iovcnt = atoi(argv[i]);
            if (_iovcnt > MAX_IOV) _iovcnt = MAX_IOV;
        }
        else if (strcmp (argv[i], "-rgba") == 0) {
	    packing = VL_PACKING_ABGR_8;
        }
        else if (strcmp (argv[i], "-rgbp") == 0) {
	    packing = VL_PACKING_BGR_8_P;
        }
        else if (strcmp (argv[i], "-yuvp") == 0) {
	    packing = VL_PACKING_UYV_8_P;
        }
        else if (strcmp (argv[i], "-yuv422") == 0) {
	    packing = VL_PACKING_YVYU_422_8;
        }
        else if (strcmp (argv[i], "-frames") == 0) {
	    _capType = VL_CAPTURE_INTERLEAVED; 
        }
	else {
	    fname = argv[i];
	}
    }

    fprintf(stderr, "Capturing %d fields,  writev vector length = %d\n", imageCount,_iovcnt);

    _f = fileCreate(fname, TRUE);
    _f->packing = packing;
    _f->compressed = FALSE;
    _f->pad = _f->headerSize;
    _f->Nfields = imageCount;;

    _DMBUFS = 2*_iovcnt;
    _readvBufs = (DMbuffer *)malloc(_iovcnt*sizeof(DMbuffer));

    initVL();
    doit();
}


void doit()
{
    /* The routine which does all the work capturing and writing 
     * fields to disk.
     */
    struct iovec iov[MAX_IOV];
    stamp_t oldmsc = -1;
    stamp_t newmsc;
    int i, toGet;
    int fieldsCaptured = 0;

    /* Start video samples flowing into dmBuffers. 
     */
    VC (vlBeginTransfer(_svr, _Cpath, 0, NULL));

    /* Now loop and capture the fields. */
    while (fieldsCaptured < _f->Nfields) {

	toGet = ((fieldsCaptured + _iovcnt) <= _f->Nfields) ?
                         _iovcnt : (_f->Nfields - fieldsCaptured);

        for (i= 0; i < toGet; i++) {
            do {    /* Get pointer to next valid field */ 
		vlDMBufferGetValid(_svr, _Cpath, _Cdrn, &(_readvBufs[i]) );
            } while (_readvBufs[i] == NULL);

	    /* Load the write size and write address for the vector */
            iov[i].iov_len = _f->readSize;
            iov[i].iov_base = dmBufferMapData(_readvBufs[i]); 

	    /* Check the MSC of the sample to see if we are dropping
             * any fields.
             */
            newmsc = vlGetFrontierMSC(_svr, _Cpath, _Cdrn);
            if ((oldmsc != -1) && (newmsc != oldmsc + 1)) {
                fprintf(stderr, "MSC drop old = %lld, new = %lld\n", oldmsc, newmsc);
            }
            oldmsc = newmsc;
	}

	OC (writev(_f->fd, iov, toGet));  /* write the fields to disk. 	*/

	/* Indicate that the buffer entries are free for re-use.	*/
        for (i=0; i < toGet; i++) {
	    dmBufferFree(_readvBufs[i]);
	}
	fieldsCaptured += toGet;
       
        if (fieldsCaptured % 600 == 0) {
	    fprintf(stderr, "...Captured %d fields\n", fieldsCaptured);
	}
    }
    fprintf(stderr, "\n");

    VC (vlEndTransfer(_svr, _Cpath));

    /* Write out the correct header info.	*/
    writeHeader(_f);
    fsync(_f->fd);
    close(_f->fd);
    fprintf(stderr, "Done writing %d fields to file\n", fieldsCaptured);
}

void initVL()
{
    /* Set up the video hardware for capturing data in YUV 4:2:2
     * format.  This includes creating a VL_VIDEO to VL_MEM path.
     */
    int i, n;
    int format;
    int memnode;
    VLDevice *devPtr;
    VLDevList devlist;
    VLControlValue val;
    DMparams* p;

    VNC (_svr = vlOpenVideo(""));
    vlGetDeviceList(_svr, &devlist);

    for (n = 0; n < devlist.numDevices; n++) {
	devPtr = &(devlist.devices[n]);
        if ((strcmp (devPtr->name, "mvp") == 0) ||
	    (strcmp (devPtr->name, "impact") == 0) ) {
            n = devlist.numDevices; 
        }
        else devPtr = NULL;
    }
    ERROR((devPtr == NULL), "No video device");

    format = GetVLFormat(_f->packing); 
    memnode = (_f->packing == VL_PACKING_YVYU_422_8) ? VL_ANY : VL_MGV_NODE_NUMBER_MEM_VGI_DL;

    /* Grab a video node and a memory node, then create a
     * path from video-in to memory.
     */
    VC (_Csrc = vlGetNode(_svr, VL_SRC, VL_VIDEO, VL_ANY));
    VC (_Cdrn = vlGetNode(_svr, VL_DRN, VL_MEM, memnode));
    VC (_Cpath = vlCreatePath(_svr, devPtr->dev, _Csrc, _Cdrn));

    if (_f->packing != VL_PACKING_YVYU_422_8) {
        VC (_csc = vlGetNode(_svr, VL_INTERNAL, VL_CSC, VL_MGV_NODE_NUMBER_CSC));
        VC (vlAddNode(_svr, _Cpath, _csc));
        fprintf(stderr, "Added CSC node\n");
    }

    VC (vlSetupPaths(_svr, (VLPathList)&_Cpath, 1, VL_SHARE, VL_SHARE));

    /* Set the attributes for the incoming video.  This is
     * done on the memory drain.
     */
    val.intVal = format; 
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_FORMAT, &val));

    val.intVal = _f->packing; 
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_PACKING, &val));

    val.intVal = _capType; 
    VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_CAP_TYPE, &val));

    if (strcmp (devPtr->name, "impact") == 0) {
	val.intVal = _f->headerSize;
	VC (vlSetControl(_svr, _Cpath, _Cdrn, VL_MGV_BUFFER_QUANTUM, &val));
    }

    if (_f->packing != VL_PACKING_YVYU_422_8) {
        /* Set up input csc controls */
        val.intVal = VL_FORMAT_DIGITAL_COMPONENT_SERIAL;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_IN_RANGE, &val));
        val.intVal = VL_PACKING_YVYU_422_8;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_IN_PACKING, &val));

        /* Set up output csc controls */
        val.intVal = format;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_OUT_RANGE, &val));
        val.intVal = _f->packing;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_OUT_PACKING, &val));

        /* Turn on or off CSC constant hue algorithm */
        val.boolVal = 0;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_CONST_HUE, &val));

        /* Turn on or off alpha correction in the alpha channel */
        val.boolVal = 0;
        VC (vlSetControl(_svr, _Cpath, _csc, VL_MGV_CSC_ALPHA_CORRECTION, &val));
    }

    vlGetControl(_svr, _Cpath, _Cdrn, VL_SIZE, &val);
    fprintf(stderr,"Field Size = %d %d\n", val.xyVal.x, val.xyVal.y);
    _f->uSize = val.xyVal.x;
    _f->vSize = val.xyVal.y;

    _f->readSize = vlGetTransferSize(_svr, _Cpath);
    val.intVal = _f->readSize;

    i = _f->readSize/_f->headerSize;
    if ((_f->readSize % _f->headerSize) != 0) ++i;
    _f->readSize = _f->headerSize*i;

    fprintf(stderr, "Transfer size  = %d Write size  = %d\n", val.intVal, _f->readSize);

    dmParamsCreate(&p); /* Create empty parameter list */

    /* Set the defaults for the dmbuffer pool, we want "_DMBUFS" buffers
     * each containing _readSize bytes.  Load this info into parameter list.
     */
    DC(dmBufferSetPoolDefaults(p,_DMBUFS,_f->readSize,DM_FALSE,DM_TRUE));

    /* Get info required for the vl to use this pool and
     * store it in the parameter list "p".  This is required
     * since buffers from this pool will be sent directly to video-out.
     */
    VC( vlDMGetParams(_svr, _Cpath, _Cdrn, p) );

    /* Flag pool for use by several threads.    */
    DC( dmParamsSetEnum(p, DM_POOL_SHARE, DM_POOL_INTERPROCESS) );

    /* Create the dmbuffer pool used for video out.     */
    DC( dmBufferCreatePool(p, &_inpool) );
    VC( vlDMPoolRegister(_svr, _Cpath, _Cdrn, _inpool) ); /* Attach pool to vl memory node.       */

    dmParamsDestroy(p);
}
